/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2019-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_AR_ALGORITHM_CPP

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include "osm_ar_algorithm.hpp"
#include "osm_armgr.hpp"

void AdaptiveRoutingAlgorithm::AllocatePersistentARGroupIDs() {
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
	DfSwData *p_switch_data;
	uint16_t sw_lid_num;
	uint16_t group_cap;
	uint8_t plft;

	for (plft = 0; plft < UCAST_MGR_MAX_PLFT; plft++) {
		for (; sw_node_it != m_sw_map_.end(); ++sw_node_it) {
			sw_lid_num = sw_node_it->second.m_general_sw_info.m_lid;
			p_switch_data = sw_node_it->second.m_p_df_data;
			group_cap = GetGroupCap(sw_node_it->second);
			if ((sw_node_it->second.m_config_rn) &&
			    p_switch_data->m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {
				m_ar_mgr_.AllocateSwArGroup(sw_lid_num, group_cap, plft);
			}
		}
	}
}

int AdaptiveRoutingAlgorithm::BuildLidMapping(LidMapping &lid_mapping)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "BuildLidMapping\n");

	int rc = 0;
	osm_node_t *p_osm_node;
	cl_qmap_t *p_node_guid_tbl = &m_ar_mgr_.m_p_osm_subn->node_guid_tbl;
	u_int16_t base_lid;
	lid_mapping.m_sw_lids.reserve(m_sw_map_.size());
	/* Go over all nodes in osm database */
	for (p_osm_node =  (osm_node_t *) cl_qmap_head(p_node_guid_tbl);
		 p_osm_node != (osm_node_t *) cl_qmap_end(p_node_guid_tbl);
		 p_osm_node = (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

		if (p_osm_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {
			base_lid = osm_node_get_base_lid_ho(p_osm_node, 0);
			OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
				   "Map Switch LID: %u\n", base_lid);

			lid_mapping.m_lid_to_sw_lid_mapping[base_lid] = base_lid;
			lid_mapping.m_lid_to_base_lid_mapping[base_lid] = base_lid;
			lid_mapping.m_sw_lids.push_back(base_lid);
			continue;
		}

		for (u_int8_t port_num = 1;
			 port_num <= p_osm_node->node_info.num_ports; port_num++) {

			osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
			osm_node_t *p_remote_node =
				osm_node_get_remote_node(p_osm_node, port_num, NULL);

			if (!p_physp || !p_remote_node || !osm_link_is_healthy(p_physp))
				continue;
			osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
			if (!p_remote_physp)
				continue;

			rc |= SetHcaLidMapping(p_physp, p_remote_node, lid_mapping);
		}
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int AdaptiveRoutingAlgorithm::SetHcaLidMapping(osm_physp_t *p_hca_physp,
					       osm_node_t *p_remote_sw_node,
					       LidMapping &lid_mapping)
{
	OSM_AR_LOG_ENTER(m_p_osm_log);

	u_int16_t node_lid = cl_ntoh16(osm_physp_get_base_lid(p_hca_physp));
	if (p_remote_sw_node->node_info.node_type != IB_NODE_TYPE_SWITCH ||
		node_lid == 0) {
		OSM_AR_LOG_RETURN(m_p_osm_log, -1);
	}

	uint8_t lmc = osm_physp_get_lmc(p_hca_physp);

	u_int16_t remote_node_lid = osm_node_get_base_lid_ho(p_remote_sw_node, 0);

	u_int16_t numLidsPerPort = (u_int16_t)((u_int16_t)1 << lmc);
	u_int16_t lid_limit = node_lid+numLidsPerPort;

	for (u_int16_t l = node_lid; l < lid_limit; l++) {
		OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
			   "Map (HCA/Router) LID: %u base LID: %u lmc:%u to SW LID: %u\n",
			   l, node_lid, lmc, remote_node_lid);

		lid_mapping.m_lid_to_sw_lid_mapping[l] = remote_node_lid;
		lid_mapping.m_lid_to_base_lid_mapping[l] = node_lid;
	}
	OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int AdaptiveRoutingAlgorithm::CalculatePortGroups()
{
	OSM_AR_LOG_ENTER(m_p_osm_log);
	int rc = m_ar_mgr_.m_port_groups_calculator.CalculatePortGroups(this);
	OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void AdaptiveRoutingAlgorithm::PrintRoute(osm_log_level_t level,
					  ARSWDataBaseEntry &src_sw_db_entry,
					  ARSWDataBaseEntry &dst_sw_db_entry,
					  RouteInfo &route_info)
{
	if (!(m_p_osm_log->level & level)) {
		return;
	}

	std::stringstream stream;

	stream << "Route from " << (char *)src_sw_db_entry.
		m_general_sw_info.m_p_osm_sw->p_node->node_desc.description;

	RouteInfo &curr_route_info = route_info;
	ARSWDataBaseEntry *premote_sw_db_entry = NULL;
	ArAlgorithmSwData *dst_sw_data = GetArAlgorithmSwData(dst_sw_db_entry);
	while (true) {

		if (curr_route_info.m_connection != NULL &&
		    curr_route_info.m_connection->m_remote_switch != NULL) {
			premote_sw_db_entry = route_info.m_connection->
				GetRemoteSwitchDbEntry();
		} else {
			break;
		}

		stream << " to " << (char *)premote_sw_db_entry->
			m_general_sw_info.m_p_osm_sw->p_node->node_desc.description;

		ArAlgorithmSwData *p_sw_data = GetArAlgorithmSwData(
			*premote_sw_db_entry);
		curr_route_info = p_sw_data->GetRouteInfo(dst_sw_data);
	}

	if (curr_route_info.m_connection != NULL) {
		stream << " to CA";
	}

	OSM_AR_LOG(m_p_osm_log, level, "%s\n",stream.str().c_str());
}

/*
 * ArAlgorithmSwData
 */
int ArAlgorithmSwData::Init(AdaptiveRoutingAlgorithm &algorithm,
			    uint32_t sw_idx,
			    uint32_t switches_number)
{
	OSM_AR_LOG_ENTER(algorithm.GetLog());

	OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
		   "Init ArAlgorithmSwData of "
		   "switch GUID: 0x%" PRIx64 ", LID: %u sw_idx: %u\n",
		   m_db_entry_.m_general_sw_info.m_guid,
		   m_db_entry_.m_general_sw_info.m_lid,
		   sw_idx);

	osm_physp_t* p_physp;
	osm_node_t *p_remote_node;
	osm_switch_t  *p_osm_sw = m_db_entry_.m_general_sw_info.m_p_osm_sw;
	osm_node_t *p_osm_node = p_osm_sw->p_node;
	SwToConnectionMap connections_map; /* One connection per Neighbor switch */
	m_connections_number_ = 0;

	m_calculate_task_.Init(&m_db_entry_, &algorithm);
	m_sw_idx_ = sw_idx;
	if (switches_number > (uint32_t)m_route_info_.size()) {
		m_route_info_.resize((uint32_t)(switches_number * 1.05));
	}

	/*
	 * Set priv to point to current ARSWDataBaseEntry
	 * This information is used also after routing for SL2VL calculation.
	 */
	p_osm_sw->priv = &(m_db_entry_);

	for (RouteInfoVec::iterator iter = m_route_info_.begin();
		iter != m_route_info_.end(); ++iter) {
		iter->Init();
	}

	m_port_num_to_connection_tbl_.assign(m_port_num_to_connection_tbl_.size(),
					     NULL);
	/* reset m_connection_pool_ */
	int connection_pool_size = (int)m_connection_pool_.size();
	for (int i = 0; i < connection_pool_size; ++i) {
		m_connection_pool_[i].Clear();
	}

	osm_switch_t  *p_remote_sw;
	/* Build connections */
	for (u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ;
		port_num++) {
		p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
		if (!p_physp) {
			continue;
		}
		osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
		if (!p_remote_physp) {
			OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
				   "Unconnected port: %d\n", port_num);
			continue;
		}
		p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
		if (!p_remote_node) {
			OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_ERROR,
				   "ERR 9001: "
				   "-A- Failed to get remote node connected to "
				   "switch GUID: 0x%" PRIx64 ", LID: %u port: %u\n",
				   m_db_entry_.m_general_sw_info.m_guid,
				   m_db_entry_.m_general_sw_info.m_lid,
				   port_num);
			continue;
		}

		u_int8_t remote_node_type = osm_node_get_type(p_remote_node);
		if (remote_node_type != IB_NODE_TYPE_SWITCH) {
			/* Create one connection to all hosts with remote_sw key == NULL */
			p_remote_sw = NULL;
		} else {
			/* remote is switch */
			p_remote_sw = p_remote_node->sw;
		}
		SwToConnectionMapInsertRes insert_res = connections_map.insert(
			SwToConnectionMapEntry(p_remote_sw,
					       m_connection_pool_[m_connections_number_]));
		SwitchConnection &connection = insert_res.first->second;

		if (insert_res.second) { /* new connection was inserted */
			connection.m_remote_switch = p_remote_sw;
			m_connections_number_++;
		}

		m_port_num_to_connection_tbl_[port_num] = &connection;
		connection.m_ports.push_back(p_physp);
		connection.m_ports_bitset.set(port_num);

		OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
			   "Add port:%u to connection to %p "
			   "Total connections number:%u \n",
			   port_num, p_remote_sw,
			   m_connections_number_);
	}
	OSM_AR_LOG_RETURN(algorithm.GetLog(), 0);
}

int ArAlgorithmSwData::SetConnectionsDirection(AdaptiveRoutingAlgorithm &algorithm)
{
	OSM_AR_LOG_ENTER(algorithm.GetLog());

	OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
		   "ArAlgorithmSwData SetConnectionsDirection of "
		   "switch GUID: 0x%" PRIx64 ", LID: %u\n",
		   m_db_entry_.m_general_sw_info.m_guid,
		   m_db_entry_.m_general_sw_info.m_lid);

	osm_switch_t  *p_osm_sw = m_db_entry_.m_general_sw_info.m_p_osm_sw;

	ConnectionVec &connections = GetConnectionsVec();
	int connections_number = GetConnectionsNumber();
	for (int connection_idx = 0; connection_idx < connections_number;
	     ++connection_idx) {

		SwitchConnection &connection = connections[connection_idx];
		if (algorithm.SetDirection(p_osm_sw, connection)) {
			OSM_AR_LOG_RETURN(algorithm.GetLog(), 1);
		}
	}

	OSM_AR_LOG_RETURN(algorithm.GetLog(), 0);
}

void ArAlgorithmSwData::Resize(uint32_t ports_num) {
	m_port_num_to_connection_tbl_.resize(ports_num, NULL);
	m_connection_pool_.resize(ports_num);
}

const SwitchConnection *ArAlgorithmSwData::GetPortConnection(uint8_t port_num) const
{
	if (port_num >= m_port_num_to_connection_tbl_.size()) {
		return NULL;
	}
	return m_port_num_to_connection_tbl_[port_num];
}

/*
 * SwitchConnection
 */

string SwitchConnection::ToString() const
{
	std::stringstream stream;

	if (m_remote_switch == NULL) {

		stream << "Connection to " << (int) m_ports.size() <<
		    " locat CAs";
	} else {

		uint16_t next_hop_lid = osm_node_get_base_lid_ho(
		    m_remote_switch->p_node, 0);

		stream << "Connection to remote switch LID:" << next_hop_lid <<
		    " on dim_index: " << (int)m_dim_idx <<
		    ", dim_sign: " << (int)m_dim_sign <<
		    ", ports: " << m_ports_bitset.to_string();
	}

	return stream.str();
}

/*
 * RouteInfo
 */

/* Select turn type using the following priority: T0, T2, T1 */
uint8_t RouteInfo::turn_type_wait[TURN_TYPE_LAST] = {
	0, /* T0 */
	2, /* T1 */
	1  /* T2 */
};

bool RouteInfo::operator<(const RouteInfo& __rhs) const
{
	if (m_connection == NULL || __rhs.m_connection == NULL) {
		return (m_connection == NULL && __rhs.m_connection != NULL);
	}

	if ((__rhs.m_connection->m_dim_idx == INVALID_DIM_INDEX ||
	     m_connection->m_dim_idx == INVALID_DIM_INDEX) &&
	    m_connection->m_dim_idx != __rhs.m_connection->m_dim_idx) {
		/* Prefer destination with valid coordinate */
		return m_connection->m_dim_idx != INVALID_DIM_INDEX;
	}
	if (m_vl_inc != __rhs.m_vl_inc) {
		return m_vl_inc < __rhs.m_vl_inc;
	}
	if (m_hops_count != __rhs.m_hops_count) {
		return m_hops_count < __rhs.m_hops_count;
	}
	if (m_global_hops_count != __rhs.m_global_hops_count) {
		return m_global_hops_count < __rhs.m_global_hops_count;
	}
	/* We prefer to use non restricted routes */
	if (m_restricted != __rhs.m_restricted) {
		return !m_restricted;
	}
	if (m_turn_type != __rhs.m_turn_type) {
		/* Select by order T0, T2, T1 */
		return turn_type_wait[m_turn_type] < turn_type_wait[__rhs.m_turn_type];
	}
	/* Among top candidate choose one laying in the maximum dimension */
	if (m_connection->m_dim_idx != __rhs.m_connection->m_dim_idx) {
		return m_connection->m_dim_idx > __rhs.m_connection->m_dim_idx;
	}
	/* The following condition is relevant only for KDOR */
	if (m_connection->m_dim_sign != __rhs.m_connection->m_dim_sign) {
		if (m_turn_type == TURN_TYPE_2) {
			/*
			 * This will guarantee that prev hop will not have KDOR TURN_TYPE_1
			 * and therefore will not increase vl
			 */
			return m_connection->m_dim_sign < __rhs.m_connection->m_dim_sign;
		} else {
			/*
			 * This will guarantee that prev hop will not have KDOR TURN_TYPE_2
			 * and therefore will not increase vl
			 */
			return m_connection->m_dim_sign > __rhs.m_connection->m_dim_sign;
		}
	}
	return false;
}

bool RouteInfo::operator==(const RouteInfo& __rhs) const
{
	return ((m_connection->m_dim_idx == __rhs.m_connection->m_dim_idx) &&
		(m_connection->m_dim_sign == __rhs.m_connection->m_dim_sign) &&
		(m_vl_inc == __rhs.m_vl_inc) &&
		(m_turn_type == __rhs.m_turn_type) &&
		(m_hops_count == __rhs.m_hops_count) &&
		(m_global_hops_count == __rhs.m_global_hops_count) &&
		(m_restricted == __rhs.m_restricted));
}

void ArGroupPerIslandData::Init(uint8_t groups_per_leaf,
				uint16_t max_island_number,
				uint16_t group_cap)
{
	if (m_max_ar_group_number > group_cap) {
		Clear();
	}

	if ((max_island_number > m_max_island_number) ||
	    (groups_per_leaf > m_max_groups_per_leaf)) {

		m_max_island_number = max(max_island_number, m_max_island_number);
		m_max_groups_per_leaf = max(groups_per_leaf, m_max_groups_per_leaf);
		m_island_to_group_id.resize(m_max_groups_per_leaf + 1);
		m_leaf_sw_id_to_ar_group.resize(m_max_groups_per_leaf + 1);

		for(int i = 1; i <=  m_max_groups_per_leaf; ++i) {
			m_island_to_group_id[i].resize(m_max_island_number + 1);
			for (int island = 1; island <=  m_max_island_number;
			     ++island) {
				if (m_island_to_group_id[i][island].empty()) {
					m_island_to_group_id[i][island].push_back(0);
				}
			}
		}
	}

	/*
	 * m_island_to_group_id[per_sw_group_index][island_id][0] holds the last
	 * used index in the vector. set to 0 on the beginning of each cycle.
	 */
	for(int i = 1; i <=  m_max_groups_per_leaf; ++i) {
		for (int island = 1; island <=  m_max_island_number; ++island) {
			m_island_to_group_id[i][island][0] = 0;
		}
	}

	m_used_groups_id.resize(group_cap, false);
	m_used_leaf_groups_id.resize(group_cap, false);
	/* Always save groups number allocated for local leafs */
	m_used_groups_id = m_used_leaf_groups_id;
	m_total_used_groups = 1 + m_total_leaves_groups; /* Group 0 is used */
	m_max_groups_per_island = group_cap / max_island_number / groups_per_leaf;
}
