/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_gpu_t.
 * This object represents an Infiniband GPU.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <iba/ib_types.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_GPU_C
#include <opensm/osm_gpu.h>
#include <opensm/osm_lid_mgr.h>
#include <opensm/osm_nvlink.h>
#include <opensm/osm_switch.h>

osm_gpu_t * osm_gpu_new(IN osm_node_t * p_node)
{
	osm_gpu_t *p_gpu;

	p_gpu = calloc(1, sizeof(osm_gpu_t));
	if (!p_gpu)
		return NULL;

	p_gpu->p_node = p_node;

	return p_gpu;
}

void osm_gpu_delete(IN OUT osm_gpu_t ** pp_gpu)
{
	free(*pp_gpu);
	*pp_gpu = NULL;
}

uint16_t osm_gpu_get_used_alid_ho(osm_gpu_t * p_gpu)
{
	return ib_alid_info_get_alid_by_index(&p_gpu->subn_alid_info, 0);
}

uint16_t osm_gpu_get_alid_by_index(osm_gpu_t * p_gpu, uint8_t index)
{
	if (index >= IB_ALID_INFO_ALIDS_PER_BLOCK)
		return 0;

	return ib_alid_info_get_alid_by_index(&p_gpu->subn_alid_info, index);
}

uint8_t osm_gpu_get_num_alids(IN osm_gpu_t * p_gpu)
{
	return p_gpu->p_node->ext_node_info.anycast_lid_table_top;
}

ib_net64_t osm_gpu_get_guid(osm_gpu_t * p_gpu)
{
	return osm_node_get_node_guid(p_gpu->p_node);
}

uint8_t osm_gpu_get_port_plane(osm_gpu_t *p_gpu, uint8_t port_num)
{
	osm_node_t *p_remote_node;
	uint8_t remote_port_num;

	if (port_num < osm_node_get_num_physp(p_gpu->p_node)) {
		p_remote_node = osm_node_get_remote_node(p_gpu->p_node, port_num, &remote_port_num);

		/*
		 * GPUs are currently supported for NVLink configuration,
		 * in which remote node of GPU is always a switch.
		 */
		if (!p_remote_node || !p_remote_node->sw)
			return 0;

		return p_remote_node->sw->nvlink_sw_data.plane;
	}

	return 0;
}
