/*
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2008 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_lid_mgr_t.
 * This file implements the LID Manager object which is responsible for
 * assigning LIDs to all ports on the subnet.
 *
 * DATA STRUCTURES:
 *  p_subn->port_lid_tbl : a vector pointing from lid to its port.
 *  osm db guid2lid domain : a hash from guid to lid (min lid).
 *  p_subn->port_guid_tbl : a map from guid to discovered port obj.
 *
 * ALGORITHM:
 *
 * 0. we define a function to obtain the correct port lid:
 *    lid_mgr_get_port_lid( p_mgr, port, &min_lid ):
 *    0.1 if the port info lid matches the guid2lid return 0
 *    0.2 if the port info has a lid and that range is empty in
 *        port_lid_tbl, return 0 and update the port_lid_tbl and
 *        guid2lid
 *    0.3 else find an empty space in port_lid_tbl, update the
 *    port_lid_tbl and guid2lid, return 1 to flag a change required.
 *
 * 1. During initialization:
 *   1.1 initialize the guid2lid database domain.
 *   1.2 if reassign_lid is not set:
 *   1.2.1 read the persistent data for the domain.
 *   1.2.2 validate no duplicate use of lids and lids are 2^(lmc-1)
 *
 * 2. During SM port lid assignment:
 *   2.1 if reassign_lids is set, make it 2^lmc
 *   2.2 cleanup all port_lid_tbl and re-fill it according to guid2lid
 *   2.3 call lid_mgr_get_port_lid the SM port
 *   2.4 set the port info
 *
 * 3. During all other ports lid assignment:
 *   3.1 go through all ports in the subnet
 *   3.1.1 call lid_mgr_get_port_min_lid
 *   3.1.2 if a change required send the port info
 *   3.2 if any change send the signal PENDING...
 *
 * 4. Store the guid2lid
 *
 *
 *
 * NVLink Assumptions:
 * ALIDs (GPU node LIDs):
 *	Number of ALIDs per GPU is the same for all GPUs of the subnet.
 *	Currently support 1 ALID per GPU.
 * 	However in the future, should support up to 2 ALIDs, that are not sequential.
 * 	Meaning for each ALID DB entry, min_lid is 1st ALID and max_lid is 2nd ALID.
 * 	Hence, currently 2nd ALID shall be 0.
 *
 * GPU port lids:
 * 	GPU LMC is for all GPUs of the subnet.
 *
 * Range validation is done after discovering planes.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_LID_MGR_C
#include <opensm/osm_lid_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_nvlink.h>

static cl_qlist_t * lid_mgr_get_free_ranges_list(osm_lid_mgr_t * p_mgr,
						 osm_lid_mgr_lid_type_t lid_type,
						 uint8_t plane)
{
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;

	OSM_ASSERT(lid_type < OSM_LID_TYPE_MAX);

	switch (lid_type) {
	case OSM_LID_TYPE_LID:
		/* Note planes are indexed 1 to N */
		if (plane > OSM_NVLINK_MAX_PLANES)
			return NULL;
		if (plane)
			return &p_nvl_mgr->p_free_ranges_per_plane[plane];

		return &p_mgr->free_ranges;
	case OSM_LID_TYPE_FLID:
		return &p_mgr->free_flid_ranges;
	case OSM_LID_TYPE_ALID:
		return &p_nvl_mgr->free_alid_ranges;
	default:
		return NULL;
	}

}

static osm_db_domain_t * lid_mgr_get_lid_db(osm_lid_mgr_t *p_mgr,
					    osm_lid_mgr_lid_type_t lid_type,
					    boolean_t gpu_port_lid_db)
{
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;

	OSM_ASSERT(lid_type < OSM_LID_TYPE_MAX);

	switch (lid_type) {
	case OSM_LID_TYPE_LID:
		if (gpu_port_lid_db)
			return p_nvl_mgr->p_g2nvlid;
		return p_mgr->p_g2l;
	case OSM_LID_TYPE_FLID:
		return p_mgr->p_g2fl;
	case OSM_LID_TYPE_ALID:
		return p_nvl_mgr->p_g2alid;
	default:
		return NULL;
	}
}

const char * osm_lid_mgr_lid_type_str(osm_lid_mgr_lid_type_t lid_type,
				      boolean_t gpu_port_lid)
{
	switch (lid_type) {
	case OSM_LID_TYPE_LID:
		return gpu_port_lid ? "nvlid" : "lid";
	case OSM_LID_TYPE_ALID:
		return "alid";
	case OSM_LID_TYPE_FLID:
		return "flid";
	default:
		return "unknown type lid";
	}
}

static uint16_t lid_mgr_nvlink_get_alid_range_len(IN osm_lid_mgr_t * p_mgr)
{
	return OSM_NVLINK_NVLID_RANGE_LEN;
}

/* Referring to GPU port lid */
static uint8_t lid_mgr_nvlink_get_gpu_lmc(IN osm_lid_mgr_t * p_mgr)
{
	return OSM_NVLINK_GPU_LMC;
}

/*
 * Returns the length of LID range for input plane,
 * as for NVLink configuration there is a separate LID range for each plane.
 */
static uint16_t lid_mgr_nvlink_get_plane_lid_range_len(IN osm_lid_mgr_t * p_mgr,
						       IN uint8_t nvlink_plane)
{
	/* Note: planes are indexed from 1 to N */
	return OSM_NVLINK_NVLID_RANGE_LEN;
}

static inline boolean_t lid_mgr_range_limits(IN osm_lid_mgr_t * p_mgr,
					     IN osm_lid_mgr_lid_type_t lid_type,
					     IN uint8_t nvlink_plane,
					     OUT uint16_t * start_lid,
					     OUT uint16_t * end_lid)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;
	uint16_t lid_range_len;

	OSM_ASSERT(lid_type < OSM_LID_TYPE_MAX);

	switch (lid_type) {
	case OSM_LID_TYPE_LID:
		/* Global LIDs range */
		*start_lid = IB_LID_UCAST_START_HO;
		*end_lid = IB_LID_UCAST_END_HO;
		if (!osm_subn_is_nvlink_enabled(p_subn))
			return TRUE;

		/* Note nvlink_planes are indexed 1 to N for NVLink LIDs */
		if (nvlink_plane && nvlink_plane <= OSM_NVLINK_MAX_PLANES) {
			*start_lid = p_nvl_mgr->p_start_ranges_per_plane[nvlink_plane];
			lid_range_len = lid_mgr_nvlink_get_plane_lid_range_len(p_mgr, nvlink_plane);
			*end_lid = (*start_lid) + lid_range_len - 1;
			return TRUE;
		} else if (nvlink_plane > OSM_NVLINK_MAX_PLANES) {
			/* Illegal nvlink_plane number */
			return FALSE;
		}

		/* Fix global LIDs range with NVLink enabled */
		*end_lid = (p_nvl_mgr->alid_range_start - 1);
		return TRUE;
	case OSM_LID_TYPE_FLID:
		*start_lid = p_subn->opt.local_flid_start;
		*end_lid = p_subn->opt.local_flid_end;
		return TRUE;
	case OSM_LID_TYPE_ALID:
		if (!osm_subn_is_nvlink_enabled(p_subn))
			return FALSE;

		*start_lid = p_nvl_mgr->alid_range_start;
		*end_lid = (*start_lid) + lid_mgr_nvlink_get_alid_range_len(p_mgr) - 1;
		return TRUE;
	default:
		return FALSE;
	}
}

static boolean_t lid_mgr_is_alid(IN osm_lid_mgr_t * p_mgr, IN uint16_t lid)
{
	uint16_t range_start, range_end;

	/* check if this LID is in ALID range */
	return lid_mgr_range_limits(p_mgr, OSM_LID_TYPE_ALID, OSM_NVLINK_NO_PLANE,
				    &range_start, &range_end) &&
	    range_start <= lid && lid <= range_end;
}

/* Note this function assumes lids of input guid are continues! */
static boolean_t lid_mgr_nvlink_is_lid_in_type_range(osm_lid_mgr_t * p_mgr,
						     uint64_t guid,
						     osm_lid_mgr_lid_type_t lid_type,
						     uint16_t min_lid,
						     uint16_t max_lid)
{
	osm_port_t *p_port = NULL;
	uint16_t range_start = 0, range_end = 0;
	uint8_t nvlink_plane = OSM_NVLINK_NO_PLANE;

	if (lid_type == OSM_LID_TYPE_ALID)
		goto Range_check;

	p_port = osm_get_port_by_guid(p_mgr->p_subn, cl_hton64(guid));
	if (!p_port) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0302: "
			"failed to find port for guid 0x%"PRIx64"\n",
			guid);
		return FALSE;
	}

	/* Use nvlink_plane on LID range search only for GPUs */
	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn) &&
	    osm_node_is_gpu(p_port->p_node))
		nvlink_plane = osm_port_get_nvlink_plane(p_port);

Range_check:
	if (!lid_mgr_range_limits(p_mgr, lid_type, nvlink_plane, &range_start, &range_end) ||
	    (min_lid < range_start || max_lid < range_start ||
	     min_lid > range_end || max_lid > range_end)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"discovered LID [%u-%u] of %s guid: 0x%016" PRIx64
			" is out of its range [%u-%u]\n",
			min_lid, max_lid, p_port ? "port" : "node",
			guid, range_start, range_end);
		return FALSE;
	}

	return TRUE;
}

/**********************************************************************
  lid range item of qlist
 **********************************************************************/
typedef struct osm_lid_mgr_range {
	cl_list_item_t item;
	uint16_t min_lid;
	uint16_t max_lid;
} osm_lid_mgr_range_t;

void osm_lid_mgr_construct(IN osm_lid_mgr_t * p_mgr)
{
	memset(p_mgr, 0, sizeof(*p_mgr));
}

static void lid_mgr_destroy_free_lid_ranges_list(osm_lid_mgr_t * p_mgr,
						 osm_lid_mgr_lid_type_t lid_type,
						 uint8_t plane)
{
	cl_list_item_t *p_item;
	cl_qlist_t *p_free_lid_ranges = NULL;

	p_free_lid_ranges = lid_mgr_get_free_ranges_list(p_mgr, lid_type, plane);
	if (!p_free_lid_ranges) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0303: "
			"Illegal LID type, cannot free LID range\n");
		return;
	}

	while ((p_item = cl_qlist_remove_head(p_free_lid_ranges)) !=
	       cl_qlist_end(p_free_lid_ranges))
		free((osm_lid_mgr_range_t *) p_item);
}

void osm_lid_mgr_destroy(IN osm_lid_mgr_t * p_mgr)
{
	uint8_t plane;

	OSM_LOG_ENTER(p_mgr->p_log);

	lid_mgr_destroy_free_lid_ranges_list(p_mgr, OSM_LID_TYPE_LID, OSM_NVLINK_NO_PLANE);
	if (osm_subn_is_flids_enabled(p_mgr->p_subn))
		lid_mgr_destroy_free_lid_ranges_list(p_mgr, OSM_LID_TYPE_FLID, OSM_NVLINK_NO_PLANE);
	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn)) {
		lid_mgr_destroy_free_lid_ranges_list(p_mgr, OSM_LID_TYPE_ALID, OSM_NVLINK_NO_PLANE);

		for (plane = 1; plane <= OSM_NVLINK_MAX_PLANES; plane++)
			lid_mgr_destroy_free_lid_ranges_list(p_mgr, OSM_LID_TYPE_LID, plane);
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

static void lid_mgr_mark_global_flids(IN osm_lid_mgr_t * p_mgr)
{
	OSM_LOG_ENTER(p_mgr->p_log);
	osm_subn_t *p_subn = p_mgr->p_subn;
	/* TODO: we currently work with 16 bits and not 20 */
	/*
	 * mark out the whole global range, so that we do not allow ucast lid
	 * to be assigned with lid from FLID range.
	 */
	uint16_t min_lid = p_subn->opt.global_flid_start;
	uint16_t max_lid = p_subn->opt.global_flid_end;
	uint16_t lid;

	/* mark it as global FLID reserved without removing previously marked values */
	for (lid = min_lid; lid <= max_lid; lid++)
		p_mgr->used_lids[lid] |= OSM_RESERVED_GLOBAL_FLID;

	OSM_LOG_EXIT(p_mgr->p_log);
}

/*
 * Special lid refers to new lids that are prioritized over the global unicast lids:
 * FLIDs, ALIDs, NVLIDs (GPU port lids)
 *
 * Returns whether lid is actually used.
 * - All global FLIDs are marked as reserved, for not to assign them to unicast LIDs
 * - All ALIDs and GPU port lids (nvlids) are marked as reserved, so SM avoids assigning
 *   them to unicast LIDs
 * - Assumption: NVLink lids and FLIDs cannot exist on the same configuration!
 */
static inline uint8_t lid_mgr_is_special_lid_used(IN osm_lid_mgr_t * p_mgr, uint16_t lid)
{
	return (p_mgr->used_lids[lid] & OSM_RESERVED_SPECIAL_LID_USED_MASK);
}

/*
 * This function is used to clear any lid as used without compromising the mark
 * of this lid being part of the configured global FLIDs range, and thus
 * can not be assigned for use for unicast LIDs
 */
static inline void lid_mgr_clear_used_lid(IN osm_lid_mgr_t * p_mgr, uint16_t lid)
{
	p_mgr->used_lids[lid] &= (OSM_RESERVED_GLOBAL_FLID | OSM_RESERVED_NVLINK_LID);
}

static void lid_mgr_mark_alids(IN osm_lid_mgr_t * p_mgr)
{
	uint16_t min_lid, max_lid, lid;

	OSM_LOG_ENTER(p_mgr->p_log);
	/* TODO: we currently work with 16 bits and not 20 */
	/*
	 * mark out the whole alid range, so that we do not allow ucast lid
	 * to be assigned with lid from ALID range.
	 */
	if (!lid_mgr_range_limits(p_mgr, OSM_LID_TYPE_ALID, OSM_NVLINK_NO_PLANE,
				  &min_lid, &max_lid)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0304: "
			"Could not mark ALIDs range as used for nvlink.\n");
		goto Exit;
	}

	/* mark it as persistent without removing previously marked values */
	for (lid = min_lid; lid <= max_lid; lid++)
		p_mgr->used_lids[lid] |= OSM_RESERVED_NVLINK_LID;

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

static void lid_mgr_mark_nvlids(IN osm_lid_mgr_t * p_mgr)
{
	uint8_t plane;
	uint16_t min_lid, max_lid, lid;

	OSM_LOG_ENTER(p_mgr->p_log);

	/* TODO: we currently work with 16 bits and not 20 */
	/*
	 * mark out the whole plane lids ranges, so that we do not allow ucast lid
	 * to be assigned with lid from GPU ports LID range.
	 */
	for (plane = 1; plane <= OSM_NVLINK_MAX_PLANES; plane++) {
		if (!lid_mgr_range_limits(p_mgr, OSM_LID_TYPE_LID, plane, &min_lid, &max_lid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0332: "
				"Could not mark plane %u lids range as used for nvlink.\n",
				plane);
			continue;
		}

		/* mark it as persistent without removing previously marked values */
		for (lid = min_lid; lid <= max_lid; lid++)
			p_mgr->used_lids[lid] |= OSM_RESERVED_NVLINK_LID;
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

boolean_t osm_lid_mgr_port_needs_prism_an_lmc(IN const osm_subn_t * p_subn,
					      IN const osm_port_t * p_port)
{
	osm_physp_t *p_physp = osm_port_get_physp(p_port);
	osm_physp_t *p_remote_physp = p_physp->p_remote_physp;

	if (!p_subn->opt.prism_an_lmc)
		return FALSE;

	/* If node isn't CA, it isn't an aggregation node */
	if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_CA)
		return FALSE;

	/* Port of aggregation node should lead to switch */
	if (!p_remote_physp)
		return FALSE;
	if (!p_remote_physp->p_node->sw)
		return FALSE;

	/* Needed to check previous conditions to use this function */
	if (osm_physp_get_special_port_type(p_physp) != IB_SPECIAL_PORT_TYPE_AN)
		return FALSE;

	/* ANs need prism_an_lmc only if they are connected to a prism switch */
	return osm_switch_is_prism_switch(p_remote_physp->p_node->sw, NULL);
}

boolean_t osm_lid_mgr_port_needs_rtr_lmc(IN const osm_subn_t * p_subn,
					 IN const osm_port_t * p_port)
{
	if (!p_subn->opt.rtr_lmc)
		return FALSE;

	return osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_ROUTER;
}

boolean_t osm_lid_mgr_port_needs_global_lmc(IN const osm_subn_t * p_subn,
					    IN const osm_port_t * p_port)
{
	if (!p_subn->opt.lmc)
		return FALSE;

	/* If switch, need to check that it's lmc capable */
	if (!p_port->p_node->sw)
		return TRUE;
	return osm_switch_sp0_is_lmc_capable(p_port->p_node->sw, p_subn);
}

/**********************************************************************
Validate the guid to lid data by making sure that under the current
LMC we did not get duplicates. If we do flag them as errors and remove
the entry.
**********************************************************************/
static void lid_mgr_validate_db(IN osm_lid_mgr_t * p_mgr)
{
	cl_qlist_t guids;
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;
	osm_db_guid_elem_t *p_item;
	uint64_t sm_port_guid;
	uint16_t lid;
	uint16_t min_lid;
	uint16_t max_lid;
	uint16_t lmc_mask;
	boolean_t lids_ok;
	uint8_t lmc_num_lids = (uint8_t) (1 << osm_subn_get_max_lmc(p_mgr->p_subn));
	uint16_t sm_min_lid;
	uint16_t sm_max_lid;

	OSM_LOG_ENTER(p_mgr->p_log);

	lmc_mask = ~(lmc_num_lids - 1);
	sm_port_guid = cl_ntoh64(p_subn->sm_port_guid);

	cl_qlist_init(&guids);

	if (osm_db_guids(p_mgr->p_g2l, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0310: "
			"could not get guid list\n");
		goto Exit;
	}

	if (p_subn->opt.master_sm_lid) {
		sm_min_lid = p_subn->opt.master_sm_lid;
		sm_max_lid = sm_min_lid + (1 << p_mgr->p_subn->opt.lmc) - 1;
	} else {
		sm_min_lid = sm_max_lid = 0;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2lid_get(p_mgr->p_g2l, p_item->guid,
					&min_lid, &max_lid))
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0311: "
				"could not get lid for guid:0x%016" PRIx64 "\n",
				p_item->guid);
		else {
			lids_ok = TRUE;

			if (min_lid > max_lid || min_lid == 0
			    || p_item->guid == 0
			    || max_lid > p_mgr->p_subn->max_ucast_lid_ho
			    || (osm_subn_is_nvlink_enabled(p_mgr->p_subn) &&
				max_lid >= p_nvl_mgr->alid_range_start)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0312: "
					"Illegal LID range [%u:%u] for "
					"guid:0x%016" PRIx64 "\n", min_lid,
					max_lid, p_item->guid);
				lids_ok = FALSE;
			} else if (min_lid != max_lid
				   && (min_lid & lmc_mask) != min_lid) {
				/* check that if the lids define a range that is
				   valid for the current LMC mask */
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0313: "
					"LID range [%u:%u] for guid:0x%016"
					PRIx64
					" is not aligned according to mask:0x%04x\n",
					min_lid, max_lid, p_item->guid,
					lmc_mask);
				lids_ok = FALSE;
			} else if (p_item->guid != sm_port_guid && sm_min_lid &&
				   ((min_lid <= sm_min_lid && sm_min_lid <= max_lid) ||
				    (sm_min_lid <= min_lid && min_lid <= sm_max_lid))) {
				/* ignore entries that will conflict with master SM LID */
				OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
					"LID range [%u:%u] of guid 0x%016" PRIx64
					" conflicts with forced master SM LID %u\n",
					min_lid, max_lid, p_item->guid,
					p_subn->opt.master_sm_lid);
				lids_ok = FALSE;
			} else if (p_subn->opt.master_sm_lid &&
				   p_item->guid == sm_port_guid &&
				   min_lid != p_subn->opt.master_sm_lid) {
				/* ignore entries that will contradict master SM LID */
				OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
					"LID range [%u:%u] of guid 0x%016" PRIx64
					" contradicts with forced master SM LID %u\n",
					min_lid, max_lid, p_item->guid,
					p_subn->opt.master_sm_lid);
				lids_ok = FALSE;
			} else {
				/* check if the lids were not previously assigned */
				for (lid = min_lid; lid <= max_lid; lid++) {
					if (p_mgr->used_lids[lid]) {
						OSM_LOG(p_mgr->p_log,
							OSM_LOG_ERROR,
							"ERR 0314: "
							"0x%04x for guid:0x%016"
							PRIx64
							" was previously used\n",
							lid, p_item->guid);
						lids_ok = FALSE;
					}
				}
			}

			if (lids_ok)
				/* mark that it was visited */
				for (lid = min_lid; lid <= max_lid; lid++) {
					if (lid < min_lid+lmc_num_lids)
						p_mgr->used_lids[lid] |= OSM_RESERVED_LID_PERSISTENT;
				}
			else if (osm_db_guid2lid_delete(p_mgr->p_g2l,
							p_item->guid))
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0315: failed to delete entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
		}		/* got a lid */
		free(p_item);
	}			/* all guids */
Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

/*
 * Validate the guid to flid data by making sure that
 * we did not get duplicates. If we do flag them as errors and remove the entry.
 */
static void lid_mgr_validate_flid_db(IN osm_lid_mgr_t * p_mgr)
{
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint16_t flid;
	boolean_t lids_ok;
	uint8_t lmc_num_lids_sm = (uint8_t) (1 << p_mgr->p_subn->opt.lmc);
	uint16_t sm_min_lid;
	uint16_t sm_max_lid;

	OSM_LOG_ENTER(p_mgr->p_log);

	if (p_mgr->p_subn->opt.master_sm_lid) {
		sm_min_lid = p_mgr->p_subn->opt.master_sm_lid;
		sm_max_lid = sm_min_lid + lmc_num_lids_sm - 1;
	} else {
		sm_min_lid = sm_max_lid = 0;
	}

	cl_qlist_init(&guids);

	if (osm_db_guids(p_mgr->p_g2fl, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 030D: "
			"could not get guid list\n");
		goto Exit;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids)) !=
		(osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2flid_get(p_mgr->p_g2fl, p_item->guid, &flid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 030E: "
				"could not get flid for guid:0x%016" PRIx64 "\n",
				p_item->guid);
		} else {
			lids_ok = TRUE;
			/* check that the FLID is inside the defined local range */
			if (flid < p_mgr->p_subn->opt.local_flid_start ||
			    flid > p_mgr->p_subn->opt.local_flid_end) {
				lids_ok = FALSE;
				OSM_LOG(p_mgr->p_log,
					OSM_LOG_ERROR,
					"ERR 0321: "
					"0x%04x for guid:0x%016"
					PRIx64
					" is not in the FLID local range "
					"[0x%" PRIx16 "-0x%" PRIx16"]\n",
					flid, p_item->guid,
					p_mgr->p_subn->opt.local_flid_start,
					p_mgr->p_subn->opt.local_flid_end);
			}

			if (lids_ok) {
				/* the FLID is ok, mark that is used */
				p_mgr->used_lids[flid] |= OSM_RESERVED_LID_PERSISTENT;
			} else if (osm_db_guid2flid_delete(p_mgr->p_g2fl, p_item->guid)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0320: failed to delete entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
			}
			/* check that the master_sm_lid doesn't conflict with FLIDs from the db */
			if (sm_min_lid && (sm_min_lid <= flid && flid <= sm_max_lid)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0322: the master_sm_lid %u with lmc num lids %u "
					"conflicts with FLID of guid 0x%" PRIx64 " from guid2flid db. "
					"Set master_sm_lid to 0.\n",
					sm_min_lid, lmc_num_lids_sm, p_item->guid);
				p_mgr->p_subn->opt.master_sm_lid = 0;
			}
		}
		free(p_item);
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

boolean_t lid_mgr_validate_is_used_nvlink_lid(IN osm_lid_mgr_t * p_mgr,
					      IN uint64_t guid,
					      IN uint16_t lid,
					      IN osm_lid_mgr_lid_type_t lid_type)
{
	if (lid_mgr_is_special_lid_used(p_mgr, lid)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 033B:"
			" %s 0x%04x for guid:0x%016" PRIx64" was previously used\n",
			osm_lid_mgr_lid_type_str(lid_type, TRUE),
			lid, guid);

		return TRUE;
	}
	return FALSE;
}

/********************************************************************
 * Validate the guid to alid data by making sure that we did not get
 * duplicates. If we do flag them as errors and remove the entry.
 ********************************************************************/
static void lid_mgr_validate_nvlid_db(IN osm_lid_mgr_t * p_mgr,
				      IN osm_lid_mgr_lid_type_t lid_type)
{
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint16_t lid;
	uint16_t min_lid;
	uint16_t max_lid;
	boolean_t lids_ok;
	uint8_t lmc_num_lids = (uint8_t) (1 << lid_mgr_nvlink_get_gpu_lmc(p_mgr));
	uint16_t lmc_mask = ~(lmc_num_lids - 1);
	uint8_t lmc_num_lids_sm = (uint8_t) (1 << p_mgr->p_subn->opt.lmc);
	uint8_t num_lids;
	uint16_t sm_min_lid;
	uint16_t sm_max_lid;
	osm_db_domain_t *guid2lid;
	const char *lid_type_str = osm_lid_mgr_lid_type_str(lid_type, TRUE);
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;

	OSM_LOG_ENTER(p_mgr->p_log);

	if (p_mgr->p_subn->opt.master_sm_lid) {
		sm_min_lid = p_mgr->p_subn->opt.master_sm_lid;
		sm_max_lid = sm_min_lid + lmc_num_lids_sm - 1;
	} else {
		sm_min_lid = sm_max_lid = 0;
	}

	cl_qlist_init(&guids);

	OSM_ASSERT(lid_type == OSM_LID_TYPE_ALID || lid_type == OSM_LID_TYPE_LID);

	guid2lid = lid_mgr_get_lid_db(p_mgr, lid_type, TRUE);
	if (!guid2lid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0328: "
			"Illegal LID type, cannot find LID DB\n");
		goto Exit;
	}

	if (osm_db_guids(guid2lid, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0305: "
			"could not get guid list\n");
		goto Exit;
	}

	/*
	 * ALIDs (GPU node lids): currently support 1 ALID per GPU (max_lid expected to be 0)
	 * 			  ALIDs will not be continuous per GPU. Meaning,
	 * 			  each DB column shall represents an alid.
	 * GPU port lids: GPU LMC is for all GPUs of the subnet.
	 *
	 * Range validation is done after discovering planes.
	 */
	num_lids = (lid_type == OSM_LID_TYPE_LID) ?
	    lmc_num_lids : p_nvl_mgr->num_alids_per_gpu;

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2lid_get(guid2lid, p_item->guid, &min_lid, &max_lid))
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0326: "
				"could not get %s for guid:0x%016" PRIx64 "\n",
				lid_type_str, p_item->guid);
		else {
			lids_ok = TRUE;

			if ((max_lid && min_lid > max_lid) || min_lid == 0 || p_item->guid == 0) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0327: "
					"Illegal %s range [%u:%u] for guid:0x%016" PRIx64 "\n",
					lid_type_str, min_lid, max_lid, p_item->guid);
				lids_ok = FALSE;
			} else if (lid_type == OSM_LID_TYPE_LID && min_lid != max_lid
				   && (min_lid & lmc_mask) != min_lid) {
				/*
				 * Check that the lids define a range that is
				 * valid for the current LMC mask
				 */
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0331: "
					"NVLID range [%u:%u] for guid:0x%016" PRIx64
					" is not aligned according to mask:0x%04x\n",
					min_lid, max_lid, p_item->guid,
					lmc_mask);
				lids_ok = FALSE;
			} else if (lid_type == OSM_LID_TYPE_ALID &&
				   osm_lid_mgr_get_lid_type_by_predefined_range(p_mgr, min_lid) !=
				       OSM_LID_TYPE_ALID) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 033D: "
					"Illegal ALID %u for guid:0x%016" PRIx64 "."
					"Out of ALIDs range [%u-%u].\n",
					min_lid, p_item->guid,
					p_nvl_mgr->alid_range_start,
					p_nvl_mgr->alid_range_start +
					    OSM_NVLINK_NVLID_RANGE_LEN - 1);
				lids_ok = FALSE;
			} else {
				/*
				 * Check if GPU lids (ALIDs or port LIDs) were
				 * not previously assigned
				 */
				lids_ok = !lid_mgr_validate_is_used_nvlink_lid(
				    p_mgr, p_item->guid, min_lid, lid_type);

				/*
				 * As explained under 'NVLink Assumptions' above,
				 * support a single ALID.
				 */
				if (lid_type == OSM_LID_TYPE_ALID)
					max_lid = min_lid;
			}

			if (lids_ok) {
				/* mark that it was visited */
				for (lid = min_lid; lid <= max_lid; lid++) {
					if (lid < min_lid + num_lids)
						p_mgr->used_lids[lid] |= OSM_RESERVED_LID_PERSISTENT;
				}
			} else if (osm_db_guid2lid_delete(guid2lid,
							  p_item->guid)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0329: failed to delete entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
			}

			/* check that the master_sm_lid doesn't conflict with LIDs from the db */
			if (sm_min_lid &&
			    ((sm_min_lid <= min_lid && min_lid <= sm_max_lid) ||
			     (sm_min_lid <= max_lid && max_lid <= sm_max_lid))) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 032A: the master_sm_lid %u with lmc num lids %u "
					"conflicts with %s of guid 0x%" PRIx64 " from "
					" guid2%s db. Set master_sm_lid to 0.\n",
					sm_min_lid, lmc_num_lids_sm,
					lid_type_str, p_item->guid, lid_type_str);
				p_mgr->p_subn->opt.master_sm_lid = 0;
			}
		}		/* got a lid */
		free(p_item);
	}			/* all guids */
Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

static uint8_t lid_mgr_is_lid_used(osm_lid_mgr_t * p_mgr,
				   osm_lid_mgr_lid_type_t lid_type,
				   uint8_t plane,
				   uint16_t lid)
{
	if (lid_type == OSM_LID_TYPE_LID && plane == OSM_NVLINK_NO_PLANE)
		return p_mgr->used_lids[lid];

	return lid_mgr_is_special_lid_used(p_mgr, lid);
}

static ib_api_status_t lid_mgr_reset_lid_free_range(IN osm_lid_mgr_t * p_mgr,
						    IN osm_lid_mgr_lid_type_t lid_type,
						    IN uint8_t plane)
{
	uint16_t lid;
	uint16_t max_lid, start_lid = 0xFFFF;
	osm_lid_mgr_range_t *p_range = NULL;
	cl_qlist_t *p_free_lid_ranges = lid_mgr_get_free_ranges_list(p_mgr, lid_type, plane);
	const char *lid_type_str = osm_lid_mgr_lid_type_str(lid_type, plane ? TRUE : FALSE);

	/* we need to cleanup the empty ranges list */
	lid_mgr_destroy_free_lid_ranges_list(p_mgr, lid_type, plane);

	if (!lid_mgr_range_limits(p_mgr, lid_type, plane, &start_lid, &max_lid)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 032B: "
			"Could not get %s ranges to clear\n",
			lid_type_str);
		return IB_ERROR;
	}

	/*
	 * For all types of lids, last range is to be up to the max allowed lid.
	 * max_ucast_lid_ho refers to the minimal max lid supported by all subnet switches,
	 * meaning its the upper limit of LFTs entries.
	 */
	if (p_mgr->p_subn->max_ucast_lid_ho < max_lid)
		max_lid = p_mgr->p_subn->max_ucast_lid_ho;

	/*
	 * For global lids,
	 * If we are in the first sweep and in reassign lids mode
	 * we should ignore all the available info and simply define one
	 * huge empty range
	 */
	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE) {
		p_range = calloc(1, sizeof(osm_lid_mgr_range_t));
		if (p_range)
			p_range->min_lid = start_lid;
	}

	/*
	 * Construct free LID ranges according to p_mgr->used_lids.
	 */
	for (lid = start_lid; lid <= max_lid; lid++) {
		if (!lid_mgr_is_lid_used(p_mgr, lid_type, plane, lid)) {
			if (p_range)
				p_range->max_lid = lid;
			else {
				p_range = calloc(1, sizeof(osm_lid_mgr_range_t));
				if (p_range) {
					p_range->min_lid = lid;
					p_range->max_lid = lid;
				}
			}
		} else {
			/* this lid is used so we need to finalize the previous free range */
			if (p_range) {
				cl_qlist_insert_tail(p_free_lid_ranges, &p_range->item);
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"new free %s range [%u:%u]\n",
					lid_type_str, p_range->min_lid, p_range->max_lid);
				p_range = NULL;
			}
		}
	}

	if (p_range) {
		p_range->max_lid = max_lid;
		cl_qlist_insert_tail(p_free_lid_ranges, &p_range->item);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"final free %s range [%u:%u]\n",
			lid_type_str, p_range->min_lid, p_range->max_lid);
	}

	return IB_SUCCESS;
}

ib_api_status_t osm_lid_mgr_init_alid_db(IN osm_lid_mgr_t * p_mgr)
{
	ib_api_status_t status = IB_SUCCESS;
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_nvl_mgr->p_g2alid = osm_db_domain_init(p_mgr->p_db, "guid2alid");
	if (!p_nvl_mgr->p_g2alid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 030C: "
			"Error initializing Guid-to-ALID persistent database\n");
		status = IB_ERROR;
		goto Exit;
	}

	cl_qlist_init(&p_nvl_mgr->free_alid_ranges);

	if (!p_mgr->p_subn->opt.reassign_lids) {
		if (osm_db_restore(p_nvl_mgr->p_g2alid)) {
#ifndef __WIN__
			/*
			 * When Windows is BSODing, it might corrupt files that
			 * were previously opened for writing, even if the files
			 * are closed, so we might see corrupted guid2alid file.
			 */
			if (p_mgr->p_subn->opt.exit_on_fatal) {
				osm_log_v2(p_mgr->p_log, OSM_LOG_SYS, FILE_ID,
					   "FATAL: Error restoring Guid-to-ALID "
					   "persistent database\n");
				status = IB_ERROR;
				goto Exit;
			} else
#endif
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0325: Error restoring Guid-to-ALID "
					"persistent database\n");
		}

		/* validate g2alid (gpu node lids) */
		lid_mgr_validate_nvlid_db(p_mgr, OSM_LID_TYPE_ALID);
	}

	/* update the free ALID range */
	lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_ALID, OSM_NVLINK_NO_PLANE);

	/*
	 * Mark alids in order to avoid using them for unicast lids.
	 * Restoring of the guid 2 alid db, may have marked
	 * only part of the range.
	 */
	lid_mgr_mark_alids(p_mgr);

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

ib_api_status_t osm_lid_mgr_init_nvlid_db(IN osm_lid_mgr_t * p_mgr)
{
	osm_sm_t *p_sm = p_mgr->sm;
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;
	ib_api_status_t status = IB_SUCCESS;
	uint8_t plane;
	cl_qlist_t *p_plane_range;

	OSM_LOG_ENTER(p_sm->p_log);

	p_nvl_mgr->p_g2nvlid = osm_db_domain_init(p_mgr->p_db, "guid2nvlid");
	if (!p_nvl_mgr->p_g2nvlid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0333: "
			"Error initializing Guid-to-NVLID persistent database\n");
		status = IB_ERROR;
		goto Exit;
	}

	for (plane = 1; plane <= OSM_NVLINK_MAX_PLANES; plane++) {
		p_plane_range = lid_mgr_get_free_ranges_list(p_mgr, OSM_LID_TYPE_LID, plane);
		cl_qlist_init(p_plane_range);
	}

	if (!p_mgr->p_subn->opt.reassign_lids) {
		if (osm_db_restore(p_nvl_mgr->p_g2nvlid)) {
#ifndef __WIN__
			/*
			 * When Windows is BSODing, it might corrupt files that
			 * were previously opened for writing, even if the files
			 * are closed, so we might see corrupted guid2nvlid file.
			 */
			if (p_mgr->p_subn->opt.exit_on_fatal) {
				osm_log_v2(p_mgr->p_log, OSM_LOG_SYS, FILE_ID,
					   "FATAL: Error restoring Guid-to-NVLID "
					   "persistent database\n");
				status = IB_ERROR;
				goto Exit;
			} else
#endif
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0334: Error restoring Guid-to-NVLID "
					"persistent database\n");
		}

		/*
		 * At this point we're yet to have GPU ALIDCap (received from ExtendedNodeInfo)
		 * validate by ALIDTop, assuming it is identical for all GPUs.
		 */
		/* Validate g2nvlid (gpu port lids) */
		lid_mgr_validate_nvlid_db(p_mgr, OSM_LID_TYPE_LID);
	}

	/* update the free NVLID range */
	for (plane = 1; plane <= OSM_NVLINK_MAX_PLANES; plane++)
		lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_LID, plane);

	/*
	 * Mark alids in order to avoid using them for unicast lids.
	 * Restoring of the guid2nvlid db, may have marked
	 * only part of the range.
	 */
	lid_mgr_mark_nvlids(p_mgr);


Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

static void osm_lid_mgr_nvlink_validate_sm_lid(IN osm_lid_mgr_t * p_mgr)
{
	uint16_t sm_min_lid, sm_max_lid;
	uint16_t start_range = 0, end_range = 0;
	uint8_t lmc_num_lids = (uint8_t) (1 << p_mgr->p_subn->opt.lmc);

	/* Check forced SM LID is on global LID range */
	if (!p_mgr->p_subn->opt.master_sm_lid)
		return;

	sm_min_lid = p_mgr->p_subn->opt.master_sm_lid;
	sm_max_lid = sm_min_lid + lmc_num_lids - 1;

	if (!lid_mgr_range_limits(p_mgr, OSM_LID_TYPE_LID, OSM_NVLINK_NO_PLANE,
				  &start_range, &end_range)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0335: "
			"Could not validate whether master_sm_lid is on correct lid range.\n");

	} else if (!(sm_min_lid >= start_range && sm_max_lid <= end_range)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0336: "
			"master_sm_lid %u with number of lids %u "
			"is out of global lid range [%u-%u]. "
			"Setting master_sm_lid to 0.\n",
			sm_min_lid, lmc_num_lids, start_range, end_range);
		p_mgr->p_subn->opt.master_sm_lid = 0;
	}
}

ib_api_status_t osm_lid_mgr_init(IN osm_lid_mgr_t * p_mgr, IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(sm->p_log);

	osm_lid_mgr_construct(p_mgr);

	p_mgr->sm = sm;
	p_mgr->p_log = sm->p_log;
	p_mgr->p_subn = sm->p_subn;
	p_mgr->p_db = sm->p_db;
	p_mgr->p_lock = sm->p_lock;

	/* we initialize and restore the db domain of guid to lid map */
	p_mgr->p_g2l = osm_db_domain_init(p_mgr->p_db, "guid2lid");
	if (!p_mgr->p_g2l) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0316: "
			"Error initializing Guid-to-Lid persistent database\n");
		status = IB_ERROR;
		goto Exit;
	}

	cl_qlist_init(&p_mgr->free_ranges);

	if (osm_subn_is_flids_enabled(p_mgr->p_subn)) {
		p_mgr->p_g2fl = osm_db_domain_init(p_mgr->p_db, "guid2flid");
		if (!p_mgr->p_g2fl) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 030A: "
					"Error initializing Guid-to-FLID persistent database\n");
			status = IB_ERROR;
			goto Exit;
		}

		cl_qlist_init(&p_mgr->free_flid_ranges);

		if (osm_db_restore(p_mgr->p_g2fl)) {
#ifndef __WIN__
			/*
			 * When Windows is BSODing, it might corrupt files that
			 * were previously opened for writing, even if the files
			 * are closed, so we might see corrupted guid2lid file.
			 */
			if (p_mgr->p_subn->opt.exit_on_fatal) {
				osm_log_v2(p_mgr->p_log, OSM_LOG_SYS, FILE_ID,
						"FATAL: Error restoring Guid-to-FLID "
						"persistent database\n");
				status = IB_ERROR;
				goto Exit;
			} else
#endif
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
						"ERR 030B: Error restoring Guid-to-FLID "
						"persistent database\n");
		}

		/* validate g2fl */
		lid_mgr_validate_flid_db(p_mgr);

		/* update the free FLID range */
		lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_FLID, OSM_NVLINK_NO_PLANE);

		/*
		 * Mark global flids in order to avoid using them for unicast lids.
		 * Restoring of the guid 2 flid db, may have marked
		 * only part of the global range.
		 */
		lid_mgr_mark_global_flids(p_mgr);
	}

	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn)) {
		osm_lid_mgr_nvlink_validate_sm_lid(p_mgr);

		status = osm_lid_mgr_init_alid_db(p_mgr);
		if (status)
			goto Exit;

		status = osm_lid_mgr_init_nvlid_db(p_mgr);
		if (status)
			goto Exit;
	}

	/* we use the stored guid to lid table if not forced to reassign */
	if (!p_mgr->p_subn->opt.reassign_lids) {
		if (osm_db_restore(p_mgr->p_g2l)) {
#ifndef __WIN__
			/*
			 * When Windows is BSODing, it might corrupt files that
			 * were previously opened for writing, even if the files
			 * are closed, so we might see corrupted guid2lid file.
			 */
			if (p_mgr->p_subn->opt.exit_on_fatal) {
				osm_log_v2(p_mgr->p_log, OSM_LOG_SYS, FILE_ID,
					   "FATAL: Error restoring Guid-to-Lid "
					   "persistent database\n");
				status = IB_ERROR;
				goto Exit;
			} else
#endif
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0317: Error restoring Guid-to-Lid "
					"persistent database\n");
		}

		/* we need to make sure we did not get duplicates with
		   current lmc */
		lid_mgr_validate_db(p_mgr);
	}


	osm_vlid_mgr_init(&p_mgr->vlid_mgr);

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

static uint16_t trim_lid(IN uint16_t lid)
{
	if (lid > IB_LID_UCAST_END_HO || lid < IB_LID_UCAST_START_HO)
		return 0;
	return lid;
}

static void lid_mgr_reset_used_lids(IN osm_lid_mgr_t * p_mgr)
{
	memset(p_mgr->used_lids, 0, sizeof(p_mgr->used_lids));
	/* make sure that the global flid range is set */
	lid_mgr_mark_global_flids(p_mgr);
	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn)) {
		lid_mgr_mark_alids(p_mgr);
		lid_mgr_mark_nvlids(p_mgr);
	}
}

/******************************************************************************
 * Recover switch FLID from discovered FLIDs of CAs connected to this switch.
 * If no connected CAs found, return 0.
 * Verify GUID does not already have FLID assigned in guid2flid.
 * Verify no conflicts with other CAs.
 * If CAs' FLIDs do not match or are out of local FLID range return 0.
 * If none of the discovered FLIDs are valid and available return 0.
******************************************************************************/
static uint16_t lid_mgr_recover_switch_flid(IN osm_lid_mgr_t *p_mgr, IN osm_switch_t *p_sw)
{
	osm_node_t *p_node = p_sw->p_node;
	osm_port_t *p_remote_port, *p_sw_port0;
	osm_physp_t *p_remote_physp, *p_physp;
	uint16_t discovered_port_flid, determined_flid = 0;
	uint64_t sw_guid, remote_guid;
	int i;

	p_sw_port0 = osm_switch_get_port0(p_sw, p_mgr->p_subn);
	sw_guid = osm_port_get_guid_ho(p_sw_port0);

	/* If switch guid is already mapped to an FLID return that FLID */
	if (!osm_db_guid2flid_get(p_mgr->p_g2fl, sw_guid, &determined_flid))
		return determined_flid;

	for (i = 1; i < osm_node_get_num_physp(p_node); i++) {
		p_physp = &p_node->physp_table[i];

		if (!p_physp || !osm_physp_has_any_link(p_physp))
			continue;

		p_remote_physp = osm_physp_get_remote(p_physp);

		/* check only CAs in active state */
		if (osm_node_get_type(p_remote_physp->p_node) != IB_NODE_TYPE_CA ||
		    osm_physp_get_port_state(p_physp) != IB_LINK_ACTIVE)
			continue;

		/* Skip special ports */
		if (osm_physp_get_special_port_type(p_remote_physp))
			continue;

		remote_guid = osm_physp_get_port_guid(p_remote_physp);
		p_remote_port = osm_get_port_by_guid(p_mgr->p_subn, remote_guid);
		if (!p_remote_port) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"guid: 0x%016" PRIx64 " is not found in port_guid_tbl\n",
				remote_guid);
			return 0;
		}

		discovered_port_flid = cl_ntoh16(
				ib_port_info_get_port_flid(&p_remote_physp->port_info));

		/* If discovered_port_flid is 0 consider as though no FLID was assigned */
		if (discovered_port_flid == 0)
			continue;

		/* check that the FLID is inside the defined local range */
		if (discovered_port_flid < p_mgr->p_subn->opt.local_flid_start ||
		    discovered_port_flid > p_mgr->p_subn->opt.local_flid_end) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Unable to recover FLID for guid: 0x%016" PRIx64
				" Discovered FLID 0x%04x is out of local range "
				"[0x%" PRIx16 "-0x%" PRIx16 "]\n",
				sw_guid, discovered_port_flid,
				p_mgr->p_subn->opt.local_flid_start,
				p_mgr->p_subn->opt.local_flid_end);
			return 0;
		}

		/* won't check if this FLID was previously assigned to different switch */

		/* check that this FLID matches the other CAs connected to this switch */
		else if (determined_flid && discovered_port_flid != determined_flid) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Unable to recover FLID for guid:0x%016"PRIx64
				" Discovered FLID 0x%04x does not match other FLIDs in this switch\n",
				sw_guid, discovered_port_flid);
			return 0;
		}

		else {
			determined_flid = discovered_port_flid;
		}
	}

	return determined_flid;
}

/**********************************************************************
 * Recover discovered FLIDs from the network.
**********************************************************************/
static void lid_mgr_recover_flids(IN osm_lid_mgr_t *p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_switch_t *p_sw;
	osm_port_t *p_sw_port0;
	uint64_t sw_guid;
	uint16_t recovered_flid;

	for (p_sw = (osm_switch_t *) cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_sw != (osm_switch_t *) cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_sw = (osm_switch_t *) cl_qmap_next(&p_sw->map_item)) {
		if (!osm_switch_has_cas(p_sw))
			continue;
		p_sw_port0 = osm_switch_get_port0(p_sw, p_subn);
		sw_guid = osm_port_get_guid_ho(p_sw_port0);
		recovered_flid = lid_mgr_recover_switch_flid(p_mgr, p_sw);

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Discovered FLID 0x%04x for guid: 0x%016" PRIx64 "\n",
			recovered_flid, sw_guid);

		if (recovered_flid) {
			p_sw_port0->flid = cl_hton16(recovered_flid);
			p_mgr->used_lids[recovered_flid] |=
				(OSM_RESERVED_LID_PERSISTENT | OSM_RESERVED_LID_DISCOVERED);
			osm_lid_mgr_set_guid_flid(p_mgr, sw_guid, recovered_flid);
		}
	}
}

static void lid_mgr_discover_alids(osm_lid_mgr_t * p_mgr)
{
	osm_gpu_t *p_gpu;
	cl_qmap_t *p_gpu_guid_tbl = &p_mgr->p_subn->gpu_guid_tbl;
	uint16_t disc_alid, db_min_lid, db_max_lid;
	osm_db_domain_t *guid2lid;
	uint64_t gpu_guid;

	OSM_LOG_ENTER(p_mgr->p_log);

	guid2lid = lid_mgr_get_lid_db(p_mgr, OSM_LID_TYPE_ALID, FALSE);
	if (!guid2lid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 032D: "
			"Could not find ALID DB\n");
		return;
	}

	for (p_gpu = (osm_gpu_t *) cl_qmap_head(p_gpu_guid_tbl);
	     p_gpu != (osm_gpu_t *) cl_qmap_end(p_gpu_guid_tbl);
	     p_gpu = (osm_gpu_t *) cl_qmap_next(&p_gpu->map_item)) {

		gpu_guid = cl_ntoh64(osm_gpu_get_guid(p_gpu));
		if (osm_db_guid2lid_get(guid2lid, gpu_guid,
					&db_min_lid, &db_max_lid)) {
			/* Could not find cached ALID for this GPU guid */

			/* Currently supports a single alid per GPU */
			disc_alid = osm_gpu_get_alid_by_index(p_gpu, 0);
			if (!disc_alid)
				continue;

			/*
			 * Check discovered ALID is on ALIDs range. Otherwise, reassigned ALID.
			 * Currently support a single alid per GPU.
			 * Once need to support multiple alids, check each alid entry accordingly
			 */
			if (!lid_mgr_nvlink_is_lid_in_type_range(p_mgr, gpu_guid, OSM_LID_TYPE_ALID,
								disc_alid, disc_alid))
				continue;

			/*
			 * Validate ALIDInfo ALIDs are free to use.
			 * ALIDs are not sequential, hence validate with range length 1.
			 */
			if (lid_mgr_is_special_lid_used(p_mgr, disc_alid)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"0x%016" PRIx64 " existing alid %u is not free\n",
					gpu_guid, disc_alid);
				continue;
			}

			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"0x%016" PRIx64 " alid %u is free\n",
				gpu_guid, disc_alid);

			/* Mark ALID as used (currently support a single alid per GPU) */
			p_mgr->used_lids[disc_alid] |= OSM_RESERVED_LID_DISCOVERED;
		}
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

/**********************************************************************
 initialize the manager for a new sweep:
 scans the known persistent assignment and port_lid_tbl
 re-calculate all empty ranges.
 cleanup invalid port_lid_tbl entries
**********************************************************************/
static int lid_mgr_init_sweep(IN osm_lid_mgr_t * p_mgr)
{
	cl_ptr_vector_t *p_discovered_vec = &p_mgr->p_subn->port_lid_tbl;
	uint16_t disc_min_lid, disc_max_lid, db_min_lid, db_max_lid;
	int status = 0;
	boolean_t is_free, is_disc_match_range, is_cached_match_range, is_gpu, recover_flids = TRUE;
	osm_port_t *p_port;
	osm_vport_t *p_vport;
	cl_qmap_t *p_port_guid_tbl;
	uint8_t lmc_num_lids = (uint8_t) (1 << osm_subn_get_max_lmc(p_mgr->p_subn));
	uint8_t lmc_num_lids_gpu = (uint8_t) (1 << lid_mgr_nvlink_get_gpu_lmc(p_mgr));
	uint16_t lmc_mask, lmc_mask_gpu, num_lids, lid, lmc_mask_cur;
	uint16_t index, top_index;
	ib_net64_t sm_port_guid = p_mgr->p_subn->sm_port_guid;
	uint16_t sm_min_lid;
	uint16_t sm_max_lid;
	uint16_t global_flid_start = p_mgr->p_subn->opt.global_flid_start;
	uint16_t global_flid_end = p_mgr->p_subn->opt.global_flid_end;
	uint8_t plane;
	osm_db_domain_t *guid2lid;
	uint64_t port_guid;
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;

	OSM_LOG_ENTER(p_mgr->p_log);
	lmc_mask = ~(lmc_num_lids - 1);
	lmc_mask_gpu = ~(lmc_num_lids_gpu - 1);

	if (p_mgr->p_subn->opt.master_sm_lid) {
		sm_min_lid = p_mgr->p_subn->opt.master_sm_lid;
		sm_max_lid = sm_min_lid + (1 << p_mgr->p_subn->opt.lmc) - 1;
	} else {
		sm_min_lid = sm_max_lid = 0;
	}


	/* We must discard previous guid2lid db if this is the first master
	 * sweep and reassign_lids option is TRUE.
	 * If we came out of standby and honor_guid2lid_file option is TRUE, we
	 * must restore guid2lid db. Otherwise if honor_guid2lid_file option is
	 * FALSE we must discard previous guid2lid db.
	 */
	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE) {
		osm_db_clear(p_mgr->p_g2l);
		lid_mgr_reset_used_lids(p_mgr);

		/* reset FLIDs on reassign lids */
		if (osm_subn_is_flids_enabled(p_mgr->p_subn)) {
			osm_db_clear(p_mgr->p_g2fl);
			recover_flids = FALSE;
		}

		/* ALIDs and NVLIDs (GPU port lids) are enabled */
		if (osm_subn_is_nvlink_enabled(p_mgr->p_subn)) {
			/* GPU node lids */
			osm_db_clear(p_nvl_mgr->p_g2alid);

			/* GPU port lids */
			osm_db_clear(p_nvl_mgr->p_g2nvlid);
		}
	} else if (p_mgr->p_subn->coming_out_of_standby == TRUE) {
		osm_db_clear(p_mgr->p_g2l);
		lid_mgr_reset_used_lids(p_mgr);

		/*
		 * If FLIDs are enabled clear guid2flids, and don't use guid2flid file
		 * to give network precedence.
		 */
		if (osm_subn_is_flids_enabled(p_mgr->p_subn))
			osm_db_clear(p_mgr->p_g2fl);

		/* ALIDs and NVLIDs (GPU port lids) are enabled */
		if (osm_subn_is_nvlink_enabled(p_mgr->p_subn)) {
			/* GPU node lids */
			osm_db_clear(p_nvl_mgr->p_g2alid);

			/* GPU port lids */
			osm_db_clear(p_nvl_mgr->p_g2nvlid);
		}

		if (p_mgr->p_subn->opt.honor_guid2lid_file == FALSE)
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Ignore guid2lid file when coming out of standby\n");
		else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Honor current guid2lid file when coming out "
				"of standby\n");
			if (osm_db_restore(p_mgr->p_g2l))
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0306: "
					"Error restoring Guid-to-Lid "
					"persistent database. Ignoring it\n");
			lid_mgr_validate_db(p_mgr);
		}
	} else if (p_mgr->p_subn->sm_port_changed == TRUE) {
		lid_mgr_reset_used_lids(p_mgr);
		lid_mgr_validate_db(p_mgr);
	} else {
		/* clear "discovered LID" bit form p_mgr->used_lids vector */
		for (lid = 1; lid <= IB_LID_UCAST_END_HO; lid++)
			p_mgr->used_lids[lid] &= ~OSM_RESERVED_LID_DISCOVERED;
	}
	/*
	 * Recover FLIDs from network.
	 * Clear FLID free ranges and build them based on used_lids again.
	 */
	if (osm_subn_is_flids_enabled(p_mgr->p_subn)) {
		if (recover_flids)
			lid_mgr_recover_flids(p_mgr);

		lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_FLID, OSM_NVLINK_NO_PLANE);
	}

	/* we need to cleanup the empty ranges list */
	lid_mgr_destroy_free_lid_ranges_list(p_mgr, OSM_LID_TYPE_LID, OSM_NVLINK_NO_PLANE);

	/* first clean up the port_by_lid_tbl */
	for (lid = 0; lid < cl_ptr_vector_get_size(p_discovered_vec); lid++)
		cl_ptr_vector_set(p_discovered_vec, lid, NULL);

	/* we if are in the first sweep and in reassign lids mode
	   we should ignore all the available info and simply define one
	   huge empty range (see range set on lid_mgr_reset_lid_free_range) */
	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Skipping all lids as we are reassigning them\n");
		goto AfterScanningLids;
	}

	/* go over all discovered ports and mark their entries */
	p_port_guid_tbl = &p_mgr->p_subn->port_guid_tbl;

	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		osm_port_get_lid_range_ho(p_port, &disc_min_lid, &disc_max_lid);
		disc_min_lid = trim_lid(disc_min_lid);

		/* On this loop we validate only the global lids (g2lid) */
		if (osm_subn_is_nvlink_enabled(p_mgr->p_subn) && osm_node_is_gpu(p_port->p_node)) {
			num_lids = lmc_num_lids_gpu;
			lmc_mask_cur = lmc_mask_gpu;
		} else if (osm_lid_mgr_port_needs_global_lmc(p_mgr->p_subn, p_port) ||
			   osm_lid_mgr_port_needs_rtr_lmc(p_mgr->p_subn, p_port) ||
			   osm_lid_mgr_port_needs_prism_an_lmc(p_mgr->p_subn, p_port)) {
			num_lids = lmc_num_lids;
			lmc_mask_cur = lmc_mask;
		} else {
			num_lids = 1;
			lmc_mask_cur = lmc_mask;
		}

		disc_max_lid = disc_min_lid + num_lids - 1;

		/*
		 * If forcing SM LID, and found port with conflicting LID, dont
		 * update it in LID-to-port table.
		 */
		if (sm_min_lid &&
		    osm_port_get_guid(p_port) != sm_port_guid &&
		    ((disc_min_lid <= sm_min_lid && sm_min_lid <= disc_max_lid) ||
		     (sm_min_lid <= disc_min_lid && disc_min_lid <= sm_max_lid)))
			continue;

		if (sm_min_lid && osm_port_get_guid(p_port) == sm_port_guid)
			continue;

		/*
		 * If the FLID feature is enabled, and found port with conflicting LID, don't
		 * update it in LID-to-port table.
		 */
		if (osm_subn_is_flids_enabled(p_mgr->p_subn) &&
		    ((disc_min_lid <= global_flid_start &&
		       global_flid_start <= disc_max_lid) ||
		     (global_flid_start <= disc_min_lid &&
		       disc_min_lid <= global_flid_end))) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"discovered LID (%u-%u) of port guid: 0x%016" PRIx64
				" conflicts with global FLID range (%u-%u)\n",
				disc_min_lid, disc_max_lid,
				osm_port_get_guid_ho(p_port),
				global_flid_start,
				global_flid_end);
			continue;
		}

		/*
		 * If NVLink is enabled, LID ranges are in the following order:
		 * global LID < ALIDs < plane 1 LIDs < ... < plane N LIDs
		 * Hence, if port LID found no to be part of the corresponding LID type
		 * range, don't update it in LID-to-port table.
		 */
		is_disc_match_range = TRUE;
		if (osm_subn_is_nvlink_enabled(p_mgr->p_subn) &&
		    !lid_mgr_nvlink_is_lid_in_type_range(p_mgr, osm_port_get_guid_ho(p_port),
							 OSM_LID_TYPE_LID,
							 disc_min_lid, disc_max_lid))
			is_disc_match_range = FALSE;

		if (is_disc_match_range && disc_min_lid && ((num_lids == 1) ||
		    (disc_min_lid & lmc_mask_cur) == disc_min_lid)) {
			for (lid = disc_min_lid; lid <= disc_max_lid; lid++) {
				cl_ptr_vector_set(p_discovered_vec, lid, p_port);
				p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;
			}
		}

		/*
		 * If virtual port has LID, but does not have persistent LID,
		 * mark current lids as used.
		 */
		top_index = osm_port_get_vport_top_index(p_port);
		for (index = 1 ; index <= top_index ; index++) {
			p_vport = osm_port_get_vport(p_port, index);
			if (p_vport &&
			    (lid = trim_lid(p_vport->vport_info.vport_lid)) &&
			    osm_db_guid2lid_get(p_mgr->p_g2l,
						cl_ntoh64(osm_vport_get_guid(p_vport)),
						NULL, NULL)) {
				p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;
			}
		}

		/*
		 * Check if port has guid2lid entry. If not, check that currently
		 * used LID is free and matches LMC, and set p_mgr->used_lids
		 * accordingly.
		 */
		is_gpu = osm_subn_is_nvlink_enabled(p_mgr->p_subn) && osm_node_is_gpu(p_port->p_node);
		guid2lid = lid_mgr_get_lid_db(p_mgr, OSM_LID_TYPE_LID, is_gpu);
		if (!guid2lid)
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0337: "
				"Illegal LID type, cannot find LID DB\n");

		port_guid = osm_port_get_guid_ho(p_port);
		if (osm_db_guid2lid_get(guid2lid, port_guid,
					&db_min_lid, &db_max_lid)) {
			if (!disc_min_lid || disc_max_lid > IB_LID_UCAST_END_HO)
				continue;

			/* Check discovered LID is on correct lid type range */
			if (!lid_mgr_nvlink_is_lid_in_type_range(p_mgr, port_guid, OSM_LID_TYPE_LID,
								 disc_min_lid, disc_max_lid))
				continue;

			/* Single LID, just mark LID as used */
			if (num_lids == 1) {
				p_mgr->used_lids[disc_min_lid] |= OSM_RESERVED_LID_DISCOVERED;
				continue;
			}

			/* Check if base LID matches LMC */
			if ((disc_min_lid & lmc_mask_cur) != disc_min_lid)
				continue;

			/*
			 * Check that LIDs range is free.
			 * If free, it can be used by current device, hence, mark as used.
			 */
			is_free = TRUE;
			for (lid = disc_min_lid; lid <= disc_max_lid; lid++) {
				if (p_mgr->used_lids[lid]) {
					is_free = FALSE;
					break;
				}
			}
			if (!is_free)
				continue;

			/* Mark LID range as used */
			for (lid = disc_min_lid; lid <= disc_max_lid; lid++)
				p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;

			continue;
		}

		/*
		 * Make sure the guid2lid entry is valid. If not, clean it.
		 * guid2lid is not valid if:
		 * - nvlink is enabled and LID does not match its LID type range.
		 * - or base LID is not aligned to LMC.
		 * - or LID range is not wide enough to match LMC.
		 */
		is_cached_match_range = TRUE;
		if (osm_subn_is_nvlink_enabled(p_mgr->p_subn)) {
			is_cached_match_range = lid_mgr_nvlink_is_lid_in_type_range(
			    p_mgr, osm_port_get_guid_ho(p_port), OSM_LID_TYPE_LID,
			    db_min_lid, db_max_lid);
		}

		if (!is_cached_match_range ||
		    (num_lids > 1 &&
		     ((db_min_lid & lmc_mask) != db_min_lid ||
		      db_max_lid - db_min_lid + 1 < num_lids))) {
			/* Not aligned, or not wide enough, then remove the entry */
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Cleaning persistent entry for guid:"
				"0x%016" PRIx64 " illegal range:[0x%x:0x%x]\n",
				osm_port_get_guid_ho(p_port),
				db_min_lid, db_max_lid);
			osm_db_guid2lid_delete(guid2lid,
					       cl_ntoh64
					       (osm_port_get_guid(p_port)));
			for (lid = db_min_lid; lid <= db_max_lid; lid++)
				p_mgr->used_lids[lid] &= OSM_RESERVED_LID_DISCOVERED;
		}
	}

	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn))
		lid_mgr_discover_alids(p_mgr);

AfterScanningLids:
	/* Make sure the defined sm lid is marked as discovered */
	if (sm_min_lid) {
		for (lid = sm_min_lid; lid <= sm_max_lid; lid++) {
			p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;
		}
	}

	/* Construct free LIDs ranges according to p_mgr->used_lids */
	/* Global lids */
	status = lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_LID, OSM_NVLINK_NO_PLANE);
	if (status)
		goto Exit;

	if (!osm_subn_is_nvlink_enabled(p_mgr->p_subn))
		goto Exit;

	/* ALIDs (GPU node lids) */
	status = lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_ALID, OSM_NVLINK_NO_PLANE);
	if (status)
		goto Exit;

	/* NVLink GPU port lids per plane */
	for (plane = 1; plane <= OSM_NVLINK_MAX_PLANES; plane++) {
		status = lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_LID, plane);
		if (status)
			goto Exit;
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

/**********************************************************************
 check if the given range of lids is free
**********************************************************************/
static boolean_t lid_mgr_is_range_not_persistent(IN osm_lid_mgr_t * p_mgr,
						 IN uint16_t lid,
						 IN uint16_t num_lids)
{
	uint16_t i;
	uint8_t used_mask = OSM_RESERVED_LID_PERSISTENT |
			    OSM_RESERVED_GLOBAL_FLID;

	for (i = lid; i < lid + num_lids; i++)
		if (p_mgr->used_lids[i] & used_mask)
			return FALSE;

	return TRUE;
}

/*
 * Get min and max values and create free range for these values
 * If p_prev_item is given, insert the new created range after it
 * If p_prev_item is not given, insert new range at tail
 *
 * return pointer for new range or NULL if failed to allocate memory
 */
static osm_lid_mgr_range_t *create_free_range(osm_lid_mgr_t *p_mgr,
					      uint16_t min, uint16_t max,
					      cl_list_item_t *p_prev_item,
					      osm_lid_mgr_lid_type_t lid_type,
					      uint8_t plane)
{
	osm_lid_mgr_range_t *p_range;
	cl_qlist_t *p_free_lid_ranges = lid_mgr_get_free_ranges_list(p_mgr, lid_type, plane);

	p_range = calloc(1, sizeof(osm_lid_mgr_range_t));
	if (p_range) {
		p_range->min_lid = min;
		p_range->max_lid = max;
		if (p_prev_item)
			cl_qlist_insert_next(p_free_lid_ranges, p_prev_item,
					     &p_range->item);
		else
			cl_qlist_insert_tail(p_free_lid_ranges, &p_range->item);

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Created new free %s range [%u:%u]\n",
			osm_lid_mgr_lid_type_str(lid_type, plane ? TRUE : FALSE),
			min, max);
	} else
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031D: "
			"Failed to allocate new free %s range\n",
			osm_lid_mgr_lid_type_str(lid_type, plane ? TRUE : FALSE));

	return p_range;
}

/* Currently support a single ALID, can ignore alid2 until we add support to a 2nd ALID */
static boolean_t lid_mgr_is_alid_used(osm_lid_mgr_t *p_mgr,
				      ib_net64_t guid,
				      uint16_t alid1,
				      uint16_t alid2)
{
	osm_gpu_t *p_gpu = osm_get_gpu_by_guid(p_mgr->p_subn, guid);

	if (!p_gpu)
		return FALSE;

	if (alid1 != osm_gpu_get_alid_by_index(p_gpu, 0))
		return FALSE;

	return TRUE;
}

static boolean_t lid_mgr_clear_ranges_list(osm_lid_mgr_t *p_mgr,
					   osm_lid_mgr_lid_type_t lid_type,
					   uint16_t lid_start,
					   uint16_t lid_end,
					   uint8_t plane)
{
	uint16_t lid, min_lid = 0;

	for (lid = lid_start; lid <= lid_end; lid++) {
		if (lid_mgr_is_lid_used(p_mgr, lid_type, plane, lid)) {
			if (min_lid) {
				if (!create_free_range(p_mgr, min_lid, lid - 1,
						       NULL, lid_type, plane)) {
					return FALSE;
				}
				min_lid = 0;
			}
		} else if (lid == IB_LID_UCAST_END_HO) {
			if (!min_lid)
				min_lid = lid;

			if (!create_free_range(p_mgr, min_lid, lid, NULL, lid_type, plane))
					return FALSE;
		} else if (!min_lid)
			min_lid = lid;
	}

	return TRUE;
}

/*
 * Clear all unused lids (lids which are mapped to physical or virtual ports
 * which are currently down)
 * Return the number of lids which were cleared
 */
static uint16_t osm_lid_mgr_clear_unused_lids(osm_lid_mgr_t *p_mgr,
					      osm_lid_mgr_lid_type_t lid_type,
					      uint8_t plane)
{
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint16_t lid, min_lid, max_lid, cleared = 0;
	uint16_t lid_start = 1, lid_end = IB_LID_UCAST_END_HO;
	osm_db_domain_t *guid2lid;
	cl_qlist_init(&guids);
	boolean_t clear_lid;
	const char *lid_type_str = osm_lid_mgr_lid_type_str(lid_type, plane ? TRUE : FALSE);

	guid2lid = lid_mgr_get_lid_db(p_mgr, lid_type, TRUE);
	if (!guid2lid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0338: "
			"Illegal LID type, cannot find LID DB\n");
		goto EXIT;
	}

	if (osm_db_guids(guid2lid, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031E: "
			"could not get guid list\n");
		goto EXIT;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2lid_get(guid2lid, p_item->guid,
					&min_lid, &max_lid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031F: "
				"could not get lid for guid:0x%016" PRIx64 "\n",
				p_item->guid);
			free(p_item);
			continue;
		}

		/*
		 * Check if DB LID for this guid is currently used by it.
		 * Otherwise, clear if from DB.
		 */
		clear_lid = FALSE;
		if (lid_type == OSM_LID_TYPE_ALID) {
			/*
			 * GPU ALIDs are not sequential within a GPU Node.
			 * Hence, for ALIDs DB, lids columns are used as alid1 and alid2.
			 */
			if (!lid_mgr_is_alid_used(p_mgr, p_item->guid, min_lid, max_lid))
				clear_lid = TRUE;

			/* As explained under 'NVLink Assumptions' above, support a single alid */
			max_lid = min_lid;
		} else if (lid_type == OSM_LID_TYPE_LID &&
			   !osm_get_port_by_lid_ho(p_mgr->p_subn, min_lid)) {
			clear_lid = TRUE;
		}

		if (clear_lid) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Clearing "
				"unused lid range [%u:%u] (mapped for guid 0x%"
				PRIx64 ") from guid2%s db\n",
				min_lid, max_lid, p_item->guid, lid_type_str);
			osm_db_guid2lid_delete(guid2lid, p_item->guid);
			for (lid = min_lid ; lid <= max_lid ; lid++)
				lid_mgr_clear_used_lid(p_mgr, lid);
			cleared++;
		}
		free(p_item);
	}

	if (!cleared)
		goto EXIT;

	/* We need to rebuild the empty ranges list */

	/* Cleanup current list */
	/* plane is non zero only when nvlink is enabled and port is GPU port */
	lid_mgr_destroy_free_lid_ranges_list(p_mgr, lid_type, plane);

	/*
	 * Note used_lids represents global LIDs and NVLink LIDs.
	 * Hence, use different start & end respectively according to lid_type
	 */
	if (!lid_mgr_range_limits(p_mgr, lid_type, plane, &lid_start, &lid_end)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0339: "
			"Could not get %s ranges to clear\n",
			lid_type_str);
		goto EXIT;
	}

	if (!lid_mgr_clear_ranges_list(p_mgr, lid_type, lid_start, lid_end, plane))
		cleared = 0;

EXIT:
	return cleared;
}

static uint16_t lid_mgr_inspect_unused_flids(osm_lid_mgr_t *p_mgr, boolean_t to_clear)
{
	osm_port_t *p_port;
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint16_t flid, cleared = 0;
	osm_router_mgr_t *p_rtr_mgr = &p_mgr->p_subn->p_osm->sm.router_mgr;

	cl_qlist_init(&guids);

	if (osm_db_guids(p_mgr->p_g2fl, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0323: "
			"could not get guid list\n");
		goto EXIT;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2flid_get(p_mgr->p_g2fl, p_item->guid, &flid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0324: "
				"could not get FLID for guid:0x%016" PRIx64 "\n",
				p_item->guid);
			free(p_item);
			continue;
		}

		p_port = osm_get_port_by_guid(p_mgr->p_subn, cl_hton64(p_item->guid));
		if (!p_port) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "%s "
				"unused FLID %u (mapped for guid 0x%" PRIx64 ") "
				"from guid2flid db\n",
				to_clear ? "Clearing" : "Inspecting", flid, p_item->guid);
			if (to_clear) {
				osm_db_guid2flid_delete(p_mgr->p_g2fl, p_item->guid);
				lid_mgr_clear_used_lid(p_mgr, flid);
				cleared++;
				/* update router mgr to remove this FLID */
				osm_router_mgr_remove_flid_not_in_subnet(p_rtr_mgr, flid);
			} else {
				/* update router mgr to add this FLID */
				osm_router_mgr_add_flid_not_in_subnet(p_rtr_mgr, flid);
			}
		}
		free(p_item);
	}

EXIT:
	return cleared;
}

void osm_lid_mgr_get_flids_not_in_subnet(osm_lid_mgr_t *p_mgr) {
	(void)lid_mgr_inspect_unused_flids(p_mgr, FALSE);
}

static uint16_t osm_lid_mgr_clear_unused_flids(osm_lid_mgr_t *p_mgr)
{
	uint16_t cleared = 0;

	cleared = lid_mgr_inspect_unused_flids(p_mgr, TRUE);

	if (cleared) {
		lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_FLID, OSM_NVLINK_NO_PLANE);
		/*
		 * no need to call 'lid_mgr_mark_global_flids', since the clearing of flid
		 * didn't remove the flid mark in 'used' array.
		 */
	}

	return cleared;
}

static uint8_t inline lid_mgr_get_lid_type_lmc(IN osm_lid_mgr_t * p_mgr,
					       IN osm_lid_mgr_lid_type_t lid_type,
					       IN uint8_t plane)
{
	if (lid_type == OSM_LID_TYPE_LID && plane)
		return lid_mgr_nvlink_get_gpu_lmc(p_mgr);

	return osm_subn_get_max_lmc(p_mgr->p_subn);
}

/*
 * Note - This is an internal function, input validation is performed beforehand.
 * We assume that p_range->min_lid <= min_lid and p_range->max_lid >= max_lid.
 */
static int lid_mgr_remove_lids_from_free_list(IN osm_lid_mgr_t * p_mgr,
					      IN cl_qlist_t * p_list,
					      IN osm_lid_mgr_range_t * p_range,
					      IN uint16_t min_lid,
					      IN uint16_t max_lid)
{
	/*
	 * Handle the specific case that our range and the current free
	 * range are identical (just remove this free range)
	 */
	if (min_lid == p_range->min_lid && max_lid == p_range->max_lid) {
		cl_qlist_remove_item(p_list, (cl_list_item_t *)p_range);
		free(p_range);
		return 0;
	}

	/*
	 * We confirmed that our range is inside the current free range,
	 * now there are three options:
	 *
	 *	- Our range at the beginning of the free range:
	 *		Just need to update a new min to the free range
	 *
	 *	- Our range at the end of the free range:
	 *		Just need to update a new max to the free range
	 *
	 *	- Our range is in the middle of the free range:
	 *		Need to split the free range into two ranges
	 */
	if (p_range->min_lid == min_lid) {
		/* Our range at the beginning of the free range */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Updated range [%u-%u] "
			"to [%u-%u]\n", p_range->min_lid, p_range->max_lid,
			max_lid + 1, p_range->max_lid);
		p_range->min_lid = max_lid + 1;
	} else if (p_range->max_lid == max_lid) {
		/* Our range at the end of the free range */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Updated range [%u-%u] "
			"to [%u-%u]\n", p_range->min_lid, p_range->max_lid,
			p_range->min_lid, min_lid - 1);
		p_range->max_lid = min_lid - 1;
	} else {
		/* Our range is in the middle of the free range */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Splitting range [%u-%u] "
			"to ranges [%u-%u] and [%u-%u]\n",
			p_range->min_lid, p_range->max_lid, p_range->min_lid,
			min_lid - 1, max_lid + 1, p_range->max_lid);

		/* First create new free range for the part of the free range
		 * which is higher than our range
		 */
		if (!create_free_range(p_mgr, max_lid + 1, p_range->max_lid,
				       (cl_list_item_t *)p_range, OSM_LID_TYPE_LID,
				       OSM_NVLINK_NO_PLANE))
			return 1;

		/* And update current range for the part of the range which
		 * is lower than our range
		 */
		p_range->max_lid = min_lid - 1;
	}

	return 0;
}

static int lid_mgr_find_free_range(IN osm_lid_mgr_t * p_mgr,
				   IN uint8_t num_lids,
				   IN osm_lid_mgr_lid_type_t lid_type,
				   IN uint8_t plane,
				   OUT uint16_t * p_min_lid,
				   OUT uint16_t * p_max_lid)
{
	cl_list_item_t *p_item;
	cl_list_item_t *p_next_item;
	osm_lid_mgr_range_t *p_range = NULL;
	uint8_t max_lmc;
	uint8_t lmc_num_lids;
	uint16_t lmc_mask;
	uint16_t lid;
	uint16_t max_lid;
	cl_qlist_t *p_free_lid_ranges;

	max_lmc = lid_mgr_get_lid_type_lmc(p_mgr, lid_type, plane);
	lmc_num_lids = (1 << max_lmc);
	lmc_mask = ~(lmc_num_lids - 1);

	p_free_lid_ranges = lid_mgr_get_free_ranges_list(p_mgr, lid_type, plane);
	if (!p_free_lid_ranges) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 032C: "
			"Illegal LID type, cannot find free LID range\n");
		return 1;
	}
	/*
	   Search the list of free lid ranges for a range which is big enough
	 */
	p_item = cl_qlist_head(p_free_lid_ranges);
	while (p_item != cl_qlist_end(p_free_lid_ranges)) {
		p_next_item = cl_qlist_next(p_item);
		p_range = (osm_lid_mgr_range_t *) p_item;

		lid = p_range->min_lid;

		/*
		 * If we require more then one lid we must align to LMC.
		 * This is relevant only for 'LID' calculation.
		 */
		if (num_lids > 1 && lid_type != OSM_LID_TYPE_ALID) {
			if ((lid & lmc_mask) != lid)
				lid = (lid + lmc_num_lids) & lmc_mask;
		}

		/* but we can be out of the range */
		max_lid = lid + num_lids - 1;
		if (max_lid <= p_range->max_lid) {
			/* ok let us use that range */
			int result = lid_mgr_remove_lids_from_free_list(p_mgr, p_free_lid_ranges, p_range, lid, max_lid);
			if (!result) {
				*p_min_lid = lid;

				/* As explained under 'NVLink Assumptions' above, support a single alid */
				if (lid_type != OSM_LID_TYPE_ALID) 
					*p_max_lid = max_lid;
			}
			
			return result;
		}
		p_item = p_next_item;
	}
	/* didn't find free range */
	return 1;
}

/**********************************************************************
find a free lid range
**********************************************************************/
static void lid_mgr_find_free_lid_range(IN osm_lid_mgr_t * p_mgr,
					IN uint8_t num_lids,
					IN osm_lid_mgr_lid_type_t lid_type,
					IN uint8_t plane,
					OUT uint16_t * p_min_lid,
					OUT uint16_t * p_max_lid)
{
	uint16_t cleared_lids;

	if (!lid_mgr_find_free_range(p_mgr, num_lids,lid_type, plane,
				     p_min_lid, p_max_lid))
		return;
	else {
		/* Couldn't find a free range of lids. */
		cleared_lids = osm_lid_mgr_clear_unused_lids(p_mgr, lid_type, plane);
		if (cleared_lids >= num_lids) {
			/* found free range after clearing unused lids */
			if (!lid_mgr_find_free_range(p_mgr, num_lids, lid_type, plane,
						     p_min_lid, p_max_lid))
				return;
		}
	}

	*p_min_lid = *p_max_lid = 0;
	/* if we run out of lids, give an error and abort! */
	OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0307: "
		"OpenSM ran out of %ss!!!\n",
		osm_lid_mgr_lid_type_str(lid_type, plane ? TRUE : FALSE));
	OSM_ASSERT(0);
}

static void lid_mgr_find_free_flid(IN osm_lid_mgr_t * p_mgr, OUT uint16_t * p_flid)
{
	uint16_t cleared_lids;

	if (!lid_mgr_find_free_range(p_mgr, 1, OSM_LID_TYPE_FLID, OSM_NVLINK_NO_PLANE,
				     p_flid, p_flid))
		return;
	else {
		/* Couldn't find a free flid range of lids. */
		cleared_lids = osm_lid_mgr_clear_unused_flids(p_mgr);
		if (cleared_lids) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"%u unused FLIDs cleared from guid2flid db.\n", cleared_lids);

			/* found free range after clearing unused lids */
			if (!lid_mgr_find_free_range(p_mgr, 1, OSM_LID_TYPE_FLID,
						     OSM_NVLINK_NO_PLANE, p_flid, p_flid))
				return;
		}
	}

	*p_flid = 0;
	OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0309: SM Ran out of FLIDs.\n");
}

static void lid_mgr_cleanup_discovered_port_lid_range(IN osm_lid_mgr_t * p_mgr,
						      IN osm_port_t * p_port)
{
	cl_ptr_vector_t *p_discovered_vec = &p_mgr->p_subn->port_lid_tbl;
	uint16_t lid, min_lid, max_lid, lmc_num_lids;
	uint16_t max_tbl_lid =
	    (uint16_t) (cl_ptr_vector_get_size(p_discovered_vec));

	osm_port_get_lid_range_ho(p_port, &min_lid, &max_lid);
	min_lid = trim_lid(min_lid);
	max_lid = trim_lid(max_lid);

	/*
	 * If current LMC is bigger than the LMC already set on the port, we might have reserved
	 * the range fitting the current LMC, and thus need to clean the bigger range.
	 */
	lmc_num_lids = 1 << osm_subn_get_max_lmc(p_mgr->p_subn);
	if (max_lid < min_lid + lmc_num_lids - 1)
		max_lid = min_lid + lmc_num_lids - 1;

	for (lid = min_lid; lid <= max_lid; lid++)
		if (lid < max_tbl_lid &&
		    p_port == cl_ptr_vector_get(p_discovered_vec, lid))
			cl_ptr_vector_set(p_discovered_vec, lid, NULL);
}

static void lid_mgr_cleanup_discovered_gpu_alid_range(IN osm_lid_mgr_t * p_mgr,
						      IN osm_gpu_t * p_gpu)
{
	cl_ptr_vector_t *p_discovered_vec = &p_mgr->p_subn->gpu_alid_tbl;
	uint16_t alid;
	uint16_t max_tbl_lid =
	    (uint16_t) (cl_ptr_vector_get_size(p_discovered_vec));

	/* currently support a single ALID per GPU */
	alid = osm_gpu_get_alid_by_index(p_gpu, 0);
	if (alid < max_tbl_lid && p_gpu == cl_ptr_vector_get(p_discovered_vec, alid))
		cl_ptr_vector_set(p_discovered_vec, alid, NULL);
}

/*
 * This function gets lid range and remove this range from the free ranges
 * list of the lid manager
 */
static int lid_mgr_set_lids_not_free(osm_lid_mgr_t *p_mgr, uint16_t min_lid,
				     uint16_t max_lid)
{
	cl_list_item_t *p_item;
	/* This function is used only for vports, hence use free_ranges of local LIDs */
	cl_qlist_t *p_list = &p_mgr->free_ranges;
	boolean_t found = FALSE;
	osm_lid_mgr_range_t *p_range = NULL;

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Removing range [%u-%u] from the "
		"free ranges list\n", min_lid, max_lid);

	/*
	 * The range we would like to remove from free ranges MUST be included
	 * in a single range item and the free ranges list.
	 * First we need to jump over all the ranges in the list which are
	 * completely lower than the range we would like to remove and get to
	 * the free range that should include our range.
	 */
	p_item = cl_qlist_head(p_list);
	while ((p_item != cl_qlist_end(p_list)) && !found) {
		p_range = (osm_lid_mgr_range_t*)p_item;

		if (min_lid > p_range->max_lid)
			p_item = cl_qlist_next(p_item);
		else
			found = TRUE;
	}

	/*
	 * If the required range was not found in the free ranges it means
	 * that we had some kind of incoherency in our db so we print an error
	 * message, however since it is not in the free ranges db and we
	 * probably just assigned the lid, there is nothing else to do as
	 * the db is now probably aligned
	 */
	if (!found) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031B: "
			"Range [%u-%u] not found in free ranges list\n",
			min_lid, max_lid);
		return 1;
	}

	/*
	 * Our range must be part of the current free range
	 * If it doesn't it means that our range was not part of the
	 * free ranges to begin with
	 */
	if (min_lid < p_range->min_lid || max_lid > p_range->max_lid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031C: "
			"Failed to remove range [%u-%u] from free "
			"ranges, this range is not part of free ranges"
			", Found at free range [%u-%u]\n",
			min_lid, max_lid, p_range->min_lid,
			p_range->max_lid);
		return 1;
	}

	return lid_mgr_remove_lids_from_free_list(p_mgr, p_list, p_range, min_lid, max_lid);
}

/**********************************************************************
 0.1 if the port info lid matches the guid2lid return 0
 0.2 if the port info has a lid and that range is empty in
     port_lid_tbl, return 0 and update the port_lid_tbl and
     guid2lid
 0.3 else find an empty space in port_lid_tbl, update the
 port_lid_tbl and guid2lid, return 1 to flag a change required.
**********************************************************************/
static int lid_mgr_get_port_lid(IN osm_lid_mgr_t * p_mgr,
				IN osm_port_t * p_port,
				IN osm_vport_t * p_vport,
				IN uint16_t req_lid,
				OUT uint16_t * p_min_lid,
				OUT uint16_t * p_max_lid)
{
	uint16_t lid, min_lid, max_lid, disc_base_lid;
	uint64_t guid;
	uint8_t num_lids;
	int lid_changed = 0;
	uint16_t lmc_mask;
	osm_db_domain_t *guid2lid = NULL;
	boolean_t is_gpu_port = osm_subn_is_nvlink_enabled(p_mgr->p_subn) && osm_node_is_gpu(p_port->p_node);

	OSM_LOG_ENTER(p_mgr->p_log);
	/* get the lid from PortInfo */
	if (p_vport) {
		guid = cl_ntoh64(osm_vport_get_guid(p_vport));
		disc_base_lid = cl_ntoh16(osm_vport_get_lid(p_vport));
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Getting lid for virtual "
			"port 0x%016" PRIx64 "\n", guid);
	} else {
		guid = osm_port_get_guid_ho(p_port);
		disc_base_lid = osm_port_get_base_lid_ho(p_port);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Getting lid for port "
			"0x%016" PRIx64 "\n",
			guid);
	}
	/* if the port is a base switch port 0 then we only need one lid */
	if (is_gpu_port)
		num_lids = ((uint8_t)1 << lid_mgr_nvlink_get_gpu_lmc(p_mgr));
	else if (!p_vport &&
		 (osm_lid_mgr_port_needs_global_lmc(p_mgr->p_subn, p_port) ||
		  osm_lid_mgr_port_needs_rtr_lmc(p_mgr->p_subn, p_port) ||
		  osm_lid_mgr_port_needs_prism_an_lmc(p_mgr->p_subn, p_port)))
		num_lids = (1 << osm_subn_get_max_lmc(p_mgr->p_subn));
	else
		num_lids = 1;

	/* if the port matches the guid2lid */
	guid2lid = lid_mgr_get_lid_db(p_mgr, OSM_LID_TYPE_LID, is_gpu_port);
	OSM_ASSERT(guid2lid);

	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE)
		goto AssignLid;

	lmc_mask = ~(num_lids - 1);

	if (!osm_db_guid2lid_get(guid2lid, guid, &min_lid, &max_lid)) {
		*p_min_lid = min_lid;
		*p_max_lid = min_lid + num_lids - 1;

		/*
		 * Validate lid from DB matches base lid
		 * and therefore matches lid type range as well)
		 */
		if (min_lid == disc_base_lid && max_lid == *p_max_lid)
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "0x%016" PRIx64
				" matches its known lid:%u\n", guid, min_lid);
		else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"0x%016" PRIx64 " with lid:%u "
				"does not match its known lid:%u\n",
				guid, disc_base_lid, min_lid);
			if (!p_vport)
				lid_mgr_cleanup_discovered_port_lid_range(p_mgr,
									  p_port);
			/* we still need to send the setting to the target port */
			lid_changed = 1;
		}
		goto NewLidSet;
	} else
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"0x%016" PRIx64 " has no persistent lid assigned\n",
			guid);

	/* Validate base lid is valid for nvlink ranges */
	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn) &&
	    !lid_mgr_nvlink_is_lid_in_type_range(p_mgr, guid, OSM_LID_TYPE_LID,
						 disc_base_lid,
						 disc_base_lid + num_lids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"0x%016" PRIx64 " configured lid does not match lid type range\n",
			guid);
		goto AssignLid;
	}

	if (req_lid && disc_base_lid != req_lid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"port 0x%016" PRIx64 " lid %u does not match "
			"requested lid %u\n",
			guid, disc_base_lid, req_lid);
		goto AssignLid;
	}

	/* if the port info carries a lid it must be lmc aligned and not mapped
	   by the pesistent storage  */
	min_lid = disc_base_lid;

	/* we want to ignore the discovered lid if we are also on first sweep of
	   reassign lids flow */
	if (min_lid) {
		/* make sure lid is valid */
		if ((min_lid & lmc_mask) == min_lid) {
			/* is it free */
			if (lid_mgr_is_range_not_persistent
			    (p_mgr, min_lid, num_lids)) {
				*p_min_lid = min_lid;
				*p_max_lid = min_lid + num_lids - 1;
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"0x%016" PRIx64
					" lid range:[%u-%u] is free\n",
					guid, *p_min_lid, *p_max_lid);
				goto NewLidSet;
			} else
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"0x%016" PRIx64 " existing lid "
					"range:[%u:%u] is not free\n",
					guid, min_lid, min_lid + num_lids - 1);
		} else
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"0x%016" PRIx64 " existing lid range:"
				"[%u:%u] is not lmc aligned\n",
				guid, min_lid, min_lid + num_lids - 1);
	}

AssignLid:
	/* first cleanup the existing discovered lid range */
	if (!p_vport)
		lid_mgr_cleanup_discovered_port_lid_range(p_mgr, p_port);

	if (req_lid) {
		*p_min_lid = req_lid;
		*p_max_lid = req_lid + num_lids - 1;
	} else {
		/* find an empty space */
		lid_mgr_find_free_lid_range(
		    p_mgr, num_lids, OSM_LID_TYPE_LID,
		    is_gpu_port ? osm_port_get_nvlink_plane(p_port) : OSM_NVLINK_NO_PLANE,
		    p_min_lid, p_max_lid);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"0x%016" PRIx64 " assigned a new lid range:[%u-%u]\n",
			guid, *p_min_lid, *p_max_lid);
	}

	lid_changed = 1;

	/* Avoid infinite loop in case of no free lids */
	if (*p_min_lid == 0)
		goto EXIT;

NewLidSet:
	if (*p_min_lid == 0)
                goto AssignLid;

	/*
	 * Update free ranges
	 * Required when called from LID manager
	 */
	if (p_vport && !lid_changed && !p_mgr->used_lids[*p_min_lid]) {
		lid_mgr_set_lids_not_free(p_mgr, *p_min_lid, *p_max_lid);
	}

	/* update the guid2lid db (or guid2nvlid) and used_lids */
	osm_db_guid2lid_set(guid2lid, guid, *p_min_lid, *p_max_lid);
	for (lid = *p_min_lid; lid <= *p_max_lid; lid++)
		p_mgr->used_lids[lid] |=
			(OSM_RESERVED_LID_PERSISTENT | OSM_RESERVED_LID_DISCOVERED);

	/* make sure the assigned lids are marked in port_lid_tbl */
	if (!p_vport)
		for (lid = *p_min_lid; lid <= *p_max_lid; lid++)
			cl_ptr_vector_set(&p_mgr->p_subn->port_lid_tbl, lid, p_port);

EXIT:
	OSM_LOG_EXIT(p_mgr->p_log);
	return lid_changed;
}

uint16_t osm_lid_mgr_query_switch_flid(IN osm_lid_mgr_t * p_mgr,
				       IN osm_port_t * p_port)
{
	uint16_t flid = 0;
	uint64_t guid;

	guid = osm_port_get_guid_ho(p_port);
	if (osm_db_guid2flid_get(p_mgr->p_g2fl, guid, &flid))
		return 0;
	return flid;
}

uint16_t osm_lid_mgr_get_switch_flid(IN osm_lid_mgr_t * p_mgr,
				     IN osm_port_t * p_port)
{
	uint16_t flid = 0;
	uint64_t guid;

	OSM_LOG_ENTER(p_mgr->p_log);

	guid = osm_port_get_guid_ho(p_port);
	
	/* try to assign new flid if not in db */
	if (osm_db_guid2flid_get(p_mgr->p_g2fl, guid, &flid)) {
		lid_mgr_find_free_flid(p_mgr, &flid);

		if (flid) {
			/*
			 * Currently all global FLIDs are marked as RESERVED_GLOBAL_FLID,
			 * mark that the flid is now also assigned (discovered).
			 */
			p_mgr->used_lids[flid] |=
				(OSM_RESERVED_LID_PERSISTENT | OSM_RESERVED_LID_DISCOVERED);
		}
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Assigned port guid 0x%016" PRIx64 " with flid %u.\n",
			osm_port_get_guid_ho(p_port), flid);

	}

	OSM_LOG_EXIT(p_mgr->p_log);
	return flid;
}

void osm_lid_mgr_set_guid_flid(IN osm_lid_mgr_t * p_mgr, IN uint64_t guid, IN uint16_t flid)
{
	OSM_LOG_ENTER(p_mgr->p_log);

	if (!guid || !flid)
		return;

	osm_db_guid2flid_set(p_mgr->p_g2fl, guid, flid);

	OSM_LOG_EXIT(p_mgr->p_log);
}

static void lid_mgr_update_alid_info_req(IN osm_lid_mgr_t * p_mgr,
					 IN osm_gpu_t * p_gpu,
					 IN uint8_t num_alids,
					 IN uint8_t alids_len,
					 IN uint16_t * p_alids)
{
	uint8_t i;
	uint8_t table_cap = p_gpu->p_node->ext_node_info.anycast_lid_table_cap;

	OSM_ASSERT(num_alids <= alids_len);

	for (i = 0; i < num_alids; i++)
		p_gpu->calc_alid_info.alid_block[i].alid = cl_hton32((uint32_t)p_alids[i]);

	/* Set the rest of ALID entries for this GPU to 0 (unused) */
	for (; i < table_cap; i++)
		memset(&p_gpu->calc_alid_info.alid_block[i], 0, sizeof(ib_net32_t));

	/* Mark the 1st ALID to be used. Used ALID might be changed later due to link failure */
	p_gpu->calc_alid_info.alid_block[0].alid |= IB_ALID_INFO_ENTRY_IS_USED_BIT;
}

/*
 * Get currently configured GPU ALID range (as for other lids, prioritize DB over subn),
 * or a new range if does not exist or does not match configuration parameters.
 * returns TRUE if new ALIDs are set to calculated alidInfo (shall be set later)
 */
static boolean_t lid_mgr_get_gpu_alid(IN osm_lid_mgr_t * p_mgr,
				      IN osm_gpu_t * p_gpu,
				      IN uint8_t num_alids,
				      IN uint8_t alids_len,
				      OUT uint16_t * p_alids)
{
	osm_db_domain_t *guid2lid = NULL;
	uint16_t disc_alid, db_alid1, db_alid2;
	uint64_t gpu_guid = cl_ntoh64(osm_gpu_get_guid(p_gpu));
	boolean_t lid_changed = FALSE;

	OSM_LOG_ENTER(p_mgr->p_log);
	OSM_ASSERT(alids_len == OSM_NVLINK_MAX_NUM_ALIDS_PER_GPU && num_alids &&
		   num_alids <= alids_len);

	disc_alid = osm_gpu_get_alid_by_index(p_gpu, 0);
	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Getting alid for GPU 0x%016" PRIx64 " (configured with %u)\n",
		gpu_guid, disc_alid);

	/* if the port matches the guid2lid */
	guid2lid = lid_mgr_get_lid_db(p_mgr, OSM_LID_TYPE_ALID, FALSE);
	OSM_ASSERT(guid2lid);

	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE)
		goto AssignALID;

	if (!osm_db_guid2lid_get(guid2lid, gpu_guid, &db_alid1, &db_alid2)) {
		/* Currently support a single ALID */
		p_alids[0] = db_alid1;

		/*
		 * Validate lid from DB matches base lid
		 * and therefore matches lid type range as well
		 */
		if (db_alid1 == disc_alid && db_alid2 == 0) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "0x%016" PRIx64
				" matches its known alid:%u\n", gpu_guid, disc_alid);
		} else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"0x%016" PRIx64 " with alid:%u "
				"does not match its known alid:%u\n",
				gpu_guid, disc_alid, db_alid1);

			lid_mgr_cleanup_discovered_gpu_alid_range(p_mgr, p_gpu);

			/* we still need to send the setting to the target GPU */
			lid_changed = TRUE;
		}

		goto NewALIDSet;
	} else {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"0x%016" PRIx64 " has no persistent alid assigned\n",
			gpu_guid);
	}

	/*
	 * If there is no ALID DB entry for this GPU,
	 * take the alid previously configured on the subnet, if valid
	 */

	/* Validate base lid is valid for nvlink ranges */
	if (!lid_mgr_nvlink_is_lid_in_type_range(p_mgr, gpu_guid, OSM_LID_TYPE_ALID,
						 disc_alid, disc_alid)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"0x%016" PRIx64 " configured alid (%u) does not "
			"match alid range\n", gpu_guid, disc_alid);
		goto AssignALID;
	}
	/* Validate this base lid is free to use */
	if (lid_mgr_is_range_not_persistent(p_mgr, disc_alid, OSM_NVLINK_NUM_ALIDS_PER_GPU)) {
		p_alids[0] = disc_alid;
		goto NewALIDSet;
	}

AssignALID:
	/* find an empty space */
	lid_mgr_find_free_lid_range(p_mgr, num_alids, OSM_LID_TYPE_ALID, OSM_NVLINK_NO_PLANE,
				    &p_alids[0], &p_alids[1]);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"0x%016" PRIx64 " assigned a new alid:[%u]\n",
		gpu_guid, p_alids[0]);

	lid_mgr_cleanup_discovered_gpu_alid_range(p_mgr, p_gpu);

	lid_changed = TRUE;

	/* Avoid infinite loop in case of no free alids */
	if (p_alids[0] == 0)
		goto EXIT;

NewALIDSet:
	if (p_alids[0] == 0)
                goto AssignALID;

	/* update the guid2alid and used_lids */

	/* Currently support a single OSM_LID_TYPE_ALID (ALIDs are not continuous within GPU) */
	p_alids[1] = 0;

	osm_db_guid2lid_set(guid2lid, gpu_guid, p_alids[0], p_alids[1]);
	p_mgr->used_lids[p_alids[0]] |=
	    (OSM_RESERVED_LID_PERSISTENT | OSM_RESERVED_LID_DISCOVERED);

EXIT:
	OSM_LOG_EXIT(p_mgr->p_log);
	return lid_changed;
}

static int lid_mgr_set_physp_pi(IN osm_lid_mgr_t * p_mgr,
				IN osm_port_t * p_port,
				IN osm_physp_t * p_physp, IN ib_net16_t lid)
{
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_port_info_t *p_pi = (ib_port_info_t *) payload;
	const ib_port_info_t *p_old_pi;
	osm_madw_context_t context;
	osm_node_t *p_node;
	ib_api_status_t status;
	uint8_t port_num;
	boolean_t send_set = FALSE;
	boolean_t update_mkey = FALSE;
	boolean_t send_client_rereg = FALSE;
	int ret = 0;

	OSM_LOG_ENTER(p_mgr->p_log);

	/*
	   Don't bother doing anything if this Physical Port is not valid.
	   This allows simplified code in the caller.
	 */
	if (!p_physp)
		goto Exit;

	port_num = osm_physp_get_port_num(p_physp);
	p_node = osm_physp_get_node_ptr(p_physp);

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH && port_num != 0) {
		/*
		   Switch ports that are not numbered 0 should not be set
		   with the following attributes as they are set later
		   (during NO_CHANGE state in link mgr).
		 */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Skipping switch port %u, GUID 0x%016" PRIx64 "\n",
			port_num, cl_ntoh64(osm_physp_get_port_guid(p_physp)));
		goto Exit;
	}

	p_old_pi = &p_physp->port_info;

	/*
	   First, copy existing parameters from the PortInfo attribute we
	   already have for this node.

	   Second, update with default values that we know must be set for
	   every Physical Port and the LID and set the neighbor MTU field
	   appropriately.

	   Third, send the SMP to this physical port.
	 */

	memcpy(payload, p_old_pi, sizeof(ib_port_info_t));

	/*
	   Should never write back a value that is bigger then 3 in
	   the PortPhysicalState field, so cannot simply copy!

	   Actually we want to write there:
	   port physical state - no change
	   link down default state = polling
	   port state - no change
	 */
	p_pi->state_info2 = 0x02;
	ib_port_info_set_port_state(p_pi, IB_LINK_NO_CHANGE);

	if (ib_port_info_get_link_down_def_state(p_pi) !=
	    ib_port_info_get_link_down_def_state(p_old_pi))
		send_set = TRUE;

	/* didn't get PortInfo before */
	if (!ib_port_info_get_port_state(p_old_pi))
		send_set = TRUE;

	p_pi->m_key = osm_physp_calc_mkey(p_mgr->p_subn, p_physp);
	if (memcmp(&p_pi->m_key, &p_old_pi->m_key, sizeof(p_pi->m_key))) {
		update_mkey = TRUE;
		send_set = TRUE;
	}

	/*
	 * Make sure that only subnet prefix part of gid prefix is set,
	 * don't change the flid part.
	 * This if should happen only first time the configured subnet prefix is not set yet.
	 */
	if (!is_equal_subnet_prefix(p_old_pi->subnet_prefix,
				    p_mgr->p_subn->opt.subnet_prefix)) {
		p_pi->subnet_prefix =
			(p_old_pi->subnet_prefix & (~IB_SITE_LOCAL_VALID_BITS_SUBNET_MASK)) |
			(p_mgr->p_subn->opt.subnet_prefix & IB_SITE_LOCAL_VALID_BITS_SUBNET_MASK);

		send_set = TRUE;
	}

	p_port->lid = lid;
	p_pi->base_lid = lid;
	if (memcmp(&p_pi->base_lid, &p_old_pi->base_lid,
		   sizeof(p_pi->base_lid))) {
		/*
		 * Reset stored base_lid.
		 * On successful send, we'll update it when we'll get a reply.
		 */
		osm_physp_set_base_lid(p_physp, 0);
		send_set = TRUE;
		p_mgr->dirty = TRUE;
	}

	/*
	   We are updating the ports with our local sm_base_lid
	   if for some reason currently received SM LID is different from our SM LID,
	   need to send client reregister to this port
	*/
	p_pi->master_sm_base_lid = p_mgr->p_subn->sm_base_lid;
	if (memcmp(&p_pi->master_sm_base_lid, &p_old_pi->master_sm_base_lid,
		   sizeof(p_pi->master_sm_base_lid))) {
		send_client_rereg = TRUE;
		send_set = TRUE;
	}

	p_pi->m_key_lease_period = p_mgr->p_subn->opt.m_key_lease_period;
	if (memcmp(&p_pi->m_key_lease_period, &p_old_pi->m_key_lease_period,
		   sizeof(p_pi->m_key_lease_period)))
		send_set = TRUE;

	p_pi->mkey_lmc = 0;
	ib_port_info_set_mpb(p_pi, p_mgr->p_subn->opt.m_key_protect_bits);
	if (ib_port_info_get_mpb(p_pi) != ib_port_info_get_mpb(p_old_pi))
		send_set = TRUE;

	/*
	   we want to set the timeout for both the switch port 0
	   and the CA ports
	 */
	ib_port_info_set_timeout(p_pi, p_mgr->p_subn->opt.subnet_timeout);
	if (ib_port_info_get_timeout(p_pi) !=
	    ib_port_info_get_timeout(p_old_pi))
		send_set = TRUE;

	if (port_num != 0) {
		/*
		   CAs don't have a port 0, and for switch port 0,
		   the state bits are ignored.
		   This is not the switch management port
		 */

		/* p_pi->mkey_lmc is initialized earlier */
		uint8_t set_lmc = 0;
		if (p_mgr->p_subn->opt.lmc &&
		    !(osm_subn_is_nvlink_enabled(p_mgr->p_subn) && osm_node_is_gpu(p_node)))
			set_lmc = p_mgr->p_subn->opt.lmc;
		else if (osm_lid_mgr_port_needs_rtr_lmc(p_mgr->p_subn, p_port))
			set_lmc = p_mgr->p_subn->opt.rtr_lmc;
		else if (osm_lid_mgr_port_needs_prism_an_lmc(p_mgr->p_subn, p_port))
			set_lmc = p_mgr->p_subn->opt.prism_an_lmc;

		ib_port_info_set_lmc(p_pi, set_lmc);
		p_physp->expected_lmc = ib_port_info_get_lmc(p_pi);
		if (ib_port_info_get_lmc(p_pi) !=
		    ib_port_info_get_lmc(p_old_pi)) {
			send_set = TRUE;
			p_mgr->dirty = TRUE;
		}

		/*
		   Several timeout mechanisms:
		 */
		ib_port_info_set_phy_and_overrun_err_thd(p_pi,
							 p_mgr->p_subn->opt.
							 local_phy_errors_threshold,
							 p_mgr->p_subn->opt.
							 overrun_errors_threshold);

		if (p_pi->error_threshold != p_old_pi->error_threshold)
			send_set = TRUE;

	} else if (ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info)) {
		/*
		 * Configure Enh. SP0:
		 * Set MTU according to the mtu_cap.
		 * Set LMC if lmc_esp0 is defined.
		 */
		ib_port_info_set_neighbor_mtu(p_pi,
					      ib_port_info_get_mtu_cap
					      (p_old_pi));
		if (ib_port_info_get_neighbor_mtu(p_pi) !=
		    ib_port_info_get_neighbor_mtu(p_old_pi))
			send_set = TRUE;

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Updating neighbor_mtu on switch GUID 0x%016" PRIx64
			" port 0 to:%u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			ib_port_info_get_neighbor_mtu(p_pi));

		/* Configure LMC on enhanced SP0 */
		if (p_mgr->p_subn->opt.lmc_esp0) {
			/* p_pi->mkey_lmc is initialized earlier */
			ib_port_info_set_lmc(p_pi, p_mgr->p_subn->opt.lmc);
			p_physp->expected_lmc = ib_port_info_get_lmc(p_pi);
			if (ib_port_info_get_lmc(p_pi) !=
			    ib_port_info_get_lmc(p_old_pi))
				send_set = TRUE;
		}
	}

	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = TRUE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;

	/*
	   For ports supporting the ClientReregistration Vol1 (v1.2) p811 14.4.11:
	   need to set the cli_rereg bit when current SM LID at the Host
	   is different from our SM LID, if we are in first_time_master_sweep
	   if this port was just now discovered, then we should also set
	   the cli_rereg bit. We know that the port was just discovered if its
	   is_new field is set.
	   Also in case of timeout for previous PortInfo with client_reregister
	   bit set, in current PortInfo it will be set again(client_rereg flag).
	   Moreover, we won't set client_reregister for a port that was discovered
	   only through an ignored port or to switch that has ignored ports,
	   though we will mark it for resending in case it becomes healthy later.
	 */
	if ((send_client_rereg ||
	     p_mgr->p_subn->first_time_master_sweep == TRUE || p_port->is_new)
	    && p_mgr->p_subn->opt.client_rereg_mode
	    && (p_old_pi->capability_mask & IB_PORT_CAP_HAS_CLIENT_REREG)) {
		/* Mark port for sending client reregister */
		p_port->client_rereg = TRUE;
	}

	if (p_port->client_rereg && p_port->disc_through != OSM_DISC_IGNORED &&
	    p_mgr->p_subn->opt.client_rereg_mode == OSM_CLIENT_REREG_MODE_LID_MGR) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Setting client rereg on %s, port %d\n",
			p_port->p_node->print_desc, p_port->p_physp->port_num);
		ib_port_info_set_client_rereg(p_pi, 1);
		context.pi_context.client_rereg = TRUE;
		send_set = TRUE;
	} else {
		ib_port_info_set_client_rereg(p_pi, 0);
		context.pi_context.client_rereg = FALSE;
	}

	/* We need to send the PortInfo Set request with the new sm_lid
	   in the following cases:
	   1. There is a change in the values (send_set == TRUE)
	   2. first_time_master_sweep flag on the subnet is TRUE. This means the
	   SM just became master, and it then needs to send a PortInfo Set to
	   every port.
	 */
	if (p_mgr->p_subn->first_time_master_sweep == TRUE)
		send_set = TRUE;

	if (!send_set)
		goto Exit;

	status = osm_req_set_pi(p_mgr->sm, p_physp,
				osm_physp_get_port_num(p_physp),
				payload, FALSE,
				ib_port_info_get_m_key(&p_physp->port_info),
				&context, TRUE);
	if (status != IB_SUCCESS)
		ret = -1;
	/* If we sent a new mkey above, update our guid2mkey map
	   now, on the assumption that the SubnSet succeeds
	*/
	if (update_mkey)
		osm_db_guid2key_set(p_mgr->p_subn->p_g2m,
				    cl_ntoh64(p_physp->port_guid),
				    cl_ntoh64(p_pi->m_key));

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

/**********************************************************************
 Processes our own node
 Lock must already be held.
**********************************************************************/
static int lid_mgr_process_our_sm_node(IN osm_lid_mgr_t * p_mgr)
{
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;
	osm_port_t *p_port;
	uint16_t min_lid_ho;
	uint16_t max_lid_ho;
	int ret;

	OSM_LOG_ENTER(p_mgr->p_log);

	/*
	   Acquire our own port object.
	 */
	p_port = osm_get_port_by_guid(p_mgr->p_subn,
				      p_mgr->p_subn->sm_port_guid);
	if (!p_port) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0308: "
			"Can't acquire SM's port object, GUID 0x%016" PRIx64
			"\n", cl_ntoh64(p_mgr->p_subn->sm_port_guid));
		ret = -1;
		goto Exit;
	}

	/*
	   Determine the LID this SM will use for its own port.
	   Be careful.  With an LMC > 0, the bottom of the LID range becomes
	   unusable, since port hardware will mask off least significant bits,
	   leaving a LID of 0 (invalid).  Therefore, make sure that we always
	   configure the SM with a LID that has non-zero bits, even after
	   LMC masking by hardware.
	 */
	lid_mgr_get_port_lid(p_mgr, p_port, NULL,
			     p_mgr->p_subn->opt.master_sm_lid,
			     &min_lid_ho, &max_lid_ho);
	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Current base LID is %u\n", min_lid_ho);
	/*
	   Update subnet object.
	 */
	p_mgr->p_subn->master_sm_base_lid = cl_hton16(min_lid_ho);
	p_mgr->p_subn->sm_base_lid = cl_hton16(min_lid_ho);

	OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
		"Assigning SM's port 0x%016" PRIx64
		"\n\t\t\t\tto LID range [%u,%u]\n",
		osm_port_get_guid_ho(p_port), min_lid_ho, max_lid_ho);

	/*
	   Set the PortInfo the Physical Port associated with this Port.
	 */
	ret = lid_mgr_set_physp_pi(p_mgr, p_port, p_port->p_physp,
				   cl_hton16(min_lid_ho));

	/* Consider SM LID for global lid range top of LFTSplit */
	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn) &&
	    max_lid_ho > p_nvl_mgr->global_lid_range_top)
		p_nvl_mgr->global_lid_range_top = max_lid_ho;

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

int osm_lid_mgr_process_sm(IN osm_lid_mgr_t * p_mgr)
{
	int ret;

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_ASSERT(p_mgr->p_subn->sm_port_guid);

	CL_PLOCK_EXCL_ACQUIRE(p_mgr->p_lock);

	/* initialize the port_lid_tbl and empty ranges list following the
	   persistent db */
	lid_mgr_init_sweep(p_mgr);

	ret = lid_mgr_process_our_sm_node(p_mgr);

	CL_PLOCK_RELEASE(p_mgr->p_lock);

	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

/*
 * Currently, a single alid per GPU is supported.
 * NVLink is planned to support up to 2 alids.
 * Global LIDs API allows us to support up to 2 lid columns, hence we can use it as long
 * as no need to support more than 2 alids.
 */
boolean_t osm_lid_mgr_process_alids(IN osm_lid_mgr_t * p_mgr)
{
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;
	osm_gpu_t *p_gpu;
	cl_qmap_t *p_gpu_guid_tbl = &p_mgr->p_subn->gpu_guid_tbl;
	boolean_t lid_changed;
	int ret = 0;
	uint8_t num_alids = p_nvl_mgr->num_alids_per_gpu;
	uint8_t max_num_alids = OSM_NVLINK_MAX_NUM_ALIDS_PER_GPU;
	uint16_t *p_alids;

	OSM_ASSERT(num_alids <= max_num_alids);

	OSM_LOG_ENTER(p_mgr->p_log);

	p_alids = (uint16_t *)calloc(max_num_alids, sizeof(uint16_t));
	if (!p_alids) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 033A: "
			"Failed to allocate alids array for %u alids\n",
			num_alids);
		return FALSE;
	}

	for (p_gpu = (osm_gpu_t *) cl_qmap_head(p_gpu_guid_tbl);
	     p_gpu != (osm_gpu_t *) cl_qmap_end(p_gpu_guid_tbl);
	     p_gpu = (osm_gpu_t *) cl_qmap_next(&p_gpu->map_item)) {
		memset(p_alids, 0, num_alids);

		/* Get currently configured ALID range for this GPU, or a newly configured */
		lid_changed = lid_mgr_get_gpu_alid(p_mgr, p_gpu, num_alids, max_num_alids, p_alids);

		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"Assigned GPU 0x%016" PRIx64 ", %s ALID [%u,%u]\n",
			osm_node_get_node_guid_ho(p_gpu->p_node),
			lid_changed ? "new" : "", p_alids[0], p_alids[1]);

		/* If we change ALID of an old GPU, then we need to invalidate ucast cache */
		if (p_mgr->p_subn->first_time_master_sweep == FALSE &&
		    lid_changed && !p_gpu->is_new) {

			OSM_LOG(p_mgr->p_log, OSM_LOG_INFO,
				"ALIDs [%u,%u] changed for existing GPU 0x%016"
				PRIx64 " \n", p_alids[0], p_alids[1],
				cl_ntoh64(osm_gpu_get_guid(p_gpu)));

			if (p_mgr->p_subn->opt.use_ucast_cache)
				p_mgr->sm->ucast_mgr.variables_invalidate_cache = TRUE;
		}

		/* Send ENI set for setting alid_top */
		if (osm_nvlink_set_gpu_extended_node_info(p_mgr->sm, p_gpu))
			ret = -1;

		lid_mgr_update_alid_info_req(p_mgr, p_gpu, num_alids, max_num_alids, p_alids);
		/* Update GPU's ALID by ALIDInfo */
		if (osm_nvlink_send_set_alid_info(p_mgr->sm, p_gpu))
			ret = -1;

		if (p_alids[1] && p_alids[1] > p_nvl_mgr->alid_range_top)
			p_nvl_mgr->alid_range_top = p_alids[1];
		else if (p_alids[0] > p_nvl_mgr->alid_range_top)
			p_nvl_mgr->alid_range_top = p_alids[0];

		cl_ptr_vector_set(&p_mgr->p_subn->gpu_alid_tbl, p_alids[0], p_gpu);
		if (p_alids[1])
			cl_ptr_vector_set(&p_mgr->p_subn->gpu_alid_tbl, p_alids[1], p_gpu);

		/*
		 * GPU port lids (nvlids) are configure at osm_lid_mgr_process_subnet by
		 * lid_mgr_get_port_lid and setting PortInfo
		 */
	}

	/* store the guid to lid table in persistent db */
	osm_db_store(p_nvl_mgr->p_g2alid, p_mgr->p_subn->opt.fsync_high_avail_files);

	lid_mgr_reset_lid_free_range(p_mgr, OSM_LID_TYPE_ALID, OSM_NVLINK_NO_PLANE);

	/*
	 * On NVLink LID space division, plane LIDs (GPU port LIDs)range is greater
	 * than ALID range. Hence should not get into this condition on healthy flow.
	 */
	if (cl_qmap_count(p_gpu_guid_tbl) &&
	    p_mgr->p_subn->max_lid < p_nvl_mgr->alid_range_top)
		p_mgr->p_subn->max_lid = p_nvl_mgr->alid_range_top;

	free(p_alids);

	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

/**********************************************************************
 1 go through all ports in the subnet.
 1.1 call lid_mgr_get_port_min_lid
 1.2 if a change is required send the port info
 2 if any change send the signal PENDING...
**********************************************************************/
int osm_lid_mgr_process_subnet(IN osm_lid_mgr_t * p_mgr)
{
	cl_qmap_t *p_port_guid_tbl;
	osm_port_t *p_port;
	ib_net64_t port_guid;
	int lid_changed, ret = 0;
	uint16_t min_lid_ho, max_lid_ho, max_flid;
	uint16_t lids, base_lid;
	osm_nvlink_mgr_t *p_nvl_mgr = &p_mgr->sm->nvl_mgr;

	OSM_ASSERT(p_mgr);

	OSM_LOG_ENTER(p_mgr->p_log);

	CL_PLOCK_EXCL_ACQUIRE(p_mgr->p_lock);

	OSM_ASSERT(p_mgr->p_subn->sm_port_guid);

	p_port_guid_tbl = &p_mgr->p_subn->port_guid_tbl;
	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		port_guid = osm_port_get_guid(p_port);

		/*
		   Our own port is a special case in that we want to
		   assign a LID to ourselves first, since we have to
		   advertise that LID value to the other ports.

		   For that reason, our node is treated separately and
		   we will not add it to any of these lists.
		 */
		if (port_guid == p_mgr->p_subn->sm_port_guid) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Skipping our own port 0x%016" PRIx64 "\n",
				cl_ntoh64(port_guid));
			continue;
		}

		/*
		   get the port lid range - we need to send it on first active
		   sweep or if there was a change (the result of
		   lid_mgr_get_port_lid)
		 */
		lid_changed = lid_mgr_get_port_lid(p_mgr, p_port, NULL, 0,
						   &min_lid_ho, &max_lid_ho);

		/* we can call the function to update the port info as it known
		   to look for any field change and will only send an updated
		   if required */
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"Assigned port 0x%016" PRIx64 ", %s LID [%u,%u]\n",
			cl_ntoh64(port_guid), lid_changed ? "new" : "",
			min_lid_ho, max_lid_ho);

		/*
		 * If we change LID of an old port, then we need to invalidate
		 * ucast cache.
		 */
		if (p_mgr->p_subn->first_time_master_sweep == FALSE && lid_changed &&
		    !p_port->is_new) {
			base_lid = osm_port_get_base_lid_ho(p_port);

			/* Report switch rebooted */
			if ((osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH) &&
			    (base_lid == 0 || base_lid == 0xffff)) {
				osm_report_notice_1312(p_mgr->p_subn, port_guid,
						       osm_node_get_node_guid(p_port->p_node),
						       osm_physp_get_port_num(p_port->p_physp));
				OSM_LOG(p_mgr->p_log, OSM_LOG_INFO,
					"Node 0x%016" PRIx64 ", Port 0x%016"
					PRIx64 " was rebooted\n", cl_ntoh64(port_guid),
					osm_node_get_node_guid_ho(p_port->p_node));
			} else {
				OSM_LOG(p_mgr->p_log,
					OSM_LOG_INFO,
					"Lid [%u,%u] changed for existing port 0x%016"
					 PRIx64 " \n",
					 min_lid_ho, max_lid_ho,
					 cl_ntoh64(port_guid));
			}

			if (p_mgr->p_subn->opt.use_ucast_cache) {
				p_mgr->sm->ucast_mgr.variables_invalidate_cache = TRUE;
			}
		}

		/*
		 * the proc returns the fact it sent a set port info.
		 * Note that a single PortInfo is sent per endport, even though planarized hosts may
		 * have multiple physical ports that need updating. That's because parameters like
		 * LID are copied to the whole aport when hosts receive PortInfo.
		 * On the SM side this copying is handled when the MAD is received successfully.
		 */
		if (lid_mgr_set_physp_pi(p_mgr, p_port, p_port->p_physp,
					 cl_hton16(min_lid_ho)))
			ret = -1;

		/* Save maximal used LID of each LID type to set later by LFTSplit MAD */
		if (osm_subn_is_nvlink_enabled(p_mgr->p_subn)) {
			uint8_t port_plane = osm_port_get_nvlink_plane(p_port);
			if (osm_node_is_gpu(p_port->p_node) &&
			    max_lid_ho > p_nvl_mgr->lid_range_top_per_plane[port_plane]) {
				p_nvl_mgr->lid_range_top_per_plane[port_plane] = max_lid_ho;
			} else if (!osm_node_is_gpu(p_port->p_node) &&
				   max_lid_ho > p_nvl_mgr->global_lid_range_top) {
				p_nvl_mgr->global_lid_range_top = max_lid_ho;
			}
		}
	}			/* all ports */

	/* store the guid to lid table in persistent db */
	osm_db_store(p_mgr->p_g2l, p_mgr->p_subn->opt.fsync_high_avail_files);
	/* the above loop handles all ports, including gpu ports */
	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn))
		osm_db_store(p_nvl_mgr->p_g2nvlid, p_mgr->p_subn->opt.fsync_high_avail_files);

	lids = cl_ptr_vector_get_size(&p_mgr->p_subn->port_lid_tbl);
	lids = lids ? lids - 1 : 0;
	/* take into consideration the greatest of configured FLID */
	max_flid = p_mgr->p_subn->opt.global_flid_end;
	lids = (max_flid > lids) ? max_flid : lids;

	if (p_mgr->p_subn->max_lid < lids)
		p_mgr->p_subn->max_lid = lids;

	CL_PLOCK_RELEASE(p_mgr->p_lock);

	/* ALIDs */
	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn))
		osm_lid_mgr_process_alids(p_mgr);

	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

ib_net16_t osm_lid_mgr_get_vport_lid(IN osm_lid_mgr_t * p_mgr,
				     IN osm_port_t * p_port,
				     IN osm_vport_t * p_vport)
{
	uint16_t min_lid, max_lid;
	ib_net16_t base_lid = osm_port_get_base_lid(p_port);

	if (p_vport->index) {
		lid_mgr_get_port_lid(p_mgr, p_port, p_vport, 0, &min_lid, &max_lid);
		if (min_lid) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Assigned lid %u "
				"for VPort 0x%" PRIx64 " (VPort %u of port 0x%"
				PRIx64 ")\n",
				min_lid, cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				osm_port_get_guid_ho(p_port));
			osm_vlid_mgr_set_vlid(&p_mgr->vlid_mgr, min_lid,
					      cl_ntoh16(base_lid));
			return cl_hton16(min_lid);
		} else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0318: "
				"Failed to assign lid for for VPort 0x%"
				PRIx64 " (VPort %u of port 0x%" PRIx64 ")\n",
				cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				osm_port_get_guid_ho(p_port));
			return 0;
		}
	} else
		return base_lid;
}

void osm_lid_mgr_clear_vport_lid(IN osm_lid_mgr_t * p_mgr,
				 IN osm_port_t * p_port,
				 IN osm_vport_t * p_vport,
				 IN boolean_t clear_guid2lid)
{
	uint64_t guid = cl_ntoh64(osm_vport_get_guid(p_vport));
	uint16_t min_lid, max_lid, lid;
	boolean_t guid2lid;

	if (osm_db_guid2lid_get(p_mgr->p_g2l, guid, &min_lid, &max_lid)) {
		/* Lid not found in guid2lid db*/
		guid2lid = FALSE;
		lid = cl_ntoh16(osm_vport_get_lid(p_vport));

		/*
		 * If clear_guid2lid, function is called without checking if
		 * already exists, in order to avoid double search. Hence missing
		 * LID is not an error.
		 */
		if (lid && !clear_guid2lid)
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0319: "
				"VPort 0x%" PRIx64 " (VPort %u of port 0x%"
				PRIx64 ") dont have a lid in guid2lid db, "
				"using lid %u from vport info\n",
				cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				osm_port_get_guid_ho(p_port), lid);
		else {
			if (clear_guid2lid)
				return;

			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031A: "
				"No lid for VPort 0x%" PRIx64 " (VPort %u of "
				"port 0x%" PRIx64 ") in guid2lid db and vport "
				"info, Nothing to clear\n",
				cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				osm_port_get_guid_ho(p_port));
			return;
		}
	} else {
		/* Lid found in guid2lid db*/
		guid2lid = TRUE;
		lid = min_lid;
	}

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Clearing lid %u "
		"for VPort 0x%" PRIx64 " (VPort %u of port 0x%"
		PRIx64 "), Clear Guid2Lid: %s\n",
		lid, guid, p_vport->index,
		osm_port_get_guid_ho(p_port),
		clear_guid2lid ? "Yes" : "No");

	osm_vlid_mgr_clear_vlid(&p_mgr->vlid_mgr, lid);

	if (guid2lid && clear_guid2lid) {
		osm_db_guid2lid_delete(p_mgr->p_g2l, guid);
		lid_mgr_clear_used_lid(p_mgr, lid);
	}
}

boolean_t osm_lid_mgr_is_flid(IN osm_lid_mgr_t * p_mgr, IN uint16_t lid)
{
	/* check if this LID falls into the FLID global range */
	return (p_mgr->p_subn->opt.global_flid_start <= lid &&
		lid <= p_mgr->p_subn->opt.global_flid_end);
}

osm_lid_mgr_lid_type_t osm_lid_mgr_get_lid_type_by_predefined_range(IN osm_lid_mgr_t * p_mgr,
								    IN uint16_t lid)
{
	/*
	 * Note that ALIDs and FLIDs do not co-exist.
	 * FLIDs are used for topologies with routers,
	 * while ALIDs are used for NVLink configuration, which does not include routers.
	 */
	if (osm_lid_mgr_is_flid(p_mgr, lid))
		return OSM_LID_TYPE_FLID;

	if (osm_subn_is_nvlink_enabled(p_mgr->p_subn) && lid_mgr_is_alid(p_mgr, lid))
		return OSM_LID_TYPE_ALID;

	return OSM_LID_TYPE_LID;
}
