/*
 * Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2018-2020 Mellanox Technologies LTD. ALL RIGHTS RESERVED.
 * See file LICENSE for terms.
 */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PARALLEL_PORT_GROUPS_CALCULATOR_CPP
#include "osm_parallel_port_groups_calculator.hpp"
#include "osm_armgr.hpp"
#include "osm_ar_algorithm.hpp"

OSMParallelPortGroupsCalculator::~OSMParallelPortGroupsCalculator()
{
	WaitForTasks();

	OSM_AR_LOG(m_osm_log_, OSM_LOG_DEBUG,
		   "OSMParallelPortGroupsCalculator destroyed\n");
}

void
OSMParallelPortGroupsCalculator::CalculatePortGroupsTree(
	TreeAlgorithmContext *context)
{
	m_tree_algorithm_context_ = context;

	/*
	 * Go through all the switches and create their port groups.
	 * In this algorithm ports groups are created out of minimal hops
	 * links - each lid will receive a group containing all the ports
	 * with minimal hops to this lid.
	 * So here's what we're doing here:
	 *   - for each switch
	 *       - for each port of the switch (port_num)
	 *           - calculate group bitmask for this lid
	 *           - asisign groups according to device limitations
	 */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
	m_sw_db_->m_sw_map.begin();
	for (; sw_node_it != m_sw_db_->m_sw_map.end(); ++sw_node_it) {

		OSM_AR_LOG(m_osm_log_, OSM_LOG_DEBUG,
			   "OSMParallelPortGroupsCalculator AddTask \n");

		AddTaskToThreadPool(
			m_thread_pool_,
			&sw_node_it->second.m_calculate_port_groups_tree_task);
	}

	WaitForTasks();

	if (!osm_log_is_active(m_osm_log_,OSM_LOG_ROUTING))
		return;

	sw_node_it = m_sw_db_->m_sw_map.begin();
	for (; sw_node_it != m_sw_db_->m_sw_map.end(); ++sw_node_it) {

		m_ar_mgr_->ARDumpSWSettings(sw_node_it->second);
	}
}

int
OSMParallelPortGroupsCalculator::CalculatePortGroups(
	AdaptiveRoutingAlgorithm *ar_algorithm)
{
	m_sub_task_ = SUB_TASK_CALCULATE_PORT_GROUPS;
	m_lid_mapping_.Clear();
	int rc = ar_algorithm->BuildLidMapping(m_lid_mapping_);
	if (rc) {
		OSM_AR_LOG_RETURN(m_osm_log_, rc);
	}

	/* 
	 * Make the allocation of group ids persistent.
	 * go over all sw_db_entries, for each LEAF switch if RN are enabled
	 * call AllocateSwARGroup
	 */
	ar_algorithm->AllocatePersistentARGroupIDs();

	/*
	 * Go through all the switches and create their port groups
	 * According to the given algorithm
	 */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_db_->m_sw_map.begin();
	for (; sw_node_it != m_sw_db_->m_sw_map.end(); ++sw_node_it) {

		OSM_AR_LOG(m_osm_log_, OSM_LOG_DEBUG,
			   "OSMParallelPortGroupsCalculator AddTask \n");

		AddTaskToThreadPool(
			m_thread_pool_,
			ar_algorithm->GetCalculateTask(sw_node_it->second));
	}

	WaitForTasks();

	if (m_ar_mgr_->IsInPermanentError()) {
		OSM_AR_LOG_RETURN(m_osm_log_, -1);
	}

	OSM_AR_LOG_RETURN(m_osm_log_, 0);
}

int
OSMParallelPortGroupsCalculator::CalculateStaticRouteInfo(
	AdaptiveRoutingAlgorithm *ar_algorithm,
	bool rebuild_route_info)
{
	m_sub_task_ = rebuild_route_info ? SUB_TASK_RECALCULATE_STATIC_ROUTE_INFO :
		SUB_TASK_CALCULATE_STATIC_ROUTE_INFO;

	/*
	 * Go through all the switches and create their port groups
	 * According to the given algorithm
	 */
	map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
	m_sw_db_->m_sw_map.begin();
	for (; sw_node_it != m_sw_db_->m_sw_map.end(); ++sw_node_it) {

		OSM_AR_LOG(m_osm_log_, OSM_LOG_DEBUG,
			   "OSMParallelPortGroupsCalculator AddTask \n");

		AddTaskToThreadPool(
			m_thread_pool_,
			ar_algorithm->GetCalculateTask(sw_node_it->second));
	}

	WaitForTasks();

	if (m_ar_mgr_->IsInPermanentError()) {
		OSM_AR_LOG_RETURN(m_osm_log_, -1);
	}

	OSM_AR_LOG_RETURN(m_osm_log_, 0);
}
