/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <dlfcn.h>
#include <stdint.h>
#include <linux/limits.h>
#include <complib/cl_passivelock.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PLUGIN_C
#include <opensm/osm_version.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_plugin.h>

#define ERROR01		"01: Failed to open plugin \'%s\' : \'%s\'\n"
#define ERROR02		"02: no %s() or %s() method.\n"
#define ERROR03		"03: version mismatch - [%s] vs [%s]\n"
#define ERROR04		"04: required symbol [%s] not found\n"
#define ERROR05		"05: memory allocation failed\n"
#define ERROR06		"06: no symbol name specified\n"
#define ERROR07		"07: no name specified\n"

#define	PLUGIN_INVALID_NAME(P_NAME)					\
	(P_NAME == NULL) || (P_NAME[0] == '\0')

#define PLUGIN_ERROR(P_OSM, FORMAT, ...)				\
        do {								\
                _plugin_error(P_OSM, FORMAT, ## __VA_ARGS__);		\
        } while (0)

#define PLUGIN_MAX_ARGS 100

/* ---------------------------------------------------------------------------------------------- */

/*
 * Plugin definition.
 */
typedef struct osm_plugin {
	cl_list_item_t		list_item;
	void			*data;
	void			*handle;
	void			*(*init)(void *);
	void			(*fini)(void *);
	int			ref_count;
	char			name[NAME_MAX];
	char			file[NAME_MAX];
} osm_plugin_t;

/* ---------------------------------------------------------------------------------------------- */

static int		osm_plugin_inited = 0;
static cl_qlist_t	osm_plugin_list;
static cl_plock_t	osm_plugin_lock;

/* ---------------------------------------------------------------------------------------------- */
/**
 * Internal functions
 */
static void _plugin_error(IN osm_opensm_t *p_osm,
			  IN const char *format,
			  IN...)
{
	osm_subn_t	*p_subn = &p_osm->subn;
	osm_log_t	*p_log = &p_subn->p_osm->log;
	va_list		args;
	char		buffer[1024];

	va_start(args, format);
	vsnprintf(buffer, sizeof(buffer), format, args);
	OSM_LOG(p_log, OSM_LOG_ERROR, "ERR %02X%s\n", FILE_ID, buffer);
	va_end(args);
}

static void __plugin_init(void)
{
	if (osm_plugin_inited == 0) {
		cl_qlist_init(&osm_plugin_list);
		cl_plock_init(&osm_plugin_lock);

		osm_plugin_inited++;
	}

	return;
}

static osm_plugin_t *__plugin_load(IN osm_opensm_t *p_osm,
				   IN const char *name,
				   IN void *context)
{
	char		init_name[NAME_MAX];
	char		fini_name[NAME_MAX];
	char		version_name[NAME_MAX];
	char		library_name[NAME_MAX];
	char		**version;
	void		*handle;
	void		*(*init)(void *);
	void		(*fini)(void *);
	osm_plugin_t	*p_plugin = NULL;

	/*
	 * Verify plugin name and setup names.
	 */
	if (PLUGIN_INVALID_NAME(name)) {
		PLUGIN_ERROR(p_osm, ERROR01, name, dlerror());
		return NULL;
	}

	snprintf(init_name, sizeof(init_name), "%s_init", name);
	snprintf(fini_name, sizeof(fini_name), "%s_fini", name);
	snprintf(version_name, sizeof(library_name), "%s_version", name);
	snprintf(library_name, sizeof(library_name), "lib%s.so", name);

	/*
  	 * See if the named library exists.
  	 */
	if ((handle = dlopen(library_name, RTLD_LAZY)) == NULL) {
		PLUGIN_ERROR(p_osm, ERROR01, name, dlerror());
		return NULL;
	}

	/*
	 * Check that the version is compatible.
	 */
	if ((version = (char **)dlsym(handle, version_name)) == NULL) {
		PLUGIN_ERROR(p_osm, ERROR04, name, version_name);
		dlclose(handle);
		return NULL;
	}

	if (strcmp(*version, OSM_PLUGIN_VERSION) != 0) {
		PLUGIN_ERROR(p_osm, ERROR03, name, OSM_PLUGIN_VERSION, *version);
		dlclose(handle);
		return NULL;
	}

	/*
  	 * Find the 'init' and 'fini' functions.
  	 */
	init = (void *)dlsym(handle, init_name);
	fini = (void *)dlsym(handle, fini_name);

	if ((init == NULL) || (fini == NULL)) {
		PLUGIN_ERROR(p_osm, ERROR02, name, init_name, fini_name);
		dlclose(handle);
		return NULL;
	}

	/*
	 * Create the plugin and fill in the fields.
	 */
	if ((p_plugin = calloc(1, sizeof(*p_plugin))) == NULL) {
		PLUGIN_ERROR(p_osm, ERROR05, name);
		dlclose(handle);
		return NULL;
	}

	p_plugin->data = (*init)(context);
	p_plugin->handle = handle;
	p_plugin->init = init;
	p_plugin->fini = fini;
	p_plugin->ref_count = 1;
	snprintf(p_plugin->name, sizeof(p_plugin->name), "%s", name);
	snprintf(p_plugin->file, sizeof(p_plugin->file), "%s", library_name);

	cl_qlist_insert_tail(&osm_plugin_list, (cl_list_item_t *)p_plugin);
	return p_plugin;
}

static void __plugin_unload(IN osm_plugin_t *p_plugin,
			    IN void *context)
{

	/*
	 * If more than one open instance, just decrement ref count and return.
	 */
	if (p_plugin->ref_count > 1) {
		p_plugin->ref_count--;
		return;
	}

	/*
	 * No more users - cleanup the plugin.
	 */
	(*p_plugin->fini)(context);
	dlclose(p_plugin->handle);
	cl_qlist_remove_item(&osm_plugin_list, (cl_list_item_t *)p_plugin);
	free(p_plugin);

	return;
}

static osm_plugin_t *__plugin_find(IN const void *name)
{
	osm_plugin_t		*p_plugin;

	foreach_plugin(p_plugin) {
		if (strcmp(p_plugin->name, name) == 0) {
			return p_plugin;
		}
	}

	return NULL;
}

/* ---------------------------------------------------------------------------------------------- */
/**
 * System functions
 */
void osm_plugin_init(IN osm_opensm_t *p_osm)
{
	__plugin_init();
}

void *osm_plugin_find(IN char *name)
{
	return (void *)__plugin_find(name);
}

osm_plugin_t *osm_plugin_first(void)
{
	if (!cl_is_qlist_empty(&osm_plugin_list)) {
		return (osm_plugin_t *)cl_qlist_head(&osm_plugin_list);
	}

	return NULL;
}

osm_plugin_t *osm_plugin_next(IN osm_plugin_t *p_plugin)
{
	cl_list_item_t	*p_item = cl_qlist_next((cl_list_item_t *)p_plugin);

	return (p_item != cl_qlist_end(&osm_plugin_list)) ? (osm_plugin_t *)p_item : NULL;
}

/* ---------------------------------------------------------------------------------------------- */
/**
 * API functions
 */
void *osm_plugin_load(IN osm_opensm_t *p_osm,
		      IN const char *plugin_name,
		      IN void *context)
{
	osm_plugin_t		*p_plugin;

	/*
	 * Check for proper parameters.
	 */
	if (PLUGIN_INVALID_NAME(plugin_name)) {
		PLUGIN_ERROR(p_osm, ERROR07);
		return NULL;
	}

	CL_PLOCK_EXCL_ACQUIRE(&osm_plugin_lock);

	/*
	 * If this is a new plugin, create it.  Otherwise, increment the reference count.
	 */
	if ((p_plugin = __plugin_find(plugin_name)) == NULL) {
		p_plugin = __plugin_load(p_osm, plugin_name, context);
	} else {
		p_plugin->ref_count++;
	}

	CL_PLOCK_RELEASE(&osm_plugin_lock);
	return (p_plugin) ? p_plugin->data : NULL;
}

void osm_plugin_unload(IN osm_opensm_t *p_osm,
		       IN const char *plugin_name,
		       IN void *context)
{
	osm_plugin_t	*p_plugin;

	/*
	 * Check for proper parameters.
	 */
	if (PLUGIN_INVALID_NAME(plugin_name)) {
		PLUGIN_ERROR(p_osm, ERROR07);
		return;
	}

	CL_PLOCK_EXCL_ACQUIRE(&osm_plugin_lock);

	/*
	 * Find the named plugin and clean it up.
	 */
	if ((p_plugin = __plugin_find(plugin_name)) != NULL) {
		__plugin_unload(p_plugin, context);
	}

	CL_PLOCK_RELEASE(&osm_plugin_lock);
}

void *osm_plugin_resolve(IN osm_opensm_t *p_osm,
			 IN const char *plugin_name,
			 IN const char *symbol_name)
{
        char		resolve_name[NAME_MAX];
	void		*symbol = NULL;
	osm_plugin_t	*p_plugin;

	/*
	 * Check for proper parameters.
	 */
	if (PLUGIN_INVALID_NAME(plugin_name)) {
		PLUGIN_ERROR(p_osm, ERROR07);
		return NULL;
	}

	if (PLUGIN_INVALID_NAME(symbol_name)) {
		PLUGIN_ERROR(p_osm, ERROR06);
		return NULL;
	}

	snprintf(resolve_name, sizeof(resolve_name), "%s_%s", plugin_name, symbol_name);

	CL_PLOCK_EXCL_ACQUIRE(&osm_plugin_lock);

	/*
	 * Find the plugin and then find the symbol.
	 */
	if ((p_plugin = __plugin_find(plugin_name)) != NULL) {
		symbol = (void *)dlsym(p_plugin->handle, resolve_name);
	}

	CL_PLOCK_RELEASE(&osm_plugin_lock);
	return symbol;
}

void *osm_plugin_data(IN void *p_plugin)
{
	return ((osm_plugin_t *)p_plugin)->data;
}

char *osm_plugin_name(IN void *p_plugin)
{
	return ((osm_plugin_t *)p_plugin)->name;
}

char **osm_plugin_get_args_array(IN osm_opensm_t *p_osm, IN const char *plugin_name, OUT int *p_argc)
{
	char *options_start = NULL;
	char *options_str = NULL;
	char *token = NULL;
	char **plugin_args = NULL;
	int argc = 0;

	if (!p_osm->subn.opt.plugin_options)
		return plugin_args;

	options_start = strstr(p_osm->subn.opt.plugin_options, plugin_name);
	if (!options_start)
		return plugin_args;

	options_str = strdup(options_start + strlen(plugin_name));
	if (!options_str)
		goto Exit;

	/* Allocate arguments array in argv format (first and last elements reserved for plugin name and NULL) */
	plugin_args = calloc(PLUGIN_MAX_ARGS + 2, sizeof(char *));
	if (!plugin_args)
		goto Exit;

	plugin_args[argc] = strdup(options_start);
	argc++;

	/*
	 * Each command line option is of the format key, value pairs.
	 * The two allowed formats are:
	 * --key value
	 * -k value
	 */
	token = strtok(options_str, " \t\n");
	while (token) {
		/* Expecting -option/--option */
		if (*token != '-')
			break;

		/* Key */
		plugin_args[argc] = strdup(token);
		argc++;

		/* value */
		token = strtok(NULL, " \t\n");
		if (!token) {
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR A508: %s plugin options have incorrect format: %s, "
				"No value was given for the option %s\n",
				plugin_name, p_osm->subn.opt.plugin_options, plugin_args[argc - 1]);
			break;
		}

		plugin_args[argc] = strdup(token);
		argc++;
		token = strtok(NULL, " \t\n");
	}

	plugin_args[argc] = NULL;
	*p_argc = argc;

Exit:
	free(options_str);

	return plugin_args;
}

void osm_plugin_release_args_array(IN char **plugin_args)
{
	int i = 0;

	if (!plugin_args)
		return;

	for (i = 0; plugin_args[i] != NULL; i++) {
		free(plugin_args[i]);
	}

	free(plugin_args);
}
