/*
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2015 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_passivelock.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTER_INFO_RCV_C
#include <opensm/osm_log.h>
#include <opensm/osm_router.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_inform.h>

static void ri_rcv_process_set(IN osm_sm_t * sm, IN osm_port_t * p_port,
			       IN ib_router_info_t *p_ri_old,
			       IN osm_ri_context_t *p_ri_context,
			       IN ib_net64_t trans_id)
{
	osm_madw_context_t context;
	osm_physp_t *p_physp = p_port->p_physp;
	uint8_t payload[IB_SMP_DATA_SIZE] = {0};
	ib_api_status_t status;
	ib_router_info_t ri;

	OSM_LOG_ENTER(sm->p_log);

	ri = *p_ri_old;
	if (ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
					   IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT)) {
		ib_router_info_set_global_router_lid_start(&ri, sm->p_subn->opt.global_flid_start);
		ib_router_info_set_global_router_lid_end(&ri, sm->p_subn->opt.global_flid_end);
		ib_router_info_set_local_router_lid_start(&ri, sm->p_subn->opt.local_flid_start);
		ib_router_info_set_local_router_lid_end(&ri, sm->p_subn->opt.local_flid_end);
	}

	/* if we are in set, then the change bit is 1, set with it */
	memcpy(payload, &ri, sizeof(ib_router_info_t));

	context.ri_context.port_guid = p_physp->port_guid;
	context.ri_context.set_method = TRUE;
	context.ri_context.sweep_type = p_ri_context->sweep_type;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Set Router Info Port GUID 0x%016" PRIx64
		",in response to TID 0x%" PRIx64 "\n",
		cl_ntoh64(p_port->guid), cl_ntoh64(trans_id));

	osm_dump_router_info(sm->p_log,
			     osm_node_get_node_guid(p_port->p_node),
                             p_port->guid, p_port->p_physp->port_num,
                             p_ri_old, FILE_ID, OSM_LOG_DEBUG);

	status = osm_req_set(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
			     payload,
			     sizeof(payload), IB_MAD_ATTR_ROUTER_INFO, 0,
			     FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F08: "
			"Failed to Send RouterInfo set for 0x%" PRIx64
			" (status = %s)\n",
			cl_ntoh64(p_port->guid), ib_get_err_str(status));
	OSM_LOG_EXIT(sm->p_log);
}

static void router_lid_info_table_get(IN osm_sm_t * sm,
				      IN osm_router_t * p_router,
				      IN osm_physp_t *p_physp,
				      IN ib_net64_t m_key,
				      IN osm_madw_context_t * p_context)
{
	ib_api_status_t    status = IB_SUCCESS;
	uint32_t num_of_blocks, block_idx;
	ib_rtr_router_lid_tbl_block_t *rtr_lid_blocks;

	OSM_LOG_ENTER(sm->p_log);

	num_of_blocks = RTR_TBL_NUM_BLOCKS(
			p_router->router_info.adjacent_subnets_router_lid_info_table_top,
			IB_ROUTER_ADJ_SUBNETS_LID_INFO_ELEMENTS);

	if (p_router->adj_router_lid_info_top <
		p_router->router_info.adjacent_subnets_router_lid_info_table_top) {

		rtr_lid_blocks = realloc(p_router->adj_router_lid_info,
					 num_of_blocks * sizeof(ib_rtr_adj_subnets_lid_info_block_t));
		if (!rtr_lid_blocks)
			return;
		p_router->adj_router_lid_info = rtr_lid_blocks;
		p_router->adj_router_lid_info_top =
			p_router->router_info.adjacent_subnets_router_lid_info_table_top;
	}

	for (block_idx = 0; block_idx < num_of_blocks; block_idx++) {
		status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
				     IB_MAD_ATTR_RTR_ADJ_ROUTER_LID_INFO,
				     cl_hton32(block_idx), FALSE,
				     m_key, 0,
				     CL_DISP_MSGID_NONE, p_context);
		if (status != IB_SUCCESS)
			/* continue despite error */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F0A: "
				"Failed to send adjacent router lid info "
				"get request (status = %s) for block %u\n",
				ib_get_err_str(status), block_idx);
	}

	OSM_LOG_EXIT(sm->p_log);
}

static void router_lid_table_get(IN osm_sm_t * sm,
				 IN osm_router_t * p_router,
				 IN osm_physp_t *p_physp,
				 IN ib_net64_t m_key,
				 IN osm_madw_context_t * p_context)
{
	ib_api_status_t    status = IB_SUCCESS;
	uint32_t block_idx, start_idx, end_idx;

	OSM_LOG_ENTER(sm->p_log);

	/* query FLIDs according to the global range */
	start_idx = RTR_TBL_ENTRY_IDX_TO_BLOCK(p_router->global_flid_start,
					       IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);
	end_idx = RTR_TBL_ENTRY_IDX_TO_BLOCK(p_router->global_flid_end,
					     IB_ROUTER_LID_TBL_BLOCK_LIDS_NUM);

	for (block_idx = start_idx; block_idx <= end_idx; block_idx++) {
		status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
				     IB_MAD_ATTR_RTR_ROUTER_LID_TBL,
				     cl_hton32(block_idx), FALSE,
				     m_key, 0,
				     CL_DISP_MSGID_NONE, p_context);
		if (status != IB_SUCCESS)
			/* continue despite error */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F0B: "
				"Failed to send router lid table "
				"get request (status = %s) for block %u\n",
				ib_get_err_str(status), block_idx);
	}
	OSM_LOG_EXIT(sm->p_log);
}

static void ri_rcv_router_table_get(IN osm_sm_t * sm,
				    IN osm_router_t * p_router)
{
	ib_api_status_t    status = IB_SUCCESS;
	osm_physp_t       *p_physp;
	osm_madw_context_t context;
	uint32_t           num_of_blocks, block_idx;
	ib_net64_t         m_key;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);
	OSM_ASSERT(p_router);

	memset(&context, 0, sizeof(context));
	p_physp = p_router->p_port->p_physp;
	context.ri_context.port_guid = p_physp->port_guid;
	context.ri_context.set_method = FALSE;
	/* remove previous MADs - router table */
	osm_router_tbl_delete(&p_router->adj_router_tbl);

	num_of_blocks =
		RTR_TBL_NUM_BLOCKS(p_router->router_info.adjacent_site_local_subnets_tbl_top,
				   IB_ROUTER_ELEMENTS_IN_ADJACENT_BLOCK);
	m_key = ib_port_info_get_m_key(&p_physp->port_info);

	for (block_idx = 0; block_idx < num_of_blocks; block_idx++) {
		status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
				     IB_MAD_ATTR_RTR_ADJ_SITE_LOCAL_TBL,
				     cl_hton32(block_idx), FALSE,
				     m_key, 0,
				     CL_DISP_MSGID_NONE, &context);
		if (status != IB_SUCCESS)
			/* continue despite error */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F04: "
				"Failed to send adjacent RouterTable "
				"get request (status = %s) for block %u\n",
				ib_get_err_str(status), block_idx);
	}

	osm_router_tbl_delete(&p_router->next_hop_router_tbl);

	num_of_blocks = RTR_TBL_NUM_BLOCKS(cl_ntoh32(p_router->router_info.next_hop_table_top),
					   IB_ROUTER_ELEMENTS_IN_NEXT_HOP_BLOCK);

	for (block_idx = 0; block_idx < num_of_blocks; block_idx++) {
		status = osm_req_get(sm, osm_physp_get_dr_path_ptr(sm, p_physp), p_physp,
				     IB_MAD_ATTR_RTR_NEXT_HOP_TBL,
				     cl_hton32(block_idx), FALSE,
				     m_key, 0,
				     CL_DISP_MSGID_NONE, &context);
		if (status != IB_SUCCESS)
			/* continue despite error */
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F07: "
				"Failed to send next hop RouterTable "
				"get request (status = %s) for block %u\n",
				ib_get_err_str(status), block_idx);
	}

	/* query only if the FLIDs feature is enabled and supported for this router */
	if (osm_subn_is_flids_enabled(sm->p_subn) &&
	    ib_general_info_is_cap_enabled(&p_router->p_port->p_node->general_info,
					   IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT) &&
	    osm_router_is_router_flid_enabled(sm->p_subn, p_router)) {
		router_lid_info_table_get(sm, p_router, p_physp, m_key, &context);

		router_lid_table_get(sm, p_router, p_physp, m_key, &context);
	}

	OSM_LOG_EXIT(sm->p_log);
}

static int osm_router_validate_router_info(osm_sm_t *sm, ib_router_info_t *p_ri,
					   ib_net64_t port_guid)
{
	int result = 0;

	if (cl_ntoh32(p_ri->next_hop_table_top) >
	    cl_ntoh32(p_ri->next_hop_table_cap)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F06: "
			" Router Info for router 0x%" PRIx64 " received with "
			"illegal next hop top value (%u), cap = %u\n",
			cl_ntoh64(port_guid), cl_ntoh32(p_ri->next_hop_table_top),
			cl_ntoh32(p_ri->next_hop_table_cap));
		result = 1;
	}

	if (p_ri->adjacent_site_local_subnets_tbl_top >
	    p_ri->adjacent_site_local_subnets_tbl_cap) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F09: "
			" Router Info for router 0x%" PRIx64 " received with "
			"illegal adjacent site local top value (%u), cap = %u\n",
			cl_ntoh64(port_guid), p_ri->adjacent_site_local_subnets_tbl_top,
			p_ri->adjacent_site_local_subnets_tbl_cap);
		result = 1;
	}
	return result;
}

static inline
boolean_t is_router_ranges_configured(IN ib_router_info_t *p_ri, IN osm_router_t *p_router)
{
	return (ib_router_info_get_global_router_lid_start(p_ri) == p_router->global_flid_start &&
		ib_router_info_get_global_router_lid_end(p_ri) == p_router->global_flid_end &&
		ib_router_info_get_local_router_lid_start(p_ri) == p_router->local_flid_start &&
		ib_router_info_get_local_router_lid_end(p_ri) == p_router->local_flid_end);
}

static const char* ri_get_sweep_type_name(IN const osm_context_sweep_type_t sweep_type)
{
	switch (sweep_type) {
	case OSM_SWEEP_CONTEXT_NONE:
		return "None";
	case OSM_SWEEP_CONTEXT_LIGHT:
		return "Light";
	case OSM_SWEEP_CONTEXT_HEAVY:
		return "Heavy";
	case OSM_SWEEP_CONTEXT_FLID:
		return "FLID";
	default:
		return "Unknown";
	}
}

void osm_ri_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_smp_t *p_smp;
	ib_net64_t port_guid;
	osm_ri_context_t *p_context;
	osm_router_t *p_router;
	ib_router_info_t *p_ri;
	ib_net16_t error_status;
	osm_router_mgr_t * p_mgr = &sm->router_mgr;

	OSM_ASSERT(sm);

	OSM_LOG_ENTER(sm->p_log);

	OSM_ASSERT(p_madw);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_ri = ib_smp_get_payload_ptr(p_smp);
	p_context = osm_madw_get_ri_context_ptr(p_madw);
	port_guid = p_context->port_guid;

	OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
		"Router GUID 0x%016" PRIx64 ", TID 0x%" PRIx64
		", received RouterInfo %s in %s sweep with change bits 0x%x.\n",
		cl_ntoh64(port_guid), cl_ntoh64(p_smp->trans_id),
		p_context->set_method ? "SET" : "GET",
		ri_get_sweep_type_name(p_context->sweep_type),
		cl_ntoh16(p_ri->table_changes_bitmask));

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	p_router = (osm_router_t *)cl_qmap_get(&sm->p_subn->rtr_guid_tbl,
					       port_guid);
	/* should exist from Node Info callback */
        if (PF(p_router ==
		(osm_router_t *)cl_qmap_end(&sm->p_subn->rtr_guid_tbl))) {
                OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F05: "
                        "Failed to find router object for "
                        "port GUID:0x%016" PRIx64 " in router table\n",
                        cl_ntoh64(port_guid));
                goto Exit;
        }

	error_status = ib_smp_get_status(p_smp);
	if (error_status) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		if (p_context->set_method) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR 6F0C: "
				"Router 0x%016" PRIx64 " responded with error status 0x%x "
				"for RouterInfo set. Check router FLIDs configuration.\n",
				cl_ntoh64(port_guid),
				cl_ntoh16(ib_smp_get_status(p_smp)));
			osm_report_notice_1313(sm->p_subn,
					       SM_VS_TRAP_RTR_LID_SUBNETS_INCONSISTENT_FLID_RANGE);
			p_router->is_router_info_set_error = TRUE;
		}
		goto Exit;
	} else if (p_context->set_method && p_router->is_router_info_set_error) {
		/*
		 * If SM missed the router reboot,
		 * clear error on first successful set of router info.
		 */
		p_router->is_router_info_set_error = FALSE;
	}

	if (osm_router_validate_router_info(sm, p_ri, port_guid))
		goto Exit;

	osm_dump_router_info(sm->p_log,
			     osm_node_get_node_guid(p_router->p_port->p_node),
                             port_guid, p_router->p_port->p_physp->port_num,
                             p_ri, FILE_ID, OSM_LOG_DEBUG);

	osm_router_set_router_info(p_router, p_ri);

	/*
	 * sweep time set to NONE is used for RouterInfo sent to set pFRN
	 * In this case, we just update local copy of the MAD
	 */
	if (p_context->sweep_type == OSM_SWEEP_CONTEXT_NONE) {
		goto Exit;
	}

	/* GET */
	if (!p_context->set_method) {
		p_mgr->ri_routers_changes_bitmask |= cl_ntoh16(p_ri->table_changes_bitmask);
		/* if any of the tables has changed */
		if (p_ri->table_changes_bitmask ||
                    (ib_general_info_is_cap_enabled(&p_router->p_port->p_node->general_info,
                                                    IB_GENERAL_INFO_IS_ROUTER_LID_SUPPORTED_BIT) &&
                     !is_router_ranges_configured(p_ri, p_router))) {

			/* Check if any other change bits besides flids is set in router info */
			if (cl_ntoh16(p_ri->table_changes_bitmask) & (~IB_ROUTER_RLTC)) {
				if (p_context->sweep_type == OSM_SWEEP_CONTEXT_HEAVY) {
					ri_rcv_process_set(sm, p_router->p_port, p_ri,
							   p_context, p_smp->trans_id);
				} else {
					/*
					 * In light sweep or FLID sweep just trigger heavy sweep and let it
					 * reset the bit.
					 */
					sm->p_subn->force_heavy_sweep = TRUE;
					osm_sm_signal(sm, OSM_SIGNAL_SWEEP);
				}

				/*
				 * Invalidate cache when there is a change in the router other than
				 * router LID table.
				 */
				if (ib_router_info_is_tbl_change_set(p_ri, IB_ROUTER_RLTC)) {
					sm->ucast_mgr.variables_invalidate_cache = TRUE;
					OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
						"Router port GUID 0x%016" PRIx64 " has a change "
						"in the router lid table, invalidating cache.\n",
						cl_ntoh64(p_router->p_port->guid));
				}
			}

			/* If only FLID bits have been set in Heavy or FLID sweep reset the bit. */
			else if (p_context->sweep_type == OSM_SWEEP_CONTEXT_HEAVY ||
				 p_context->sweep_type == OSM_SWEEP_CONTEXT_FLID) {
				ri_rcv_process_set(sm, p_router->p_port, p_ri,
						   p_context, p_smp->trans_id);
			}

		/* If none of the tables have been changed, in Heavy or FLID sweep send get */
		} else if (p_context->sweep_type == OSM_SWEEP_CONTEXT_HEAVY ||
			   p_context->sweep_type == OSM_SWEEP_CONTEXT_FLID) {
			ri_rcv_router_table_get(sm, p_router);
		}

	/* SET */
	} else {
		/* send router info before reset of change bits in set, to check change bits */
		ri_rcv_router_table_get(sm, p_router);
	}

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);

	OSM_LOG_EXIT(sm->p_log);
}
