/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of TENANT virtualization.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <iba/ib_types.h>
#include <opensm/osm_base.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_TENANT_MGR_C
#include <opensm/osm_tenant_mgr.h>

extern unsigned osm_match_ipv6_snm_mgid(ib_gid_t * mgid);

/* -------------------------------------------------------------------------------------------------
 *
 * Internal tenant db functions
 */

int tenant_guidlist_init(IN osm_tenant_guid_list_t *p_list)
{
	return cl_hashmap_init_uint64(p_list);
}

void tenant_guidlist_destroy(IN osm_tenant_guid_list_t *p_list)
{
	cl_hashmap_destroy(p_list);
}

int tenant_list_init(IN osm_tenant_list_t *p_list)
{
	return cl_hashmap_init_uint64(p_list);
}

void tenant_list_free(cl_qhashmap_item_t *p_item, void *context) {
	free((void *)p_item->value);
}

void tenant_list_destroy(IN osm_tenant_list_t *p_list)
{
	cl_hashmap_apply_func(p_list, tenant_list_free, NULL);
	cl_hashmap_destroy(p_list);
}

static osm_tenant_database_t *tenant_mgr_guid_map_alloc(IN osm_subn_t *p_subn)
{
	osm_tenant_database_t *p_tenant_db = NULL;

	if ((p_tenant_db = calloc(1, sizeof(osm_tenant_database_t))) != NULL) {
		tenant_guidlist_init(&p_tenant_db->virtual_guids);
		tenant_guidlist_init(&p_tenant_db->physical_guids);
		tenant_list_init(&p_tenant_db->tenant_list);
	} else {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AE05: allocation error\n");
	}

	return p_tenant_db;
}

static void tenant_db_free(IN osm_tenant_database_t *p_tenant_db)
{
	if (p_tenant_db != NULL) {
		tenant_guidlist_destroy(&p_tenant_db->virtual_guids);
		tenant_guidlist_destroy(&p_tenant_db->physical_guids);
		tenant_list_destroy(&p_tenant_db->tenant_list);
		free(p_tenant_db);
	}
}

static void tenant_mgr_fatal_error(IN osm_tenant_mgr_t *p_tenant_mgr)
{
	osm_opensm_set_exit_flag(p_tenant_mgr->p_subn->p_osm);
	OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_ERROR, "ERR AE04: fatal error - exiting\n");
}

/* -------------------------------------------------------------------------------------------------
 *
 * Parser interface.
 */
typedef struct yy_buffer_state *YY_BUFFER_STATE;

extern void	__tenant_parser_set_in(FILE *);
extern void	__tenant_parser_restart(FILE *);
extern void	__tenant_parser_set_lineno(int);
extern void	__tenant_parser__flush_buffer(FILE *);
extern void	__tenant_parser__delete_buffer(YY_BUFFER_STATE);
extern int	__tenant_parser_parse(osm_tenant_parser_t *);
extern int	__tenant_parser_lex_destroy (void);


static int tenant_parse_file(IN osm_tenant_mgr_t *p_tenant_mgr, IN osm_tenant_database_t *p_tenant_db, FILE *std_tenant)
{
	int			status = -1;
	osm_tenant_parser_t	parser;
	struct timeval		start, stop;
	float			elapsed_time;

	memset((void *)&parser, 0, sizeof(parser));
	parser.p_log = p_tenant_mgr->p_log;
	parser.p_tenant_db = p_tenant_db;

	/*
	 * Parse the file.
	 */
	__tenant_parser_restart(std_tenant);
	__tenant_parser_set_in(std_tenant);
	__tenant_parser__flush_buffer(NULL);
	__tenant_parser_set_lineno(1);

	gettimeofday(&start, NULL);
	status = __tenant_parser_parse(&parser);
	gettimeofday(&stop, NULL);

	__tenant_parser_lex_destroy();

	if (status == 0) {
		elapsed_time = (double)(stop.tv_usec - start.tv_usec) / 1000000 + (double)(stop.tv_sec - start.tv_sec);
		OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_DEBUG, "tenants_policy_file parsing complete.  Time = %11.8f secs\n", elapsed_time);
	}

	return status;
}

/* -------------------------------------------------------------------------------------------------
 *
 * API interface.
 */

int osm_tenant_db_insert_guid(IN osm_tenant_database_t *p_tenant_db,
				    IN uint64_t guid,
				    IN osm_tenant_t *p_tenant,
				    IN boolean_t is_virtual_guid)
{
	osm_tenant_list_t *p_list = is_virtual_guid ? &p_tenant_db->virtual_guids : &p_tenant_db->physical_guids;

	return cl_hashmap_insert(p_list, guid, (uintptr_t)p_tenant);
}

int osm_tenant_db_insert_tenant(IN osm_tenant_database_t *p_tenant_db,
				IN uint64_t tenant_id,
				IN osm_tenant_t *p_tenant)
{
	return cl_hashmap_insert(&p_tenant_db->tenant_list, tenant_id, (uintptr_t)p_tenant);
}

osm_tenant_t *osm_tenant_db_get_tenant_by_physical_guid(IN osm_tenant_database_t *p_db, IN uint64_t guid)
{
	return (osm_tenant_t *)cl_hashmap_find(&p_db->physical_guids, guid);
}

osm_tenant_t *osm_tenant_db_get_tenant_by_virtual_guid(IN osm_tenant_database_t *p_db, IN uint64_t guid)
{
	return (osm_tenant_t *)cl_hashmap_find(&p_db->virtual_guids, guid);
}

int osm_tenant_delete_by_id(IN osm_tenant_mgr_t *p_mgr, IN uint64_t tenant_id)
{
	return cl_hashmap_remove(&p_mgr->p_tenant_db->tenant_list, tenant_id);
}

osm_tenant_t *osm_tenant_list_find_by_id(IN osm_tenant_database_t *p_tenant_db, IN uint64_t tenant_id)
{
	return (osm_tenant_t *)cl_hashmap_find(&p_tenant_db->tenant_list, tenant_id);
}

osm_tenant_t *osm_tenant_create(IN osm_log_t *p_log, IN uint64_t tenant_id)
{
	osm_tenant_t *p_tenant = NULL;

	if ((p_tenant = calloc(1, sizeof(osm_tenant_t))) != NULL) {
		p_tenant->tenant_id = tenant_id;
		p_tenant->max_tenant_mc_groups = OSM_DEFAULT_MAX_TENANT_MC_GROUPS;
		p_tenant->tenant_vport_lids = FALSE;
	} else {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR AE05: allocation error\n");
	}

	return p_tenant;
}

int osm_tenant_mgr_validate(IN osm_sm_t *p_sm, IN uint64_t virtual_guid, IN uint64_t physical_guid, uint16_t vport_index)
{
	int			status;
	uint64_t		virtual_group_id = 0;
	uint64_t		physical_group_id = 0;
	osm_tenant_database_t	*p_tenant_db = p_sm->tenant_mgr.p_tenant_db;
	osm_tenant_t		*p_tenant;

	if (!p_sm->p_subn->opt.tenants_policy_enabled) {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG, "tenant mgr is not enabled for validation\n");
		return 1;
	}

	/*
	 * If the physical port is managed and the virtual port is managed, then the GUIDS must
	 * be in the same group.  There are 5 possibilities:
	 *
	 *           is managed?      | same
	 *	virtual  |  physical  | group  |  status     |  reason
	 *	---------+------------+--------+-------------+------------------------------
	 *	  yes    |    yes     |  yes   |  good       |  both ports in same group
	 *	  yes    |    yes     |  no    |  violation  |  ports in different groups
	 *	  yes    |    no      |  n/a   |  violation  |  only one port managed
	 *	  no     |    yes     |  n/a   |  violation  |  only one port managed
	 *	  no     |    no      |  n/a   |  good       |  neither port managed
	 */
	p_tenant = (osm_tenant_t *)osm_tenant_db_get_tenant_by_virtual_guid(p_tenant_db, virtual_guid);
	virtual_group_id = (p_tenant != NULL) ? p_tenant->tenant_id : 0;
	p_tenant = (osm_tenant_t *)osm_tenant_db_get_tenant_by_physical_guid(p_tenant_db, physical_guid);
	physical_group_id = (p_tenant != NULL) ? p_tenant->tenant_id : 0;

	if ((status = (virtual_group_id == physical_group_id)) == 0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR,
			"ERR AE01: tenant violation: virt=0x%016lx (%ld) phys=0x%016lx (%ld)  vport=%d\n",
			virtual_guid, virtual_group_id, physical_guid, physical_group_id, vport_index);

		osm_report_notice_1601(p_sm->p_subn,
			virtual_guid, virtual_group_id,
			physical_guid, physical_group_id, vport_index,
			(uint8_t)OSM_TENANT_MGR_NOTICE_GUID_DEVIATION);
	} else {
		OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
			"tenant validation successful: virt=0x%016lx (%ld) phys=0x%016lx (%ld)  vport=%d\n",
			virtual_guid, virtual_group_id, physical_guid, physical_group_id, vport_index);
	}

	return status;
}

int osm_tenant_mgr_validate_mcgroup_create(IN osm_tenant_mgr_t *p_mgr, IN ib_member_rec_t *mcm_rec)
{
	uint64_t guid = cl_hton64(ib_gid_get_guid(&mcm_rec->port_gid));
	osm_tenant_t *p_tenant = osm_tenant_db_get_tenant_by_physical_guid(p_mgr->p_tenant_db, guid);
	uint16_t signature;

	memcpy(&signature, &(mcm_rec->mgid.multicast.raw_group_id),
	       sizeof(signature));
	signature = cl_ntoh16(signature);

	/* If the guid does not belong to any tenant approve without tracking */
	if (p_tenant == NULL)
		return 0;
	/* If the group is an IPv6 Solicited node multicast (SNM) group approve without tracking */
	if (osm_match_ipv6_snm_mgid(&mcm_rec->mgid))
		return 0;

	if (p_tenant->mc_groups_counter < p_tenant->max_tenant_mc_groups) {
		p_tenant->mc_groups_counter++;
		return 0;
	}

	osm_report_notice_1601(p_mgr->p_subn, 0, 0, guid, p_tenant->tenant_id, 0,
			       (uint8_t)OSM_TENANT_MGR_NOTICE_MC_GROUP_VIOLATION);

	OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR AE07: "
		"Requested MGRP creation of guid 0x%" PRIx64 ", tenant id %ld exceeds tenant limit\n",
		 guid, p_tenant->tenant_id);
	return 1;
}

void osm_tenant_mgr_decrease_counter(IN osm_tenant_mgr_t *p_mgr, IN ib_member_rec_t *mcm_rec)
{
	uint64_t guid = cl_hton64(ib_gid_get_guid(&mcm_rec->port_gid));
	osm_tenant_t *p_tenant = osm_tenant_db_get_tenant_by_physical_guid(p_mgr->p_tenant_db, guid);
	uint16_t signature;

	memcpy(&signature, &(mcm_rec->mgid.multicast.raw_group_id),
	       sizeof(signature));
	signature = cl_ntoh16(signature);

	/* If the guid does not belong to any tenant do not track */
	if (p_tenant == NULL)
		return;
	/* If the group is an IPv6 Solicited node multicast (SNM) group do not track */
	if (osm_match_ipv6_snm_mgid(&mcm_rec->mgid))
		return;

	CL_ASSERT(p_tenant->mc_groups_counter > 0);
	p_tenant->mc_groups_counter--;
}
void osm_tenant_mgr_log_mc_group_counts(IN cl_qhashmap_item_t *p_item, IN void *context)
{
	osm_log_t *p_log = (osm_log_t *)context;
	osm_tenant_t *p_tenant = (osm_tenant_t *)p_item->value;
	OSM_LOG(p_log, OSM_LOG_DEBUG, "Tenant %ld has %d MC groups open in the fabric\n",
				      p_tenant->tenant_id, p_tenant->mc_groups_counter);
}

void osm_tenant_mgr_zero_tenants_mc_counter(IN cl_qhashmap_item_t *p_item, IN void *context)
{
	osm_tenant_t *p_tenant = (osm_tenant_t *)p_item->value;
	p_tenant->mc_groups_counter = 0;
}

void osm_tenant_mgr_count_mc_groups_in_fabric(IN osm_subn_t *p_subn)
{
	osm_mgrp_t *p_mgrp;
	osm_tenant_mgr_t *p_mgr = &p_subn->p_osm->sm.tenant_mgr;
	uint64_t port_guid;
	osm_tenant_t *p_tenant;
	osm_mcm_port_t *p_mcm_port;
	osm_tenant_list_t *p_tenant_list = &p_mgr->p_tenant_db->tenant_list;

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG, "Tenant Mgr: Recounting MC groups in fabric\n"); 

	cl_hashmap_apply_func(p_tenant_list, osm_tenant_mgr_zero_tenants_mc_counter, NULL);

	/* Iterate over multicast groups and count per tenant */
	for (p_mgrp = (osm_mgrp_t *)cl_fmap_head(&p_subn->mgrp_mgid_tbl);
	     p_mgrp != (osm_mgrp_t *)cl_fmap_end(&p_subn->mgrp_mgid_tbl);
	     p_mgrp = (osm_mgrp_t *)cl_fmap_next(&p_mgrp->map_item)) {

		if (p_mgrp->well_known)
			continue;

		if (cl_qmap_count(&p_mgrp->mcm_port_tbl) == 0)
			continue;

		p_mcm_port = (osm_mcm_port_t *)cl_qmap_head(&p_mgrp->mcm_port_tbl);
		if (p_mcm_port == NULL) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Tenant manager sanity check, untracked group MLID 0x%04X\n",
				cl_ntoh16(p_mgrp->mlid));
			continue;
		}

		/* If the group is an IPv6 Solicited node multicast (SNM) group don't track */
		if (osm_match_ipv6_snm_mgid(&p_mgrp->mcmember_rec.mgid)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Tenant manager sanity check, untracked group MLID 0x%04X\n",
				cl_ntoh16(p_mgrp->mlid));
			continue;
		}

		port_guid = cl_ntoh64(osm_port_get_guid(p_mcm_port->port));
		p_tenant = osm_tenant_db_get_tenant_by_physical_guid(p_mgr->p_tenant_db, port_guid);

		/* This multicast group has been created by a port not belonging to any tenant */
		if (p_tenant == NULL) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Tenant manager sanity check, group MLID 0x%04X "
				"created by port 0x%" PRIx64 " not belonging to any tenant.\n",
				cl_ntoh16(p_mgrp->mlid), port_guid);
			continue;
		}

		p_tenant->mc_groups_counter++;
	}

	cl_hashmap_apply_func(p_tenant_list, osm_tenant_mgr_log_mc_group_counts, &p_subn->p_osm->log);
}

int osm_tenant_mgr_rescan(IN osm_subn_t *p_subn)
{
	int			status = 0;
	uint32_t		crc = 0;
	osm_sm_t		*p_sm = NULL;
	osm_tenant_database_t	*p_tenant_db = NULL;
	osm_tenant_database_t	*p_tenant_db_old = NULL;
	osm_tenant_mgr_t	*p_tenant_mgr = NULL;
	FILE			*std_tenant = NULL;

	/*
	 * If the tenant mgr is not enabled, just return.
	 */
	if (!p_subn->opt.tenants_policy_enabled) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG, "tenant mgr is not enabled for rescanning\n");
		return 0;
	}

	/*
	 * It is a fatal error if we can't read the file.
	 */
	p_sm = &p_subn->p_osm->sm;
	p_tenant_mgr = &p_sm->tenant_mgr;

	if ((std_tenant = fopen(p_tenant_mgr->filename, "r")) == NULL) {
		OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_ERROR, "ERR AE03: tenants policy file - %s\n", strerror(errno));
		tenant_mgr_fatal_error(p_tenant_mgr);
		return status;
	}

	/*
	 * Check if the file has changed.  If it has, parse it.
	 */
	crc = osm_get_crc_value(p_sm, OSM_CRC_TENANTS_POLICY_CONF_FILE);
	if (crc != p_tenant_mgr->crc) {
		OSM_LOG(p_tenant_mgr->p_log, OSM_LOG_DEBUG, "rescanning tenants policy file\n");
		if ((p_tenant_db = tenant_mgr_guid_map_alloc(p_subn)) == NULL) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR AE06: allocation error\n");
			fclose(std_tenant);
			tenant_mgr_fatal_error(p_tenant_mgr);
			return status;
		}

		if (tenant_parse_file(p_tenant_mgr, p_tenant_db, std_tenant) != 0) {
			fclose(std_tenant);
			tenant_db_free(p_tenant_db);
			tenant_mgr_fatal_error(p_tenant_mgr);
			return status;
		}

		/*
		 * The file was parsed correctly.  Update state and signal the virt_mgr to revalidate.
		 */
		p_tenant_db_old = p_tenant_mgr->p_tenant_db;
		p_tenant_mgr->p_tenant_db = p_tenant_db;
		p_tenant_mgr->crc = crc;
		tenant_db_free(p_tenant_db_old);
		status = 1;
	}

	fclose(std_tenant);
	return status;
}

void osm_tenant_mgr_config_change(osm_subn_t *p_subn, void *p_value)
{
	boolean_t		value = *(boolean_t *)p_value;
	osm_tenant_mgr_t	*p_tenant_mgr = NULL;

	p_tenant_mgr = &p_subn->p_osm->sm.tenant_mgr;

	if (value) {
		/*
		 * We were disabled, but now we are enabled.  We need to setup for rescanning.
		 */
		p_tenant_mgr->crc = (uint32_t)-1;
	} else {
		/*
		 * We were enabled, but now we are disabled.  We need to clean up our resources.
		 */
		tenant_db_free(p_tenant_mgr->p_tenant_db);
		p_tenant_mgr->p_tenant_db = NULL;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "tenant mgr is now %s\n", value ? "enabled" : "disabled");
}

ib_api_status_t osm_tenant_mgr_init(IN osm_tenant_mgr_t *p_tenant_mgr, osm_sm_t *p_sm)
{
	ib_api_status_t		status = IB_SUCCESS;
	osm_subn_t		*p_subn = p_sm->p_subn;

	/*
	 * Initialize the tenant mgr.  We set the crc to -1 so that the first time through
	 * we rescan the conf file correctly.
	 */
	memset((void *)p_tenant_mgr, 0, sizeof(osm_tenant_mgr_t));
	p_tenant_mgr->p_log = p_sm->p_log;
	p_tenant_mgr->crc = (uint32_t)-1;
	p_tenant_mgr->p_subn = p_subn;

	/*
	 * Get configuration filename.
	 */
	if (p_subn->opt.tenants_policy_file != NULL) {
		p_tenant_mgr->filename = p_subn->opt.tenants_policy_file;
	} else {
		p_tenant_mgr->filename = OSM_DEFAULT_TENANTS_POLICY_CONF_FILE;
	}

	OSM_LOG(p_sm->p_log, OSM_LOG_INFO, "tenant mgr is %s\n", p_subn->opt.tenants_policy_enabled ? "enabled" : "disabled");
	return status;
}

void osm_tenant_mgr_destroy(IN osm_tenant_mgr_t *p_tenant_mgr, osm_sm_t *p_sm)
{
	tenant_db_free(p_tenant_mgr->p_tenant_db);
	memset((void *)p_tenant_mgr, 0, sizeof(osm_tenant_mgr_t));

	OSM_LOG(p_sm->p_log, OSM_LOG_INFO, "osm_tenant_mgr_destroy complete\n");
}
