/*
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2010 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2020 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2008 Xsigo Systems Inc.  All rights reserved.
 * Copyright (c) 2009 System Fabric Works, Inc. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 * Copyright (c) 2009-2015 ZIH, TU Dresden, Federal Republic of Germany. All rights reserved.
 * Copyright (C) 2012-2017 Tokyo Institute of Technology. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_subn_t.
 * This object represents an IBA subnet.
 * This object is part of the opensm family of objects.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <limits.h>
#include <errno.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <complib/cl_debug.h>
#include <complib/cl_log.h>
#include <complib/cl_math.h>
#include <complib/cl_vector.h>
#include <complib/cl_sha1.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_SUBNET_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_log.h>
#include <opensm/osm_madw.h>
#include <opensm/osm_port.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_router.h>
#include <opensm/osm_remote_sm.h>
#include <opensm/osm_partition.h>
#include <opensm/osm_node.h>
#include <opensm/osm_guid.h>
#include <opensm/osm_multicast.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_inform_mgr.h>
#include <opensm/osm_console.h>
#include <opensm/osm_perfmgr.h>
#include <opensm/osm_congestion_control.h>
#include <opensm/osm_event_plugin.h>
#include <opensm/osm_qos_policy.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_service.h>
#include <opensm/osm_db.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_mcast_tbl.h>
#include <opensm/osm_router_policy.h>
#include <opensm/osm_enhanced_qos_mgr.h>
#include <opensm/osm_root_detection.h>
#include <opensm/osm_subscriber.h>
#include <opensm/osm_router_mgr.h>
#include <opensm/osm_gpu.h>
#include <opensm/osm_tenant_mgr.h>

extern int parse_verbose_bypass_policy(struct osm_opensm *osm_ptr);
static inline
void regex_get_substr(char * dest, const char * source, regmatch_t * p_match);

typedef enum device_generation {
	GENERIC_DEVICE,
	MLNX_SWITCH_X,
	MLNX_SWITCH_IB1,
	MLNX_SWITCH_IB2,
	MLNX_CONNECTX_3,
	MLNX_CONNECT_IB,
	MLNX_CONNECTX_4,
	MLNX_CONNECTX_5
} device_generation_t;

static const char null_str[] = "(null)";

#define OPT_OFFSET(opt) offsetof(osm_subn_opt_t, opt)
#define ARR_SIZE(a) (sizeof(a)/sizeof((a)[0]))

typedef struct opt_rec {
	const char *name;
	unsigned long opt_offset;
	void (*parse_fn)(osm_subn_t *p_subn, char *p_key, char *p_val_str,
			 void *p_val1, void *p_val2,
			 boolean_t log,
			 void (*)(osm_subn_t *, void *));
	void (*setup_fn)(osm_subn_t *p_subn, void *p_val);
	int  can_update;
} opt_rec_t;

typedef enum held_back_reason {		// These are bitmap values, not ints
        HELD_BACK_REASON_FILE = 1,
        HELD_BACK_REASON_ISSU = 2,
} held_back_reason_t;

typedef struct held_back_item {
	cl_map_item_t item;
	held_back_reason_t reason;
} held_back_item_t;

static void log_report(const char *fmt, ...)
{
	char buf[1024];
	va_list args;
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	va_end(args);
	printf("%s", buf);
	cl_log_event("OpenSM", CL_LOG_INFO, buf, NULL, 0);
}

static void log_config_value(char *name, const char *fmt, ...)
{
	char buf[128];
	va_list args;
	unsigned n;
	va_start(args, fmt);
	n = snprintf(buf, sizeof(buf), " Loading Cached Option:%s = ", name);

	if (n < sizeof(buf))
		n += vsnprintf(buf + n, sizeof(buf) - n, fmt, args);

	if (n > sizeof(buf) - 2)
		n = sizeof(buf) - 2;
	snprintf(buf + n, sizeof(buf) - n, "\n");
	va_end(args);
	printf("%s", buf);
	cl_log_event("OpenSM", CL_LOG_INFO, buf, NULL, 0);
}

static void opts_setup_change_dump_limit(osm_subn_t *p_subn, void *p_val)
{
	/* convert size in MB to bytes */
	p_subn->p_osm->statistic_max_size = *(unsigned long*)p_val << 20;
}

static void opts_setup_change_perflog_dump_limit(osm_subn_t *p_subn, void *p_val)
{
	/* convert size in MB to bytes */
	p_subn->p_osm->perflog_mgr.perflog_max_size = *(unsigned long*)p_val << 20;
}

static void opts_setup_ucast_reinit_threads(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->sm.ucast_mgr.variables_reinit_thread_pool = TRUE;
}

static void opts_setup_syslog_log_flags(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->log.syslog_level = *((uint8_t *) p_val) | OSM_LOG_SYS;
}

static void opts_setup_log_flags(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->log.level = *((uint8_t *) p_val);
}

static void opts_setup_force_log_flush(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->log.flush = *((boolean_t *) p_val);
}

static void opts_setup_accum_log_file(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->log.accum_log_file = *((boolean_t *) p_val);
}

static void opts_setup_log_max_size(osm_subn_t *p_subn, void *p_val)
{
	uint32_t log_max_size = *((uint32_t *) p_val);

	p_subn->p_osm->log.max_size = (unsigned long)log_max_size << 20; /* convert from MB to bytes */
}

static void opts_setup_sminfo_polling_timeout(osm_subn_t *p_subn, void *p_val)
{
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	uint32_t sminfo_polling_timeout = *((uint32_t *) p_val);

	cl_timer_stop(&p_sm->polling_timer);
	cl_timer_start(&p_sm->polling_timer, sminfo_polling_timeout);
}

static void opts_setup_sm_priority(osm_subn_t *p_subn, void *p_val)
{
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	uint8_t sm_priority = *((uint8_t *) p_val);

	osm_set_sm_priority(p_sm, sm_priority);
}

static void opts_setup_routing_engine(osm_subn_t *p_subn, void *p_val)
{
	osm_update_routing_engines(p_subn, p_val);
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
	p_subn->force_multicast_rerouting = TRUE;
}

static void opts_setup_allowed_sm_guids(osm_subn_t *p_subn, void *p_val)
{
	cl_u64_vector_remove_all(&p_subn->allowed_sm_guid_tbl);
	osm_update_allowed_sm_guids(p_subn, p_val);
}

static void opts_setup_enable_inc_mc_routing(osm_subn_t *p_subn, void *p_val)
{
	/* force rerouting when the parameter is changed */
	p_subn->force_multicast_rerouting = TRUE;
}

static void opts_setup_ucast_invalidate_options(osm_subn_t *p_subn, void *p_val)
{
	/* force rerouting when the parameter is changed */
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
}

static void opts_setup_force_rerouting(osm_subn_t *p_subn, void *p_val)
{
	/* force rerouting when the parameter is changed */
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
	p_subn->force_multicast_rerouting = TRUE;
}

static void opts_setup_change_osm_stats_interval(osm_subn_t *p_subn, void *p_val)
{
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	uint32_t osm_stats_interval = *(uint32_t *) p_val;
	p_subn->opt.osm_stats_interval = osm_stats_interval;
	cl_timer_stop(&p_sm->dump_statistics_timer);
	cl_timer_start(&p_sm->dump_statistics_timer, osm_stats_interval * MINUTES_TO_MILLISEC);
}

static void opts_setup_adv_routing_engine(osm_subn_t *p_subn, void *p_val)
{
	osm_setup_advanced_routing_engine(p_subn, (const char *)p_val);
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
	p_subn->need_update = 1;
}

static void opts_setup_adv_routing_mode(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
	p_subn->need_update = 1;
}

static void opts_setup_ar_mask(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
}

static void opts_setup_enable_ar_by_device_cap(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
}

static void opts_setup_mlnx_cc(osm_subn_t *p_subn, void *p_val)
{
	osm_congestion_control_invalidate(&p_subn->p_osm->cc);
}

static void opts_setup_rtr_selection_algo_parameters(osm_subn_t *p_subn, void *p_val)
{
	osm_update_rtr_selection_algo_parameters(p_subn, (char *)p_val);
}

static void opts_setup_rtr_selection_function(osm_subn_t *p_subn, void *p_val)
{
	osm_update_rtr_selection_function(p_subn, (char *)p_val);
}

static void opts_setup_key_mgr_seed(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->cc.recalculate_keys =
	    p_subn->p_osm->vs.recalculate_keys =
	    p_subn->p_osm->n2n.recalculate_keys = TRUE;
}

static void opts_setup_cc_key_enable(osm_subn_t *p_subn, void *p_val)
{
	uint8_t enable_bit = *((uint8_t *) p_val);

	/* update only if new value is not ignore */
	p_subn->p_osm->cc.recalculate_keys = (enable_bit != 0);
	p_subn->opt.cc_key_enable = enable_bit;
}

static void opts_setup_vs_key_enable(osm_subn_t *p_subn, void *p_val)
{
	uint8_t enable_bits = *((uint8_t *) p_val);

	/* update only if new value is not ignore */
	p_subn->p_osm->vs.recalculate_keys = (enable_bits != 0);
	p_subn->opt.vs_key_enable = enable_bits;
}

static void opts_setup_n2n_key_enable(osm_subn_t *p_subn, void *p_val)
{
	uint8_t enable_bits = *((uint8_t *) p_val);

	/* update only if new value is not ignore */
	p_subn->p_osm->n2n.recalculate_keys = (enable_bits != 0);
	p_subn->opt.n2n_key_enable = enable_bits;
}

static void opts_setup_cc_key_protect_bit(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->cc.cc_key_param_update = TRUE;
}

static void opts_setup_cc_key_lease_period(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->cc.cc_key_param_update = TRUE;
}

static void opts_setup_vs_key_ci_protect_bits(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->vs.vs_key_param_update = TRUE;
}

static void opts_setup_vs_key_lease_period(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->vs.vs_key_param_update = TRUE;
}

static void opts_setup_n2n_key_params(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->n2n.n2n_key_param_update = TRUE;
}

static void opts_setup_m_key(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->cc.recalculate_keys =
	    p_subn->p_osm->vs.recalculate_keys =
	    p_subn->p_osm->n2n.recalculate_keys =
	    (p_subn->opt.key_mgr_seed == 0);
}

static void opts_setup_m_key_per_port(osm_subn_t *p_subn, void *p_val)
{
	p_subn->p_osm->cc.recalculate_keys =
	    p_subn->p_osm->vs.recalculate_keys =
	    p_subn->p_osm->n2n.recalculate_keys =
	    (p_subn->opt.key_mgr_seed == 0);
}

static void opts_setup_pfrn_config_change(osm_subn_t *p_subn, void *p_val)
{
	p_subn->pfrn_config_change = TRUE;
}

static int opts_strtoul(uint32_t *val,
			IN char *p_val_str,
			IN char *p_key,
			uint32_t max_value)
{
	errno = 0;
	char * endptr;
	unsigned long int tmp_val;
	tmp_val = strtoul(p_val_str, &endptr, 0);
	*val = tmp_val;
	if (*p_val_str == '\0' || *endptr != '\0') {
		log_report("-E- Parsing error in field %s, expected "
			   "numeric input received: %s\n", p_key, p_val_str);
		return -1;
	}
	if (tmp_val > max_value ||
  	   ((tmp_val == ULONG_MAX) &&
	     errno == ERANGE))
	{
		log_report("-E- Parsing error in field %s, value out of range\n", p_key);
		return -1;
	}
	return 0;
}

static int opts_strtoull(uint64_t *val,
			 IN char *p_val_str,
			 IN char *p_key,
			 uint64_t max_value)
{
	errno = 0;
	char * endptr;
	unsigned long long int tmp_val;
	tmp_val = strtoull(p_val_str, &endptr, 0);
	*val = tmp_val;
	if (*p_val_str == '\0' || *endptr != '\0') {
		log_report("-E- Parsing error in field %s, expected "
			   "numeric input received: %s\n", p_key, p_val_str);
		return -1;
	}
	if (tmp_val > max_value ||
	   (tmp_val == ULLONG_MAX &&
	    errno == ERANGE))
	{
		log_report("-E- Parsing error in field %s, value out of range\n", p_key);
		return -1;
	}
	return 0;
}

static int opts_strtol(int32_t *val,
		       IN char *p_val_str,
		       IN char *p_key,
		       int32_t max_value)
{
	errno = 0;
	char * endptr;
	*val = strtol(p_val_str, &endptr, 0);
	if (*p_val_str == '\0' || *endptr != '\0') {
		log_report("-E- Parsing error in field %s, expected "
			   "numeric input received: %s\n", p_key, p_val_str);
		return -1;
	}
	if (*val > max_value ||
	   ((*val == INT32_MAX || *val == INT32_MIN) && errno == ERANGE) ||
	    (*val == 0 && errno !=0))
	{
		log_report("-E- Parsing error in field %s, value out of range\n", p_key);
		return -1;
	}
	return 0;
}

static void opts_parse_uint64(IN osm_subn_t *p_subn, IN char *p_key,
			      IN char *p_val_str, void *p_v1, void *p_v2,
			      boolean_t log,
			      void (*pfn)(osm_subn_t *, void *))
{
	uint64_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint64_t val;

	if (opts_strtoull(&val, p_val_str, p_key, UINT64_MAX))
		return;

	if (val != *p_val1) {
		if (log)
			log_config_value(p_key, "0x%016" PRIx64, val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = val;
	}
}

static void opts_parse_net64(IN osm_subn_t *p_subn, IN char *p_key,
			     IN char *p_val_str, void *p_v1, void *p_v2,
			     boolean_t log,
			     void (*pfn)(osm_subn_t *, void *))
{
	uint64_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint64_t val;

	if (opts_strtoull(&val, p_val_str, p_key, UINT64_MAX))
		return;

	if (cl_hton64(val) != *p_val1) {
		if (log)
			log_config_value(p_key, "0x%016" PRIx64, val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = cl_ntoh64(val);
	}
}

static void opts_parse_net64_guid_list(IN osm_subn_t *p_subn, IN char *p_key,
				       IN char *p_val_list_str, void *p_v1, void *p_v2,
				       boolean_t log,
				       void (*pfn)(osm_subn_t *, void *))
{
	char *val_str, *str, *p;
	uint64_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint64_t val[OSM_MAX_BINDING_PORTS];
	boolean_t is_changed = FALSE;
	uint8_t i = 0;
	char buf[128];

	str = strdup(p_val_list_str);
	val_str = strtok_r(str, ", \t\n", &p);
	while (val_str && *val_str && i < OSM_MAX_BINDING_PORTS) {
		if (opts_strtoull(&val[i], val_str, p_key, UINT64_MAX))
			goto Exit;

		if (cl_hton64(val[i]) != p_val1[i]) {
			is_changed = TRUE;

			if (log) {
				snprintf(buf, sizeof(buf), "%s_%u", p_key, i + 1);
				log_config_value(buf, "0x%016" PRIx64, val[i]);
			}

			p_val1[i] = p_val2[i] = cl_ntoh64(val[i]);
		}

		val_str = strtok_r(NULL, ", \t\n", &p);
		i++;
	}

	if (is_changed && pfn)
		pfn(p_subn, &val);

Exit:
	free(str);
}

static void opts_parse_uint32(IN osm_subn_t *p_subn, IN char *p_key,
			      IN char *p_val_str, void *p_v1, void *p_v2,
			      boolean_t log,
			      void (*pfn)(osm_subn_t *, void *))
{
	uint32_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint32_t val;

	if (opts_strtoul(&val, p_val_str, p_key, UINT32_MAX))
		return;

	if (val != *p_val1) {
		if (log)
			log_config_value(p_key, "%u", val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = val;
	}
}

static void opts_parse_uint32_minutes_to_milisec(IN osm_subn_t *p_subn, IN char *p_key,
					         IN char *p_val_str, void *p_v1, void *p_v2,
					         boolean_t log,
					         void (*pfn)(osm_subn_t *, void *))
{
	uint32_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint32_t val;

	if (opts_strtoul(&val, p_val_str, p_key, UINT32_MAX / MINUTES_TO_MILLISEC)){
		val = OSM_DEFAULT_STATISTIC_DUMP_TIME_MINUTES;
		log_report("got illegal value, setting osm_stats_interval to default\n");
	}
	
	if (val != *p_val1) {
		if (log)
			log_config_value(p_key, "%u", val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = val;
	}
}

static void opts_parse_net32(IN osm_subn_t *p_subn, IN char *p_key,
			     IN char *p_val_str, void *p_v1, void *p_v2,
			     boolean_t log,
			     void (*pfn)(osm_subn_t *, void *))
{
	uint32_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint32_t val;

	if (opts_strtoul(&val, p_val_str, p_key, UINT32_MAX))
		return;

	if (cl_hton32(val) != *p_val1) {
		if (log)
			log_config_value(p_key, "%u", val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = cl_hton32(val);
	}
}

static void opts_parse_int32(IN osm_subn_t *p_subn, IN char *p_key,
			     IN char *p_val_str, void *p_v1, void *p_v2,
			     boolean_t log,
			     void (*pfn)(osm_subn_t *, void *))
{
	int32_t *p_val1 = p_v1, *p_val2 = p_v2;
	int32_t val;

	if (opts_strtol(&val, p_val_str, p_key, INT32_MAX))
		return;

	if (val != *p_val1) {
		if (log)
			log_config_value(p_key, "%d", val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = val;
	}
}

static void opts_parse_uint16(IN osm_subn_t *p_subn, IN char *p_key,
			      IN char *p_val_str, void *p_v1, void *p_v2,
			      boolean_t log,
			      void (*pfn)(osm_subn_t *, void *))
{
	uint16_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint32_t tmp_val;

	if (opts_strtoul(&tmp_val, p_val_str, p_key, UINT16_MAX))
		return;

	uint16_t val = (uint16_t) tmp_val;
	if (val != *p_val1) {
		if (log)
			log_config_value(p_key, "%u", val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = val;
	}
}

static void opts_parse_net16(IN osm_subn_t *p_subn, IN char *p_key,
			     IN char *p_val_str, void *p_v1, void *p_v2,
			     boolean_t log,
			     void (*pfn)(osm_subn_t *, void *))
{
	uint16_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint32_t tmp_val;

	if (opts_strtoul(&tmp_val, p_val_str, p_key, UINT16_MAX))
		return;

	uint16_t val = (uint16_t) tmp_val;
	if (cl_hton16(val) != *p_val1) {
		if (log)
			log_config_value(p_key, "0x%04x", val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = cl_hton16(val);
	}
}

static void opts_parse_uint8(IN osm_subn_t *p_subn, IN char *p_key,
			     IN char *p_val_str, void *p_v1, void *p_v2,
			     boolean_t log,
			     void (*pfn)(osm_subn_t *, void *))
{
	uint8_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint32_t tmp_val;

	if (opts_strtoul(&tmp_val, p_val_str, p_key, UINT8_MAX))
		return;

	uint8_t val = (uint8_t) tmp_val;
	if (val != *p_val1) {
		if (log)
			log_config_value(p_key, "%u", val);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = val;
	}
}

static void opts_parse_boolean(IN osm_subn_t *p_subn, IN char *p_key,
			       IN char *p_val_str, void *p_v1, void *p_v2,
			       boolean_t log,
			       void (*pfn)(osm_subn_t *, void *))
{
	boolean_t *p_val1 = p_v1, *p_val2 = p_v2;
	boolean_t val;

	if (!p_val_str)
		return;

	if (!strcasecmp("TRUE", p_val_str)) {
		val = TRUE;
	} else if (!strcasecmp("FALSE", p_val_str)) {
		val = FALSE;
	} else {
		log_report("-E- Parsing error in field %s, expected "
			   "boolean input received: %s\n", p_key, p_val_str);
		return;
	}

	if (val != *p_val1) {
		if (log)
			log_config_value(p_key, "%s", p_val_str);
		if (pfn)
			pfn(p_subn, &val);
		*p_val1 = *p_val2 = val;
	}
}

static void opts_parse_charp(IN osm_subn_t *p_subn, IN char *p_key,
			     IN char *p_val_str, void *p_v1, void *p_v2,
			     boolean_t log,
			     void (*pfn)(osm_subn_t *, void *))
{
	char **p_val1 = p_v1, **p_val2 = p_v2;
	const char *current_str = *p_val1 ? *p_val1 : null_str ;

	if (p_val_str && strcmp(p_val_str, current_str)) {
		char *new;
		if (log)
			log_config_value(p_key, "%s", p_val_str);
		/* special case the "(null)" string */
		new = strcmp(null_str, p_val_str) ? strdup(p_val_str) : NULL;
		if (pfn)
			pfn(p_subn, new);
		if (*p_val1 && *p_val1 != *p_val2)
			free(*p_val1);
		if (*p_val2)
			free(*p_val2);
		*p_val1 = *p_val2 = new;
	}
}

static void opts_parse_256bit(IN osm_subn_t *p_subn, IN char *p_key,
			      IN char *p_val_str, void *p_v1, void *p_v2,
			      boolean_t log,
			      void (*pfn)(osm_subn_t *, void *))
{
	uint8_t *p_val1 = p_v1, *p_val2 = p_v2;
	uint8_t val[IB_CC_PORT_MASK_DATA_SIZE] = { 0 };
	char tmpbuf[3] = { 0 };
	uint8_t tmpint;
	int numdigits = 0;
	int startindex;
	char *strptr = p_val_str;
	char *ptr;
	int i;

	/* parse like it's hypothetically a 256 bit integer code
	 *
	 * store "big endian"
	 */

	if (!strncmp(strptr, "0x", 2) || !strncmp(strptr, "0X", 2))
		strptr+=2;

	for (ptr = strptr; *ptr; ptr++) {
		if (!isxdigit(*ptr)) {
			log_report("invalid hex digit in bitmask\n");
			return;
		}
		numdigits++;
	}

	if (!numdigits) {
		log_report("invalid length bitmask\n");
		return;
	}

	/* max of 2 hex chars per byte */
	if (numdigits > IB_CC_PORT_MASK_DATA_SIZE * 2)
		numdigits = IB_CC_PORT_MASK_DATA_SIZE * 2;

	startindex = IB_CC_PORT_MASK_DATA_SIZE - ((numdigits - 1) / 2) - 1;

	if (numdigits % 2) {
		memcpy(tmpbuf, strptr, 1);
		strptr += 1;
	}
	else {
		memcpy(tmpbuf, strptr, 2);
		strptr += 2;
	}

	tmpint = strtoul(tmpbuf, NULL, 16);
	val[startindex] = tmpint;

	for (i = (startindex + 1); i < IB_CC_PORT_MASK_DATA_SIZE; i++) {
		memcpy(tmpbuf, strptr, 2);
		strptr += 2;
		tmpint = strtoul(tmpbuf, NULL, 16);
		val[i] = tmpint;
	}

	if (memcmp(val, p_val1, IB_CC_PORT_MASK_DATA_SIZE)) {
		if (log)
			log_config_value(p_key, "%s", p_val_str);
		if (pfn)
			pfn(p_subn, val);
		memcpy(p_val1, val, IB_CC_PORT_MASK_DATA_SIZE);
		memcpy(p_val2, val, IB_CC_PORT_MASK_DATA_SIZE);
	}

}

static void opts_setup_hbf_weights(osm_subn_t *p_subn, void *p_val)
{
	osm_setup_hbf_weights(p_subn, (const char *)p_val);
}

static const opt_rec_t opt_tbl[] = {
	{ "guid", OPT_OFFSET(guid), opts_parse_net64_guid_list, NULL, 0 },
	{ "m_key", OPT_OFFSET(m_key), opts_parse_net64, opts_setup_m_key , 1 },
	{ "sm_key", OPT_OFFSET(sm_key), opts_parse_net64, NULL, 1 },
	{ "sa_key", OPT_OFFSET(sa_key), opts_parse_net64, NULL, 1 },
	{ "allowed_sm_guids", OPT_OFFSET(allowed_sm_guids), opts_parse_charp, opts_setup_allowed_sm_guids, 1 },
	{ "subnet_prefix", OPT_OFFSET(subnet_prefix), opts_parse_net64, NULL, 0 },
	{ "m_key_lease_period", OPT_OFFSET(m_key_lease_period), opts_parse_net16, NULL, 1 },
	{ "m_key_protection_level", OPT_OFFSET(m_key_protect_bits), opts_parse_uint8, NULL, 1 },
	{ "m_key_lookup", OPT_OFFSET(m_key_lookup), opts_parse_boolean, NULL, 1 },
	{ "m_key_per_port", OPT_OFFSET(m_key_per_port), opts_parse_boolean, opts_setup_m_key_per_port, 1 },
	{ "sweep_interval", OPT_OFFSET(sweep_interval), opts_parse_uint32, NULL, 1 },
	{ "max_wire_smps", OPT_OFFSET(max_wire_smps), opts_parse_uint32, NULL, 1 },
	{ "max_wire_smps2", OPT_OFFSET(max_wire_smps2), opts_parse_uint32, NULL, 1 },
	{ "max_smps_timeout", OPT_OFFSET(max_smps_timeout), opts_parse_uint32, NULL, 1 },
	{ "max_sa_reports_queued", OPT_OFFSET(max_sa_reports_queued), opts_parse_uint32, NULL, 1 },
	{ "max_sa_reports_on_wire", OPT_OFFSET(max_sa_reports_on_wire), opts_parse_uint32, NULL, 1 },
	{ "sa_enhanced_trust_model", OPT_OFFSET(sa_enhanced_trust_model), opts_parse_boolean, NULL, 0 },
	{ "sa_etm_allow_untrusted_proxy_requests", OPT_OFFSET(sa_etm_allow_untrusted_proxy_requests), opts_parse_boolean, NULL, 1 },
	{ "sa_etm_allow_untrusted_guidinfo_rec", OPT_OFFSET(sa_etm_allow_untrusted_guidinfo_rec), opts_parse_boolean, NULL, 1 },
	{ "sa_etm_allow_guidinfo_rec_by_vf", OPT_OFFSET(sa_etm_allow_guidinfo_rec_by_vf), opts_parse_boolean, NULL, 1 },
	{ "sa_etm_max_num_mcgs", OPT_OFFSET(sa_etm_max_num_mcgs), opts_parse_uint32, NULL, 1 },
	{ "sa_etm_max_num_srvcs", OPT_OFFSET(sa_etm_max_num_srvcs), opts_parse_uint32, NULL, 1 },
	{ "sa_etm_max_num_event_subs", OPT_OFFSET(sa_etm_max_num_event_subs), opts_parse_uint32, NULL, 1 },
	{ "sa_rate_threshold", OPT_OFFSET(sa_rate_threshold), opts_parse_uint32, NULL, 1 },
	{ "sa_check_sgid_spoofing", OPT_OFFSET(sa_check_sgid_spoofing), opts_parse_boolean, NULL, 1 },
	{ "console", OPT_OFFSET(console), opts_parse_charp, NULL, 0 },
	{ "console_port", OPT_OFFSET(console_port), opts_parse_uint16, NULL, 0 },
	{ "transaction_timeout", OPT_OFFSET(transaction_timeout), opts_parse_uint32, NULL, 0 },
	{ "transaction_retries", OPT_OFFSET(transaction_retries), opts_parse_uint32, NULL, 0 },
	{ "long_transaction_timeout", OPT_OFFSET(long_transaction_timeout), opts_parse_uint32, NULL, 0 },
	{ "max_msg_fifo_timeout", OPT_OFFSET(max_msg_fifo_timeout), opts_parse_uint32, NULL, 1 },
	{ "sm_priority", OPT_OFFSET(sm_priority), opts_parse_uint8, opts_setup_sm_priority, 1 },
	{ "master_sm_priority", OPT_OFFSET(master_sm_priority), opts_parse_uint8, NULL, 0},
	{ "master_sm_lid", OPT_OFFSET(master_sm_lid), opts_parse_uint16, NULL, 0},
	{ "enable_sm_port_failover", OPT_OFFSET(enable_sm_port_failover), opts_parse_boolean, NULL, 0 },
	{ "lmc", OPT_OFFSET(lmc), opts_parse_uint8, NULL, 0 },
	{ "lids_per_rtr", OPT_OFFSET(lids_per_rtr), opts_parse_uint8, NULL, 0 },
	{ "prism_an_lmc", OPT_OFFSET(prism_an_lmc), opts_parse_uint8, NULL, 0 },
	{ "lmc_esp0", OPT_OFFSET(lmc_esp0), opts_parse_boolean, NULL, 0 },
	{ "qos_config_vl_enabled", OPT_OFFSET(qos_config_vl_enabled), opts_parse_boolean, NULL, 1 },
	{ "max_op_vls", OPT_OFFSET(max_op_vls), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1 },
	{ "max_op_vls_ca", OPT_OFFSET(max_op_vls_ca), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1 },
	{ "max_op_vls_sw", OPT_OFFSET(max_op_vls_sw), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1 },
	{ "max_op_vls_rtr", OPT_OFFSET(max_op_vls_rtr), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1 },
	{ "max_neighbor_mtu", OPT_OFFSET(max_neighbor_mtu), opts_parse_uint8, NULL, 0 },
	{ "suppress_mc_pkey_traps", OPT_OFFSET(suppress_mc_pkey_traps), opts_parse_uint8, NULL, 1},
	{ "force_link_speed", OPT_OFFSET(force_link_speed), opts_parse_uint8, NULL, 1 },
	{ "force_link_speed_ext", OPT_OFFSET(force_link_speed_ext), opts_parse_uint8, NULL, 1 },
	{ "force_link_speed_ext2", OPT_OFFSET(force_link_speed_ext2), opts_parse_uint8, NULL, 1 },
	{ "force_link_width", OPT_OFFSET(force_link_width), opts_parse_uint8, NULL, 1 },
	{ "fdr10", OPT_OFFSET(fdr10), opts_parse_uint8, NULL, 1 },
	{ "support_mepi_speeds", OPT_OFFSET(support_mepi_speeds), opts_parse_boolean, NULL, 1 },
	{ "mepi_enabled_speeds", OPT_OFFSET(mepi_enabled_speeds), opts_parse_uint8, NULL, 1 },
	{ "reassign_lids", OPT_OFFSET(reassign_lids), opts_parse_boolean, NULL, 1 },
	{ "ignore_other_sm", OPT_OFFSET(ignore_other_sm), opts_parse_boolean, NULL, 1 },
	{ "single_thread", OPT_OFFSET(single_thread), opts_parse_boolean, NULL, 0 },
	{ "disable_multicast", OPT_OFFSET(disable_multicast), opts_parse_boolean, NULL, 1 },
	{ "subnet_timeout", OPT_OFFSET(subnet_timeout), opts_parse_uint8, NULL, 1 },
	{ "packet_life_time", OPT_OFFSET(packet_life_time), opts_parse_uint8, NULL, 1 },
	{ "vl_stall_count", OPT_OFFSET(vl_stall_count), opts_parse_uint8, NULL, 1 },
	{ "leaf_vl_stall_count", OPT_OFFSET(leaf_vl_stall_count), opts_parse_uint8, NULL, 1 },
	{ "head_of_queue_lifetime", OPT_OFFSET(head_of_queue_lifetime), opts_parse_uint8, NULL, 1 },
	{ "leaf_head_of_queue_lifetime", OPT_OFFSET(leaf_head_of_queue_lifetime), opts_parse_uint8, NULL, 1 },
	{ "local_phy_errors_threshold", OPT_OFFSET(local_phy_errors_threshold), opts_parse_uint8, NULL, 1 },
	{ "overrun_errors_threshold", OPT_OFFSET(overrun_errors_threshold), opts_parse_uint8, NULL, 1 },
	{ "use_mfttop", OPT_OFFSET(use_mfttop), opts_parse_boolean, NULL, 1},
	{ "sminfo_polling_timeout", OPT_OFFSET(sminfo_polling_timeout), opts_parse_uint32, opts_setup_sminfo_polling_timeout, 1 },
	{ "polling_retry_number", OPT_OFFSET(polling_retry_number), opts_parse_uint32, NULL, 1 },
	{ "force_heavy_sweep", OPT_OFFSET(force_heavy_sweep), opts_parse_boolean, NULL, 1 },
	{ "port_prof_ignore_file", OPT_OFFSET(port_prof_ignore_file), opts_parse_charp, NULL, 0 },
	{ "hop_weights_file", OPT_OFFSET(hop_weights_file), opts_parse_charp, NULL, 0 },
	{ "dimn_ports_file", OPT_OFFSET(port_search_ordering_file), opts_parse_charp, NULL, 0 },
	{ "port_search_ordering_file", OPT_OFFSET(port_search_ordering_file), opts_parse_charp, NULL, 0 },
	{ "port_profile_switch_nodes", OPT_OFFSET(port_profile_switch_nodes), opts_parse_boolean, opts_setup_force_rerouting, 1 },
	{ "sweep_on_trap", OPT_OFFSET(sweep_on_trap), opts_parse_boolean, NULL, 1 },
	{ "routing_engine", OPT_OFFSET(routing_engine_names), opts_parse_charp, opts_setup_routing_engine, 1 },
	{ "enable_queries_during_routing", OPT_OFFSET(enable_queries_during_routing), opts_parse_boolean, NULL, 1 },
	{ "avoid_throttled_links", OPT_OFFSET(avoid_throttled_links), opts_parse_boolean, NULL, 0 },
	{ "connect_roots", OPT_OFFSET(connect_roots), opts_parse_boolean, opts_setup_force_rerouting, 1 },
	{ "calculate_missing_routes", OPT_OFFSET(calc_missing_routes), opts_parse_boolean, opts_setup_force_rerouting, 1 },
	{ "max_cas_on_spine", OPT_OFFSET(max_cas_on_spine), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1 },
	{ "find_roots_color_algorithm", OPT_OFFSET(find_roots_color_algorithm), opts_parse_boolean, NULL, 1},
	{ "dfp_find_roots_color_algorithm", OPT_OFFSET(dfp_find_roots_color_algorithm), opts_parse_boolean, NULL, 1},
	{ "use_ucast_cache", OPT_OFFSET(use_ucast_cache), opts_parse_boolean, NULL, 0 },
	{ "log_file", OPT_OFFSET(log_file), opts_parse_charp, NULL, 0 },
	{ "log_max_size", OPT_OFFSET(log_max_size), opts_parse_uint32, opts_setup_log_max_size, 1 },
	{ "log_flags", OPT_OFFSET(log_flags), opts_parse_uint8, opts_setup_log_flags, 1 },
	{ "force_log_flush", OPT_OFFSET(force_log_flush), opts_parse_boolean, opts_setup_force_log_flush, 1 },
	{ "accum_log_file", OPT_OFFSET(accum_log_file), opts_parse_boolean, opts_setup_accum_log_file, 1 },
	{ "partition_config_file", OPT_OFFSET(partition_config_file), opts_parse_charp, NULL, 0 },
	{ "no_partition_enforcement", OPT_OFFSET(no_partition_enforcement), opts_parse_boolean, NULL, 1 },
	{ "part_enforce", OPT_OFFSET(part_enforce), opts_parse_charp, NULL, 1 },
	{ "allow_both_pkeys", OPT_OFFSET(allow_both_pkeys), opts_parse_boolean, NULL, 0 },
	{ "keep_pkey_indexes", OPT_OFFSET(keep_pkey_indexes), opts_parse_boolean, NULL, 1 },
	{ "sm_assigned_guid", OPT_OFFSET(sm_assigned_guid), opts_parse_uint8, NULL, 1 },
	{ "qos", OPT_OFFSET(qos), opts_parse_boolean, opts_setup_ucast_invalidate_options, 1 },
	{ "qos_policy_file", OPT_OFFSET(qos_policy_file), opts_parse_charp, NULL, 0 },
	{ "suppress_sl2vl_mad_status_errors", OPT_OFFSET(suppress_sl2vl_mad_status_errors), opts_parse_boolean, NULL, 1 },
	{ "override_create_mcg_sl", OPT_OFFSET(override_create_mcg_sl), opts_parse_uint8, NULL, 1 },
	{ "dump_files_dir", OPT_OFFSET(dump_files_dir), opts_parse_charp, NULL, 0 },
	{ "lid_matrix_dump_file", OPT_OFFSET(lid_matrix_dump_file), opts_parse_charp, NULL, 0 },
	{ "lfts_file", OPT_OFFSET(lfts_file), opts_parse_charp, NULL, 0 },
	{ "root_guid_file", OPT_OFFSET(root_guid_file), opts_parse_charp, NULL, 0 },
	{ "pgrp_policy_file", OPT_OFFSET(pgrp_policy_file), opts_parse_charp, NULL, 0 },
	{ "topo_policy_file", OPT_OFFSET(topo_policy_file), opts_parse_charp, NULL, 0 },
	{ "rch_policy_file", OPT_OFFSET(rch_policy_file), opts_parse_charp, NULL, 0 },
	{ "device_configuration_file", OPT_OFFSET(device_configuration_file), opts_parse_charp, NULL, 0 },
	{ "fast_recovery_enabled", OPT_OFFSET(fast_recovery_enabled), opts_parse_uint8, NULL, 0 },
	{ "fast_recovery_conf_file", OPT_OFFSET(fast_recovery_conf_file), opts_parse_charp, NULL, 0 },
	{ "ftree_ca_order_dump_file", OPT_OFFSET(ftree_ca_order_dump_file), opts_parse_charp, NULL, 0 },
	{ "held_back_sw_file", OPT_OFFSET(held_back_sw_file), opts_parse_charp, NULL, 0 },
	{ "cn_guid_file", OPT_OFFSET(cn_guid_file), opts_parse_charp, NULL, 0 },
	{ "io_guid_file", OPT_OFFSET(io_guid_file), opts_parse_charp, NULL, 0 },
	{ "port_shifting", OPT_OFFSET(port_shifting), opts_parse_boolean, opts_setup_force_rerouting, 1 },
	{ "scatter_ports", OPT_OFFSET(scatter_ports), opts_parse_uint32, opts_setup_force_rerouting, 1 },
	{ "updn_lid_tracking_mode", OPT_OFFSET(updn_lid_tracking_mode), opts_parse_boolean, opts_setup_ucast_invalidate_options, 1 },
	{ "updn_lid_tracking_converge_routes", OPT_OFFSET(updn_lid_tracking_converge_routes),
			opts_parse_boolean, opts_setup_ucast_invalidate_options, 1 },
	{ "updn_lid_tracking_prefer_total_routes", OPT_OFFSET(updn_lid_tracking_prefer_total_routes),
			opts_parse_boolean, opts_setup_ucast_invalidate_options, 1 },
	{ "dfp_max_cas_on_spine", OPT_OFFSET(dfp_max_cas_on_spine), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1 },
	{ "multicast_optimization", OPT_OFFSET(const_multicast), opts_parse_boolean, NULL, 0},
	{ "max_seq_redisc", OPT_OFFSET(max_seq_redisc), opts_parse_uint32, NULL, 1 },
	{ "rereg_on_guid_migr", OPT_OFFSET(rereg_on_guid_migr), opts_parse_boolean, NULL, 0},
	{ "aguid_inout_notice", OPT_OFFSET(aguid_inout_notice), opts_parse_boolean, NULL, 1},
	{ "sm_assign_guid_func", OPT_OFFSET(sm_assign_guid_func), opts_parse_charp, NULL, 1},
	{ "mc_primary_root_guid", OPT_OFFSET(mc_primary_root_guid), opts_parse_net64, NULL, 1},
	{ "mc_secondary_root_guid", OPT_OFFSET(mc_secondary_root_guid), opts_parse_net64, NULL, 1},
	{ "mc_roots_file", OPT_OFFSET(mc_roots_file), opts_parse_charp, NULL, 0 },
	{ "max_reverse_hops", OPT_OFFSET(max_reverse_hops), opts_parse_uint16, opts_setup_force_rerouting, 1 },
	{ "routing_threads_num", OPT_OFFSET(routing_threads_num), opts_parse_uint16, opts_setup_ucast_reinit_threads, 1 },
	{ "max_threads_per_core", OPT_OFFSET(max_threads_per_core), opts_parse_uint16, opts_setup_ucast_reinit_threads, 1 },
	{ "ids_guid_file", OPT_OFFSET(ids_guid_file), opts_parse_charp, NULL, 0 },
	{ "guid_routing_order_file", OPT_OFFSET(guid_routing_order_file), opts_parse_charp, NULL, 0 },
	{ "guid_routing_order_no_scatter", OPT_OFFSET(guid_routing_order_no_scatter), opts_parse_boolean,
					   opts_setup_force_rerouting, 1 },
	{ "use_scatter_for_switch_lid", OPT_OFFSET(use_scatter_for_switch_lid), opts_parse_boolean,
					opts_setup_force_rerouting, 1 },
	{ "offsweep_balancing_enabled", OPT_OFFSET(offsweep_balancing_enabled), opts_parse_boolean,
					opts_setup_force_rerouting, 1},
	{ "offsweep_balancing_window", OPT_OFFSET(offsweep_balancing_window), opts_parse_uint32,
					NULL, 1},
	{ "pqft_structure", OPT_OFFSET(pqft_structure), opts_parse_charp, NULL, 0 },
	{ "sa_db_file", OPT_OFFSET(sa_db_file), opts_parse_charp, NULL, 0 },
	{ "sa_db_dump", OPT_OFFSET(sa_db_dump), opts_parse_boolean, NULL, 1 },
	{ "sm_db_dump", OPT_OFFSET(sm_db_dump), opts_parse_boolean, NULL, 1 },
	{ "torus_config", OPT_OFFSET(torus_conf_file), opts_parse_charp, NULL, 1 },
	{ "do_mesh_analysis", OPT_OFFSET(do_mesh_analysis), opts_parse_boolean, NULL, 1 },
	{ "exit_on_fatal", OPT_OFFSET(exit_on_fatal), opts_parse_boolean, NULL, 1 },
	{ "honor_guid2lid_file", OPT_OFFSET(honor_guid2lid_file), opts_parse_boolean, NULL, 1 },
	{ "daemon", OPT_OFFSET(daemon), opts_parse_boolean, NULL, 0 },
	{ "sm_inactive", OPT_OFFSET(sm_inactive), opts_parse_boolean, NULL, 1 },
	{ "babbling_port_policy", OPT_OFFSET(babbling_port_policy), opts_parse_boolean, NULL, 1 },
	{ "drop_subscr_on_report_fail", OPT_OFFSET(drop_subscr_on_report_fail), opts_parse_boolean, NULL, 1 },
	{ "drop_event_subscriptions", OPT_OFFSET(drop_event_subscriptions), opts_parse_boolean, NULL, 1 },
	{ "drop_unreachable_event_subscriptions", OPT_OFFSET(drop_unreachable_event_subscriptions), opts_parse_boolean, NULL, 1 },
	{ "ipoib_mcgroup_creation_validation", OPT_OFFSET(ipoib_mcgroup_creation_validation), opts_parse_boolean, NULL, 1 },
	{ "mcgroup_join_validation", OPT_OFFSET(mcgroup_join_validation), opts_parse_boolean, NULL, 1 },
	{ "use_original_extended_sa_rates_only", OPT_OFFSET(use_original_extended_sa_rates_only), opts_parse_boolean, NULL, 1 },
	{ "max_rate_enum", OPT_OFFSET(max_rate_enum), opts_parse_uint8, NULL, 1},
	{ "reports", OPT_OFFSET(reports), opts_parse_uint8, NULL, 1 },
	{ "use_optimized_slvl", OPT_OFFSET(use_optimized_slvl), opts_parse_boolean, NULL, 1 },
	{ "use_optimized_port_mask_slvl", OPT_OFFSET(use_optimized_port_mask_slvl), opts_parse_boolean, NULL, 1 },
	{ "fsync_high_avail_files", OPT_OFFSET(fsync_high_avail_files), opts_parse_boolean, NULL, 1 },
	{ "default_mcg_mtu", OPT_OFFSET(default_mcg_mtu), opts_parse_uint8, NULL, 1 },
	{ "default_mcg_rate", OPT_OFFSET(default_mcg_rate), opts_parse_uint8, NULL, 1 },
#ifdef ENABLE_OSM_PERF_MGR
	{ "perfmgr", OPT_OFFSET(perfmgr), opts_parse_boolean, NULL, 0 },
	{ "perfmgr_redir", OPT_OFFSET(perfmgr_redir), opts_parse_boolean, NULL, 0 },
	{ "perfmgr_sweep_time_s", OPT_OFFSET(perfmgr_sweep_time_s), opts_parse_uint16, NULL, 0 },
	{ "perfmgr_max_outstanding_queries", OPT_OFFSET(perfmgr_max_outstanding_queries), opts_parse_uint32, NULL, 0 },
	{ "perfmgr_ignore_cas", OPT_OFFSET(perfmgr_ignore_cas), opts_parse_boolean, NULL, 0 },
	{ "event_db_dump_file", OPT_OFFSET(event_db_dump_file), opts_parse_charp, NULL, 0 },
	{ "perfmgr_rm_nodes", OPT_OFFSET(perfmgr_rm_nodes), opts_parse_boolean, NULL, 0 },
	{ "perfmgr_log_errors", OPT_OFFSET(perfmgr_log_errors), opts_parse_boolean, NULL, 0 },
	{ "perfmgr_query_cpi", OPT_OFFSET(perfmgr_query_cpi), opts_parse_boolean, NULL, 0 },
	{ "perfmgr_xmit_wait_log", OPT_OFFSET(perfmgr_xmit_wait_log), opts_parse_boolean, NULL, 0 },
	{ "perfmgr_xmit_wait_threshold", OPT_OFFSET(perfmgr_xmit_wait_threshold), opts_parse_uint32, NULL, 0 },
#endif				/* ENABLE_OSM_PERF_MGR */
	{ "plugin_name", OPT_OFFSET(plugin_name), opts_parse_charp, NULL, 0 },
	{ "plugin_options", OPT_OFFSET(plugin_options), opts_parse_charp, NULL, 0 },
	{ "event_plugin_name", OPT_OFFSET(event_plugin_name), opts_parse_charp, NULL, 0 },
	{ "event_plugin_options", OPT_OFFSET(event_plugin_options), opts_parse_charp, NULL, 0 },
	{ "node_name_map_name", OPT_OFFSET(node_name_map_name), opts_parse_charp, NULL, 0 },
	{ "qos_max_vls", OPT_OFFSET(qos_options.max_vls), opts_parse_uint32, NULL, 1 },
	{ "qos_high_limit", OPT_OFFSET(qos_options.high_limit), opts_parse_int32, NULL, 1 },
	{ "qos_vlarb_high", OPT_OFFSET(qos_options.vlarb_high), opts_parse_charp, NULL, 1 },
	{ "qos_vlarb_low", OPT_OFFSET(qos_options.vlarb_low), opts_parse_charp, NULL, 1 },
	{ "qos_sl2vl", OPT_OFFSET(qos_options.sl2vl), opts_parse_charp, NULL, 1 },
	{ "qos_ca_max_vls", OPT_OFFSET(qos_ca_options.max_vls), opts_parse_uint32, NULL, 1 },
	{ "qos_ca_high_limit", OPT_OFFSET(qos_ca_options.high_limit), opts_parse_int32, NULL, 1 },
	{ "qos_ca_vlarb_high", OPT_OFFSET(qos_ca_options.vlarb_high), opts_parse_charp, NULL, 1 },
	{ "qos_ca_vlarb_low", OPT_OFFSET(qos_ca_options.vlarb_low), opts_parse_charp, NULL, 1 },
	{ "qos_ca_sl2vl", OPT_OFFSET(qos_ca_options.sl2vl), opts_parse_charp, NULL, 1 },
	{ "qos_sw0_max_vls", OPT_OFFSET(qos_sw0_options.max_vls), opts_parse_uint32, NULL, 1 },
	{ "qos_sw0_high_limit", OPT_OFFSET(qos_sw0_options.high_limit), opts_parse_int32, NULL, 1 },
	{ "qos_sw0_vlarb_high", OPT_OFFSET(qos_sw0_options.vlarb_high), opts_parse_charp, NULL, 1 },
	{ "qos_sw0_vlarb_low", OPT_OFFSET(qos_sw0_options.vlarb_low), opts_parse_charp, NULL, 1 },
	{ "qos_sw0_sl2vl", OPT_OFFSET(qos_sw0_options.sl2vl), opts_parse_charp, NULL, 1 },
	{ "qos_swe_max_vls", OPT_OFFSET(qos_swe_options.max_vls), opts_parse_uint32, NULL, 1 },
	{ "qos_swe_high_limit", OPT_OFFSET(qos_swe_options.high_limit), opts_parse_int32, NULL, 1 },
	{ "qos_swe_vlarb_high", OPT_OFFSET(qos_swe_options.vlarb_high), opts_parse_charp, NULL, 1 },
	{ "qos_swe_vlarb_low", OPT_OFFSET(qos_swe_options.vlarb_low), opts_parse_charp, NULL, 1 },
	{ "qos_swe_sl2vl", OPT_OFFSET(qos_swe_options.sl2vl), opts_parse_charp, NULL, 1 },
	{ "qos_sw2sw_max_vls", OPT_OFFSET(qos_sw2sw_options.max_vls), opts_parse_uint32, NULL, 1 },
	{ "qos_sw2sw_high_limit", OPT_OFFSET(qos_sw2sw_options.high_limit), opts_parse_int32, NULL, 1 },
	{ "qos_sw2sw_vlarb_high", OPT_OFFSET(qos_sw2sw_options.vlarb_high), opts_parse_charp, NULL, 1 },
	{ "qos_sw2sw_vlarb_low", OPT_OFFSET(qos_sw2sw_options.vlarb_low), opts_parse_charp, NULL, 1 },
	{ "qos_sw2sw_sl2vl", OPT_OFFSET(qos_sw2sw_options.sl2vl), opts_parse_charp, NULL, 1 },
	{ "qos_rtr_max_vls", OPT_OFFSET(qos_rtr_options.max_vls), opts_parse_uint32, NULL, 1 },
	{ "qos_rtr_high_limit", OPT_OFFSET(qos_rtr_options.high_limit), opts_parse_int32, NULL, 1 },
	{ "qos_rtr_vlarb_high", OPT_OFFSET(qos_rtr_options.vlarb_high), opts_parse_charp, NULL, 1 },
	{ "qos_rtr_vlarb_low", OPT_OFFSET(qos_rtr_options.vlarb_low), opts_parse_charp, NULL, 1 },
	{ "qos_rtr_sl2vl", OPT_OFFSET(qos_rtr_options.sl2vl), opts_parse_charp, NULL, 1 },
	{ "mlnx_congestion_control", OPT_OFFSET(mlnx_congestion_control), opts_parse_uint8, opts_setup_mlnx_cc, 1 },
	{ "cc_key", OPT_OFFSET(cc_key), opts_parse_net64, NULL, 0},
	{ "cc_max_outstanding_mads", OPT_OFFSET(cc_max_outstanding_mads), opts_parse_uint32, NULL, 0 },
	{ "cc_key_enable", OPT_OFFSET(cc_key_enable), opts_parse_uint8, opts_setup_cc_key_enable, 1 },
	{ "cc_key_lease_period", OPT_OFFSET(cc_key_lease_period), opts_parse_uint16, opts_setup_cc_key_lease_period, 1 },
	{ "cc_key_protect_bit", OPT_OFFSET(cc_key_protect_bit), opts_parse_uint8, opts_setup_cc_key_protect_bit, 1 },
	{ "vs_key_enable", OPT_OFFSET(vs_key_enable), opts_parse_uint8, opts_setup_vs_key_enable, 1 },
	{ "vs_key_lease_period", OPT_OFFSET(vs_key_lease_period), opts_parse_uint16, opts_setup_vs_key_lease_period, 1 },
	{ "vs_key_ci_protect_bits", OPT_OFFSET(vs_key_ci_protect_bits), opts_parse_uint8, opts_setup_vs_key_ci_protect_bits, 1 },
	{ "vs_max_outstanding_mads", OPT_OFFSET(vs_max_outstanding_mads), opts_parse_uint32, NULL, 0 },
	{ "n2n_key_enable", OPT_OFFSET(n2n_key_enable), opts_parse_uint8, opts_setup_n2n_key_enable, 1 },
	{ "n2n_key_lease_period", OPT_OFFSET(n2n_key_lease_period), opts_parse_uint16, opts_setup_n2n_key_params, 1 },
	{ "n2n_key_protect_bit", OPT_OFFSET(n2n_key_protect_bit), opts_parse_uint8, opts_setup_n2n_key_params, 1 },
	{ "n2n_max_outstanding_mads", OPT_OFFSET(n2n_max_outstanding_mads), opts_parse_uint32, NULL, 0 },
	{ "key_mgr_seed", OPT_OFFSET(key_mgr_seed), opts_parse_net64, opts_setup_key_mgr_seed, 1 },
	{ "congestion_control_policy_file", OPT_OFFSET(congestion_control_policy_file), opts_parse_charp, NULL, 0 },
	{ "ppcc_algo_dir", OPT_OFFSET(ppcc_algo_dir), opts_parse_charp, NULL, 0 },
	{ "enable_quirks", OPT_OFFSET(enable_quirks), opts_parse_boolean, NULL, 1 },
	{ "no_clients_rereg", OPT_OFFSET(no_clients_rereg), opts_parse_boolean, NULL, 1 },
	{ "client_rereg_mode", OPT_OFFSET(client_rereg_mode), opts_parse_uint8, NULL, 1 },
	{ "prefix_routes_file", OPT_OFFSET(prefix_routes_file), opts_parse_charp, NULL, 0 },
	{ "consolidate_ipv6_snm_req", OPT_OFFSET(consolidate_ipv6_snm_req), opts_parse_boolean, NULL, 1 },
	{ "consolidate_ipv4_mask", OPT_OFFSET(consolidate_ipv4_mask), opts_parse_uint32, NULL, 1 },
	{ "pid_file", OPT_OFFSET(pid_file), opts_parse_charp, NULL, 0},
#ifdef ENABLE_MLNX_OS
	{ "ca_port_num_limit", OPT_OFFSET(ca_port_num_limit), opts_parse_uint32, NULL, 1 },
	{ "ca_port_soft_limit", OPT_OFFSET(ca_port_soft_limit), opts_parse_uint32, NULL, 1 },
	{ "switch_name", OPT_OFFSET(switch_name), opts_parse_charp, NULL, 1 },
#endif
	{ "lash_start_vl", OPT_OFFSET(lash_start_vl), opts_parse_uint8, opts_setup_force_rerouting, 1 },
	{ "sm_sl", OPT_OFFSET(sm_sl), opts_parse_uint8, NULL, 1 },
	{ "log_prefix", OPT_OFFSET(log_prefix), opts_parse_charp, NULL, 1 },
	{ "per_module_logging_file", OPT_OFFSET(per_module_logging_file), opts_parse_charp, NULL, 0 },
	{ "max_msg_fifo_len", OPT_OFFSET(max_msg_fifo_len), opts_parse_uint32, NULL, 1},
	{ "max_alt_dr_path_retries", OPT_OFFSET(max_alt_dr_path_retries), opts_parse_uint32, NULL, 1 },
	{ "sa_pr_full_world_queries_allowed", OPT_OFFSET(sa_pr_full_world_queries_allowed), opts_parse_boolean, NULL, 1},
	{ "enable_crashd", OPT_OFFSET(enable_crashd), opts_parse_boolean, NULL, 0 },
	{ "hm_unhealthy_ports_checks", OPT_OFFSET(hm_unhealthy_ports_checks), opts_parse_boolean, NULL, 0},
	{ "hm_ports_health_policy_file",  OPT_OFFSET(hm_ports_health_policy_file), opts_parse_charp, NULL, 1},
	{ "hm_ca_reboot_action", OPT_OFFSET(hm_ca_reboot_action_s), opts_parse_charp, NULL, 0},
	{ "hm_ca_unresponsive_action", OPT_OFFSET(hm_ca_unresponsive_action_s), opts_parse_charp, NULL, 0},
	{ "hm_ca_noisy_action", OPT_OFFSET(hm_ca_noisy_action_s), opts_parse_charp, NULL, 0},
	{ "hm_ca_seterr_action", OPT_OFFSET(hm_ca_seterr_action_s), opts_parse_charp, NULL, 0},
	{ "hm_ca_flapping_action", OPT_OFFSET(hm_ca_flapping_action_s), opts_parse_charp, NULL, 0},
	{ "hm_ca_illegal_action", OPT_OFFSET(hm_ca_illegal_action_s), opts_parse_charp, NULL, 0},
	{ "hm_ca_manual_action", OPT_OFFSET(hm_ca_manual_action_s), opts_parse_charp, NULL, 0},
	{ "hm_sw_reboot_action", OPT_OFFSET(hm_sw_reboot_action_s), opts_parse_charp, NULL, 0},
	{ "hm_sw_unresponsive_action", OPT_OFFSET(hm_sw_unresponsive_action_s), opts_parse_charp, NULL, 0},
	{ "hm_sw_noisy_action", OPT_OFFSET(hm_sw_noisy_action_s), opts_parse_charp, NULL, 0},
	{ "hm_sw_seterr_action", OPT_OFFSET(hm_sw_seterr_action_s), opts_parse_charp, NULL, 0},
	{ "hm_sw_flapping_action", OPT_OFFSET(hm_sw_flapping_action_s), opts_parse_charp, NULL, 0},
	{ "hm_sw_illegal_action", OPT_OFFSET(hm_sw_illegal_action_s), opts_parse_charp, NULL, 0},
	{ "hm_sw_manual_action", OPT_OFFSET(hm_sw_manual_action_s), opts_parse_charp, NULL, 0},
	{ "hm_num_reboots", OPT_OFFSET(hm_num_reboots), opts_parse_uint32, NULL, 0},
	{ "hm_reboots_period_secs", OPT_OFFSET(hm_reboots_period_secs), opts_parse_uint32, NULL, 0},
	{ "hm_num_set_err_sweeps", OPT_OFFSET(hm_num_set_err_sweeps), opts_parse_uint32, NULL, 0},
	{ "hm_num_set_err_sweeps_window", OPT_OFFSET(hm_num_set_err_sweeps_window), opts_parse_uint32, NULL, 0},
	{ "hm_num_no_resp_sweeps", OPT_OFFSET(hm_num_no_resp_sweeps), opts_parse_uint32, NULL, 0},
	{ "hm_num_no_resp_sweeps_window", OPT_OFFSET(hm_num_no_resp_sweeps_window), opts_parse_uint32, NULL, 0},
	{ "hm_num_traps", OPT_OFFSET(hm_num_traps), opts_parse_uint32, NULL, 0},
	{ "hm_num_traps_period_secs", OPT_OFFSET(hm_num_traps_period_secs), opts_parse_uint32, NULL, 0},
	{ "hm_num_flapping_sweeps", OPT_OFFSET(hm_num_flapping_sweeps), opts_parse_uint32, NULL, 0},
	{ "hm_num_flapping_sweeps_window", OPT_OFFSET(hm_num_flapping_sweeps_window), opts_parse_uint32, NULL, 0},
	{ "hm_num_illegal", OPT_OFFSET(hm_num_illegal), opts_parse_uint32, NULL, 0},
	{ "force_heavy_sweep_window", OPT_OFFSET(force_heavy_sweep_window), opts_parse_int32, NULL, 1},
	{ "validate_smps", OPT_OFFSET(validate_smps), opts_parse_boolean, NULL, 1},
	{ "max_topologies_per_sw", OPT_OFFSET(max_topologies_per_sw), opts_parse_uint8, NULL, 0},
	{ "auto_join_hca_wo_topo", OPT_OFFSET(auto_join_hca_wo_topo), opts_parse_boolean, NULL, 0},
	{ "enable_inc_mc_routing", OPT_OFFSET(enable_inc_mc_routing), opts_parse_boolean, opts_setup_enable_inc_mc_routing, 1},
	{ "allow_sm_port_reset", OPT_OFFSET(allow_sm_port_reset), opts_parse_boolean, NULL, 1},
	{ "port_speed_change_action", OPT_OFFSET(port_speed_change_action_s), opts_parse_charp, NULL, 1},
	{ "port_ext_speed_change_action", OPT_OFFSET(port_ext_speed_change_action_s), opts_parse_charp, NULL, 1},
	{ "port_ext_speed2_change_action", OPT_OFFSET(port_ext_speed2_change_action_s), opts_parse_charp, NULL, 1},
	{ "port_mepi_speed_change_action", OPT_OFFSET(port_mepi_speed_change_action_s), opts_parse_charp, NULL, 1},
	{ "port_mtu_change_action", OPT_OFFSET(port_mtu_change_action_s), opts_parse_charp, NULL, 1},
	{ "port_vl_change_action", OPT_OFFSET(port_vl_change_action_s), opts_parse_charp, NULL, 1},
	{ "port_ame_bit_change_action", OPT_OFFSET(port_ame_bit_change_action_s), opts_parse_charp, NULL, 1},
	{ "support_mlnx_enhanced_link", OPT_OFFSET(support_mlnx_enhanced_link), opts_parse_boolean, NULL, 1 },
	{ "mlnx_enhanced_link_enable", OPT_OFFSET(mlnx_enhanced_link_enable), opts_parse_boolean, NULL, 1 },
	{ "adaptive_timeout_sl_mask", OPT_OFFSET(adaptive_timeout_sl_mask), opts_parse_uint16, NULL, 1 },
	{ "mepi_cache_enabled", OPT_OFFSET(mepi_cache_enabled), opts_parse_boolean, NULL, 1 },
	{ "virt_enabled", OPT_OFFSET(virt_enabled), opts_parse_uint32, NULL, 0},
	{ "virt_max_ports_in_process", OPT_OFFSET(virt_max_ports_in_process), opts_parse_uint16, NULL, 1 },
	{ "virt_default_hop_limit", OPT_OFFSET(virt_default_hop_limit), opts_parse_uint8, NULL, 1 },
	{ "enable_virt_rec_ext", OPT_OFFSET(enable_virt_rec_ext), opts_parse_boolean, NULL, 1},
	{ "improved_lmc_path_distribution", OPT_OFFSET(improved_lmc_path_distribution),opts_parse_boolean,
								opts_setup_ucast_invalidate_options, 1},
	{ "syslog_log_flags", OPT_OFFSET(syslog_log_flags), opts_parse_uint8, opts_setup_syslog_log_flags, 1 },
	{ "sweep_every_hup_signal", OPT_OFFSET(sweep_every_hup_signal), opts_parse_boolean, NULL, 1 },
	{ "osm_stats_interval", OPT_OFFSET(osm_stats_interval), opts_parse_uint32_minutes_to_milisec, opts_setup_change_osm_stats_interval, 1 },
	{ "osm_stats_dump_limit", OPT_OFFSET(osm_stats_dump_limit), opts_parse_uint32, opts_setup_change_dump_limit, 1 },
	{ "osm_stats_dump_per_sm_port", OPT_OFFSET(osm_stats_dump_per_sm_port), opts_parse_boolean, NULL, 0 },
	{ "osm_perflog_dump_limit", OPT_OFFSET(osm_perflog_dump_limit), opts_parse_uint32, opts_setup_change_perflog_dump_limit, 1 },
	{ "enable_performance_logging", OPT_OFFSET(enable_performance_logging), opts_parse_boolean, NULL, 1  },
	{ "quasi_ftree_indexing", OPT_OFFSET(quasi_ftree_indexing), opts_parse_boolean, opts_setup_force_rerouting, 1 },
	{ "sharp_enabled", OPT_OFFSET(sharp_enabled), opts_parse_uint8, NULL, 1},
	{ "rtr_aguid_enable", OPT_OFFSET(rtr_aguid_enable), opts_parse_uint8, NULL, 1 },
	{ "rtr_pr_flow_label", OPT_OFFSET(rtr_pr_flow_label), opts_parse_uint32, NULL, 1},
	{ "rtr_pr_tclass", OPT_OFFSET(rtr_pr_tclass), opts_parse_uint8, NULL, 1},
	{ "rtr_pr_sl", OPT_OFFSET(rtr_pr_sl), opts_parse_uint8, NULL, 1},
	{ "rtr_pr_mtu", OPT_OFFSET(rtr_pr_mtu), opts_parse_uint8, NULL, 1},
	{ "rtr_pr_rate", OPT_OFFSET(rtr_pr_rate), opts_parse_uint8, NULL, 1},
	{ "aguid_default_hop_limit", OPT_OFFSET(aguid_default_hop_limit), opts_parse_uint8, NULL, 1 },
	{ "router_policy_file", OPT_OFFSET(router_policy_file), opts_parse_charp, NULL, 0 },
	{ "router_policy_enabled", OPT_OFFSET(router_policy_enabled), opts_parse_boolean, NULL, 0 },
	{ "load_inform_info_from_sadb", OPT_OFFSET(load_inform_info_from_sadb), opts_parse_boolean, NULL, 0 },
	{ "verbose_bypass_policy_file", OPT_OFFSET(verbose_bypass_policy_file), opts_parse_charp, NULL, 1 },
	{ "sa_db_dump_file", OPT_OFFSET(sa_db_dump_file), opts_parse_charp, NULL, 0 },
	{ "enable_subnet_lst", OPT_OFFSET(enable_subnet_lst), opts_parse_boolean, NULL, 1 },
	{ "subnet_lst_file", OPT_OFFSET(subnet_lst_file), opts_parse_charp, NULL, 0 },
	{ "virt_dump_file", OPT_OFFSET(virt_dump_file), opts_parse_charp, NULL, 0 },
	{ "dor_hyper_cube_mode", OPT_OFFSET(dor_hyper_cube_mode), opts_parse_boolean, NULL, 1},
	{ "standalone", OPT_OFFSET(stand_alone), opts_parse_uint8, NULL, 0 },
	{ "additional_gi_supporting_devices", OPT_OFFSET(additional_gi_supporting_devices), opts_parse_charp, NULL, 1 },
	{ "additional_mepi_force_devices", OPT_OFFSET(additional_mepi_force_devices), opts_parse_charp, NULL, 1 },
	{ "disable_default_gi_support", OPT_OFFSET(disable_default_gi_support), opts_parse_boolean, NULL, 0 },
	{ "activity_report_subjects", OPT_OFFSET(activity_report_subjects), opts_parse_charp, NULL, 1 },
	{ "enhanced_qos_policy_file", OPT_OFFSET(enhanced_qos_policy_file), opts_parse_charp, NULL, 0 },
	{ "enhanced_qos_vport0_unlimit_default_rl", OPT_OFFSET(enhanced_qos_vport0_unlimit_default_rl), opts_parse_boolean, NULL, 1 },
	{ "adv_routing_engine", OPT_OFFSET(adv_routing_engine_name), opts_parse_charp, opts_setup_adv_routing_engine, 1 },
	{ "ar_mode", OPT_OFFSET(ar_mode), opts_parse_uint8, opts_setup_adv_routing_mode, 1 },
	{ "shield_mode", OPT_OFFSET(shield_mode), opts_parse_uint8, opts_setup_adv_routing_mode, 1 },
	{ "ar_sl_mask", OPT_OFFSET(ar_sl_mask), opts_parse_uint16, opts_setup_ar_mask, 1 },
	{ "enable_ar_by_device_cap", OPT_OFFSET(enable_ar_by_device_cap), opts_parse_boolean, opts_setup_enable_ar_by_device_cap, 1 },
	{ "enable_ar_group_copy", OPT_OFFSET(enable_ar_group_copy), opts_parse_boolean, opts_setup_ar_mask, 1},
	{ "ar_transport_mask", OPT_OFFSET(ar_transport_mask), opts_parse_uint8, opts_setup_ar_mask, 1},
	{ "ar_tree_asymmetric_flow", OPT_OFFSET(ar_tree_asymmetric_flow), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1 },
	{ "ar_tree_asymmetric_flow_threshold", OPT_OFFSET(ar_tree_asymmetric_flow_threshold), opts_parse_uint32, opts_setup_ucast_invalidate_options, 1 },
	{ "ar_tree_asymmetric_flow_threshold_limit", OPT_OFFSET(ar_tree_asymmetric_flow_threshold_limit), opts_parse_uint32, opts_setup_ucast_invalidate_options, 1 },
	{ "routing_flags", OPT_OFFSET(routing_flags), opts_parse_uint64, opts_setup_ucast_invalidate_options, 1 },
	{ "smp_threads", OPT_OFFSET(smp_threads), opts_parse_uint16, NULL, 0 },
	{ "gmp_threads", OPT_OFFSET(gmp_threads), opts_parse_uint16, NULL, 0 },
	{ "gmp_traps_threads_num", OPT_OFFSET(gmp_traps_threads_num), opts_parse_uint16, NULL, 0 },
	{ "dump_ar", OPT_OFFSET(dump_ar), opts_parse_boolean, NULL, 1},
	{ "cache_ar_group_id", OPT_OFFSET(cache_ar_group_id), opts_parse_boolean, NULL, 1},
	{ "dfp_down_up_turns_mode", OPT_OFFSET(dfp_down_up_turns_mode), opts_parse_uint8, opts_setup_ucast_invalidate_options, 1},
	{ "enable_vl_packing", OPT_OFFSET(enable_vl_packing), opts_parse_boolean, NULL, 1},
	{ "service_name2key_map_file", OPT_OFFSET(service_name2key_map_file), opts_parse_charp, NULL, 0},
	{ "topo_config_file", OPT_OFFSET(topo_config_file), opts_parse_charp, NULL, 0},
	{ "topo_config_enabled", OPT_OFFSET(topo_config_enabled), opts_parse_boolean, NULL, 1},
	{ "max_wire_smps_per_device", OPT_OFFSET(max_wire_smps_per_device), opts_parse_uint8, NULL, 0},
	{ "rtr_selection_function", OPT_OFFSET(rtr_selection_function_s), opts_parse_charp, opts_setup_rtr_selection_function, 1},
	{ "rtr_selection_seed", OPT_OFFSET(rtr_selection_seed), opts_parse_net64, NULL, 1},
	{ "rtr_selection_algo_parameters", OPT_OFFSET(rtr_selection_algo_parameters_s), opts_parse_charp , opts_setup_rtr_selection_algo_parameters, 1},
	{ "get_mft_tables", OPT_OFFSET(get_mft_tables), opts_parse_boolean, NULL, 1},
	{ "hbf_sl_mask", OPT_OFFSET(hbf_sl_mask), opts_parse_uint16, opts_setup_ar_mask, 1 },
	{ "hbf_hash_type", OPT_OFFSET(hbf_hash_type), opts_parse_uint8, opts_setup_ar_mask, 1},
	{ "hbf_seed_type", OPT_OFFSET(hbf_seed_type), opts_parse_uint8, opts_setup_ar_mask, 1},
	{ "hbf_seed", OPT_OFFSET(hbf_seed), opts_parse_uint32, opts_setup_ar_mask, 1},
	{ "hbf_hash_fields", OPT_OFFSET(hbf_hash_fields), opts_parse_net64, opts_setup_ar_mask, 1},
	{ "hbf_weights", OPT_OFFSET(hbf_weights), opts_parse_charp, opts_setup_hbf_weights, 1},
	{ "pfrn_sl", OPT_OFFSET(pfrn_sl), opts_parse_uint8, opts_setup_pfrn_config_change, 1 },
	{ "pfrn_mask_clear_timeout", OPT_OFFSET(pfrn_mask_clear_timeout), opts_parse_uint16, opts_setup_pfrn_config_change, 1  },
	{ "pfrn_mask_force_clear_timeout", OPT_OFFSET(pfrn_mask_force_clear_timeout), opts_parse_uint16, opts_setup_pfrn_config_change, 1  },
	{ "pfrn_over_router_enabled", OPT_OFFSET(pfrn_over_router_enabled), opts_parse_uint8, opts_setup_pfrn_config_change, 1},
#if OSM_ENABLE_NVLINK
	{ "nvlink_enable", OPT_OFFSET(nvlink_enable), opts_parse_boolean, NULL, 0},
	{ "nvl_reassign_rails", OPT_OFFSET(nvl_reassign_rails), opts_parse_boolean, NULL, 1 },
	{ "nvl_hbf_hash_type", OPT_OFFSET(nvl_hbf_hash_type), opts_parse_uint8, opts_setup_ar_mask, 1},
	{ "nvl_hbf_packet_hash_bitmask", OPT_OFFSET(nvl_hbf_packet_hash_bitmask), opts_parse_uint8, opts_setup_ar_mask, 1},
	{ "nvl_hbf_seed", OPT_OFFSET(nvl_hbf_seed), opts_parse_uint32, opts_setup_ar_mask, 1},
	{ "nvl_hbf_hash_fields", OPT_OFFSET(nvl_hbf_fields_enable), opts_parse_uint64, opts_setup_ar_mask, 1},
	{ "nvl_partition_reroute_wait", OPT_OFFSET(nvl_partition_reroute_wait), opts_parse_uint8, NULL, 1},
#endif
	{ "tenants_policy_enabled", OPT_OFFSET(tenants_policy_enabled), opts_parse_boolean, osm_tenant_mgr_config_change, 1},
	{ "tenants_policy_file", OPT_OFFSET(tenants_policy_file), opts_parse_charp, NULL, 0},
	{ "reply_lid_smps_in_dr", OPT_OFFSET(reply_lid_smps_in_dr), opts_parse_boolean, NULL, 1},
	{ "respond_unknown_lid_traps", OPT_OFFSET(respond_unknown_lid_traps), opts_parse_boolean, NULL, 1},
	{ "fabric_mode_profile", OPT_OFFSET(fabric_mode_profile), opts_parse_charp, NULL, 1},
	{ "fabric_mode_policy_file", OPT_OFFSET(fabric_mode_policy_file), opts_parse_charp, NULL, 0},
	{ "issu_mode", OPT_OFFSET(issu_mode), opts_parse_uint32, osm_issu_mgr_mode_change, 1},
	{ "issu_timeout", OPT_OFFSET(issu_timeout), opts_parse_uint32, osm_issu_mgr_timeout_change, 1},
	{ "issu_pre_upgrade_time", OPT_OFFSET(issu_pre_upgrade_time), opts_parse_uint32, osm_issu_mgr_pre_upgrade_time_change, 1},
	{0}
};

static int compare_traps(const void *t1, const void *t2)
{
	osm_inform_work_obj_t *trap1, *trap2;

	trap1 = (* (osm_inform_work_obj_t **) t1);
	trap2 = (* (osm_inform_work_obj_t **) t2);

	/* TODO - worry about GSI traps !!! */

	/* First, compare issuer LIDs */
	if (cl_ntoh16(trap1->ntc.issuer_lid) < cl_hton16(trap2->ntc.issuer_lid))
		return -1;
	if (cl_ntoh16(trap1->ntc.issuer_lid) > cl_hton16(trap2->ntc.issuer_lid))
		return 1;

	/* If same issuer LID, compare TIDs */
	if (trap1->tid < trap2->tid)
		return -1;
	if (trap1->tid > trap2->tid)
		return 1;

	return 0;
}

static int compare_unhealthy_ports(const void *p1, const void *p2)
{
	osm_unhealthy_port_t *port1, *port2;
	port1 = (* (osm_unhealthy_port_t **) p1);
	port2 = (* (osm_unhealthy_port_t **) p2);

	/* Compare GUIDs */
	if (port1->peer_guid < port2->peer_guid)
		return -1;
	if (port1->peer_guid > port2->peer_guid)
		return 1;

	/* If same GUIDs, compare port number */
	if (port1->peer_port_num < port2->peer_port_num)
		return -1;
	if (port1->peer_port_num > port2->peer_port_num)
		return 1;

	return 0;
}

static int compare_marked_healthy_ports(const void *p1, const void *p2)
{
	osm_healthy_port_t *port1, *port2;
	port1 = (* (osm_healthy_port_t **) p1);
	port2 = (* (osm_healthy_port_t **) p2);

	/* Compare GUIDs */
	if (port1->peer_guid < port2->peer_guid)
		return -1;
	if (port1->peer_guid > port2->peer_guid)
		return 1;

	/* If same GUIDs, compare port number */
	if (port1->peer_port_num < port2->peer_port_num)
		return -1;
	if (port1->peer_port_num > port2->peer_port_num)
		return 1;

	return 0;
}

static void subn_validate_g2key(osm_subn_t *p_subn, osm_db_domain_t * p_g2key)
{
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint64_t key;
	boolean_t valid_entry;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	cl_qlist_init(&guids);

	if (osm_db_guids(p_g2key, &guids)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7506: could not get key guid list\n");
		goto Exit;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		valid_entry = TRUE;

		if (p_item->guid == 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7507: found invalid zero guid");
			valid_entry = FALSE;
		} else if (osm_db_guid2key_get(p_g2key, p_item->guid,
					       &key)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7508: could not get key for guid:0x%016"
				PRIx64 "\n", p_item->guid);
			valid_entry = FALSE;
		}

		if (valid_entry == FALSE) {
			if (osm_db_guid2key_delete(p_g2key,
						   p_item->guid))
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7509: failed to delete entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
		}
		free(p_item);
	}

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void subn_validate_neighbor(osm_subn_t *p_subn)
{
	cl_qlist_t entries;
	osm_db_guid_portnum_elem_t *p_item;
	boolean_t valid_entry;
	uint64_t guid;
	uint8_t port;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	cl_qlist_init(&entries);

	if (osm_db_guid_portnum_keys_list(p_subn->p_neighbor, &entries)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 7512: "
			"could not get neighbor entry list\n");
		goto Exit;
	}

	while ((p_item =
		(osm_db_guid_portnum_elem_t *) cl_qlist_remove_head(&entries))
	       != (osm_db_guid_portnum_elem_t *) cl_qlist_end(&entries)) {
		valid_entry = TRUE;

		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Validating neighbor for guid:0x%016" PRIx64
			", port %d\n",
			p_item->guid, p_item->portnum);
		if (p_item->guid == 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7513: found invalid zero guid\n");
			valid_entry = FALSE;
		} else if (p_item->portnum == 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7514: found invalid zero port for "
				"guid: 0x%016" PRIx64 "\n",
				p_item->guid);
			valid_entry = FALSE;
		} else if (osm_db_neighbor_get(p_subn->p_neighbor,
					       p_item->guid, p_item->portnum,
					       &guid, &port)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7515: could not find neighbor for "
				"guid: 0x%016" PRIx64 ", port %d\n",
				p_item->guid, p_item->portnum);
			valid_entry = FALSE;
		} else if (guid == 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Found invalid neighbor "
				"zero guid for guid: 0x%016" PRIx64
				", port %d\n",
				p_item->guid, p_item->portnum);
			valid_entry = FALSE;
		} else if (port == 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Found invalid neighbor "
				"zero port for guid: 0x%016" PRIx64
				", port %d\n",
				p_item->guid, p_item->portnum);
			valid_entry = FALSE;
		} else if (osm_db_neighbor_get(p_subn->p_neighbor,
					       guid, port, &guid, &port) ||
			   guid != p_item->guid || port != p_item->portnum) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Neighbor does not point "
				"back at us (guid: 0x%016" PRIx64
				", port %d)\n",
				p_item->guid, p_item->portnum);
			valid_entry = FALSE;
		}

		if (valid_entry == FALSE) {
			if (osm_db_guid_portnum_delete(p_subn->p_neighbor,
						       p_item->guid,
						       p_item->portnum))
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7519: failed to delete entry for "
					"guid:0x%016" PRIx64 " port:%u\n",
					p_item->guid, p_item->portnum);
		}
		free(p_item);
	}

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static boolean_t
subn_validate_an2an_entry(osm_subn_t *p_subn, osm_db_an2an_elem_t *p_item)
{
	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"Validating an2an cache entry"
		" (0x%016" PRIx64 ":%u <==> 0x%016" PRIx64 ":%u)\n",
		cl_ntoh64(p_item->guid1), p_item->portnum1,
		cl_ntoh64(p_item->guid2), p_item->portnum2);

	if ((p_item->guid1 == 0) || (p_item->guid2 == 0)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7534: found invalid zero guid\n");
		return FALSE;
	}

	if ((p_item->portnum1 == 0) || (p_item->portnum2 == 0)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7535: found invalid zero port\n");
		return FALSE;
	}

	if ((p_item->portnum1 == OSM_NO_PATH) ||
	    (p_item->portnum2 == OSM_NO_PATH)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7536: found invalid 0xFF port\n");
		return FALSE;
	}

	/*
	 * To avoid double records (and loopbacks), we want to make sure that
	 * guid1 is always less than guid2.
	 */
	if ((p_item->guid1 >= p_item->guid2)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7537: invalid pair - guid1 >= guid2\n");
		return FALSE;
	}

	return TRUE;
}

static void subn_validate_an2an(osm_subn_t *p_subn)
{
	cl_qlist_t entries;
	osm_db_an2an_elem_t *p_item;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	cl_qlist_init(&entries);

	if (osm_db_an2an_get_all(p_subn->p_an2an, &entries)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 7533: "
			"could not get an2an entry list\n");
		goto Exit;
	}

	while ((p_item =
		(osm_db_an2an_elem_t *) cl_qlist_remove_head(&entries))
	       != (osm_db_an2an_elem_t *) cl_qlist_end(&entries)) {

		if (!subn_validate_an2an_entry(p_subn, p_item)) {
			if (osm_db_an2an_delete(p_subn->p_an2an,
						p_item->guid1, p_item->guid2))
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7536: failed to delete an2an entry"
					" (0x%016" PRIx64 ", 0x%016" PRIx64 ")\n",
					p_item->guid1, p_item->guid2);
		}
		free(p_item);
	}
Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static int compare_service_name(IN const void * p_service_name1,
				IN const void * p_service_name2)
{
	return memcmp(p_service_name1, p_service_name2, sizeof(ib_svc_name_t));
}

static int compare_fabric_profiles(IN const void * p_profile_name1,
				   IN const void * p_profile_name2)
{
	return memcmp(p_profile_name1, p_profile_name2, sizeof(OSM_FABRIC_PROFILE_NAME_MAX_LEN));
}

void osm_subn_construct(IN osm_subn_t * p_subn)
{
	memset(p_subn, 0, sizeof(*p_subn));
	cl_ptr_vector_construct(&p_subn->port_lid_tbl);
	cl_ptr_vector_construct(&p_subn->gpu_alid_tbl);
	cl_u64_vector_construct(&p_subn->allowed_sm_guid_tbl);
	cl_qmap_init(&p_subn->sw_guid_tbl);
	osm_aports_systems_tables_init(&p_subn->aports_systems_tables);
	p_subn->is_planarized = FALSE;
	cl_qmap_init(&p_subn->node_guid_tbl);
	cl_qmap_init(&p_subn->gpu_guid_tbl);
	cl_qmap_init(&p_subn->port_guid_tbl);
	cl_qmap_init(&p_subn->alias_port_guid_tbl);
	cl_qmap_init(&p_subn->assigned_guids_tbl);
	cl_qmap_init(&p_subn->sm_guid_tbl);
	cl_qlist_init(&p_subn->sa_sr_list);
	cl_qlist_init(&p_subn->sa_infr_list);
	cl_qlist_init(&p_subn->alias_guid_list);
	cl_qlist_init(&p_subn->perflog_list);
	cl_fmap_init(&p_subn->rcvd_traps_tbl, compare_traps);
	cl_fmap_init(&p_subn->unhealthy_ports_tbl, compare_unhealthy_ports);
	cl_fmap_init(&p_subn->healthy_ports_tbl, compare_marked_healthy_ports);
	cl_qlist_init(&p_subn->rcvd_traps_list);
	cl_qlist_init(&p_subn->sa_reports_list);
	cl_qlist_init(&p_subn->prefix_routes_list);
	cl_qmap_init(&p_subn->rtr_guid_tbl);
	cl_qmap_init(&p_subn->prtn_pkey_tbl);
	cl_qmap_init(&p_subn->held_back_sw_tbl);
	cl_fmap_init(&p_subn->mgrp_mgid_tbl, ib_compare_gids);
	cl_fmap_init(&p_subn->mc_roots_tbl, ib_compare_gids);
	p_subn->min_path_bit = 0;
	p_subn->max_path_bit = 0;
	cl_qmap_init(&p_subn->border_sw_tbl);
	cl_list_construct(&p_subn->ignored_nodes_list);
	cl_list_init(&p_subn->ignored_nodes_list, 1);
	osm_subn_clear_hub_switch(p_subn);
	p_subn->hup_signal = FALSE;
	cl_vector_init(&p_subn->bad_mkey_context_list, 0, 10,
		       sizeof(osm_bad_mkey_context_t),
		       NULL, NULL, NULL);
	cl_fmap_init(&p_subn->service_keys_tbl, compare_service_name);
	cl_qlist_init(&p_subn->event_subscriber_list);
	cl_qlist_init(&p_subn->routing_subscriber_list);
	cl_qlist_init(&p_subn->partitions_subscriber_list);
	cl_fmap_init(&p_subn->fabric_mode_data.fabric_mode_profiles_map, compare_fabric_profiles);
}

static void subn_destroy_qos_options(osm_qos_options_t *opt)
{
	free(opt->vlarb_high);
	free(opt->vlarb_low);
	free(opt->sl2vl);
}

void subn_opt_destroy(IN osm_subn_opt_t * p_opt)
{
	free(p_opt->console);
	free(p_opt->port_prof_ignore_file);
	free(p_opt->hop_weights_file);
	free(p_opt->port_search_ordering_file);
	free(p_opt->routing_engine_names);
	free(p_opt->allowed_sm_guids);
	free(p_opt->log_file);
	free(p_opt->partition_config_file);
	free(p_opt->qos_policy_file);
	free(p_opt->dump_files_dir);
	free(p_opt->part_enforce);
	free(p_opt->lid_matrix_dump_file);
	free(p_opt->lfts_file);
	free(p_opt->root_guid_file);
	free(p_opt->pgrp_policy_file);
	free(p_opt->topo_policy_file);
	free(p_opt->rch_policy_file);
	free(p_opt->device_configuration_file);
	free(p_opt->congestion_control_policy_file);
	free(p_opt->ppcc_algo_dir);
	free(p_opt->fast_recovery_conf_file);
	free(p_opt->ftree_ca_order_dump_file);
	free(p_opt->held_back_sw_file);
	free(p_opt->cn_guid_file);
	free(p_opt->io_guid_file);
	free(p_opt->ids_guid_file);
	free(p_opt->guid_routing_order_file);
	free(p_opt->pqft_structure);
	free(p_opt->sa_db_file);
	free(p_opt->torus_conf_file);
	free(p_opt->sm_assign_guid_func);
	free(p_opt->pid_file);
	free(p_opt->per_module_logging_file);
	free(p_opt->mc_roots_file);
#ifdef ENABLE_OSM_PERF_MGR
	free(p_opt->event_db_dump_file);
#endif /* ENABLE_OSM_PERF_MGR */
	free(p_opt->plugin_name);
	free(p_opt->plugin_options);
	free(p_opt->event_plugin_name);
	free(p_opt->event_plugin_options);
	free(p_opt->node_name_map_name);
	free(p_opt->prefix_routes_file);
	free(p_opt->log_prefix);
	free(p_opt->switch_name);
	subn_destroy_qos_options(&p_opt->qos_options);
	subn_destroy_qos_options(&p_opt->qos_ca_options);
	subn_destroy_qos_options(&p_opt->qos_sw0_options);
	subn_destroy_qos_options(&p_opt->qos_swe_options);
	subn_destroy_qos_options(&p_opt->qos_sw2sw_options);
	subn_destroy_qos_options(&p_opt->qos_rtr_options);
	subn_destroy_qos_options(&p_opt->qos_options_prev);
	subn_destroy_qos_options(&p_opt->qos_ca_options_prev);
	subn_destroy_qos_options(&p_opt->qos_sw0_options_prev);
	subn_destroy_qos_options(&p_opt->qos_swe_options_prev);
	subn_destroy_qos_options(&p_opt->qos_sw2sw_options_prev);
	subn_destroy_qos_options(&p_opt->qos_rtr_options_prev);
	free(p_opt->hm_ports_health_policy_file);
	free(p_opt->hm_ca_reboot_action_s);
	free(p_opt->hm_ca_unresponsive_action_s);
	free(p_opt->hm_ca_noisy_action_s);
	free(p_opt->hm_ca_seterr_action_s);
	free(p_opt->hm_ca_flapping_action_s);
	free(p_opt->hm_ca_illegal_action_s);
	free(p_opt->hm_ca_manual_action_s);
	free(p_opt->hm_sw_reboot_action_s);
	free(p_opt->hm_sw_unresponsive_action_s);
	free(p_opt->hm_sw_noisy_action_s);
	free(p_opt->hm_sw_seterr_action_s);
	free(p_opt->hm_sw_flapping_action_s);
	free(p_opt->hm_sw_illegal_action_s);
	free(p_opt->hm_sw_manual_action_s);
	free(p_opt->port_speed_change_action_s);
	free(p_opt->port_ext_speed_change_action_s);
	free(p_opt->port_ext_speed2_change_action_s);
	free(p_opt->port_mepi_speed_change_action_s);
	free(p_opt->port_mtu_change_action_s);
	free(p_opt->port_vl_change_action_s);
	free(p_opt->port_ame_bit_change_action_s);
	free(p_opt->router_policy_file);
	free(p_opt->verbose_bypass_policy_file);
	free(p_opt->sa_db_dump_file);
	free(p_opt->subnet_lst_file);
	free(p_opt->virt_dump_file);
	free(p_opt->additional_gi_supporting_devices);
	free(p_opt->additional_mepi_force_devices);
	free(p_opt->activity_report_subjects);
	free(p_opt->enhanced_qos_policy_file);
	free(p_opt->adv_routing_engine_name);
	free(p_opt->service_name2key_map_file);
	free(p_opt->topo_config_file);
	if (p_opt->opts_def) {
		subn_opt_destroy(p_opt->opts_def);
		free(p_opt->opts_def);
	}
        free(p_opt->rtr_selection_function_s);
        free(p_opt->rtr_selection_algo_parameters_s);
	free(p_opt->hbf_weights);
	free(p_opt->tenants_policy_file);
	free(p_opt->fabric_mode_profile);
	free(p_opt->fabric_mode_policy_file);
	free(p_opt->conf_template);
}

static int clear_tbl_entry_val(st_data_t key, st_data_t val, st_data_t arg)
{
	free((osm_neighbors_info_t *)val);
	return ST_DELETE;
}

static int topoconfig_tbl_clear(st_table *p_topoconfig_hash_tbl)
{
	st_foreach(p_topoconfig_hash_tbl, clear_tbl_entry_val, (st_data_t) NULL);
	st_free_table(p_topoconfig_hash_tbl);
	return 0;
}

void osm_subn_destroy_aports_systems_tables(IN osm_subn_t * p_subn)
{
	osm_aports_systems_tables_destroy(&p_subn->aports_systems_tables);
	p_subn->is_planarized = FALSE;
}

void osm_subn_destroy(IN osm_subn_t * p_subn)
{
	int i;
	osm_node_t *p_node;
	osm_gpu_t *p_gpu;
	osm_assigned_guids_t *p_assigned_guids;
	osm_alias_guid_t *p_alias_guid;
	osm_port_t *p_port;
	osm_switch_t *p_sw;
	osm_remote_sm_t *p_rsm;
	osm_prtn_t *p_prtn;
	osm_infr_t *p_infr;
	osm_svcr_t *p_svcr;
	osm_unhealthy_port_t *p_unhealthy_port;
	osm_held_back_switch_t *p_held_back_sw;
	osm_madw_t *p_madw;
	cl_fmap_item_t *p_item;
	cl_map_item_t * p_map_item;
	osm_router_t *p_rtr;
	osm_perflog_event_t *p_perflog_event;

	/* it might be a good idea to de-allocate all known objects */
	while (cl_qmap_count(&p_subn->node_guid_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->node_guid_tbl);
		cl_qmap_remove_item(&p_subn->node_guid_tbl, p_map_item);
		p_node = (osm_node_t *)p_map_item;
		osm_node_delete(&p_node);
	}

	osm_subn_destroy_aports_systems_tables(p_subn);

	while (cl_qmap_count(&p_subn->gpu_guid_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->gpu_guid_tbl);
		cl_qmap_remove_item(&p_subn->gpu_guid_tbl, p_map_item);
		p_gpu = (osm_gpu_t *)p_map_item;
		osm_gpu_delete(&p_gpu);
	}

	while (cl_qmap_count(&p_subn->assigned_guids_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->assigned_guids_tbl);
		cl_qmap_remove_item(&p_subn->assigned_guids_tbl, p_map_item);
		p_assigned_guids = (osm_assigned_guids_t *)p_map_item;
		osm_assigned_guids_delete(&p_assigned_guids);
	}

	while (cl_qlist_count(&p_subn->sa_reports_list)) {
		p_madw = (osm_madw_t *) cl_qlist_remove_head(&p_subn->sa_reports_list);
		osm_mad_pool_put(p_subn->p_osm->sa.p_mad_pool, p_madw);
	}
	cl_qlist_remove_all(&p_subn->rcvd_traps_list);
	cl_fmap_remove_all(&p_subn->rcvd_traps_tbl);

	while(!cl_is_fmap_empty(&p_subn->unhealthy_ports_tbl)) {
		p_item = cl_fmap_head(&p_subn->unhealthy_ports_tbl);
		cl_fmap_remove_item(&p_subn->unhealthy_ports_tbl, p_item);
		p_unhealthy_port = (osm_unhealthy_port_t *) p_item;
		osm_unhealthy_port_delete(&p_unhealthy_port);
	}

	osm_hm_clear_marked_healthy_ports(&p_subn->p_osm->sm);

	while (cl_qmap_count(&p_subn->prtn_pkey_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->prtn_pkey_tbl);
		cl_qmap_remove_item(&p_subn->prtn_pkey_tbl, p_map_item);
		p_prtn = (osm_prtn_t *)p_map_item;
		osm_prtn_delete(p_subn, &p_prtn, OSM_ACT_MGR_PRTN_SHUTDOWN);
	}

	cl_fmap_remove_all(&p_subn->mgrp_mgid_tbl);

	for (i = 0; i <= p_subn->max_mcast_lid_ho - IB_LID_MCAST_START_HO;
	     i++)
		if (p_subn->mboxes[i])
			osm_mgrp_box_delete(p_subn, p_subn->mboxes[i]);


	p_map_item = cl_qmap_head(&p_subn->port_guid_tbl);
	while (p_map_item != cl_qmap_end(&p_subn->port_guid_tbl)) {
		p_port = (osm_port_t *)p_map_item;
		osm_port_delete_virtualization(&p_subn->p_osm->sm, p_port);
		p_map_item = cl_qmap_next(p_map_item);
	}
	virt_mgr_handle_empty_vnodes(&p_subn->p_osm->sm, FALSE, TRUE);

	while (cl_qmap_count(&p_subn->alias_port_guid_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->alias_port_guid_tbl);
		cl_qmap_remove_item(&p_subn->alias_port_guid_tbl, p_map_item);
		p_alias_guid = (osm_alias_guid_t *)p_map_item;
		osm_alias_guid_delete(&p_alias_guid);
	}

	while (cl_qlist_count(&p_subn->alias_guid_list))
		osm_guid_work_obj_delete((osm_guidinfo_work_obj_t *) cl_qlist_remove_head(&p_subn->alias_guid_list));

	while (cl_qlist_count(&p_subn->perflog_list)) {
		p_perflog_event = (osm_perflog_event_t *) cl_qlist_remove_head(&p_subn->perflog_list);
		free(p_perflog_event);
	}

	while (cl_qmap_count(&p_subn->port_guid_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->port_guid_tbl);
		cl_qmap_remove_item(&p_subn->port_guid_tbl, p_map_item);
		p_port = (osm_port_t *)p_map_item;
		osm_port_delete(&p_subn->p_osm->sm, &p_port);
	}

	while (cl_qmap_count(&p_subn->sw_guid_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->sw_guid_tbl);
		cl_qmap_remove_item(&p_subn->sw_guid_tbl, p_map_item);
		p_sw = (osm_switch_t *)p_map_item;
		osm_switch_delete(&p_sw);
	}

	while (cl_qmap_count(&p_subn->sm_guid_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->sm_guid_tbl);
		cl_qmap_remove_item(&p_subn->sm_guid_tbl, p_map_item);
		p_rsm = (osm_remote_sm_t *)p_map_item;
		free(p_rsm);
	}

	while (cl_qmap_count(&p_subn->held_back_sw_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->held_back_sw_tbl);
		cl_qmap_remove_item(&p_subn->held_back_sw_tbl, p_map_item);
		p_held_back_sw = (osm_held_back_switch_t *)p_map_item;
		free(p_held_back_sw);
	}

	while (cl_qlist_count(&p_subn->sa_infr_list)) {
		p_infr =
		    (osm_infr_t *)cl_qlist_remove_head(&p_subn->sa_infr_list);
		osm_infr_delete(p_infr);
	}

	while (cl_qlist_count(&p_subn->sa_sr_list)) {
		p_svcr =
		    (osm_svcr_t *)cl_qlist_remove_head(&p_subn->sa_sr_list);
		osm_svcr_delete(p_svcr);
	}

	cl_ptr_vector_destroy(&p_subn->port_lid_tbl);
	cl_ptr_vector_destroy(&p_subn->gpu_alid_tbl);

	osm_qos_policy_destroy(p_subn->p_qos_policy);

	while (!cl_is_qlist_empty(&p_subn->prefix_routes_list)) {
		cl_list_item_t *item = cl_qlist_remove_head(&p_subn->prefix_routes_list);
		free(item);
	}

	/* remove router db and their prefixes */
	while (cl_qmap_count(&p_subn->rtr_guid_tbl)) {
		p_map_item = cl_qmap_head(&p_subn->rtr_guid_tbl);
		cl_qmap_remove_item(&p_subn->rtr_guid_tbl, p_map_item);
		p_rtr = (osm_router_t *)p_map_item;
		osm_router_delete(&p_rtr);
	}

	cl_list_remove_all(&p_subn->ignored_nodes_list);
	cl_list_destroy(&p_subn->ignored_nodes_list);

	cl_u64_vector_destroy(&p_subn->allowed_sm_guid_tbl);
	osm_subn_clear_leafs(p_subn);
	cl_list_destroy(&p_subn->leaf_switches_list);

	osm_subn_clear_mc_roots_tbl(p_subn);

	osm_destroy_routing_engines(p_subn);
	osm_destroy_default_routing_engine(p_subn);
	osm_destroy_advanced_routing_engine(p_subn);
	subn_opt_destroy(&p_subn->opt);
	free(p_subn->opt.file_opts);
	cl_vector_destroy(&p_subn->bad_mkey_context_list);
	osm_subn_clear_service_name_key_tbl(p_subn);

	if (p_subn->p_topoconfig_hash_tbl)
		topoconfig_tbl_clear(p_subn->p_topoconfig_hash_tbl);

	while (cl_qlist_count(&p_subn->event_subscriber_list)) {
		cl_qlist_remove_head(&p_subn->event_subscriber_list);
	}

	while (cl_qlist_count(&p_subn->routing_subscriber_list)) {
		cl_qlist_remove_head(&p_subn->routing_subscriber_list);
	}

	while (cl_qlist_count(&p_subn->partitions_subscriber_list)) {
		cl_qlist_remove_head(&p_subn->partitions_subscriber_list);
	}

	osm_fabric_mode_clear_profiles(p_subn);
}

/*
 * TODO - Should use this function to resolve the code duplication done with the
 *        rest of the tables (in osm_subn_init).
 */
 static
osm_db_domain_t *osm_subn_init_db(osm_opensm_t * p_osm, char * name,
				  char * description)
{
	/* Initialize the given database */
	osm_db_domain_t *domain = osm_db_domain_init(&p_osm->db, name);
	if (!domain) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 7510: "
			"Error initializing %s persistent database\n",
			description);
		return NULL;
	}

	if (osm_db_restore(domain)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted guid2mkey file.
		 */
		if (p_osm->subn.opt.exit_on_fatal) {
			osm_log(&p_osm->log, OSM_LOG_SYS,
				"FATAL: Error restoring %s "
				"persistent database\n", description);
			return NULL;
		} else
#endif
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 7511: Error restoring %s "
				"persistent database\n", description);
	}
	return domain;
}

static ib_api_status_t subn_validate_key_db(osm_opensm_t * p_osm,
					    uint8_t mclass,
					    boolean_t neighbor_key)
{
	osm_subn_t *p_subn = &p_osm->subn;
	osm_db_domain_t *p_g2key;

	switch (mclass) {
	case IB_MCLASS_CC:
		p_g2key = p_subn->p_g2cckey =
		    osm_db_domain_init(&p_osm->db, "guid2cckey");
		break;
	case IB_MCLASS_VS:
		p_g2key = p_subn->p_g2vskey =
		    osm_db_domain_init(&p_osm->db, "guid2vskey");
		break;
	case IB_MCLASS_N2N:
		if (neighbor_key)
			p_g2key = p_subn->p_g2n2nkey =
			    osm_db_domain_init(&p_osm->db, "guid2_n2n_key");
		else
			p_g2key = p_subn->p_g2n2n_mgrkey =
			    osm_db_domain_init(&p_osm->db, "guid2_m2n_key");
		break;
	default:
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 754E: "
			"Key db not supported for class 0x%x%s\n", mclass,
			neighbor_key ? " (neighbor key)" : "");
		return IB_ERROR;
	}

	if (!p_g2key) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 754C: "
			"Error initializing Guid-to-Keys persistent "
			"database\n");
		return IB_ERROR;
	}

	if (osm_db_restore(p_g2key)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted neighbors file.
		 */
		if (p_subn->opt.exit_on_fatal) {
			osm_log(&p_osm->log, OSM_LOG_SYS,
				"FATAL: Error restoring key "
				"persistent database\n");
			return IB_ERROR;
		} else
#endif
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 754D: Error restoring key "
				"persistent database\n");
	}
	subn_validate_g2key(p_subn, p_g2key);

	return IB_SUCCESS;
}

uint8_t osm_get_sm_last_port_state(struct osm_sm *p_sm)
{
    osm_subn_t *p_subn = p_sm->p_subn;
    uint8_t last_sm_port_state;

    CL_SPINLOCK_ACQUIRE(&p_sm->state_lock);
    last_sm_port_state = p_subn->sm_state_info.last_sm_port_state;
    CL_SPINLOCK_RELEASE(&p_sm->state_lock);

    return last_sm_port_state;
}

osm_sm_state_info_t osm_get_sm_state_info(struct osm_sm *p_sm)
{
    osm_subn_t *p_subn = p_sm->p_subn;
    osm_sm_state_info_t state_info;

    CL_SPINLOCK_ACQUIRE(&p_sm->state_lock);
    state_info = p_subn->sm_state_info;
    CL_SPINLOCK_RELEASE(&p_sm->state_lock);

    return state_info;
}

void osm_set_sm_last_port_state(struct osm_sm *p_sm, uint8_t last_sm_port_state)
{
    osm_subn_t *p_subn = p_sm->p_subn;

    CL_SPINLOCK_ACQUIRE(&p_sm->state_lock);
    p_subn->sm_state_info.last_sm_port_state = last_sm_port_state;
    CL_SPINLOCK_RELEASE(&p_sm->state_lock);
}

ib_api_status_t osm_subn_init(IN osm_subn_t * p_subn, IN osm_opensm_t * p_osm,
			      IN const osm_subn_opt_t * p_opt)
{
	cl_status_t status;

	p_subn->p_osm = p_osm;

	status = cl_ptr_vector_init(&p_subn->port_lid_tbl,
				    OSM_SUBNET_VECTOR_MIN_SIZE,
				    OSM_SUBNET_VECTOR_GROW_SIZE);
	if (status != CL_SUCCESS)
		return IB_UNKNOWN_ERROR;

	status = cl_ptr_vector_init(&p_subn->gpu_alid_tbl,
				    OSM_SUBNET_VECTOR_MIN_SIZE,
				    OSM_SUBNET_VECTOR_GROW_SIZE);
	if (status != CL_SUCCESS)
		return IB_UNKNOWN_ERROR;

	status = cl_u64_vector_init(&p_subn->allowed_sm_guid_tbl, 0, TRUE);
	if (status != CL_SUCCESS)
		return IB_UNKNOWN_ERROR;

	status = cl_list_init(&p_subn->leaf_switches_list, 10);
	if (status != CL_SUCCESS)
		return IB_UNKNOWN_ERROR;

	status = cl_ptr_vector_set_capacity(&p_subn->port_lid_tbl,
					    OSM_SUBNET_VECTOR_CAPACITY);
	if (status != CL_SUCCESS)
		return IB_UNKNOWN_ERROR;

	status = cl_ptr_vector_set_capacity(&p_subn->gpu_alid_tbl,
					    OSM_SUBNET_VECTOR_CAPACITY);
	if (status != CL_SUCCESS)
		return IB_UNKNOWN_ERROR;

	/*
	   LID zero is not valid.  NULL out this entry for the
	   convenience of other code.
	 */
	cl_ptr_vector_set(&p_subn->port_lid_tbl, 0, NULL);
	cl_ptr_vector_set(&p_subn->gpu_alid_tbl, 0, NULL);

	p_subn->opt = *p_opt;
	p_subn->max_ucast_lid_ho = IB_LID_UCAST_END_HO;
	p_subn->max_mcast_lid_ho = IB_LID_MCAST_END_HO;
	p_subn->min_ca_mtu = IB_MAX_MTU;
	p_subn->min_ca_rate = IB_PATH_RECORD_RATE_300_GBS;
	p_subn->min_data_vls = IB_MAX_NUM_VLS - 1;
	p_subn->min_sw_data_vls = IB_MAX_NUM_VLS - 1;
	p_subn->ignore_existing_lfts = TRUE;
	p_subn->light_sweep_counter = 0;
	p_subn->force_multicast_rerouting = TRUE;
	p_subn->reroute_all_mlids = FALSE;

	/* we assume master by default - so we only need to set it true if STANDBY */
	p_subn->coming_out_of_standby = FALSE;
	p_subn->sm_port_changed = FALSE;
	p_subn->sweeping_enabled = TRUE;
	p_subn->sm_state_info.last_sm_port_state = OSM_SM_PORT_STATE_UP;

	p_subn->min_path_bit = 0;
	p_subn->max_path_bit = (1 << osm_subn_get_max_lmc(p_subn)) - 1;

	p_subn->pfrn_supported = FALSE;
	p_subn->pfrn_config_change = FALSE;

	p_subn->enable_sweep = TRUE;

	/* Initialize the guid2mkey database */
	p_subn->p_g2m = osm_db_domain_init(&p_osm->db, "guid2mkey");
	if (!p_subn->p_g2m) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 7510: "
			"Error initializing Guid-to-MKey persistent database\n");
		return IB_ERROR;
	}

	if (osm_db_restore(p_subn->p_g2m)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted guid2mkey file.
		 */
		if (p_subn->opt.exit_on_fatal) {
			osm_log(&p_osm->log, OSM_LOG_SYS,
				"FATAL: Error restoring Guid-to-Mkey "
				"persistent database\n");
			return IB_ERROR;
		} else
#endif
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 7511: Error restoring Guid-to-Mkey "
				"persistent database\n");
	}

	subn_validate_g2key(p_subn, p_subn->p_g2m);

	/* Initialize the neighbor database */
	p_subn->p_neighbor = osm_db_domain_init(&p_osm->db, "neighbors");
	if (!p_subn->p_neighbor) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR 7520: Error "
			"initializing neighbor link persistent database\n");
		return IB_ERROR;
	}

	if (osm_db_restore(p_subn->p_neighbor)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted neighbors file.
		 */
		if (p_subn->opt.exit_on_fatal) {
			osm_log(&p_osm->log, OSM_LOG_SYS,
				"FATAL: Error restoring neighbor link "
				"persistent database\n");
			return IB_ERROR;
		} else
#endif
			OSM_LOG(&p_osm->log, OSM_LOG_ERROR,
				"ERR 7521: Error restoring neighbor link "
				"persistent database\n");
	}

	subn_validate_neighbor(p_subn);

	p_subn->p_an2an = osm_subn_init_db(p_osm, "an2an", "an2an routing");
	if (!p_subn->p_an2an)
		return IB_ERROR;

	subn_validate_an2an(p_subn);

	if (p_subn->opt.cc_key_enable != OSM_CAPABILITY_IGNORED &&
	    subn_validate_key_db(p_osm, IB_MCLASS_CC, FALSE))
		return IB_ERROR;

	if (p_subn->opt.vs_key_enable != OSM_CAPABILITY_IGNORED &&
	    subn_validate_key_db(p_osm, IB_MCLASS_VS, FALSE))
		return IB_ERROR;

	if (p_subn->opt.n2n_key_enable != OSM_CAPABILITY_IGNORED &&
	    (subn_validate_key_db(p_osm, IB_MCLASS_N2N, FALSE) ||
	     subn_validate_key_db(p_osm, IB_MCLASS_N2N, TRUE)))
		return IB_ERROR;

	p_subn->p_topoconfig_hash_tbl = NULL;

	return IB_SUCCESS;
}

osm_port_t *osm_get_port_by_mad_addr(IN osm_log_t * p_log,
				     IN const osm_subn_t * p_subn,
				     IN osm_mad_addr_t * p_mad_addr)
{
	osm_port_t *port = osm_get_port_by_lid(p_subn, p_mad_addr->dest_lid);
	if (!port)
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 7504: "
			"Lid is out of range: %u\n",
			cl_ntoh16(p_mad_addr->dest_lid));

	return port;
}

ib_api_status_t osm_get_gid_by_mad_addr(IN osm_log_t * p_log,
					IN const osm_subn_t * p_subn,
					IN osm_mad_addr_t * p_mad_addr,
					OUT ib_gid_t * p_gid)
{
	ib_net64_t guid;

	if (p_gid == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 7505: "
			"Provided output GID is NULL\n");
		return IB_INVALID_PARAMETER;
	}

	if (p_mad_addr->addr_type.gsi.global_route)
		*p_gid = p_mad_addr->addr_type.gsi.grh_info.dest_gid;
	else {
		guid = osm_get_guid_by_lid(p_subn, p_mad_addr->dest_lid);

		if (guid) {
			p_gid->unicast.interface_id = guid;
			p_gid->unicast.prefix = p_subn->opt.subnet_prefix;
		} else
			return IB_INVALID_PARAMETER;
	}

	return IB_SUCCESS;
}

osm_physp_t *osm_get_physp_by_mad_addr(IN osm_log_t * p_log,
				       IN const osm_subn_t * p_subn,
				       IN osm_mad_addr_t * p_mad_addr)
{
	osm_port_t *p_port;

	p_port = osm_get_port_by_mad_addr(p_log, p_subn, p_mad_addr);
	if (!p_port)
		return NULL;

	return p_port->p_physp;
}

static void report_sgid_spoofing(osm_subn_t *p_subn, const osm_madw_t *p_madw,
				 ib_gid_t *p_gid, osm_alias_guid_t *p_alias_guid)
{
	osm_mad_addr_t *p_mad_addr = osm_madw_get_mad_addr_ptr(p_madw);
	ib_sa_mad_t *p_sa_mad = osm_madw_get_sa_mad_ptr(p_madw);
	struct _vendor_data_details ntc_data;

	ntc_data.v_data_details.ntc_1301.slid = p_mad_addr->dest_lid;
	ntc_data.v_data_details.ntc_1301.attr_id = p_sa_mad->attr_id;
	ntc_data.v_data_details.ntc_1301.method = p_sa_mad->method;
	memcpy(&ntc_data.v_data_details.ntc_1301.spoofed_sgid, p_gid, sizeof(ib_gid_t));

	if (p_alias_guid &&
	    (p_alias_guid != (osm_alias_guid_t*)cl_qmap_end(&p_subn->alias_port_guid_tbl)))
		ntc_data.v_data_details.ntc_1301.physp_sguid = p_alias_guid->base_port_guid;
	else
		ntc_data.v_data_details.ntc_1301.physp_sguid = 0;

	ntc_data.trap_num = SA_VS_TRAP_NUM_SGID_SPOOFED;
	osm_report_sa_security_notice(p_subn, &ntc_data);
}

boolean_t osm_validate_mad_addr(IN osm_log_t * p_log,
				IN osm_subn_t * p_subn,
				IN const osm_madw_t * p_madw)
{
	ib_net16_t attr_id = (osm_madw_get_sa_mad_ptr(p_madw))->attr_id;
	osm_mad_addr_t *p_mad_addr = osm_madw_get_mad_addr_ptr(p_madw);
	osm_alias_guid_t *p_alias_guid = NULL;
	boolean_t res = FALSE;
	ib_net64_t guid;
	ib_gid_t gid;
	osm_vport_t *p_vport;

	/* No need to check if their is no grh */
	if (!p_mad_addr->addr_type.gsi.global_route) {
		res = TRUE;
		goto Exit;
	}

	gid = p_mad_addr->addr_type.gsi.grh_info.dest_gid;
	if (!ib_gid_is_link_local(&gid) &&
	    (!is_equal_subnet_prefix(ib_gid_get_subnet_prefix(&gid),
				     p_subn->opt.subnet_prefix))) {
		if (!ib_gid_is_site_local(&gid)) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 753F: Global GID subnet prefix 0x%016"
				PRIx64 " is not supported\n",
				cl_ntoh64(ib_gid_get_subnet_prefix(&gid)));
		} else {
			osm_port_t *p_port = osm_get_port_by_lid(p_subn, p_mad_addr->dest_lid);
			if (!p_port)
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 753E: "
					"Cannot find requesting port GUID - unknown lid: %u\n",
					cl_ntoh16(p_mad_addr->dest_lid));
			else if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_ROUTER)
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 753A: "
					"Cannot find requesting port - GUID 0x%" PRIx64
					", unknown subnet prefix from non-router port "
					"0x%" PRIx64 "\n",
					cl_ntoh64(ib_gid_get_guid(&gid)),
					cl_ntoh64(ib_gid_get_subnet_prefix(&gid)));
			else
				res = TRUE;
		}
		goto Exit;
	}

	guid = osm_get_guid_by_lid(p_subn, p_mad_addr->dest_lid);
	if (!guid) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 753B: "
			"Cannot find requesting port GUID - unknown lid: %u\n",
			cl_ntoh16(p_mad_addr->dest_lid));
		goto Exit;
	}

	if (ib_gid_get_guid(&gid) != guid) {
		p_alias_guid =
			(osm_alias_guid_t*)cl_qmap_get(&p_subn->alias_port_guid_tbl,
						       gid.unicast.interface_id);
		if (p_alias_guid ==
			(osm_alias_guid_t*)cl_qmap_end(&p_subn->alias_port_guid_tbl)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 753C: "
				"Cannot find requesting port by grh - unknown guid "
				"0x%" PRIx64 "\n",
				cl_ntoh64(gid.unicast.interface_id));
			goto Exit;
		} else if (p_alias_guid->p_base_port !=
			   osm_get_port_by_lid(p_subn, p_mad_addr->dest_lid)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 753D: "
				"Sent %s - GUID 0x%" PRIx64 " does "
				"not match lid %u\n",
				ib_get_sm_attr_str(attr_id),
				cl_ntoh64(gid.unicast.interface_id),
				cl_ntoh16(p_mad_addr->dest_lid));
			goto Exit;
		} else if (!osm_is_alias_guid(p_alias_guid)) {
			p_vport = osm_port_get_vport(p_alias_guid->p_base_port,
						     p_alias_guid->vport_index);
			if (!p_vport ||
			    (p_mad_addr->dest_lid != osm_vport_calc_lid(p_vport))) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 7548: "
					"Sent %s - GUID 0x%" PRIx64 " does "
					"not match lid %u\n",
					ib_get_sm_attr_str(attr_id),
					cl_ntoh64(gid.unicast.interface_id),
					cl_ntoh16(p_mad_addr->dest_lid));
				goto Exit;
			}
		}
	}

	res = TRUE;
Exit:
	if (res == FALSE)
		report_sgid_spoofing(p_subn, p_madw, &gid, p_alias_guid);

	return res;
}

osm_alias_guid_t * osm_get_alias_guid_by_mad_addr(IN osm_log_t * p_log,
						  IN const osm_subn_t * p_subn,
						  osm_mad_addr_t * p_mad_addr)
{
	osm_alias_guid_t *p_alias_guid = NULL;
	ib_gid_t gid;

	if (osm_get_gid_by_mad_addr(p_log, p_subn, p_mad_addr, &gid)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 750A: "
			"Cannot find requesting port - unknown lid: %u\n",
			cl_ntoh16(p_mad_addr->dest_lid));
		goto Exit;
	}

	if (!is_equal_subnet_prefix(gid.unicast.prefix, p_subn->opt.subnet_prefix) &&
	    gid.unicast.prefix != IB_DEFAULT_SUBNET_PREFIX) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 750B: "
			"Cannot find requesting port - GUID 0x%" PRIx64
			", unknown subnet prefix "
			"0x%" PRIx64 "\n",
			cl_ntoh64(gid.unicast.interface_id),
			cl_ntoh64(gid.unicast.prefix));
		goto Exit;
	}

	p_alias_guid =
	    (osm_alias_guid_t*)cl_qmap_get(&p_subn->alias_port_guid_tbl,
					   gid.unicast.interface_id);
	if (p_alias_guid ==
	    (osm_alias_guid_t*)cl_qmap_end(&p_subn->alias_port_guid_tbl)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 750C: "
			"Cannot find requesting port - unknown guid "
			"0x%" PRIx64 "\n",
			cl_ntoh64(gid.unicast.interface_id));
		p_alias_guid = NULL;
		goto Exit;
	}

	if (p_mad_addr->addr_type.gsi.global_route &&
	    p_alias_guid->p_base_port !=
	    osm_get_port_by_lid(p_subn, p_mad_addr->dest_lid)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 750D: "
			"Cannot find requesting port - GUID 0x%" PRIx64 " does "
			"not match lid %u\n",
			cl_ntoh64(gid.unicast.interface_id),
			cl_ntoh16(p_mad_addr->dest_lid));
		p_alias_guid = NULL;
	}

Exit:
	return p_alias_guid;
}

ib_net64_t osm_get_guid_by_lid(IN osm_subn_t const *p_subn, IN ib_net16_t lid)
{
	osm_port_t *p_port = osm_get_port_by_lid(p_subn, lid);
	osm_vport_t *p_vport = NULL;
	ib_net64_t guid = 0;
	uint16_t min_lid, max_lid, lid_ho;

	if (p_port) {
		osm_port_get_lid_range_ho(p_port, &min_lid, &max_lid);
		lid_ho = cl_ntoh16(lid);
		if ((lid_ho >= min_lid) && (lid_ho <= max_lid)) {
			/* Lid belongs to the port itself */
			guid = osm_port_get_guid(p_port);
		} else {
			/* Lid belongs to one of the vports of the port */
			p_vport = osm_port_get_vport_by_lid(p_port, lid);
			if (p_vport)
				guid = osm_vport_get_guid(p_vport);
		}
	}

	if (guid)
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG, "Found %s guid 0x%"
			PRIx64 " for lid %u\n", p_vport ? "virtual" : "physical",
			cl_ntoh64(guid), cl_ntoh16(lid));
	else
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"Couldn't find guid for lid %u\n", cl_ntoh16(lid));

	return guid;
}

struct osm_physp * osm_get_sm_physp_by_dr(IN const osm_subn_t * p_subn,
					  IN const osm_dr_path_t * p_path,
					  IN uint8_t sm_binding_port_index)
{
	osm_port_t *p_port;
	ib_net64_t sm_guid;
	uint8_t port_num;

	sm_guid = (sm_binding_port_index == p_subn->sm_port_idx) ?
		  p_subn->sm_port_guid : p_subn->opt.guid[sm_binding_port_index];
	p_port = osm_get_port_by_guid(p_subn, sm_guid);
	if (!p_port || !p_port->p_node)
		return NULL;

	/*
	 * Use port_num specified in path.
	 * If path doesn't specify the port number, use the default one.
	 */
	if (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH) {
		port_num = 0;
	} else if (p_path->exit_port_num == 0) {
		port_num = p_subn->p_osm->p_vendor[sm_binding_port_index]->umad_port[0].portnum;
	} else {
		port_num = p_path->exit_port_num;
	}

	return osm_node_get_physp_ptr(p_port->p_node, port_num);
}

osm_physp_t *osm_get_ext_physp_by_dr(IN osm_sm_t * sm, IN const osm_dr_path_t * p_path)
{
	osm_node_t *p_node;
	osm_physp_t *p_sm_physp, *p_physp = NULL;
	uint8_t hop, hop_count;

	OSM_LOG_ENTER(sm->p_log);

	hop_count = p_path->hop_count;

	p_sm_physp = osm_get_sm_physp_by_dr(sm->p_subn, p_path, sm->p_subn->sm_port_idx);

	if (!p_sm_physp)
		goto Exit;

	/* hop_count == 0: destination port guid is SM */
	if (hop_count == 0) {
		p_physp = p_sm_physp;
		goto Exit;
	}

	p_node = osm_physp_get_node_ptr(p_sm_physp);
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH)
		p_physp = osm_node_get_physp_ptr(p_node, p_path->path[1]);
	else
		p_physp = p_sm_physp;

	/* hop_count == 1: outgoing physp is SM physp */
	for (hop = 2; p_physp && hop <= hop_count; hop++) {
		p_physp = p_physp->p_remote_physp;
		if (!p_physp)
			break;
		p_node = p_physp->p_node;

		if (!p_node || p_path->path[hop] >= p_node->physp_tbl_size)
			p_physp = NULL;
		else
			p_physp = osm_node_get_physp_ptr(p_node, p_path->path[hop]);
	}

	/*
	 * At this point, p_physp points at the outgoing physp on the
	 * last hop, or NULL if we don't know it.
	 */

Exit:
	OSM_LOG_EXIT(sm->p_log);

	return p_physp;
}

osm_physp_t *osm_get_physp_by_dr(IN osm_log_t * p_log,
				 IN const osm_subn_t * p_subn,
				 IN const osm_dr_path_t * p_path,
				 IN uint8_t sm_binding_port_index)
{
	uint8_t pn;
	uint8_t hop;
	osm_physp_t *physp = NULL;
	osm_physp_t *rem_physp = NULL;
	osm_physp_t *p_sm_physp;
	osm_node_t *node = NULL;

	/* we use the SM port guid as starting port */
	p_sm_physp = osm_get_sm_physp_by_dr(p_subn, p_path, sm_binding_port_index);

	if (!p_sm_physp) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 7526: "
			"Couldn't find SM port by guid: 0x%016" PRIx64 "\n",
			cl_ntoh64(p_subn->sm_port_guid));
		return NULL;
	}
	rem_physp = p_sm_physp;
	node = osm_physp_get_node_ptr(p_sm_physp);

	for (hop = 1; hop <= p_path->hop_count; hop++) {
		if (!node) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 7527: "
				"Couldn't find node for port with guid: 0x%016" PRIx64 "\n",
				osm_physp_get_port_guid(rem_physp));
			return NULL;
		}

		pn = p_path->path[hop];
		physp = osm_node_get_physp_ptr(node, pn);
		if (PF(!physp)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR 7528: "
				"Couldn't find physp %d for node with guid: 0x%016" PRIx64
				"\n", pn, osm_node_get_node_guid_ho(node));
			return NULL;
		}

		rem_physp = osm_physp_get_remote(physp);
		if (!rem_physp) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Couldn't find remote physp on node with guid: 0x%016" PRIx64
				" port %d\n", osm_node_get_node_guid_ho(node), pn);
			return NULL;
		}

		node = osm_physp_get_node_ptr(rem_physp);
	}

	/* if the destination is a switch return physp 0 -
		else the port we got in through */
	if (osm_node_get_type(node) == IB_NODE_TYPE_SWITCH)
		rem_physp = osm_node_get_physp_ptr(node, 0);

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"Got from SM guid: 0x%016" PRIx64
		" to node with guid: 0x%016" PRIx64 " port %d\n",
		cl_ntoh64(p_subn->sm_port_guid),
		osm_node_get_node_guid_ho(node),
		osm_physp_get_port_num(rem_physp));

	return(rem_physp);
}

osm_switch_t *osm_get_switch_by_guid(IN const osm_subn_t * p_subn,
				     IN ib_net64_t guid)
{
	osm_switch_t *p_switch;

	p_switch = (osm_switch_t *) cl_qmap_get(&p_subn->sw_guid_tbl, guid);
	if (p_switch == (osm_switch_t *) cl_qmap_end(&p_subn->sw_guid_tbl))
		p_switch = NULL;
	return p_switch;
}

osm_node_t *osm_get_node_by_guid(IN osm_subn_t const *p_subn, IN ib_net64_t guid)
{
	osm_node_t *p_node;

	p_node = (osm_node_t *) cl_qmap_get(&p_subn->node_guid_tbl, guid);
	if (p_node == (osm_node_t *) cl_qmap_end(&p_subn->node_guid_tbl))
		p_node = NULL;
	return p_node;
}

osm_gpu_t *osm_get_gpu_by_guid(IN osm_subn_t const *p_subn, IN ib_net64_t guid)
{
	osm_gpu_t *p_gpu;

	p_gpu = (osm_gpu_t *) cl_qmap_get(&p_subn->gpu_guid_tbl, guid);
	if (p_gpu == (osm_gpu_t *) cl_qmap_end(&p_subn->gpu_guid_tbl))
		p_gpu = NULL;

	return p_gpu;
}

osm_port_t *osm_get_port_by_guid(IN osm_subn_t const *p_subn, IN ib_net64_t guid)
{
	osm_port_t *p_port;

	p_port = (osm_port_t *) cl_qmap_get(&p_subn->port_guid_tbl, guid);
	if (p_port == (osm_port_t *) cl_qmap_end(&p_subn->port_guid_tbl))
		p_port = NULL;
	return p_port;
}

osm_alias_guid_t *osm_get_alias_guid_by_guid(IN osm_subn_t const *p_subn,
					     IN ib_net64_t guid)
{
	osm_alias_guid_t *p_alias_guid;

	p_alias_guid = (osm_alias_guid_t *) cl_qmap_get(&p_subn->alias_port_guid_tbl, guid);
	if (p_alias_guid == (osm_alias_guid_t *) cl_qmap_end(&p_subn->alias_port_guid_tbl))
		return NULL;

	if (p_alias_guid->dirty)
		return NULL;
	else
		return p_alias_guid;
}

osm_port_t *osm_get_port_by_alias_guid(IN osm_subn_t const *p_subn,
				       IN ib_net64_t guid)
{
	osm_alias_guid_t *p_alias_guid;

	p_alias_guid = osm_get_alias_guid_by_guid(p_subn, guid);
	if (!p_alias_guid)
		return NULL;
	return p_alias_guid->p_base_port;
}

boolean_t osm_is_alias_guid(IN const struct osm_alias_guid *p_alias_guid)
{
	ib_net64_t base_guid;

	if (p_alias_guid->vport_index) {
		/* vport */
		osm_vport_t *p_vport;
		p_vport = osm_port_get_vport(p_alias_guid->p_base_port,
				p_alias_guid->vport_index);
		if (p_vport)
			base_guid = osm_vport_get_guid(p_vport);
		else
			return FALSE;
	} else {
		/* physical port */
		base_guid = osm_port_get_guid(p_alias_guid->p_base_port);
	}

	return (p_alias_guid->alias_guid != base_guid);
}

osm_assigned_guids_t *osm_assigned_guids_new(IN const ib_net64_t port_guid,
					     IN const uint32_t num_guids)
{
	osm_assigned_guids_t *p_assigned_guids;

	p_assigned_guids = calloc(1, sizeof(*p_assigned_guids) +
				     sizeof(ib_net64_t) * (num_guids - 1));
	if (p_assigned_guids)
		p_assigned_guids->port_guid = port_guid;
	return p_assigned_guids;
}

void osm_assigned_guids_delete(IN OUT osm_assigned_guids_t ** pp_assigned_guids)
{
	free(*pp_assigned_guids);
	*pp_assigned_guids = NULL;
}

osm_assigned_guids_t *osm_get_assigned_guids_by_guid(IN osm_subn_t const *p_subn,
						     IN ib_net64_t port_guid)
{
	osm_assigned_guids_t *p_assigned_guids;

	p_assigned_guids = (osm_assigned_guids_t *) cl_qmap_get(&p_subn->assigned_guids_tbl, port_guid);
	if (p_assigned_guids == (osm_assigned_guids_t *) cl_qmap_end(&p_subn->assigned_guids_tbl))
		return NULL;
	return p_assigned_guids;
}

osm_port_t *osm_get_port_by_lid_ho(IN osm_subn_t const * subn, IN uint16_t lid)
{
	osm_vlid_mgr_t *p_vlid_mgr = &subn->p_osm->sm.lid_mgr.vlid_mgr;
	osm_port_t *p_port = NULL;
	uint16_t port_lid;

	if (lid < cl_ptr_vector_get_size(&subn->port_lid_tbl))
		p_port = cl_ptr_vector_get(&subn->port_lid_tbl, lid);

	if (!p_port) {
		port_lid = osm_vlid_mgr_get_lid_by_vlid(p_vlid_mgr, lid);
		if (port_lid)
			p_port = osm_get_port_by_lid_ho(subn, port_lid);
	}

	return p_port;
}

osm_gpu_t *osm_get_gpu_by_alid(IN osm_subn_t const * subn, IN uint16_t alid)
{
	osm_gpu_t *p_gpu = NULL;

	if (alid < cl_ptr_vector_get_size(&subn->gpu_alid_tbl))
		p_gpu = cl_ptr_vector_get(&subn->gpu_alid_tbl, alid);

	return p_gpu;
}

uint8_t osm_get_type_by_lid_ho(osm_subn_t const * p_subn, uint16_t lid)
{
	osm_port_t *p_port = osm_get_port_by_lid_ho(p_subn, lid);
	if (!p_port)
		return 0;

	return osm_node_get_type(p_port->p_node);
}

osm_mgrp_t *osm_get_mgrp_by_mgid(IN osm_subn_t * subn, IN ib_gid_t * mgid)
{
	osm_mgrp_t *mgrp;

	mgrp = (osm_mgrp_t *)cl_fmap_get(&subn->mgrp_mgid_tbl, mgid);
	if (mgrp != (osm_mgrp_t *)cl_fmap_end(&subn->mgrp_mgid_tbl))
		return mgrp;
	return NULL;
}

boolean_t is_mepi_supported(struct osm_sm *p_sm, struct osm_node *p_node)
{
	/*
	 * mlnx ext port info is supported if matching capability bit in
	 * general info is enabled or if the capability to force mepi
	 * is enabled in the capability manager for this node
	 */
	return (ib_general_info_is_cap_enabled(&p_node->general_info,
			IB_GENERAL_INFO_EXTENDED_PORT_INFO_SUPPORTED_BIT) ||
		osm_cap_mgr_is_enabled(&p_sm->cap_mgr, MEPI, p_node));
}

static void set_default_qos_options(IN osm_qos_options_t * opt)
{
        opt->max_vls = OSM_DEFAULT_QOS_MAX_VLS;
        opt->high_limit = OSM_DEFAULT_QOS_HIGH_LIMIT;
        opt->vlarb_high = strdup(OSM_DEFAULT_QOS_VLARB_HIGH);
        opt->vlarb_low = strdup(OSM_DEFAULT_QOS_VLARB_LOW);
        opt->sl2vl = strdup(OSM_DEFAULT_QOS_SL2VL);
}

static void subn_init_qos_options(osm_qos_options_t *opt, osm_qos_options_t *f)
{
	opt->max_vls = 0;
	opt->high_limit = -1;
	if (opt->vlarb_high)
		free(opt->vlarb_high);
	opt->vlarb_high = NULL;
	if (opt->vlarb_low)
		free(opt->vlarb_low);
	opt->vlarb_low = NULL;
	if (opt->sl2vl)
		free(opt->sl2vl);
	opt->sl2vl = NULL;
	if (f)
		memcpy(f, opt, sizeof(*f));
}

static size_t comp_opt(osm_log_t *p_log, char *buf, const opt_rec_t *r, void *p_src, void *p_dst, size_t max_size)
{
	char **pp_src, **pp_dst;
	int n;

	if ((r->parse_fn == opts_parse_256bit) &&
	    memcmp(p_src, p_dst, 32)) {
		n = snprintf(buf, max_size, "%s\n", r->name);
		return n;
	}

	if ((r->parse_fn == opts_parse_net64) &&
	    memcmp(p_src, p_dst, 8)) {
		 n = snprintf(buf, max_size, "%s (Def:|0x%016" PRIx64 "|, "
			"Current:|0x%016" PRIx64 "|)\n", r->name,
			cl_ntoh64(*((ib_net64_t*)(p_dst))),
			cl_ntoh64(*((ib_net64_t*)(p_src))));
		return n;
	}

	if (((r->parse_fn == opts_parse_net32) ||
	    (r->parse_fn == opts_parse_uint32) ||
	    (r->parse_fn == opts_parse_uint32_minutes_to_milisec) ||
	    (r->parse_fn == opts_parse_int32) ||
	    (r->parse_fn == opts_parse_boolean)) &&
	    memcmp(p_src, p_dst, 4)) {
		if (r->parse_fn == opts_parse_net32)
			n = snprintf(buf, max_size, "%s (Def:|0x%08" PRIx32 "|, "
				"Current:|0x%08" PRIx32 "|)\n", r->name,
				cl_ntoh32(*((ib_net32_t*)(p_dst))),
				cl_ntoh32(*((ib_net32_t*)(p_src))));
		else if (r->parse_fn == opts_parse_int32)
			n = snprintf(buf, max_size,  "%s (Def:|%d|, Current:|%d|\n",
				r->name, *((int*)(p_dst)), *((int*)(p_src)));
		else if (r->parse_fn == opts_parse_boolean)
			n = snprintf(buf, max_size, "%s (Def:|%s|, Current:|%s|\n",
				r->name, *((boolean_t*)(p_dst)) ? "True" : "False" ,
				*((boolean_t*)(p_src)) ? "True" : "False");
		else
			n = snprintf(buf, max_size, "%s (Def:|0x%08" PRIx32 "|, "
				"Current:|0x%08" PRIx32 "|)\n", r->name,
				*((ib_net32_t*)(p_dst)), *((ib_net32_t*)(p_src)));
		return n;
	}

	if (((r->parse_fn == opts_parse_net16) ||
	    (r->parse_fn == opts_parse_uint16)) &&
	    memcmp(p_src, p_dst, 2)) {
		if (r->parse_fn == opts_parse_net16)
			n = snprintf(buf, max_size, "%s (def:|0x%04" PRIx16 "|, "
				"current:|0x%04" PRIx16 "|)\n", r->name,
				cl_ntoh16(*((ib_net16_t*)(p_dst))),
				cl_ntoh16(*((ib_net16_t*)(p_src))));
		else
			n = snprintf(buf, max_size,  "%s (def:|0x%04" PRIx16 "|, "
				"current:|0x%04" PRIx16 "|)\n", r->name,
				*((ib_net16_t*)(p_dst)), *((ib_net16_t*)(p_src)));
		return n;
	}

	if ((r->parse_fn == opts_parse_uint8) &&
	    memcmp(p_src, p_dst, 1)) {
		n = snprintf(buf, max_size, "%s (def:|%" PRIu8 "|, "
			"current:|%" PRIu8 "|)\n", r->name,
			*((uint8_t*)(p_dst)), *((uint8_t*)(p_src)));
		return n;
	}

	if (r->parse_fn == opts_parse_charp) {
		pp_src = p_src;
		pp_dst = p_dst;
		if (!*pp_src || !*pp_dst) {
			if (!*pp_src && !*pp_dst)
				return 0;
			else {
				if (!*pp_src)
					n = snprintf(buf, max_size, "%s (Def:|%s|, Current:|null|)\n", r->name, *pp_dst);
				else
					n = snprintf(buf, max_size, "%s (Def:|null|, Current:|%s|)\n", r->name, *pp_src);
				return n;
			}
		}

		if (strcmp(*pp_src, *pp_dst)) {
			n = snprintf(buf, max_size, "%s (Def:|%s|, Current:|%s|)\n", r->name, *pp_dst, *pp_src);
			return n;
		}
	}

	return 0;
}

void osm_subn_print_conf_changes(osm_subn_t *p_subn)
{
	osm_subn_opt_t *p_opts_src = &p_subn->opt;
	osm_subn_opt_t *p_opts_dst = p_opts_src->opts_def;
	osm_log_t *p_log = &p_subn->p_osm->log;
	const opt_rec_t *r;
	size_t n = 0, buf_len = 0;
	char buf[LOG_ENTRY_SIZE_MAX - 256];
	char tmp_buf[1024];

	if (!OSM_LOG_IS_ACTIVE_V2(p_log, OSM_LOG_DEBUG))
		return;

	memset(buf, 0, sizeof(buf));
	buf_len = sprintf(buf, "List of parameters not in default value:\n");

	for (r = opt_tbl; r->name; r++) {
		n = comp_opt(p_log, tmp_buf, r,
			     ((void *)p_opts_src + r->opt_offset),
			     ((void *)p_opts_dst + r->opt_offset),
			      sizeof(tmp_buf) - 1);
		if (n) {
			if (buf_len + n < sizeof(buf)) {
				strcat(buf, tmp_buf);
				buf_len += n;
			} else {
				OSM_LOG(p_log, OSM_LOG_DEBUG, "Too many changes\n");
				break;
			}
		}
	}
	if (buf_len)
		OSM_LOG(p_log, OSM_LOG_DEBUG, "%s\n", buf);
	else
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"All parameters in default value\n");

}

static void osm_subn_set_default_opt_internal(IN osm_subn_opt_t * p_opt)
{
	memset(p_opt, 0, sizeof(osm_subn_opt_t));

	memset(p_opt->guid, 0, sizeof(p_opt->guid));
	p_opt->num_guids = 1;
	p_opt->m_key = OSM_DEFAULT_M_KEY;
	p_opt->sm_key = OSM_DEFAULT_SM_KEY;
	p_opt->sa_key = OSM_DEFAULT_SA_KEY;
	p_opt->subnet_prefix = IB_DEFAULT_SUBNET_PREFIX;
	p_opt->m_key_lease_period = 0;
	p_opt->m_key_protect_bits = 0;
	p_opt->m_key_lookup = TRUE;
	p_opt->m_key_per_port = FALSE;
	p_opt->sweep_interval = OSM_DEFAULT_SWEEP_INTERVAL_SECS;
	p_opt->max_wire_smps = OSM_DEFAULT_SMP_MAX_ON_WIRE;
	p_opt->max_wire_smps2 = p_opt->max_wire_smps;
	p_opt->max_sa_reports_queued = OSM_DEFAULT_MAX_SA_REPORTS;
	p_opt->max_sa_reports_on_wire = OSM_DEFAULT_MAX_SA_REPORTS_ON_WIRE;
	p_opt->sa_enhanced_trust_model = FALSE;
	p_opt->sa_etm_allow_untrusted_proxy_requests = TRUE;
	p_opt->sa_etm_allow_untrusted_guidinfo_rec = TRUE;
	p_opt->sa_etm_allow_guidinfo_rec_by_vf = TRUE;
	p_opt->sa_etm_max_num_mcgs = 0;
	p_opt->sa_etm_max_num_srvcs = 0;
	p_opt->sa_etm_max_num_event_subs = 0;
	p_opt->sa_rate_threshold = 0;
	p_opt->sa_check_sgid_spoofing = TRUE;
	p_opt->console = strdup(OSM_DEFAULT_CONSOLE);
	p_opt->console_port = OSM_DEFAULT_CONSOLE_PORT;
	p_opt->transaction_timeout = OSM_DEFAULT_TRANS_TIMEOUT_MILLISEC;
	p_opt->transaction_retries = OSM_DEFAULT_RETRY_COUNT;
	p_opt->long_transaction_timeout = OSM_DEFAULT_LONG_TRANS_TIMEOUT_MILLISEC;
	p_opt->max_smps_timeout = 1000 * p_opt->transaction_timeout *
				  p_opt->transaction_retries;
	/* by default we will consider waiting for 50x transaction timeout normal */
	p_opt->max_msg_fifo_timeout = 50 * OSM_DEFAULT_TRANS_TIMEOUT_MILLISEC;
	p_opt->sm_priority = OSM_DEFAULT_SM_PRIORITY;
	p_opt->master_sm_priority = OSM_DEFAULT_SM_MASTER_PRIORITY;
	p_opt->master_sm_lid = 0;
	p_opt->enable_sm_port_failover = FALSE;
	p_opt->lmc = OSM_DEFAULT_LMC;
	p_opt->rtr_lmc = OSM_DEFAULT_LMC;
	p_opt->prism_an_lmc = OSM_DEFAULT_PRISM_AN_LMC;
	p_opt->lids_per_rtr = 0;
	p_opt->lmc_esp0 = FALSE;
	p_opt->max_op_vls = OSM_DEFAULT_MAX_OP_VLS;
	p_opt->max_op_vls_ca = 0;
	p_opt->max_op_vls_sw = 0;
	p_opt->max_op_vls_rtr = 0;
	p_opt->qos_config_vl_enabled = FALSE;
	p_opt->max_neighbor_mtu = 0;
	p_opt->force_link_speed = IB_LINK_SPEED_SET_LSS;
	p_opt->force_link_speed_ext = IB_LINK_SPEED_EXT_SET_LSES;
	p_opt->force_link_speed_ext2 = IB_LINK_SPEED_EXT2_SET_LSES2;
	p_opt->force_link_width = IB_LINK_WIDTH_SET_LWS;
	/*
	 * Set default value to deprecated parameter fdr10 in order to
	 * notice cases in which this parameter is still specifically
	 * used in the configuration file
	 */
	p_opt->fdr10 = OSM_DEFAULT_FDR10;
	p_opt->support_mepi_speeds = TRUE;
	p_opt->mepi_enabled_speeds = OSM_DEFAULT_MEPI_ENABLED_SPEED;
	p_opt->reassign_lids = FALSE;
	p_opt->ignore_other_sm = FALSE;
	p_opt->single_thread = FALSE;
	p_opt->disable_multicast = FALSE;
	p_opt->force_log_flush = TRUE;
	p_opt->subnet_timeout = OSM_DEFAULT_SUBNET_TIMEOUT;
	p_opt->packet_life_time = OSM_DEFAULT_SWITCH_PACKET_LIFE;
	p_opt->vl_stall_count = OSM_DEFAULT_VL_STALL_COUNT;
	p_opt->leaf_vl_stall_count = OSM_DEFAULT_LEAF_VL_STALL_COUNT;
	p_opt->head_of_queue_lifetime = OSM_DEFAULT_HEAD_OF_QUEUE_LIFE;
	p_opt->leaf_head_of_queue_lifetime =
	    OSM_DEFAULT_LEAF_HEAD_OF_QUEUE_LIFE;
	p_opt->local_phy_errors_threshold = OSM_DEFAULT_ERROR_THRESHOLD;
	p_opt->overrun_errors_threshold = OSM_DEFAULT_ERROR_THRESHOLD;
	p_opt->use_mfttop = TRUE;
	p_opt->sminfo_polling_timeout =
	    OSM_SM_DEFAULT_POLLING_TIMEOUT_MILLISECS;
	p_opt->polling_retry_number = OSM_SM_DEFAULT_POLLING_RETRY_NUMBER;
	p_opt->force_heavy_sweep = FALSE;
	p_opt->log_flags = OSM_LOG_DEFAULT_LEVEL;
	p_opt->honor_guid2lid_file = FALSE;
	p_opt->daemon = FALSE;
	p_opt->sm_inactive = FALSE;
	p_opt->babbling_port_policy = FALSE;
	p_opt->drop_subscr_on_report_fail = TRUE;
	p_opt->drop_event_subscriptions = TRUE;
	p_opt->drop_unreachable_event_subscriptions = TRUE;
	p_opt->ipoib_mcgroup_creation_validation = TRUE;
	p_opt->mcgroup_join_validation = TRUE;
	p_opt->use_original_extended_sa_rates_only = FALSE;
	p_opt->reports = 2;
	p_opt->use_optimized_slvl = TRUE;
	p_opt->use_optimized_port_mask_slvl = TRUE;
	p_opt->fsync_high_avail_files = TRUE;
	p_opt->default_mcg_mtu = 0;
	p_opt->default_mcg_rate = 0;
#ifdef ENABLE_OSM_PERF_MGR
	p_opt->perfmgr = FALSE;
	p_opt->perfmgr_redir = TRUE;
	p_opt->perfmgr_sweep_time_s = OSM_PERFMGR_DEFAULT_SWEEP_TIME_S;
	p_opt->perfmgr_max_outstanding_queries =
	    OSM_PERFMGR_DEFAULT_MAX_OUTSTANDING_QUERIES;
	p_opt->perfmgr_ignore_cas = FALSE;
	p_opt->event_db_dump_file = NULL; /* use default */
	p_opt->perfmgr_rm_nodes = TRUE;
	p_opt->perfmgr_log_errors = TRUE;
	p_opt->perfmgr_query_cpi = TRUE;
	p_opt->perfmgr_xmit_wait_log = FALSE;
	p_opt->perfmgr_xmit_wait_threshold = OSM_PERFMGR_DEFAULT_XMIT_WAIT_THRESHOLD;
#endif				/* ENABLE_OSM_PERF_MGR */

	p_opt->plugin_name = NULL;
	p_opt->plugin_options = NULL;
	p_opt->event_plugin_name = NULL;
	p_opt->event_plugin_options = NULL;
	p_opt->node_name_map_name = NULL;

	p_opt->dump_files_dir = getenv("OSM_TMP_DIR");
	if (!p_opt->dump_files_dir || !(*p_opt->dump_files_dir))
		p_opt->dump_files_dir = strdup(OSM_DEFAULT_TMP_DIR);
	else
		p_opt->dump_files_dir = strdup(p_opt->dump_files_dir);

	p_opt->log_file = strdup(OSM_DEFAULT_LOG_FILE);
	p_opt->log_max_size = 0;
	p_opt->partition_config_file = strdup(OSM_DEFAULT_PARTITION_CONFIG_FILE);
	p_opt->no_partition_enforcement = FALSE;
	p_opt->part_enforce = strdup(OSM_PARTITION_ENFORCE_BOTH);
	p_opt->allow_both_pkeys = FALSE;
	p_opt->keep_pkey_indexes = TRUE;
	p_opt->sm_assigned_guid = 0;
	p_opt->qos = TRUE;
	p_opt->qos_policy_file = NULL;
	p_opt->suppress_sl2vl_mad_status_errors = FALSE;
	p_opt->override_create_mcg_sl = OSM_MCMR_USER_DEFINED_SL;
	p_opt->accum_log_file = TRUE;
	p_opt->port_prof_ignore_file = NULL;
	p_opt->hop_weights_file = NULL;
	p_opt->port_search_ordering_file = NULL;
	p_opt->port_profile_switch_nodes = FALSE;
	p_opt->sweep_on_trap = TRUE;
	p_opt->use_ucast_cache = FALSE;
	p_opt->enable_queries_during_routing = TRUE;
	p_opt->allowed_sm_guids = NULL;
	p_opt->avoid_throttled_links = FALSE;
	p_opt->connect_roots = FALSE;
	p_opt->calc_missing_routes = TRUE;
	p_opt->enhanced_missing_routes = OSM_ENH_MISSING_ROUTES_DISABLED;
	p_opt->max_cas_on_spine = OSM_DEFAULT_MAX_CAS_ON_SPINE;
	p_opt->find_roots_color_algorithm = TRUE;
	p_opt->dfp_find_roots_color_algorithm = FALSE;
	p_opt->lid_matrix_dump_file = NULL;
	p_opt->lfts_file = NULL;
	p_opt->root_guid_file = NULL;
	p_opt->pgrp_policy_file = NULL;
	p_opt->topo_policy_file = NULL;
	p_opt->rch_policy_file = NULL;
	p_opt->device_configuration_file = NULL;
	p_opt->congestion_control_policy_file = NULL;
	p_opt->ppcc_algo_dir = strdup(OSM_DEFAULT_PPCC_ALGO_DIR);
	p_opt->fast_recovery_enabled = OSM_FAST_RECOVERY_CAP_DEFAULT;
	p_opt->fast_recovery_conf_file = strdup(OSM_DEFAULT_FAST_RECOVERY_CONF_FILE);
	p_opt->ftree_ca_order_dump_file = strdup(OSM_DEFAULT_FTREE_CA_ORDER_DUMP_FILE);
	p_opt->held_back_sw_file = NULL;
	p_opt->cn_guid_file = NULL;
	p_opt->io_guid_file = NULL;
	p_opt->port_shifting = FALSE;
	p_opt->mc_primary_root_guid = 0;
	p_opt->mc_secondary_root_guid = 0;
	p_opt->mc_roots_file = NULL;
	p_opt->scatter_ports = OSM_DEFAULT_SCATTER_PORTS;
	p_opt->updn_lid_tracking_mode = FALSE;
	p_opt->updn_lid_tracking_converge_routes = TRUE;
	p_opt->updn_lid_tracking_prefer_total_routes = FALSE;
	p_opt->dfp_max_cas_on_spine = OSM_DEFAULT_MAX_CAS_ON_SPINE;
	p_opt->const_multicast = FALSE;
	p_opt->max_seq_redisc = OSM_DEFAULT_MAX_SEQ_REDISC;
	p_opt->rereg_on_guid_migr = FALSE;
	p_opt->aguid_inout_notice = FALSE;
	p_opt->sm_assign_guid_func = strdup(OSM_UNIQ_CNT_FUNC);
	p_opt->max_reverse_hops = 0;
	p_opt->routing_threads_num = OSM_DEFAULT_ROUTING_THREADS_NUM;
	p_opt->max_threads_per_core = OSM_DEFAULT_MAX_THREADS_PER_CORE;
	p_opt->ids_guid_file = NULL;
	p_opt->guid_routing_order_file = NULL;
	p_opt->offsweep_balancing_enabled = FALSE;
	p_opt->offsweep_balancing_window = OSM_DEFAULT_OFFSWEEP_BALANCING_WINDOW;
	p_opt->pqft_structure = NULL;
	p_opt->guid_routing_order_no_scatter = FALSE;
	p_opt->use_scatter_for_switch_lid = FALSE;
	p_opt->sa_db_file = NULL;
	p_opt->sa_db_dump = FALSE;
	p_opt->sm_db_dump = TRUE;
	p_opt->torus_conf_file = strdup(OSM_DEFAULT_TORUS_CONF_FILE);
	p_opt->do_mesh_analysis = FALSE;
	p_opt->exit_on_fatal = TRUE;
	p_opt->mlnx_congestion_control = OSM_MLNX_CC_CAP_DEFAULT;
	p_opt->cc_key = OSM_DEFAULT_CC_KEY;
	p_opt->cc_max_outstanding_mads = OSM_CC_DEFAULT_MAX_OUTSTANDING_QUERIES;
	p_opt->cc_key_enable = OSM_KEY_DEFAULT_ENABLEMENT;
	p_opt->cc_key_lease_period = OSM_KEY_DEFAULT_LEASE_PERIOD;
	p_opt->cc_key_protect_bit = 1;
	p_opt->vs_key_enable = OSM_KEY_DEFAULT_ENABLEMENT;
	p_opt->vs_key_lease_period = OSM_KEY_DEFAULT_LEASE_PERIOD;
	p_opt->vs_key_ci_protect_bits = OSM_VS_KEY_DEFAULT_CI_PROTECT;
	p_opt->vs_max_outstanding_mads = OSM_VS_DEFAULT_MAX_OUTSTANDING_QUERIES;
	p_opt->n2n_key_enable = OSM_KEY_DEFAULT_ENABLEMENT;
	p_opt->n2n_key_lease_period = OSM_KEY_DEFAULT_LEASE_PERIOD;
	p_opt->n2n_key_protect_bit = OSM_N2N_KEY_DEFAULT_PROTECT_BIT;
	p_opt->n2n_max_outstanding_mads = OSM_N2N_DEFAULT_MAX_OUTSTANDING_QUERIES;
	p_opt->key_mgr_seed = 0;
	p_opt->enable_quirks = FALSE;
	p_opt->no_clients_rereg = FALSE;
	p_opt->client_rereg_mode = OSM_DEFAULT_CLIENT_REREG_MODE;
	p_opt->prefix_routes_file = strdup(OSM_DEFAULT_PREFIX_ROUTES_FILE);
	p_opt->consolidate_ipv6_snm_req = TRUE;
	p_opt->consolidate_ipv4_mask = 0xffffffff;
	p_opt->pid_file = strdup(OSM_DEFAULT_PID_FILE);
	p_opt->ca_port_num_limit = 0;
	p_opt->ca_port_soft_limit = 0;
	p_opt->switch_name = strdup(OSM_DEFAULT_SWITCH_NAME);
	p_opt->lash_start_vl = 0;
	p_opt->sm_sl = OSM_DEFAULT_SL;
	p_opt->log_prefix = NULL;
	p_opt->per_module_logging_file = strdup(OSM_DEFAULT_PER_MOD_LOGGING_CONF_FILE);
	p_opt->max_msg_fifo_len = OSM_DEFAULT_MAX_SA_GET_FIFO_SIZE;
	p_opt->max_alt_dr_path_retries = 0;
	subn_init_qos_options(&p_opt->qos_options, NULL);
	subn_init_qos_options(&p_opt->qos_ca_options, NULL);
	subn_init_qos_options(&p_opt->qos_sw0_options, NULL);
	subn_init_qos_options(&p_opt->qos_swe_options, NULL);
	subn_init_qos_options(&p_opt->qos_sw2sw_options, NULL);
	subn_init_qos_options(&p_opt->qos_rtr_options, NULL);
	set_default_qos_options(&p_opt->qos_options_prev);
	set_default_qos_options(&p_opt->qos_ca_options_prev);
	set_default_qos_options(&p_opt->qos_sw0_options_prev);
	set_default_qos_options(&p_opt->qos_swe_options_prev);
	set_default_qos_options(&p_opt->qos_sw2sw_options_prev);
	set_default_qos_options(&p_opt->qos_rtr_options_prev);
	p_opt->sa_pr_full_world_queries_allowed = FALSE;
	p_opt->enable_crashd = TRUE;
	p_opt->hm_unhealthy_ports_checks = TRUE;
	p_opt->hm_ports_health_policy_file = NULL;
	p_opt->hm_ca_reboot_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_ca_unresponsive_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_ca_noisy_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_ca_seterr_action = OSM_HM_ACTION_IGNORE;
	p_opt->hm_ca_flapping_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_ca_illegal_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_ca_manual_action = OSM_HM_ACTION_IGNORE;
	p_opt->hm_sw_reboot_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_sw_unresponsive_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_sw_noisy_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_sw_seterr_action = OSM_HM_ACTION_IGNORE;
	p_opt->hm_sw_flapping_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_sw_illegal_action = OSM_HM_ACTION_REPORT;
	p_opt->hm_sw_manual_action = OSM_HM_ACTION_NO_DISCOVER;
	p_opt->hm_ca_reboot_action_s = strdup("report");
	p_opt->hm_ca_unresponsive_action_s = strdup("report");
	p_opt->hm_ca_noisy_action_s = strdup("report");
	p_opt->hm_ca_seterr_action_s = strdup("ignore");
	p_opt->hm_ca_flapping_action_s = strdup("report");
	p_opt->hm_ca_illegal_action_s = strdup("report");
	p_opt->hm_ca_manual_action_s = strdup("ignore");
	p_opt->hm_sw_reboot_action_s = strdup("report");
	p_opt->hm_sw_unresponsive_action_s = strdup("report");
	p_opt->hm_sw_noisy_action_s = strdup("report");
	p_opt->hm_sw_seterr_action_s = strdup("ignore");
	p_opt->hm_sw_flapping_action_s = strdup("report");
	p_opt->hm_sw_illegal_action_s = strdup("report");
	p_opt->hm_sw_manual_action_s = strdup("no_discover");
	p_opt->hm_num_reboots = OSM_DEFAULT_HM_NUM_REBOOTS;
	p_opt->hm_reboots_period_secs = OSM_DEFAULT_HM_REBOOTS_PERIOD_SECS;
	p_opt->hm_num_set_err_sweeps = OSM_DEFAULT_HM_NUM_SET_ERR_SWEEPS;
	p_opt->hm_num_set_err_sweeps_window = OSM_DEFAULT_HM_NUM_ERR_SWEEPS_WINDOW;
	p_opt->hm_num_no_resp_sweeps = OSM_DEFAULT_HM_NUM_NO_RESP_SWEEPS;
	p_opt->hm_num_no_resp_sweeps_window = OSM_DEFAULT_HM_NUM_NO_RESP_SWEEPS_WINDOW;
	p_opt->hm_num_traps = OSM_DEFAULT_HM_NUM_TRAPS;
	p_opt->hm_num_traps_period_secs = OSM_DEFAULT_HM_NUM_TRAPS_PERIOD_SECS;
	p_opt->hm_num_flapping_sweeps = OSM_DEFAULT_HM_NUM_FLAPPING_SWEEPS;
	p_opt->hm_num_flapping_sweeps_window = OSM_DEFAULT_HM_NUM_FLAPPING_SWEEPS_WINDOW;
	p_opt->hm_num_illegal = OSM_DEFAULT_HM_NUM_ILLEGAL;
	p_opt->force_heavy_sweep_window = OSM_DEFAULT_FORCE_HEAVY_SWEEP_WINDOW;
	p_opt->validate_smps = TRUE;
	p_opt->max_topologies_per_sw = OSM_DEFAULT_TOPOLOGIES_PER_SW;
	p_opt->auto_join_hca_wo_topo = TRUE;
	p_opt->enable_inc_mc_routing = FALSE;
	p_opt->allow_sm_port_reset = FALSE;
	p_opt->port_speed_change_action = OSM_PORT_RESET_TYPE_NONE;
	p_opt->port_ext_speed_change_action = OSM_PORT_RESET_TYPE_NONE;
	p_opt->port_ext_speed2_change_action = OSM_PORT_RESET_TYPE_NONE;
	p_opt->port_mepi_speed_change_action = OSM_PORT_RESET_TYPE_NONE;
	p_opt->port_mtu_change_action = OSM_PORT_RESET_TYPE_LOGICAL;
	p_opt->port_vl_change_action = OSM_PORT_RESET_TYPE_LOGICAL;
	p_opt->port_ame_bit_change_action = OSM_PORT_RESET_TYPE_PHYSICAL;
	p_opt->port_speed_change_action_s = strdup("no_reset");
	p_opt->port_ext_speed_change_action_s = strdup("no_reset");
	p_opt->port_ext_speed2_change_action_s = strdup("no_reset");
	p_opt->port_mepi_speed_change_action_s = strdup("no_reset");
	p_opt->port_mtu_change_action_s = strdup("logical");
	p_opt->port_vl_change_action_s = strdup("logical");
	p_opt->port_ame_bit_change_action_s = strdup("physical");
	p_opt->support_mlnx_enhanced_link = FALSE;
	p_opt->mlnx_enhanced_link_enable = TRUE;
	p_opt->adaptive_timeout_sl_mask = OSM_DEFAULT_ADAPTIVE_TIMEOUT_SL_MASK;
	p_opt->mepi_cache_enabled = TRUE;
	p_opt->virt_enabled = OSM_CAPABILITY_ENABLED;
	p_opt->virt_max_ports_in_process = OSM_DEFAULT_VIRT_MAX_PORTS_IN_PROCESS;
	p_opt->virt_default_hop_limit = OSM_DEFAULT_VIRT_HOP_LIMIT;
	p_opt->enable_virt_rec_ext = TRUE;
	p_opt->improved_lmc_path_distribution = TRUE;
	p_opt->syslog_log_flags = 0;
	p_opt->sweep_every_hup_signal = TRUE;
	p_opt->osm_stats_interval = OSM_DEFAULT_STATISTIC_DUMP_TIME_MINUTES;
	p_opt->osm_stats_dump_limit = OSM_DEFAULT_STATISTIC_DUMP_FILE_SIZE;
	p_opt->osm_stats_dump_per_sm_port = TRUE;
	p_opt->osm_perflog_dump_limit = OSM_DEFAULT_PERFLOG_DUMP_FILE_SIZE;
	p_opt->enable_performance_logging = TRUE;
	p_opt->quasi_ftree_indexing = FALSE;
	p_opt->sharp_enabled = OSM_SHARP_CAP_DEFAULT;
	p_opt->aguid_default_hop_limit = OSM_DEFAULT_AGUID_HOP_LIMIT;
	p_opt->rtr_aguid_enable = OSM_RTR_AGUID_DEFAULT;
	p_opt->rtr_pr_flow_label = OSM_DEFAULT_RTR_PR_FLOWLABEL;
	p_opt->rtr_pr_tclass = OSM_DEFAULT_RTR_PR_TCLASS;
	p_opt->rtr_pr_sl = OSM_DEFAULT_SL;
	p_opt->rtr_pr_mtu = OSM_DEFAULT_RTR_PR_MTU;
	p_opt->rtr_pr_rate = OSM_DEFAULT_RTR_PR_RATE;
	p_opt->router_policy_file = NULL;
	p_opt->router_policy_enabled = FALSE;
	p_opt->load_inform_info_from_sadb = TRUE;
	p_opt->verbose_bypass_policy_file = NULL;
	p_opt->sa_db_dump_file = NULL;
	p_opt->enable_subnet_lst = FALSE;
	p_opt->subnet_lst_file = NULL;
	p_opt->dor_hyper_cube_mode = FALSE;
	p_opt->virt_dump_file = NULL;
	p_opt->additional_mepi_force_devices = NULL;
	p_opt->disable_default_gi_support = FALSE;
	p_opt->stand_alone = OSM_STANDALONE_NA;
	p_opt->activity_report_subjects = strdup(OSM_ACTIVITY_DEFAULT_SUBJECTS);
	p_opt->enhanced_qos_policy_file = NULL;
	p_opt->enhanced_qos_vport0_unlimit_default_rl = TRUE;
	p_opt->adv_routing_engine_name = strdup("auto");
	p_opt->ar_mode = OSM_ADV_ROUTING_MODE_AUTO;
	p_opt->shield_mode = OSM_ADV_ROUTING_MODE_AUTO;
	p_opt->ar_sl_mask = OSM_ALL_SL_VL_AR_ENABLED;
	p_opt->ar_tree_asymmetric_flow = OSM_ASYMMETRIC_TREE_1_SUB_GROUP_ENABLED;
	p_opt->ar_tree_asymmetric_flow_threshold = 0;
	p_opt->ar_tree_asymmetric_flow_threshold_limit = 2;
	p_opt->routing_flags = 0;
	p_opt->enable_ar_by_device_cap = TRUE;
	p_opt->enable_ar_group_copy = TRUE;
	p_opt->ar_transport_mask = 0xA; /* RC, DC are enabled the rest are not */
	p_opt->smp_threads = 0;
	p_opt->gmp_threads = 0;
	p_opt->gmp_traps_threads_num = OSM_DEFAULT_GMP_TRAPS_THREADS_NUM;
	p_opt->dump_ar = FALSE;
	p_opt->cache_ar_group_id = TRUE;
	p_opt->dfp_down_up_turns_mode = OSM_DFP_DOWN_PATH_INTERMEDIATE_SPINE_NO_SHORTER_PATH;
	p_opt->enable_vl_packing = TRUE;
	/* Adjust defaults to mlnx_os */
#ifdef ENABLE_MLNX_OS
	p_opt->master_sm_priority = 0;
	p_opt->validate_smps = FALSE;
	p_opt->hm_unhealthy_ports_checks = FALSE;
	p_opt->calc_missing_routes = FALSE;
	p_opt->load_inform_info_from_sadb = FALSE;
	p_opt->sm_db_dump = FALSE;
	p_opt->routing_engine_names = strdup("minhop");
#else
	p_opt->routing_engine_names = strdup(OSM_DEFAULT_ROUTING_ENGINE_NAMES);
#endif
	p_opt->service_name2key_map_file = NULL;
	p_opt->topo_config_file = NULL;
	p_opt->topo_config_enabled = FALSE;
	p_opt->max_wire_smps_per_device = OSM_DEFAULT_MAX_WIRE_SMPS_PER_DEVICE;
	p_opt->rtr_selection_function = OSM_CRC32;
	p_opt->rtr_selection_function_s = strdup("crc32");
	p_opt->rtr_selection_seed = 0;
	p_opt->rtr_selection_algo_parameters_s = strdup("sgid,dgid");
	p_opt->rtr_selection_params_mask =
	    OSM_RTR_SELECTION_PARAM_SGID | OSM_RTR_SELECTION_PARAM_DGID;
	p_opt->global_flid_start = 0;
	p_opt->global_flid_end = 0;
	p_opt->local_flid_start = 0;
	p_opt->local_flid_end = 0;
	p_opt->flid_compression = OSM_DEFAULT_FLID_COMPRESSION;
	p_opt->max_rate_enum = SA_RATE_MAX_ENUM;
	p_opt->get_mft_tables = TRUE;
	p_opt->hbf_sl_mask = OSM_DEFAULT_HBF_SL_MASK;
	p_opt->hbf_hash_type = OSM_HBF_HASH_TYPE_CRC;
	p_opt->hbf_seed_type = OSM_HBF_SEED_TYPE_CONFIGURABLE;
	p_opt->hbf_seed = OSM_HBF_HASH_SEED_GUID;
	p_opt->hbf_hash_fields = OSM_DEFAULT_HBF_HASH_FIELDS;
	p_opt->hbf_weights = strdup(OSM_DEFAULT_WHBF_CONFIG);
	p_opt->pfrn_sl = OSM_DEFAULT_PFRN_SL;
	p_opt->pfrn_mask_clear_timeout = OSM_DEFAULT_PFRN_MASK_CLEAR;
	p_opt->pfrn_mask_force_clear_timeout = OSM_DEFAULT_PFRN_MASK_FORCE_CLEAR;
	p_opt->pfrn_over_router_enabled = OSM_CAPABILITY_ENABLED;
	p_opt->suppress_mc_pkey_traps = OSM_DEFAULT_SUPPRESS_MC_PKEY_TRAPS;
	p_opt->nvlink_enable = FALSE;
	p_opt->nvl_reassign_rails = FALSE;
	p_opt->nvl_hbf_hash_type = OSM_DEFAULT_NVL_HBF_HASH_TYPE;
	p_opt->nvl_hbf_packet_hash_bitmask = OSM_DEFAULT_NVL_HBF_PACKET_HASH_BITMASK;
	p_opt->nvl_hbf_seed = OSM_DEFAULT_NVL_HBF_SEED;
	p_opt->nvl_hbf_fields_enable = OSM_DEFAULT_NVL_HBF_FIELDS_ENABLE;
	p_opt->nvl_partition_reroute_wait = OSM_DEFAULT_NVL_PRTN_REROUTE_WAIT;
	p_opt->tenants_policy_enabled = FALSE;
	p_opt->tenants_policy_file = strdup(OSM_DEFAULT_TENANTS_POLICY_CONF_FILE);
	p_opt->reply_lid_smps_in_dr = FALSE;
	p_opt->respond_unknown_lid_traps = FALSE;
	p_opt->fabric_mode_profile = strdup(OSM_DEFAULT_FABRIC_MODE_PROFILE);
	p_opt->fabric_mode_policy_file = strdup(OSM_DEFAULT_FABRIC_MODE_POLICY_CONF_FILE);
	p_opt->issu_mode = ISSU_MODE_DISABLED;
	p_opt->issu_timeout = 0;
	p_opt->issu_pre_upgrade_time = 0;
	p_opt->conf_template = NULL;
	p_opt->dbg_lvl = 0;
	p_opt->validate_conf_files = FALSE;
}

void osm_subn_set_default_opt(IN osm_subn_opt_t * p_opt)
{
	osm_subn_set_default_opt_internal(p_opt);

	if (!p_opt->opts_def &&
	    !(p_opt->opts_def = calloc(1, sizeof(*p_opt))))
		return;

	osm_subn_set_default_opt_internal(p_opt->opts_def);
}

static char *clean_val(char *val)
{
	char *p = val;
	/* clean leading spaces */
	while (isspace(*p))
		p++;
	val = p;
	if (!*val)
		return val;
	/* clean trailing spaces */
	p = val + strlen(val) - 1;
	while (p > val && isspace(*p))
		p--;
	p[1] = '\0';
	/* clean quotas */
	if ((*val == '\"' && *p == '\"') || (*val == '\'' && *p == '\'')) {
		val++;
		*p-- = '\0';
	}
	return val;
}

static int subn_dump_qos_options(FILE * file, const char *set_name,
				 const char *prefix, osm_qos_options_t * opt)
{
	return fprintf(file, "# %s\n"
		       "%s_max_vls %u\n"
		       "%s_high_limit %d\n"
		       "%s_vlarb_high %s\n"
		       "%s_vlarb_low %s\n"
		       "%s_sl2vl %s\n\n",
		       set_name,
		       prefix, opt->max_vls,
		       prefix, opt->high_limit,
		       prefix, opt->vlarb_high,
		       prefix, opt->vlarb_low, prefix, opt->sl2vl);
}

static ib_api_status_t append_prefix_route(IN osm_subn_t * p_subn,
					   uint64_t prefix, uint64_t guid)
{
	osm_prefix_route_t *route;

	route = malloc(sizeof *route);
	if (! route) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "out of memory");
		return IB_ERROR;
	}

	route->prefix = cl_hton64(prefix);
	route->guid = cl_hton64(guid);
	cl_qlist_insert_tail(&p_subn->prefix_routes_list, &route->list_item);
	return IB_SUCCESS;
}

static ib_api_status_t add_neighbor_info(osm_subn_t *p_subn, st_table *p_topoconfig_hash_tbl,
					 uint64_t sw_guid, uint8_t port,
					 uint64_t neighbor_guid, uint8_t neighbor_port,
					 uint8_t neighbor_node_type, osm_topoconfig_state_enum_t state)
{
	osm_neighbors_info_t *p_neighbors_info = NULL;
	osm_neighbor_info_t *p_neighbor_info = NULL;
	osm_log_t *p_log = &p_subn->p_osm->log;
	cl_status_t status = IB_SUCCESS;

	if (st_lookup(p_topoconfig_hash_tbl, (st_data_t) sw_guid, (void *)&p_neighbors_info)) {
		if (p_neighbors_info->neighbors[port].state != OSM_TOPOCONFIG_NONE) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7568: GUID: 0x%016" PRIx64
				":%d already exists in the table\n", cl_ntoh64(sw_guid), port);
			status = IB_ERROR;
			goto Exit;
		}
	} else {
		p_neighbors_info = (osm_neighbors_info_t *) calloc(1, sizeof(osm_neighbors_info_t));
		if (!p_neighbors_info) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR 7569: Couldn't allocate neighbors_info for topo config\n");
			status = IB_ERROR;
			goto Exit;
		}
		st_insert(p_topoconfig_hash_tbl, (st_data_t) sw_guid, (st_data_t) p_neighbors_info);
	}

	p_neighbors_info->max_port_num = MAX(p_neighbors_info->max_port_num, port);
	p_neighbor_info = &p_neighbors_info->neighbors[port];
	p_neighbor_info->node_type = neighbor_node_type;
	p_neighbor_info->guid = neighbor_guid;
	p_neighbor_info->port = neighbor_port;
	p_neighbor_info->state = state;

Exit:
	return status;
}

osm_neighbors_info_t *osm_topoconfig_lookup(st_table *p_topoconfig_hash_tbl, uint64_t guid)
{
	osm_neighbors_info_t *p_neighbors_info = NULL;

	if (!st_lookup(p_topoconfig_hash_tbl, (st_data_t) guid, (void *)&p_neighbors_info))
		p_neighbors_info = NULL;

	return p_neighbors_info;
}

osm_neighbor_info_t *osm_topoconfig_lookup_port(st_table *p_topoconfig_hash_tbl,
						uint64_t guid, uint8_t port)
{
	osm_neighbors_info_t *p_neighbors_info;
	osm_neighbor_info_t *p_neighbor_info = NULL;

	if (st_lookup(p_topoconfig_hash_tbl, (st_data_t) guid, (void *)&p_neighbors_info))
		p_neighbor_info = &p_neighbors_info->neighbors[port];

	return p_neighbor_info;
}

boolean_t osm_topoconfig_neighbor_is_disabled(osm_neighbors_info_t *p_neighbors_info, uint8_t port)
{
	return p_neighbors_info && (p_neighbors_info->neighbors[port].state == OSM_TOPOCONFIG_DISABLED);
}

static ib_api_status_t parse_prefix_routes_file(IN osm_subn_t * p_subn)
{
	osm_log_t *log = &p_subn->p_osm->log;
	FILE *fp;
	char buf[1024];
	int line = 0;
	int errors = 0;

	while (!cl_is_qlist_empty(&p_subn->prefix_routes_list)) {
		cl_list_item_t *item = cl_qlist_remove_head(&p_subn->prefix_routes_list);
		free(item);
	}

	fp = fopen(p_subn->opt.prefix_routes_file, "r");
	if (! fp) {
		if (errno == ENOENT)
			return IB_SUCCESS;

		OSM_LOG(log, OSM_LOG_ERROR, "fopen(%s) failed: %s",
			p_subn->opt.prefix_routes_file, strerror(errno));
		return IB_ERROR;
	}

	while (fgets(buf, sizeof buf, fp) != NULL) {
		char *p_prefix, *p_guid, *p_extra, *p_last, *p_end;
		uint64_t prefix, guid;

		line++;
		if (errors > 10)
			break;

		p_prefix = strtok_r(buf, " \t\n", &p_last);
		if (! p_prefix)
			continue; /* ignore blank lines */

		if (*p_prefix == '#')
			continue; /* ignore comment lines */

		p_guid = strtok_r(NULL, " \t\n", &p_last);
		if (! p_guid) {
			OSM_LOG(log, OSM_LOG_ERROR, "%s:%d: missing GUID\n",
				p_subn->opt.prefix_routes_file, line);
			errors++;
			continue;
		}

		p_extra = strtok_r(NULL, " \t\n", &p_last);
		if (p_extra && *p_extra != '#') {
			OSM_LOG(log, OSM_LOG_INFO, "%s:%d: extra tokens ignored\n",
				p_subn->opt.prefix_routes_file, line);
		}

		if (strcmp(p_prefix, "*") == 0)
			prefix = 0;
		else {
			prefix = strtoull(p_prefix, &p_end, 16);
			if (*p_end != '\0') {
				OSM_LOG(log, OSM_LOG_ERROR, "%s:%d: illegal prefix: %s\n",
					p_subn->opt.prefix_routes_file, line, p_prefix);
				errors++;
				continue;
			}
		}

		if (strcmp(p_guid, "*") == 0)
			guid = 0;
		else {
			guid = strtoull(p_guid, &p_end, 16);
			if (*p_end != '\0' && *p_end != '#') {
				OSM_LOG(log, OSM_LOG_ERROR, "%s:%d: illegal GUID: %s\n",
					p_subn->opt.prefix_routes_file, line, p_guid);
				errors++;
				continue;
			}
		}

		if (append_prefix_route(p_subn, prefix, guid) != IB_SUCCESS) {
			errors++;
			break;
		}
	}

	fclose(fp);
	return (errors == 0) ? IB_SUCCESS : IB_ERROR;
}

static int topoconfig_get_key_of_tbl_entry(st_data_t key, st_data_t val, st_data_t arg)
{
	cl_u64_vector_t *p_vector = (cl_u64_vector_t *) arg;
	uint64_t guid = (uint64_t) key;

	cl_u64_vector_insert(p_vector, guid, NULL);
	return ST_CONTINUE;
}

static char *topoconfig_state_to_str(osm_topoconfig_state_enum_t state)
{
	switch (state) {
	case OSM_TOPOCONFIG_DISABLED:
		return "disabled";
	case OSM_TOPOCONFIG_NO_DISCOVER:
		return "no-discover";
	case OSM_TOPOCONFIG_ACTIVE:
		return "active";
	case OSM_TOPOCONFIG_NONE:
		return "none";
	}

	return "none";
}

static ib_api_status_t topoconfig_validate_topo(IN osm_subn_t * p_subn, st_table *p_topoconfig_hash_tbl)
{
	osm_neighbors_info_t *p_neighbors_info;
	osm_neighbor_info_t *p_neighbor_info, *p_info;
	ib_api_status_t status = IB_SUCCESS;
	cl_u64_vector_t vector;
	uint32_t vector_size;
	uint8_t port;
	int i;

	cl_u64_vector_construct(&vector);
	cl_u64_vector_init(&vector, 0, TRUE);

	st_foreach(p_topoconfig_hash_tbl, topoconfig_get_key_of_tbl_entry,
		   (st_data_t) &vector);

	vector_size = cl_u64_vector_get_size(&vector);
	for (i = 0; i < vector_size; i++) {
		uint64_t guid = cl_u64_vector_get(&vector, i);
		p_neighbors_info = osm_topoconfig_lookup(p_topoconfig_hash_tbl, guid);

		for (port = 1; port <= p_neighbors_info->max_port_num; port++) {
			p_neighbor_info = &p_neighbors_info->neighbors[port];
			if (!p_neighbor_info->guid || p_neighbor_info->node_type != IB_NODE_TYPE_SWITCH)
				continue;

			/* Check that both sides of the link have the same data */
			p_info = osm_topoconfig_lookup_port(p_topoconfig_hash_tbl, p_neighbor_info->guid, p_neighbor_info->port);
			if (!p_info) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7562: Validation of topoconfig file failed. "
					"Only one side of the link exists. "
					"0x%" PRIx64 ":%u -> 0x%" PRIx64 ":%u\n",
					cl_ntoh64(guid), port, cl_ntoh64(p_neighbor_info->guid), p_neighbor_info->port);
				status = IB_ERROR;
				goto Exit;
			}

			if (guid != p_info->guid || port != p_info->port || p_neighbor_info->state != p_info->state) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7563: Validation of topoconfig file failed. "
					"Inconsistent link data: "
					"0x%" PRIx64 ":%u -> 0x%" PRIx64 ":%u state: %s, "
					"0x%" PRIx64 ":%u -> 0x%" PRIx64 ":%u state: %s\n",
					cl_ntoh64(guid), port, cl_ntoh64(p_neighbor_info->guid),
					p_neighbor_info->port, topoconfig_state_to_str(p_neighbor_info->state),
					cl_ntoh64(p_neighbor_info->guid), p_neighbor_info->port,
					cl_ntoh64(p_info->guid), p_info->port, topoconfig_state_to_str(p_info->state));
				status = IB_ERROR;
				goto Exit;
			}
		}
	}

Exit:
	cl_u64_vector_remove_all(&vector);
	cl_u64_vector_destroy(&vector);

	return status;
}

static ib_api_status_t parse_topo_config_file(IN osm_subn_t * p_subn)
{
	ib_api_status_t validation_status = IB_SUCCESS;
	osm_log_t *log = &p_subn->p_osm->log;
	FILE *fp;
	char buf[1024];
	int line = 0;
	int errors = 0;
	st_table *p_topoconfig_hash_tbl;
	boolean_t is_file_empty = FALSE;
	osm_switch_t *p_sw;
	cl_map_item_t *item;

	p_subn->topo_config_enabled = p_subn->opt.topo_config_enabled;
	if (!p_subn->opt.topo_config_enabled)
		return IB_SUCCESS;

	fp = fopen(p_subn->opt.topo_config_file, "r");
	if (!fp) {
		OSM_LOG(log, OSM_LOG_ERROR, "ERR 7550: Failed to open %s: %s. %s\n",
			p_subn->opt.topo_config_file, strerror(errno),
			p_subn->p_topoconfig_hash_tbl ?
			"Keeping current topo configuration" : "Removing all the nodes");

		if (!p_subn->p_topoconfig_hash_tbl)
			p_subn->p_topoconfig_hash_tbl = st_init_numtable_with_size(5000);

		osm_report_notice_1315(p_subn, SM_VS_TRAP_TOPOCONFIG_FAILED_TO_OPEN_FILE);
		return IB_ERROR;
	}

	is_file_empty = TRUE;
	p_topoconfig_hash_tbl = st_init_numtable_with_size(5000);
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		char *p_sw_guid, *p_port, *p_neighbor_guid, *p_neighbor_port, *p_state;
		char *p_neighbor_node_type;
		char *p_extra, *p_last, *p_end;
		uint64_t sw_guid, neighbor_guid;
		uint8_t port, neighbor_port, neighbor_node_type;
		osm_topoconfig_state_enum_t state;

		is_file_empty = FALSE;
		line++;
		if (errors > 10)
			break;

		p_sw_guid = strtok_r(buf, " ,\t\n", &p_last);
		if (!p_sw_guid)
			continue; /* ignore blank lines */

		if (*p_sw_guid == '#')
			continue; /* ignore comment lines */

		p_port = strtok_r(NULL, " ,\t\n", &p_last);
		if (!p_port) {
			OSM_LOG(log, OSM_LOG_ERROR, "ERR 7551: %s:%d missing port\n",
				p_subn->opt.topo_config_file, line);
			errors++;
			continue;
		}

		p_neighbor_guid = strtok_r(NULL, " ,\t\n", &p_last);
		if (!p_neighbor_guid) {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7552: %s:%d missing neighbor node GUID\n",
				p_subn->opt.topo_config_file, line);
			errors++;
			continue;
		}

		p_neighbor_port = strtok_r(NULL, " ,\t\n", &p_last);
		if (!p_neighbor_port) {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7553: %s:%d missing neighbor port\n",
				p_subn->opt.topo_config_file, line);
			errors++;
			continue;
		}

		p_neighbor_node_type = strtok_r(NULL, " ,\t\n", &p_last);
		if (!p_neighbor_node_type) {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7554: %s:%d missing neighbor node type\n",
				p_subn->opt.topo_config_file, line);
			errors++;
			continue;
		}

		p_state = strtok_r(NULL, " ,\t\n", &p_last);
		if (!p_state) {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7555: %s:%d missing state\n",
				p_subn->opt.topo_config_file, line);
			errors++;
			continue;
		}

		p_extra = strtok_r(NULL, " \t\n", &p_last);
		if (p_extra && *p_extra != '#') {
			OSM_LOG(log, OSM_LOG_INFO, "%s:%d: extra tokens ignored\n",
				p_subn->opt.topo_config_file, line);
		}

		sw_guid = cl_hton64(strtoull(p_sw_guid, &p_end, 16));
		if (*p_end != '\0' || !sw_guid) {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7556: %s:%d: illegal GUID: %s\n",
				p_subn->opt.topo_config_file, line, p_sw_guid);
				errors++;
			continue;
		}

		port = strtoul(p_port, &p_end, 0);
		if (*p_end != '\0' && *p_end != '#') {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7557: %s:%d: illegal first port: %s\n",
				p_subn->opt.topo_config_file, line, p_port);
				errors++;
			continue;
		}

		if (!strcmp(p_neighbor_guid, "-")) {
			neighbor_guid = 0;
			if (strcmp(p_neighbor_port, "-")) {
				OSM_LOG(log, OSM_LOG_ERROR,
					"ERR 7566: %s:%d: neighbor port defined: %s, but neighbor GUID not.\n",
					p_subn->opt.topo_config_file, line, p_neighbor_port);
				errors++;
				continue;
			}
		} else {
			neighbor_guid = cl_hton64(strtoull(p_neighbor_guid, &p_end, 16));
			if (*p_end != '\0') {
				OSM_LOG(log, OSM_LOG_ERROR,
					"ERR 7558: %s:%d: illegal neighbor GUID: %s\n",
					p_subn->opt.topo_config_file, line, p_neighbor_guid);
				errors++;
				continue;
			}
		}

		if (!strcmp(p_neighbor_port, "-")) {
			neighbor_port = 0;
			if (strcmp(p_neighbor_guid, "-")) {
				OSM_LOG(log, OSM_LOG_ERROR,
					"ERR 7567: %s:%d: neighbor GUID defined: %s, but port not.\n",
					p_subn->opt.topo_config_file, line, p_neighbor_guid);
				errors++;
				continue;
			}
		} else {
			neighbor_port = strtoul(p_neighbor_port, &p_end, 0);
			if (*p_end != '\0') {
				OSM_LOG(log, OSM_LOG_ERROR,
					"ERR 7559: %s:%d: illegal neighbor port: %s\n",
					p_subn->opt.topo_config_file, line, p_neighbor_port);
				errors++;
				continue;
			}
		}

		if (!strcasecmp(p_neighbor_node_type, "HCA")) {
			neighbor_node_type = IB_NODE_TYPE_CA;
		} else if (!strcasecmp(p_neighbor_node_type, "Switch")) {
			neighbor_node_type = IB_NODE_TYPE_SWITCH;
		} else if (!strcasecmp(p_neighbor_node_type, "Any")) {
			neighbor_node_type = 0;
		} else {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7560: Illegal neighbor node type format: %s, on line %d\n",
				p_neighbor_node_type, line);
			errors++;
			continue;
		}

		if (!strcasecmp(p_state, "Active")) {
			state = OSM_TOPOCONFIG_ACTIVE;
		} else if (!strcasecmp(p_state, "Disabled")) {
			state = OSM_TOPOCONFIG_DISABLED;
		} else if (!strcasecmp(p_state, "No-Discover")) {
			state = OSM_TOPOCONFIG_NO_DISCOVER;
		} else {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7561: Illegal state format: %s, on line %d\n",
				p_state, line);
			errors++;
			continue;
		}

		if (add_neighbor_info(p_subn, p_topoconfig_hash_tbl, sw_guid, port,
				      neighbor_guid, neighbor_port,
				      neighbor_node_type, state) != IB_SUCCESS) {
			errors++;
			break;
		}

		OSM_LOG(log, OSM_LOG_DEBUG,
			"Parsed:  0x%016" PRIx64 ":%d 0x%016" PRIx64 ":%d state: %s\n",
			cl_ntoh64(sw_guid), port, cl_ntoh64(neighbor_guid),
			neighbor_port, p_state);
	}

	validation_status = topoconfig_validate_topo(p_subn, p_topoconfig_hash_tbl);
	if (errors || validation_status) {
		topoconfig_tbl_clear(p_topoconfig_hash_tbl);
		if (p_subn->p_topoconfig_hash_tbl) {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7564: Parsing of topoconfig file failed. "
				"Keeping current configuration\n");
		} else {
			OSM_LOG(log, OSM_LOG_ERROR,
				"ERR 7565: Parsing of topoconfig file failed. "
				"Removing all the nodes\n");
			p_subn->p_topoconfig_hash_tbl = st_init_numtable_with_size(5000);
		}

		if (errors)
			osm_report_notice_1315(p_subn, SM_VS_TRAP_TOPOCONFIG_BAD_SYNTAX);
		else
			osm_report_notice_1315(p_subn, SM_VS_TRAP_TOPOCONFIG_VALIDATION_FAILED);

		goto Exit;
	}

	if (p_subn->p_topoconfig_hash_tbl)
		topoconfig_tbl_clear(p_subn->p_topoconfig_hash_tbl);

	p_subn->p_topoconfig_hash_tbl = p_topoconfig_hash_tbl;

	if (osm_is_crc_changed(&p_subn->p_osm->sm, OSM_CRC_TOPO_CONFIGURATION_FILE)) {
		OSM_LOG(log, OSM_LOG_INFO,
			"The topo config %s changed. Send PortInfo to all ports\n",
			p_subn->opt.topo_config_file);
		for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
		     item != cl_qmap_end(&p_subn->sw_guid_tbl);
		     item = cl_qmap_next(item)) {
			p_sw = (osm_switch_t *) item;
			p_sw->send_port_info = TRUE;
		}
	}

Exit:
	/* If file is empty disable the feature */
	if (is_file_empty) {
		p_subn->topo_config_enabled = FALSE;
		OSM_LOG(log, OSM_LOG_INFO,
			"The topo config %s is empty. Disabling topo config feature\n",
			p_subn->opt.topo_config_file);
	}

	fclose(fp);
	return (errors == 0) && (validation_status == IB_SUCCESS) ? IB_SUCCESS : IB_ERROR;
}

static ib_api_status_t insert_per_module_debug(IN osm_subn_t * p_subn,
					       char *mod_name,
					       osm_log_level_t level)
{
	uint8_t index;

	if (find_module_name(mod_name, &index)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"Module name %s not found\n", mod_name);
		return IB_ERROR;
	}
	osm_set_log_per_module(&p_subn->p_osm->log, index, level);
	return IB_SUCCESS;
}

static ib_api_status_t parse_per_mod_logging_file(IN osm_subn_t * p_subn)
{
	osm_log_t *log = &p_subn->p_osm->log;
	FILE *fp;
	char buf[1024];
	int line = 0;
	int errors = 0;

	osm_reset_log_per_module(log);

	if (p_subn->opt.per_module_logging_file == NULL)
		return IB_SUCCESS;

	fp = fopen(p_subn->opt.per_module_logging_file, "r");
	if (!fp) {
		if (errno == ENOENT)
			return IB_SUCCESS;

		OSM_LOG(log, OSM_LOG_ERROR, "fopen(%s) failed: %s",
			p_subn->opt.per_module_logging_file, strerror(errno));
		return IB_ERROR;
	}

	while (fgets(buf, sizeof buf, fp) != NULL) {
		char *p_mod_name, *p_level, *p_extra, *p_last;
		osm_log_level_t level;

		line++;
		if (errors > 10)
			break;

		p_mod_name = strtok_r(buf, " =,\t\n", &p_last);
		if (!p_mod_name)
			continue; /* ignore blank lines */

		if (*p_mod_name == '#')
			continue; /* ignore comment lines */

		p_level = strtok_r(NULL, " \t\n", &p_last);
		if (!p_level) {
			OSM_LOG(log, OSM_LOG_ERROR, "%s:%d: missing log level\n",
				p_subn->opt.per_module_logging_file, line);
			errors++;
			continue;
		}
		p_extra = strtok_r(NULL, " \t\n", &p_last);
		if (p_extra && *p_extra != '#') {
			OSM_LOG(log, OSM_LOG_INFO, "%s:%d: extra tokens ignored\n",
				p_subn->opt.per_module_logging_file, line);
		}

		level = strtoul(p_level, NULL, 0);
		if (insert_per_module_debug(p_subn, p_mod_name, level) != IB_SUCCESS) {
			errors++;
			break;
		}
	}

	fclose(fp);
	return (errors == 0) ? IB_SUCCESS : IB_ERROR;
}

/*
 * Create unhealthy port, add it to the fmap and return pointer to
 * new added port
 */
static osm_unhealthy_port_t *add_manual_port(IN osm_subn_t *p_subn,
				      IN uint64_t node_guid,
				      IN uint8_t port_num,
				      IN cl_fmap_t *f_map,
				      IN osm_hm_action_type_enum manual_action,
				      IN const char *mode)
{
	osm_log_t *log = &p_subn->p_osm->log;
	cl_fmap_item_t *p_item;
	osm_unhealthy_port_t *p_unhealthy_port;

	p_unhealthy_port = osm_unhealthy_port_new(0, 0, NULL, 0,
						  cl_hton64(node_guid),
						  port_num,
						  NULL);

	if (!p_unhealthy_port) {
		OSM_LOG(log, OSM_LOG_ERROR, "ERR 7522: Out of memory\n");
		goto EXIT;
	}
	p_unhealthy_port->manual_action = manual_action;
	p_item = cl_fmap_insert(f_map, &p_unhealthy_port->p_obj,
				&p_unhealthy_port->fmap_item);

	if (p_item != &p_unhealthy_port->fmap_item) {
		OSM_LOG(log, OSM_LOG_ERROR, "ERR 7524: Fail to add new "
			" unhealthy port, guid 0x%" PRIx64 " port %d, "
			" unhealthy port already exist\n", cl_ntoh64(node_guid),
			port_num);
		free(p_unhealthy_port);
		p_unhealthy_port = NULL;
	} else
		OSM_LOG(log, OSM_LOG_DEBUG, "Adding manual %s NodeGUID 0x%016"
			PRIx64 " port %d\n", mode, node_guid, port_num);
EXIT:
	return p_unhealthy_port;
}

static ib_api_status_t parse_hm_policy_file(IN osm_subn_t *p_subn)
{
	osm_log_t *log = &p_subn->p_osm->log;
	char line[256];
	FILE *f;
	boolean_t recover = FALSE;

	/* Clear those marked as healthy, before parsing */
	osm_hm_clear_marked_healthy_ports(&p_subn->p_osm->sm);
	p_subn->all_healthy = FALSE;

	if (p_subn->opt.hm_unhealthy_ports_checks == FALSE)
		return IB_SUCCESS;

	if (p_subn->opt.hm_ports_health_policy_file == NULL)
		return IB_SUCCESS;

	if (!(f = fopen(p_subn->opt.hm_ports_health_policy_file, "r"))) {
		if (errno == ENOENT)
			return IB_SUCCESS;

		OSM_LOG(log, OSM_LOG_ERROR, "fopen(%s) failed: %s",
			p_subn->opt.hm_ports_health_policy_file, strerror(errno));
		return IB_ERROR;
	}

	while (fgets(line, 255, f)) {
		uint64_t guid;
		uint8_t port_num;
		char *arg1, *arg2, *arg3, *arg4,*p_last, *e;
		osm_unhealthy_port_t *p_uhp; /* uhp = UnHealthyPort */
		boolean_t healthy, unhealthy;
		osm_sm_t * sm;
		boolean_t new;

		new = FALSE;
		healthy = FALSE;
		unhealthy = FALSE;
		arg1 = strtok_r(line, " \t\n", &p_last);
		if (!arg1)
			continue; /* ignore blank lines */

		if (*arg1 == '#')
			continue; /* ignore comment lines */

		arg2 = strtok_r(NULL, " \t\n", &p_last);
		if (!arg2)
			continue;

		/*
		 * Parse "all healthy" string first, but recover
		 * at the end so can keep timestamp of those that remain unhealthy
		 */
		if (!strcasecmp(arg1, "all")) {
			if (!strcasecmp(arg2, "healthy")) {
				recover = TRUE;
				p_subn->all_healthy = TRUE;
				osm_hm_mark_recover_all(p_subn);
			} else {
				fclose(f);
				return IB_ERROR;
			}
		}

		arg3 = strtok_r(NULL, " \t\n", &p_last);
		if (!arg3)
			continue;

		guid = strtoull(arg1, &e, 0);
		if (e == arg1 || (!isspace(*e) && *e != '#' && *e != '\0'))
			continue;

		port_num = (uint8_t) strtoul(arg2, &e, 0);
		if (e == arg2 || (!isspace(*e) && *e != '#' && *e != '\0'))
			continue;

		healthy = !strcasecmp(arg3, "healthy");
		unhealthy = !strcasecmp(arg3, "unhealthy");
		if (!healthy && !unhealthy)
			continue;

		sm = &p_subn->p_osm->sm;
		cl_fmap_t *p_unhealthy_tbl = &p_subn->unhealthy_ports_tbl;
		p_uhp = osm_hm_get_unhealthy_port_by_guid(sm, cl_hton64(guid), port_num);
		if (healthy) {
			/*
			 * Add port to healthy_ports_tbl so in case found as unhealthy,
			 * any unhealthy_reason will be ignored.
			 */
			osm_hm_insert_marked_healthy(sm, cl_hton64(guid), port_num);

			if (p_uhp) {
				OSM_LOG(log, OSM_LOG_DEBUG, "Recovering "
					"unhealthy port 0x%" PRIx64
					"[%d]\n", guid, port_num);
				cl_fmap_remove_item(p_unhealthy_tbl,
						    (cl_fmap_item_t * const)p_uhp);

				/* Remove from unhealthy */
				osm_unhealthy_port_remove(sm, &p_uhp);
				sm->hm_dirty = TRUE;
			} else
				OSM_LOG(log, OSM_LOG_DEBUG, "Ignoring "
					"unhealthy recover request for "
					"non-exist unhealthy port 0x%" PRIx64
					"[%d]\n", guid, port_num);
		} else {
			osm_hm_action_type_enum manual_action;
			arg4 = strtok_r(NULL, " \t\n", &p_last);

			if (!arg4)
				manual_action = OSM_HM_ACTION_IGNORE;
			else if (!(strcasecmp(arg4, "report")))
				manual_action  = OSM_HM_ACTION_REPORT;
			else if (!(strcasecmp(arg4, "isolate")))
				manual_action  = OSM_HM_ACTION_ISOLATE;
			else if (!(strcasecmp(arg4, "disable")))
				manual_action  = OSM_HM_ACTION_DISABLE;
			else if (!(strcasecmp(arg4, "no_discover")))
				manual_action = OSM_HM_ACTION_NO_DISCOVER;
			else {
				OSM_LOG(log, OSM_LOG_ERROR, "ERR 7540: "
					"Ignoring manual unhealthy set "
					"request for port 0x%" PRIx64
					"[%d], invalid action \"%s\"\n",
					guid, port_num, arg4);
				continue;
			}

			/* unhealthy */
			if (!p_uhp) {
				new = TRUE;
				p_uhp = add_manual_port(p_subn, guid, port_num,
							p_unhealthy_tbl,
							manual_action,
							"unhealthy");
			}
			if (p_uhp) {
				if (!new) {
					p_uhp->manual_action = manual_action;
					if (osm_hm_is_cond_set(p_uhp, OSM_HM_COND_MANUAL))
						osm_hm_update_actions(sm, p_uhp);
				}

				osm_hm_set_cond(sm, p_uhp, OSM_HM_COND_MANUAL);
				if (new)
					sm->hm_dirty = TRUE;

				/* So won't be recovered by 'all healthy' if still in unhealthy file */
				p_uhp->recover = FALSE;
			}
		}
	}

	/*
	 * Recover only those missing from unhealthy configuration file,
	 * so if port is already unhealthy, original timestamp is kept
	 */
	if (recover)
		osm_hm_recover_all_healthy(p_subn);

	fclose(f);
	return IB_SUCCESS;
}

/* ----------------------------------------------------------------------------------------------
 * Held back switch section.
 */

#define foreach_held_back_switch(SUBN,P_ITEM)                                           \
	for (P_ITEM  = (held_back_item_t *)cl_qmap_head(&((SUBN)->held_back_sw_tbl));   \
	     P_ITEM != (held_back_item_t *)cl_qmap_end(&((SUBN)->held_back_sw_tbl));    \
	     P_ITEM  = (held_back_item_t *)cl_qmap_next((cl_map_item_t *)P_ITEM))


/* The exclusive lock must be held prior to calling this function. */
static void restore_held_back_switch(IN osm_subn_t *p_subn, IN held_back_item_t *p_held_back_sw)
{
	ib_net64_t guid;
	osm_switch_t *p_sw;

	/* If there is no reason to isolate the switch, restore it. */
	if (!p_held_back_sw->reason) {
		guid = cl_qmap_key(&p_held_back_sw->item);
		p_sw = osm_get_switch_by_guid(p_subn, guid);
		if (p_sw) {
			p_sw->need_update = 2;
		}
		cl_qmap_remove_item(&p_subn->held_back_sw_tbl, &p_held_back_sw->item);
		free(p_held_back_sw);
	}
}

/* The exclusive lock is already held by osm_subn_rescan_conf_files */
static void restore_held_back_switches_file(IN osm_subn_t *p_subn)
{
	cl_map_item_t *p_item;
	held_back_item_t *p_held_back_sw;

	p_item = cl_qmap_head(&p_subn->held_back_sw_tbl);
	while (p_item != cl_qmap_end(&p_subn->held_back_sw_tbl)) {
		p_held_back_sw = (held_back_item_t *)p_item;
		p_item = cl_qmap_next(p_item);
		restore_held_back_switch(p_subn, p_held_back_sw);
	}
}

int osm_restore_held_back_sw_issu(IN osm_subn_t *p_subn, IN ib_net64_t guid)
{
	cl_map_item_t *p_item;
	held_back_item_t *p_held_back_sw;

	CL_PLOCK_EXCL_ACQUIRE(p_subn->p_osm->sm.p_lock);

	p_item = cl_qmap_get(&p_subn->held_back_sw_tbl, guid);
	if (p_item != cl_qmap_end(&p_subn->held_back_sw_tbl)) {
		p_held_back_sw = (held_back_item_t *)p_item;
		p_held_back_sw->reason &= ~HELD_BACK_REASON_ISSU;
		restore_held_back_switch(p_subn, p_held_back_sw);
	}

	CL_PLOCK_RELEASE(p_subn->p_osm->sm.p_lock);

	return 0;
}

static int add_held_back_sw(void *cxt, uint64_t guid, held_back_reason_t reason)
{
	cl_qmap_t *map = cxt;
	cl_map_item_t *inserted_item;
	held_back_item_t *new_item;

	new_item = (held_back_item_t *)calloc(1, sizeof(held_back_item_t));
	if (!new_item)
		return -1;

	inserted_item = cl_qmap_insert(map, cl_hton64(guid), &new_item->item);
	if (&new_item->item != inserted_item) {
		/* Item already existed, change pointer to the existing item */
		free(new_item);
		new_item = (held_back_item_t *)inserted_item;
	}

	/* Add the new reason for holding back the switch */
	new_item->reason |= reason;

	return 0;
}

int osm_add_held_back_sw_issu(IN osm_subn_t *p_subn, uint64_t guid)
{
	return add_held_back_sw(&p_subn->held_back_sw_tbl, guid, HELD_BACK_REASON_ISSU);
}

static int add_held_back_sw_file(void *cxt, uint64_t guid, char *p)
{
	return add_held_back_sw(cxt, guid, HELD_BACK_REASON_FILE);
}

static ib_api_status_t parse_held_back_file(IN osm_subn_t *p_subn)
{
	held_back_item_t *p_held_back_sw;
	ib_api_status_t status = IB_SUCCESS;

	/* Remove the FILE reason for all held back switches */
	foreach_held_back_switch(p_subn, p_held_back_sw) {
		p_held_back_sw->reason &= ~HELD_BACK_REASON_FILE;
	}

	/* Parse the file */
	if (p_subn->opt.held_back_sw_file &&
	    parse_node_map(p_subn->opt.held_back_sw_file,
			   add_held_back_sw_file, &p_subn->held_back_sw_tbl)) {
		status = IB_ERROR;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"Unable to parse Heldback file %s\n",
			p_subn->opt.held_back_sw_file);
	}

	/* Mark switches leaving the heldback file for re-updating */
	restore_held_back_switches_file(p_subn);

	return status;
}

/* ---------------------------------------------------------------------------------------------- */

static void subn_verify_max_vls(unsigned *max_vls, const char *prefix)
{
	if (!*max_vls || *max_vls > 15) {
		if (*max_vls)
			log_report(" Invalid Cached Option: %s_max_vls=%u: "
				   "Using Default = %u\n",
				   prefix, *max_vls, OSM_DEFAULT_QOS_MAX_VLS);
		*max_vls = 0;
	}
}

static void subn_verify_high_limit(int *high_limit, const char *prefix)
{
	if (*high_limit < 0 || *high_limit > 255) {
		if (*high_limit > 255)
			log_report(" Invalid Cached Option: %s_high_limit=%d: "
				   "Using Default: %d\n",
				   prefix, *high_limit,
				   OSM_DEFAULT_QOS_HIGH_LIMIT);
		*high_limit = -1;
	}
}

static void subn_verify_vlarb(char **vlarb, const char *prefix,
			      const char *suffix)
{
	char *str, *tok, *end, *ptr;
	int count = 0;

	if (*vlarb == NULL)
		return;

	str = strdup(*vlarb);

	tok = strtok_r(str, ",\n", &ptr);
	while (tok) {
		char *vl_str, *weight_str;

		vl_str = tok;
		weight_str = strchr(tok, ':');

		if (weight_str) {
			long vl, weight;

			*weight_str = '\0';
			weight_str++;

			vl = strtol(vl_str, &end, 0);

			if (*end)
				log_report(" Warning: Cached Option "
					   "%s_vlarb_%s:vl=%s"
					   " improperly formatted\n",
					   prefix, suffix, vl_str);
			else if (vl < 0 || vl > 14)
				log_report(" Warning: Cached Option "
					   "%s_vlarb_%s:vl=%ld out of range\n",
					   prefix, suffix, vl);

			weight = strtol(weight_str, &end, 0);

			if (*end)
				log_report(" Warning: Cached Option "
					   "%s_vlarb_%s:weight=%s "
					   "improperly formatted\n",
					   prefix, suffix, weight_str);
			else if (weight < 0 || weight > 255)
				log_report(" Warning: Cached Option "
					   "%s_vlarb_%s:weight=%ld "
					   "out of range\n",
					   prefix, suffix, weight);
		} else
			log_report(" Warning: Cached Option "
				   "%s_vlarb_%s:vl:weight=%s "
				   "improperly formatted\n",
				   prefix, suffix, tok);

		count++;
		tok = strtok_r(NULL, ",\n", &ptr);
	}

	if (count > 64)
		log_report(" Warning: Cached Option %s_vlarb_%s: > 64 listed:"
			   " excess vl:weight pairs will be dropped\n",
			   prefix, suffix);

	free(str);
}

static void subn_verify_sl2vl(char **sl2vl, const char *prefix)
{
	char *str, *tok, *end, *ptr;
	int count = 0;

	if (*sl2vl == NULL)
		return;

	str = strdup(*sl2vl);

	tok = strtok_r(str, ",\n", &ptr);
	while (tok) {
		long vl = strtol(tok, &end, 0);

		if (*end)
			log_report(" Warning: Cached Option %s_sl2vl:vl=%s "
				   "improperly formatted\n", prefix, tok);
		else if (vl < 0 || vl > 15)
			log_report(" Warning: Cached Option %s_sl2vl:vl=%ld "
				   "out of range\n", prefix, vl);

		count++;
		tok = strtok_r(NULL, ",\n", &ptr);
	}

	if (count < 16)
		log_report(" Warning: Cached Option %s_sl2vl: < 16 VLs "
			   "listed\n", prefix);
	else if (count > 16)
		log_report(" Warning: Cached Option %s_sl2vl: > 16 listed: "
			   "excess VLs will be dropped\n", prefix);

	free(str);
}

static void subn_verify_qos_set(osm_qos_options_t *set, const char *prefix)
{
	subn_verify_max_vls(&set->max_vls, prefix);
	subn_verify_high_limit(&set->high_limit, prefix);
	subn_verify_vlarb(&set->vlarb_low, prefix, "low");
	subn_verify_vlarb(&set->vlarb_high, prefix, "high");
	subn_verify_sl2vl(&set->sl2vl, prefix);
}

static void set_prev_qos_options(osm_qos_options_t *previous, osm_qos_options_t *current)
{
	previous->max_vls = current->max_vls;
	previous->high_limit = current->high_limit;
	if (previous->vlarb_low)
		free(previous->vlarb_low);
	previous->vlarb_low = current->vlarb_low ? strdup(current->vlarb_low) : NULL;
	if (previous->vlarb_high)
		free(previous->vlarb_high);
	previous->vlarb_high = current->vlarb_high ? strdup(current->vlarb_high) : NULL;
	if (previous->sl2vl)
		free(previous->sl2vl);
	previous->sl2vl = current->sl2vl ? strdup(current->sl2vl) : NULL;
}

static void update_prev_qos_options(IN osm_subn_opt_t * p_opts)
{
	/* Copy current QoS options to previous */
	set_prev_qos_options(&p_opts->qos_options_prev, &p_opts->qos_options);
	set_prev_qos_options(&p_opts->qos_ca_options_prev, &p_opts->qos_ca_options);
	set_prev_qos_options(&p_opts->qos_sw0_options_prev, &p_opts->qos_sw0_options);
	set_prev_qos_options(&p_opts->qos_swe_options_prev, &p_opts->qos_swe_options);
	set_prev_qos_options(&p_opts->qos_sw2sw_options_prev, &p_opts->qos_sw2sw_options);
	set_prev_qos_options(&p_opts->qos_rtr_options_prev, &p_opts->qos_rtr_options);
}

static void subn_verify_hm_action_set(IN char *hm_action_s,
				      IN osm_hm_action_type_enum *hm_action,
				      IN const char *action_opt,
				      IN boolean_t allow_no_discover)
{
	if (!strcmp(hm_action_s, "ignore"))
		*hm_action = OSM_HM_ACTION_IGNORE;
	else if (!strcmp(hm_action_s, "report"))
		*hm_action = OSM_HM_ACTION_REPORT;
	else if (!strcmp(hm_action_s, "isolate"))
		*hm_action = OSM_HM_ACTION_ISOLATE;
	else if (!strcmp(hm_action_s, "disable"))
		*hm_action = OSM_HM_ACTION_DISABLE;
	else if (allow_no_discover && !strcmp(hm_action_s, "no_discover"))
		*hm_action = OSM_HM_ACTION_NO_DISCOVER;
	else
		log_report(" Invalid Cached Option Value:%s = %s\n",
			   action_opt, hm_action_s);
}

static void subn_verify_hm_set(IN osm_subn_opt_t * p_opts)
{
	subn_verify_hm_action_set(p_opts->hm_ca_reboot_action_s,
				  &p_opts->hm_ca_reboot_action,
				  "hm_ca_reboot_action", FALSE);
	subn_verify_hm_action_set(p_opts->hm_ca_unresponsive_action_s,
				  &p_opts->hm_ca_unresponsive_action,
				  "hm_ca_unresponsive_action", FALSE);
	subn_verify_hm_action_set(p_opts->hm_ca_noisy_action_s,
				  &p_opts->hm_ca_noisy_action,
				  "hm_ca_noisy_action", FALSE);
	subn_verify_hm_action_set(p_opts->hm_ca_seterr_action_s,
				  &p_opts->hm_ca_seterr_action,
				  "hm_ca_seterr_action", FALSE);
	subn_verify_hm_action_set(p_opts->hm_ca_flapping_action_s,
				  &p_opts->hm_ca_flapping_action,
				  "hm_ca_flapping_action", FALSE);
	subn_verify_hm_action_set(p_opts->hm_ca_illegal_action_s,
				  &p_opts->hm_ca_illegal_action,
				  "hm_ca_illegal_action", FALSE);
	subn_verify_hm_action_set(p_opts->hm_ca_manual_action_s,
				  &p_opts->hm_ca_manual_action,
				  "hm_ca_manual_action", FALSE);
	subn_verify_hm_action_set(p_opts->hm_sw_reboot_action_s,
				  &p_opts->hm_sw_reboot_action,
				  "hm_sw_reboot_action", TRUE);
	subn_verify_hm_action_set(p_opts->hm_sw_unresponsive_action_s,
				  &p_opts->hm_sw_unresponsive_action,
				  "hm_sw_unresponsive_action", TRUE);
	subn_verify_hm_action_set(p_opts->hm_sw_noisy_action_s,
				  &p_opts->hm_sw_noisy_action,
				  "hm_sw_noisy_action,", TRUE);
	subn_verify_hm_action_set(p_opts->hm_sw_seterr_action_s,
				  &p_opts->hm_sw_seterr_action,
				  "hm_sw_seterr_action", TRUE);
	subn_verify_hm_action_set(p_opts->hm_sw_flapping_action_s,
				  &p_opts->hm_sw_flapping_action,
				  "hm_sw_flapping_action", TRUE);
	subn_verify_hm_action_set(p_opts->hm_sw_illegal_action_s,
				  &p_opts->hm_sw_illegal_action,
				  "hm_sw_illegal_action", TRUE);
	subn_verify_hm_action_set(p_opts->hm_sw_manual_action_s,
				  &p_opts->hm_sw_manual_action,
				  "hm_sw_manual_action", TRUE);

	if (p_opts->hm_num_set_err_sweeps_window > OSM_MAX_HM_WINDOW) {
		log_report(" Invalid Cached Option "
			   "Value:hm_num_set_err_sweeps_window = %u:"
			   "Using Default:%u\n", p_opts->hm_num_set_err_sweeps_window,
			   OSM_DEFAULT_HM_NUM_ERR_SWEEPS_WINDOW);
		p_opts->hm_num_set_err_sweeps_window =
		    OSM_DEFAULT_HM_NUM_ERR_SWEEPS_WINDOW;
	}

	if (p_opts->hm_num_set_err_sweeps >
	    p_opts->hm_num_set_err_sweeps_window) {
		log_report(" Invalid Cached Option "
			   "Value:hm_num_set_err_sweeps = %u:Using:%u\n",
			   p_opts->hm_num_set_err_sweeps,
			   p_opts->hm_num_set_err_sweeps_window);
		p_opts->hm_num_set_err_sweeps =
		    p_opts->hm_num_set_err_sweeps_window;
	}

	if (p_opts->hm_num_no_resp_sweeps_window > OSM_MAX_HM_WINDOW) {
		log_report(" Invalid Cached Option Value:"
			   "hm_num_no_resp_sweeps_window = %u:"
			   "Using Default:%u\n",
			   p_opts->hm_num_no_resp_sweeps_window,
			   OSM_DEFAULT_HM_NUM_NO_RESP_SWEEPS_WINDOW);
		p_opts->hm_num_no_resp_sweeps_window =
		    OSM_DEFAULT_HM_NUM_NO_RESP_SWEEPS_WINDOW;
	}

	if (p_opts->hm_num_no_resp_sweeps >
	    p_opts->hm_num_no_resp_sweeps_window) {
		log_report(" Invalid Cached Option Value:"
			   "hm_num_no_resp_sweeps = %u:Using:%u\n",
			   p_opts->hm_num_no_resp_sweeps,
			   p_opts->hm_num_no_resp_sweeps_window);
		p_opts->hm_num_no_resp_sweeps =
		    p_opts->hm_num_no_resp_sweeps_window;
	}

	if (p_opts->hm_num_flapping_sweeps_window > OSM_MAX_HM_WINDOW) {
		log_report(" Invalid Cached Option Value:"
			   "hm_num_flapping_sweeps_window = %u:"
			   "Using Default:%u\n",
			   p_opts->hm_num_flapping_sweeps_window,
			   OSM_DEFAULT_HM_NUM_FLAPPING_SWEEPS_WINDOW);

		p_opts->hm_num_flapping_sweeps_window =
		    OSM_DEFAULT_HM_NUM_FLAPPING_SWEEPS_WINDOW;
	}

	if (p_opts->hm_num_flapping_sweeps >
	    p_opts->hm_num_flapping_sweeps_window) {
		log_report(" Invalid Cached Option Value:"
			   "hm_num_flapping_sweeps = %u:Using:%u\n",
			   p_opts->hm_num_flapping_sweeps,
			   p_opts->hm_num_flapping_sweeps_window);
		p_opts->hm_num_flapping_sweeps =
		    p_opts->hm_num_flapping_sweeps_window;
	}
}

static
void subn_verify_port_reset_type_set(IN char * reset_type_s,
				     IN osm_port_reset_type_enum * reset_type,
				     IN const char * reset_type_opt)
{
	osm_port_reset_type_enum e;

	for (e = OSM_PORT_RESET_TYPE_NONE; e < OSM_PORT_RESET_TYPE_MAX; e++) {
		if (!strcmp(reset_type_s, osm_port_reset_type_str(e))) {
			*reset_type = e;
			break;
		}
	}

	if (e == OSM_PORT_RESET_TYPE_MAX) {
                log_report(" Invalid Cached Option Value:%s = %s\n",
                           reset_type_opt, reset_type_s);
		*reset_type = OSM_PORT_RESET_TYPE_NONE;
	}
}


static void subn_verify_port_reset_action_set(IN osm_subn_opt_t * p_opts)
{
	subn_verify_port_reset_type_set(p_opts->port_speed_change_action_s,
					&p_opts->port_speed_change_action,
					"port_speed_change_action");

	subn_verify_port_reset_type_set(p_opts->port_ext_speed_change_action_s,
					&p_opts->port_ext_speed_change_action,
					"port_ext_speed_change_action");

	subn_verify_port_reset_type_set(p_opts->port_ext_speed2_change_action_s,
					&p_opts->port_ext_speed2_change_action,
					"port_ext_speed2_change_action");

	subn_verify_port_reset_type_set(p_opts->port_mepi_speed_change_action_s,
					&p_opts->port_mepi_speed_change_action,
					"port_mepi_speed_change_action");

	subn_verify_port_reset_type_set(p_opts->port_mtu_change_action_s,
					&p_opts->port_mtu_change_action,
					"port_mtu_change_action");

	subn_verify_port_reset_type_set(p_opts->port_vl_change_action_s,
					&p_opts->port_vl_change_action,
					"port_vl_change_action");

	subn_verify_port_reset_type_set(p_opts->port_ame_bit_change_action_s,
					&p_opts->port_ame_bit_change_action,
					"port_ame_bit_change_action");

}

static void subn_verify_advanced_routing_set(IN osm_subn_opt_t * p_opts)
{
	if (p_opts->ar_mode >= OSM_ADV_ROUTING_MODE_MAX) {
		log_report(" Invalid Cached Option Value:"
			   "ar_mode = %u:"
			   "Using Default:%u\n",
			   p_opts->ar_mode,
			   OSM_ADV_ROUTING_MODE_DISABLED);
		p_opts->ar_mode = OSM_ADV_ROUTING_MODE_DISABLED;
	}

	if (p_opts->shield_mode >= OSM_ADV_ROUTING_MODE_MAX) {
		log_report(" Invalid Cached Option Value:"
			   "shield_mode = %u:"
			   "Using Default:%u\n",
			   p_opts->shield_mode,
			   OSM_ADV_ROUTING_MODE_DISABLED);
		p_opts->shield_mode = OSM_ADV_ROUTING_MODE_DISABLED;
	}
}

static void subn_verify_max_op_vls_set(IN osm_subn_opt_t * p_opts)
{
	/* the maximal valid value in IB spec for op vls */
	uint8_t max_valid_op_vls = 5;

	if (p_opts->max_op_vls > max_valid_op_vls) {
		log_report(" Invalid Cached Option: max_op_vls=%u: "
			   "Using Default = %u\n",
			   p_opts->max_op_vls, OSM_DEFAULT_MAX_OP_VLS);
		p_opts->max_op_vls = OSM_DEFAULT_MAX_OP_VLS;
	}
	if (p_opts->max_op_vls_sw > max_valid_op_vls) {
		log_report(" Invalid Cached Option: max_op_vls_sw=%u: "
			   "Using Default = %u\n",
			   p_opts->max_op_vls_sw, 0);
		p_opts->max_op_vls_sw = 0;
	}
	if (p_opts->max_op_vls_ca > max_valid_op_vls) {
		log_report(" Invalid Cached Option: max_op_vls_ca=%u: "
			   "Using Default = %u\n",
			   p_opts->max_op_vls_ca, 0);
		p_opts->max_op_vls_ca = 0;
	}
	if (p_opts->max_op_vls_rtr > max_valid_op_vls) {
		log_report(" Invalid Cached Option: max_op_vls_rtr=%u: "
			   "Using Default = %u\n",
			   p_opts->max_op_vls_rtr, 0);
		p_opts->max_op_vls_rtr = 0;
	}
}

int osm_subn_verify_config(IN osm_subn_opt_t * p_opts)
{
	/* If first guid is defined, count how many guids parsed */
	if (p_opts->guid[0]) {
		int i = 1;
		p_opts->num_guids = 1;

		for (; i < OSM_MAX_BINDING_PORTS; i++) {
			if (!p_opts->guid[i])
				break;

			p_opts->num_guids++;
		}
	}

	if (p_opts->lmc > 7) {
		log_report(" Invalid Cached Option Value:lmc = %u:"
			   "Using Default:%u\n", p_opts->lmc, OSM_DEFAULT_LMC);
		p_opts->lmc = OSM_DEFAULT_LMC;
	}

	/* check range */
	if (p_opts->lids_per_rtr > 128) {
		log_report(" Invalid Cached Option Value:lids_per_rtr = %u:"
			   "Using Default:%u\n", p_opts->lids_per_rtr, 0);
		p_opts->lids_per_rtr = 0;
	}

	/* if global LMC is set, limit lids_per_rtr according to global LMC */
	if (p_opts->lmc && p_opts->lids_per_rtr > (1 << p_opts->lmc)) {
		log_report(" Non-zero lids_per_rtr = %u does not match non-zero lmc = %u:"
			   " Non-zero lmc will be used for lids_per_rtr\n",
				p_opts->lids_per_rtr, p_opts->lmc );
		p_opts->lids_per_rtr = (1 << p_opts->lmc);
	}

	/* calculate LMC for routers (rtr_lmc) */
	if (!p_opts->lmc && p_opts->lids_per_rtr) {
		p_opts->rtr_lmc = 0;
		while ((1 << p_opts->rtr_lmc) < p_opts->lids_per_rtr)
			p_opts->rtr_lmc++;
		log_report(" Setting routers LMC to %u to support lids_per_rtr = %u\n",
			   p_opts->rtr_lmc, p_opts->lids_per_rtr);
	} else {
		p_opts->rtr_lmc = p_opts->lmc;
	}

	/* Check prism_an_lmc range */
	if (p_opts->prism_an_lmc > 7) {
		log_report(" Invalid Cached Option Value: prism_an_lmc = %u. Using Default: %u\n",
			   p_opts->prism_an_lmc, OSM_DEFAULT_PRISM_AN_LMC);
		p_opts->prism_an_lmc = OSM_DEFAULT_PRISM_AN_LMC;
	}

	/* If another LMC is set, override prism_an_lmc */
	if (p_opts->lmc || p_opts->rtr_lmc) {
		p_opts->prism_an_lmc = p_opts->lmc;
		log_report(" Overriding prism_an_lmc, since another LMC is set."
			   " Global LMC = %u, rtr_lmc = %u\n",
			   p_opts->lmc, p_opts->rtr_lmc);
	}

	/* Force using persistent SM LID */
	if (p_opts->enable_sm_port_failover && !p_opts->master_sm_lid) {
		p_opts->master_sm_lid = 1 << p_opts->lmc;
		/* Use the first valid LID */
		log_report(" Setting master_sm_lid to %u to support sm_port_failover\n",
			   p_opts->master_sm_lid);
	}

	if (p_opts->master_sm_lid) {
		uint16_t lmc_mask = ~((1 << p_opts->lmc) - 1);
		/*
		 * if requested master_sm_lid does not match subnet LMC, use
		 * the first valid LID instead.
		 */
		if ((p_opts->master_sm_lid & lmc_mask) != p_opts->master_sm_lid) {
			uint16_t master_sm_lid = 1 << p_opts->lmc;
			log_report(" master_sm_lid %u does not match lmc = %u:"
				   " Setting master_sm_lid to %u\n",
				   p_opts->master_sm_lid,
				   p_opts->lmc,
				   master_sm_lid);
			p_opts->master_sm_lid = master_sm_lid;
		}
	}

	if (15 < p_opts->sm_priority) {
		log_report(" Invalid Cached Option Value:sm_priority = %u:"
			   "Using Default:%u\n",
			   p_opts->sm_priority, OSM_DEFAULT_SM_PRIORITY);
		p_opts->sm_priority = OSM_DEFAULT_SM_PRIORITY;
	}

	if (15 < p_opts->master_sm_priority) {
		log_report(" Invalid Cached Option Value:master_sm_priority = %u:"
			   "Using Default:%u\n",
			   p_opts->master_sm_priority,
			   OSM_DEFAULT_SM_MASTER_PRIORITY);
		p_opts->master_sm_priority = OSM_DEFAULT_SM_MASTER_PRIORITY;
	}

	if (p_opts->master_sm_lid > IB_LID_UCAST_END_HO) {
		log_report(" Invalid Cached Option Value:master_sm_lid = %u:"
			   "Using Default:%u\n",
			   p_opts->master_sm_lid,
			   0);
		p_opts->master_sm_priority = 0;
	}

	if ((IB_LINK_SPEED_SET_LSS < p_opts->force_link_speed) ||
	    (p_opts->force_link_speed > IB_LINK_SPEED_2_5_5_OR_10 &&
	     p_opts->force_link_speed < IB_LINK_SPEED_SET_LSS)) {
		log_report(" Invalid Cached Option Value:force_link_speed = %u:"
			   "Using Default:%u\n", p_opts->force_link_speed,
			   IB_LINK_SPEED_SET_LSS);
		p_opts->force_link_speed = IB_LINK_SPEED_SET_LSS;
	}

	if ((IB_LINK_SPEED_EXT_SET_LSES < p_opts->force_link_speed_ext) ||
	    (p_opts->force_link_speed_ext > IB_LINK_SPEED_EXT_MAX_VALUE &&
	     p_opts->force_link_speed_ext < IB_LINK_SPEED_EXT_DISABLE)) {
		log_report(" Invalid Cached Option Value:force_link_speed_ext = %u:"
			   "Using Default:%u\n", p_opts->force_link_speed_ext,
			   IB_LINK_SPEED_EXT_SET_LSES);
		p_opts->force_link_speed_ext = IB_LINK_SPEED_EXT_SET_LSES;
	}

	if ((IB_LINK_SPEED_EXT2_SET_LSES2 < p_opts->force_link_speed_ext2) ||
	    (p_opts->force_link_speed_ext2 > IB_LINK_SPEED_EXT2_MAX_VALUE &&
	     p_opts->force_link_speed_ext2 < IB_LINK_SPEED_EXT2_SET_LSES2)) {
		log_report(" Invalid Cached Option Value:force_link_speed_ext2 = %u:"
			   "Using Default:%u\n", p_opts->force_link_speed_ext2,
			   IB_LINK_SPEED_EXT2_SET_LSES2);
		p_opts->force_link_speed_ext2 = IB_LINK_SPEED_EXT2_SET_LSES2;
	}

	if ((IB_LINK_WIDTH_SET_LWS < p_opts->force_link_width) ||
	    (p_opts->force_link_width > IB_LINK_WIDTH_1X_2X_4X_8X_OR_12X &&
	     p_opts->force_link_width < IB_LINK_WIDTH_SET_LWS)) {
		log_report(" Invalid Cached Option Value:force_link_width = %u:"
			   "Using Default:%u\n", p_opts->force_link_width,
			   IB_LINK_WIDTH_SET_LWS);
		p_opts->force_link_width = IB_LINK_WIDTH_SET_LWS;
	}

	/* Currently the fdr10 is the only supported mlnx extended speed */
	if (p_opts->mepi_enabled_speeds &&
	    p_opts->mepi_enabled_speeds != OSM_DEFAULT_MEPI_ENABLED_SPEED) {
		log_report(" Invalid Cached Option Value:mepi_enabled_speeds = %u:"
			   "Using Default:%u\n", p_opts->mepi_enabled_speeds,
			   OSM_DEFAULT_MEPI_ENABLED_SPEED);
		p_opts->mepi_enabled_speeds = OSM_DEFAULT_MEPI_ENABLED_SPEED;
	}

	if (2 < p_opts->fdr10 && p_opts->fdr10 != OSM_DEFAULT_FDR10) {
		log_report(" Invalid Cached Option Value:fdr10 = %u:"
			   "Using Default:%u\n", p_opts->fdr10, 1);
		p_opts->fdr10 = 1;
	} else {
		/*
		 * this parameter is deprecated but still if used need to
		 * effect new parameters which replaced it
		 */
		if (p_opts->fdr10 != OSM_DEFAULT_FDR10) {
			if (p_opts->fdr10) {
				p_opts->support_mepi_speeds = TRUE;
				if (p_opts->fdr10 == 1)
					/* enable fdr10 */
					p_opts->mepi_enabled_speeds |= FDR10_BIT;
				else
					/* disable fdr10 */
					p_opts->mepi_enabled_speeds &= ~FDR10_BIT;
			} else
				/* do not use mepi MAD */
				p_opts->support_mepi_speeds = FALSE;
		}
	}

	if (p_opts->max_wire_smps == 0)
		p_opts->max_wire_smps = 0x7FFFFFFF;
	else if (p_opts->max_wire_smps > 0x7FFFFFFF) {
		log_report(" Invalid Cached Option Value: max_wire_smps = %u,"
			   " Using Default: %u\n",
			   p_opts->max_wire_smps, OSM_DEFAULT_SMP_MAX_ON_WIRE);
		p_opts->max_wire_smps = OSM_DEFAULT_SMP_MAX_ON_WIRE;
	}

	if (p_opts->max_wire_smps2 > 0x7FFFFFFF) {
		log_report(" Invalid Cached Option Value: max_wire_smps2 = %u,"
			   " Using Default: %u",
			   p_opts->max_wire_smps2, p_opts->max_wire_smps);
		p_opts->max_wire_smps2 = p_opts->max_wire_smps;
	}

	if (p_opts->long_transaction_timeout < p_opts->transaction_timeout) {
		log_report(" Invalid Cached Option Value: long_transaction_timeout = %u,"
			   " Using transaction_timeout: %u",
			   p_opts->long_transaction_timeout, p_opts->transaction_timeout);
		p_opts->long_transaction_timeout = p_opts->transaction_timeout;
	}

	if (p_opts->max_sa_reports_queued > 0x7FFFFFFF) {
		log_report(" Invalid Cached Option Value: max_sa_reports_queued = %u,"
			   " Using Default: %u\n",
			   p_opts->max_sa_reports_queued, 0);
		p_opts->max_sa_reports_queued = OSM_DEFAULT_MAX_SA_REPORTS;
	}

	if (p_opts->max_sa_reports_on_wire > 0x7FFFFFFF) {
		log_report(" Invalid Cached Option Value: max_sa_reports_on_wire = %u,"
			   " Using Default: %u\n",
			   p_opts->max_sa_reports_on_wire, 0);
		p_opts->max_sa_reports_on_wire = OSM_DEFAULT_MAX_SA_REPORTS_ON_WIRE;
	}

	if (strcmp(p_opts->console, OSM_DISABLE_CONSOLE)
	    && strcmp(p_opts->console, OSM_LOCAL_CONSOLE)
#ifdef ENABLE_OSM_CONSOLE_LOOPBACK
	    && strcmp(p_opts->console, OSM_LOOPBACK_CONSOLE)
#endif
#ifdef ENABLE_OSM_CONSOLE_SOCKET
	    && strcmp(p_opts->console, OSM_REMOTE_CONSOLE)
#endif
	    ) {
		log_report(" Invalid Cached Option Value:console = %s"
			   ", Using Default:%s\n",
			   p_opts->console, OSM_DEFAULT_CONSOLE);
		free(p_opts->console);
		p_opts->console = strdup(OSM_DEFAULT_CONSOLE);
	}

	if (!strcmp(p_opts->console, OSM_LOCAL_CONSOLE) && p_opts->daemon) {
		log_report(" Local console is invalid when running as daemon."
			   " Disabling local console.\n");
		free(p_opts->console);
		p_opts->console = strdup(OSM_DISABLE_CONSOLE);
	}

	if (strcmp(p_opts->sm_assign_guid_func, OSM_UNIQ_CNT_FUNC) &&
	    strcmp(p_opts->sm_assign_guid_func, OSM_PHYS_BAS_FUNC)) {
		log_report(" Invalid Cached Option Value:sm_assign_guid_func = %s"
			   ", Using Default:%s\n",
			   p_opts->sm_assign_guid_func, OSM_UNIQ_CNT_FUNC);
		p_opts->sm_assign_guid_func = strdup(OSM_UNIQ_CNT_FUNC);
	}

	if (p_opts->no_partition_enforcement == TRUE) {
		strcpy(p_opts->part_enforce, OSM_PARTITION_ENFORCE_OFF);
		p_opts->part_enforce_enum = OSM_PARTITION_ENFORCE_TYPE_OFF;
	} else {
		if (strcmp(p_opts->part_enforce, OSM_PARTITION_ENFORCE_BOTH) == 0)
			p_opts->part_enforce_enum = OSM_PARTITION_ENFORCE_TYPE_BOTH;
		else if (strcmp(p_opts->part_enforce, OSM_PARTITION_ENFORCE_IN) == 0)
			p_opts->part_enforce_enum = OSM_PARTITION_ENFORCE_TYPE_IN;
		else if (strcmp(p_opts->part_enforce, OSM_PARTITION_ENFORCE_OUT) == 0)
			p_opts->part_enforce_enum = OSM_PARTITION_ENFORCE_TYPE_OUT;
		else if (strcmp(p_opts->part_enforce, OSM_PARTITION_ENFORCE_OFF) == 0)
			p_opts->part_enforce_enum = OSM_PARTITION_ENFORCE_TYPE_OFF;
		else {
			log_report(" Invalid Cached Option Value:part_enforce = %s"
	                           ", Using Default:%s\n",
	                           p_opts->part_enforce, OSM_PARTITION_ENFORCE_BOTH);
			strcpy(p_opts->part_enforce, OSM_PARTITION_ENFORCE_BOTH);
			p_opts->part_enforce_enum = OSM_PARTITION_ENFORCE_TYPE_BOTH;
		}
	}

	if (p_opts->hm_unhealthy_ports_checks)
		subn_verify_hm_set(p_opts);

	if (p_opts->qos) {
		subn_verify_qos_set(&p_opts->qos_options, "qos");
		subn_verify_qos_set(&p_opts->qos_ca_options, "qos_ca");
		subn_verify_qos_set(&p_opts->qos_sw0_options, "qos_sw0");
		subn_verify_qos_set(&p_opts->qos_swe_options, "qos_swe");
		subn_verify_qos_set(&p_opts->qos_sw2sw_options, "qos_sw2sw");
		subn_verify_qos_set(&p_opts->qos_rtr_options, "qos_rtr");
	}

	subn_verify_max_op_vls_set(p_opts);

	if (p_opts->default_mcg_mtu) {
		if (!ib_mtu_is_valid(p_opts->default_mcg_mtu)) {
			log_report(" Invalid Cached Option Value:default_mcg_mtu = %u"
				   ", Using Default:%u\n",
				   p_opts->default_mcg_mtu, 0);
			p_opts->default_mcg_mtu = 0;
		}
	}

	if (p_opts->default_mcg_rate) {
		if (!ib_rate_is_valid(p_opts->default_mcg_rate)) {
			log_report(" Invalid Cached Option Value:default_mcg_rate = %u"
				   ", Using Default:%u\n",
				   p_opts->default_mcg_rate, 0);
			p_opts->default_mcg_rate = 0;
		}
	}

#ifdef ENABLE_OSM_PERF_MGR
	if (p_opts->perfmgr_sweep_time_s < 1) {
		log_report(" Invalid Cached Option Value:perfmgr_sweep_time_s "
			   "= %u Using Default:%u\n",
			   p_opts->perfmgr_sweep_time_s,
			   OSM_PERFMGR_DEFAULT_SWEEP_TIME_S);
		p_opts->perfmgr_sweep_time_s = OSM_PERFMGR_DEFAULT_SWEEP_TIME_S;
	}
	if (p_opts->perfmgr_max_outstanding_queries < 1) {
		log_report(" Invalid Cached Option Value:"
			   "perfmgr_max_outstanding_queries = %u"
			   " Using Default:%u\n",
			   p_opts->perfmgr_max_outstanding_queries,
			   OSM_PERFMGR_DEFAULT_MAX_OUTSTANDING_QUERIES);
		p_opts->perfmgr_max_outstanding_queries =
		    OSM_PERFMGR_DEFAULT_MAX_OUTSTANDING_QUERIES;
	}
#endif

	if (p_opts->m_key_protect_bits > 3) {
		log_report(" Invalid Cached Option Value:"
			   "m_key_protection_level = %u Setting to %u "
			   "instead\n", p_opts->m_key_protect_bits, 2);
		p_opts->m_key_protect_bits = 2;
	}
	if (p_opts->m_key_protect_bits && p_opts->m_key_lease_period) {
		if (!p_opts->sweep_interval) {
			log_report(" Sweep disabled with protected mkey "
				   "leases in effect; re-enabling sweeping "
				   "with interval %u\n",
				   cl_ntoh16(p_opts->m_key_lease_period) - 1);
			p_opts->sweep_interval =
				cl_ntoh16(p_opts->m_key_lease_period) - 1;
		}
		if (p_opts->sweep_interval >=
			cl_ntoh16(p_opts->m_key_lease_period)) {
			log_report(" Sweep interval %u >= mkey lease period "
				   "%u. Setting lease period to %u\n",
				   p_opts->sweep_interval,
				   cl_ntoh16(p_opts->m_key_lease_period),
				   p_opts->sweep_interval + 1);
			p_opts->m_key_lease_period =
				cl_hton16(p_opts->sweep_interval + 1);
		}
	}

	if (p_opts->m_key_per_port && p_opts->m_key_protect_bits < 2) {
		log_report(" M_Key per port enabled, setting m_key_protection_level "
			   "to 2\n");
		p_opts->m_key_protect_bits = 2;
	}

	if (p_opts->m_key_per_port && p_opts->m_key_lookup == FALSE) {
		log_report(" M_Key per port enabled, setting m_key_lookup "
			   "to TRUE\n");
		p_opts->m_key_lookup = TRUE;
	}

	if (p_opts->m_key_per_port && p_opts->m_key == 0) {
		log_report(" M_Key per port enabled, setting m_key "
			   "to non-zero\n");
		p_opts->m_key = 1;
	}

	if (p_opts->no_clients_rereg) {
		log_report(" no_clients_rereg is enabled, setting "
			   "client_rereg_mode to 0\n");
		p_opts->client_rereg_mode = OSM_CLIENT_REREG_MODE_DISABLE;
	}

	if (p_opts->client_rereg_mode > OSM_CLIENT_REREG_MODE_LINK_MGR) {
		log_report(" Invalid Cached Option Value:"
			   "client_rereg_mode = %u"
			   " Using Default:%u\n",
			   p_opts->client_rereg_mode,
			   OSM_DEFAULT_CLIENT_REREG_MODE);
		p_opts->client_rereg_mode = OSM_DEFAULT_CLIENT_REREG_MODE;
	}

	if (p_opts->pgrp_policy_file != NULL) {
		FILE *policy_file = fopen(p_opts->pgrp_policy_file, "r");
		if (!policy_file)
			log_report("Port Groups policy file provided: "
				   "%s doesn't exist.\n", p_opts->pgrp_policy_file);
		else
			fclose(policy_file);
	}

	if (p_opts->topo_policy_file != NULL) {
		FILE *policy_file = fopen(p_opts->topo_policy_file, "r");
		if (!policy_file)
			log_report("Topology policy file provided: "
				   "%s doesn't exist.\n", p_opts->topo_policy_file);
		else
			fclose(policy_file);
	}

	if (p_opts->rch_policy_file != NULL) {
		FILE *policy_file = fopen(p_opts->rch_policy_file, "r");
		if (!policy_file)
			log_report("Routing chains policy file provided: "
				   "%s doesn't exist.\n", p_opts->rch_policy_file);
		else
			fclose(policy_file);
	}

	if (p_opts->device_configuration_file != NULL) {
		FILE *policy_file = fopen(p_opts->device_configuration_file, "r");
		if (!policy_file)
			log_report("Device configuration file provided: "
				   "%s doesn't exist.\n", p_opts->device_configuration_file);
		else
			fclose(policy_file);
	}

	if (p_opts->topo_config_enabled) {
		if (p_opts->topo_config_file != NULL) {
			FILE *topo_config_file = fopen(p_opts->topo_config_file, "r");
			if (!topo_config_file)
				log_report("Topo configuration file provided: "
					   "%s doesn't exist.\n", p_opts->topo_config_file);
			else
				fclose(topo_config_file);
		}
	}

	if (p_opts->mlnx_congestion_control > OSM_CAPABILITY_ENABLED) {
		log_report(" Invalid Cached Option Value: mlnx_congestion_control = %u,"
			   " Using Default: %u\n",
			   p_opts->mlnx_congestion_control, OSM_MLNX_CC_CAP_DEFAULT);
		p_opts->mlnx_congestion_control = OSM_MLNX_CC_CAP_DEFAULT;
	}

	if ((p_opts->congestion_control_policy_file != NULL) &&
	    (p_opts->mlnx_congestion_control == OSM_CAPABILITY_ENABLED)) {
		FILE *policy_file = fopen(p_opts->congestion_control_policy_file, "r");
		if (!policy_file)
			log_report("Congestion control policy file provided: "
				   "%s doesn't exist.\n", p_opts->congestion_control_policy_file);
		else
			fclose(policy_file);
	}

	if (p_opts->fast_recovery_enabled > OSM_CAPABILITY_ENABLED) {
		log_report(" Invalid Cached Option Value: fast_recovery_enabled = %u,"
			   " Using Default: %u\n",
			   p_opts->fast_recovery_enabled, OSM_FAST_RECOVERY_CAP_DEFAULT);
		p_opts->fast_recovery_enabled = OSM_FAST_RECOVERY_CAP_DEFAULT;
	}

	if (p_opts->fast_recovery_conf_file != NULL &&
	    p_opts->fast_recovery_enabled == OSM_CAPABILITY_ENABLED) {
		FILE *policy_file = fopen(p_opts->fast_recovery_conf_file, "r");
		if (!policy_file)
			log_report("Fast recovery configuration file provided: "
				   "%s doesn't exist.\n", p_opts->fast_recovery_conf_file);
		else
			fclose(policy_file);
	}

	if (p_opts->ca_port_soft_limit > p_opts->ca_port_num_limit) {
		log_report(" Invalid Cached Option Value:ca_port_soft_limit "
			   "= %u Using Default:0\n",
			   p_opts->ca_port_soft_limit);
		p_opts->ca_port_soft_limit = 0;
	}

	if (p_opts->virt_enabled > OSM_CAPABILITY_ENABLED) {
		log_report(" Invalid Cached Option Value: virt_enabled = %u,"
			   " Using Default: %u\n",
			   p_opts->virt_enabled, OSM_CAPABILITY_IGNORED);
		p_opts->virt_enabled = OSM_CAPABILITY_IGNORED;
	}

	if (p_opts->override_create_mcg_sl > 15 &&
	    p_opts->override_create_mcg_sl != OSM_MCMR_USER_DEFINED_SL) {
		log_report(" Invalid Cached Option Value:"
			   " override_create_mcg_sl = %u,"
			   " Using Default: 0x%x\n",
			   p_opts->override_create_mcg_sl,
			   OSM_MCMR_USER_DEFINED_SL);
		p_opts->override_create_mcg_sl = OSM_MCMR_USER_DEFINED_SL;
	}

	if (p_opts->osm_stats_dump_limit > OSM_MAX_STATISTIC_DUMP_FILE_SIZE){
		log_report(" Statistic dump file max size exceeding max value,"
			   " setting osm_stats_dump_limit to %u\n",
			   OSM_MAX_STATISTIC_DUMP_FILE_SIZE);
		p_opts->osm_stats_dump_limit = OSM_MAX_STATISTIC_DUMP_FILE_SIZE;
	}

	if (p_opts->osm_perflog_dump_limit > OSM_MAX_PERFLOG_DUMP_FILE_SIZE) {
		log_report(" Performance log dump file max size exceeding max value,"
			   " setting osm_perflog_dump_limit to %u\n",
			   OSM_MAX_PERFLOG_DUMP_FILE_SIZE);
		p_opts->osm_perflog_dump_limit = OSM_MAX_PERFLOG_DUMP_FILE_SIZE;
	}

	if (p_opts->sharp_enabled > OSM_CAPABILITY_ENABLED){
		log_report(" SHArP capability type exceeding max value,"
			   " setting SHArP capability to default %u\n",
			   OSM_SHARP_CAP_DEFAULT );
		p_opts->sharp_enabled = OSM_SHARP_CAP_DEFAULT;
	}
	/* if 10 most significant bits are site local format, check validity */
	if ((p_opts->subnet_prefix & IB_10_MSB_MASK) ==
		  IB_SITE_LOCAL_SUBNET_PREFIX) {
		/* check that it's format 0xfec0...<16 free bits> */
		if ((p_opts->subnet_prefix & IB_LOCAL_SUBNET_MASK) !=
		     IB_SITE_LOCAL_SUBNET_PREFIX) {
			log_report(" Illegal subnet prefix 0x%016" PRIx64
				   " has 10 most significant bits in"
				   " site local format.\n But it's not strict"
				   " 0xfec000000000<16-bit Subnet Prefix>"
				   " format.\n Setting subnet_prefix to"
				   " 0x%016" PRIx64 "\n",
				   cl_ntoh64(p_opts->subnet_prefix),
				   IB_DEFAULT_SUBNET_PREFIX_HO,
				   IB_DEFAULT_SUBNET_PREFIX_HO);
			p_opts->subnet_prefix = IB_DEFAULT_SUBNET_PREFIX;
		}
	}
	/* if 10 most significant bits are
	 * not link local neither site local formats, it's global! */
	else if ((p_opts->subnet_prefix & IB_10_MSB_MASK) !=
	             IB_DEFAULT_SUBNET_PREFIX) {
			log_report(" Subnet prefix 0x%016" PRIx64
				   " is in the global scope,"
				   " which is not supported."
				   " Setting subnet_prefix"
				   " to 0x%016" PRIx64 "\n",
				   cl_ntoh64(p_opts->subnet_prefix),
				   IB_DEFAULT_SUBNET_PREFIX_HO);
			p_opts->subnet_prefix = IB_DEFAULT_SUBNET_PREFIX;
	}

	if (p_opts->rtr_pr_rate != OSM_RTR_PR_RATE_USE_RATE_ON_PATH &&
	    (p_opts->rtr_pr_rate < IB_MIN_RATE || p_opts->rtr_pr_rate > IB_MAX_RATE)) {
		log_report(" Illegal router path record rate rtr_pr_rate (%u)"
			   ", is out of valid rate range (min %u, max %u),"
			   " setting rtr_pr_rate to default (%u)\n",
			   p_opts->rtr_pr_rate,
			   IB_MIN_RATE, IB_MAX_RATE, OSM_DEFAULT_RTR_PR_RATE);
		p_opts->rtr_pr_rate = OSM_DEFAULT_RTR_PR_RATE;
	}

	if (p_opts->rtr_pr_mtu != OSM_RTR_PR_MTU_USE_MTU_ON_PATH &&
	    (p_opts->rtr_pr_mtu < IB_MIN_MTU || p_opts->rtr_pr_mtu > IB_MAX_MTU)) {
		log_report(" Illegal router path record mtu rtr_pr_mtu (%u)"
			   ", is out of valid mtu range (min %u, max %u),"
			   " setting rtr_pr_mtu to default (%u)\n",
			   p_opts->rtr_pr_mtu,
			   IB_MIN_MTU, IB_MAX_MTU, OSM_DEFAULT_RTR_PR_MTU);
		p_opts->rtr_pr_mtu = OSM_DEFAULT_RTR_PR_MTU;
	}

	if (p_opts->router_policy_file != NULL) {
		FILE *policy_file = fopen(p_opts->router_policy_file, "r");
		if (!policy_file) {
			log_report("Router policy file provided: "
				   "%s doesn't exist.\n", p_opts->router_policy_file);
			free(p_opts->router_policy_file);
			p_opts->router_policy_file = NULL;
		} else
			fclose(policy_file);
	}

	if (p_opts->router_policy_enabled) {
		log_report("Router policy enabled\n");

		/* Router FLIDs and NVLink LIDs are not supported to coexist */
		if (p_opts->nvlink_enable) {
			log_report("NVLink configuration is not supported while "
				   "Router policy is enabled, "
				   "setting nvlink_enable to FALSE\n");
			p_opts->nvlink_enable = FALSE;
		}
	}

	if (p_opts->verbose_bypass_policy_file != NULL) {
		FILE *policy_file = fopen(p_opts->verbose_bypass_policy_file, "r");
		if (!policy_file)
			log_report("Verbose Bypass policy file provided: "
				   "%s doesn't exist.\n", p_opts->verbose_bypass_policy_file);
		else
			fclose(policy_file);
	}
       if (p_opts->stand_alone < OSM_STANDALONE_SA ||
            p_opts->stand_alone > OSM_STANDALONE_BOTH) {
                log_report(" Illegal SM mode (%u)"
                           ", is out of valid SM mode (min %u, max %u),"
                           " setting stand_alone to default (%u)\n",
                           p_opts->stand_alone,
                           OSM_STANDALONE_SA, OSM_STANDALONE_BOTH, OSM_STANDALONE_NA);
                p_opts->stand_alone = OSM_STANDALONE_NA;
        }

	subn_verify_port_reset_action_set(p_opts);
	subn_verify_advanced_routing_set(p_opts);

	if (p_opts->enhanced_qos_policy_file != NULL) {
		FILE *policy_file = fopen(p_opts->enhanced_qos_policy_file, "r");
		if (!policy_file)
			log_report("Enhanced Qos policy file provided: "
				   "%s doesn't exist.\n", p_opts->enhanced_qos_policy_file);
		else
			fclose(policy_file);
	}

	if (p_opts->sa_enhanced_trust_model)
		p_opts->drop_event_subscriptions = TRUE;

	if(p_opts->cc_key_enable > OSM_CAPABILITY_ENABLED) {
		log_report(" Invalid Cached Option Value: "
			   "cc_key_enable = %u Setting to %u "
			   "instead\n", p_opts->cc_key_enable,
			   OSM_KEY_DEFAULT_ENABLEMENT);
		p_opts->cc_key_enable = OSM_KEY_DEFAULT_ENABLEMENT;
	}

	if (p_opts->cc_key_protect_bit > 1) {
		log_report(" Invalid Cached Option Value: "
			   "cc_key_protect_bit = %u Setting to %u "
			   "instead\n", p_opts->cc_key_protect_bit, 1);
		p_opts->cc_key_protect_bit = 1;
	}

	if(p_opts->vs_key_enable > OSM_CAPABILITY_ENABLED) {
		log_report(" Invalid Cached Option Value: "
			   "vs_key_enable = %u Setting to %u "
			   "instead\n", p_opts->vs_key_enable,
			   OSM_KEY_DEFAULT_ENABLEMENT);
		p_opts->vs_key_enable = OSM_KEY_DEFAULT_ENABLEMENT;
	}

	if (p_opts->vs_key_ci_protect_bits > 1) {
		log_report(" Invalid Cached Option Value: "
			   "vs_key_ci_protect_bits = %u Setting to %u "
			   "instead\n", p_opts->vs_key_ci_protect_bits,
			   OSM_VS_KEY_DEFAULT_CI_PROTECT);
		p_opts->vs_key_ci_protect_bits = OSM_VS_KEY_DEFAULT_CI_PROTECT;
	}

	if (p_opts->n2n_key_enable > OSM_CAPABILITY_ENABLED) {
		log_report(" Invalid Cached Option Value: "
			   "n2n_key_enable = %u Setting to %u "
			   "instead\n", p_opts->n2n_key_enable,
			   OSM_KEY_DEFAULT_ENABLEMENT);
		p_opts->n2n_key_enable = OSM_KEY_DEFAULT_ENABLEMENT;
	}

	if (p_opts->n2n_key_protect_bit > 1) {
		log_report(" Invalid Cached Option Value: "
			   "n2n_key_protect_bit = %u Setting to %u "
			   "instead\n", p_opts->n2n_key_protect_bit, 1);
		p_opts->n2n_key_protect_bit = 1;
	}

	if (p_opts->gmp_traps_threads_num == 0) {
		log_report(" Invalid Cached Option Value: "
			   "gmp_traps_threads_num = %u Setting to %u instead\n",
			   p_opts->gmp_traps_threads_num,
			   OSM_DEFAULT_GMP_TRAPS_THREADS_NUM);
		p_opts->gmp_traps_threads_num = OSM_DEFAULT_GMP_TRAPS_THREADS_NUM;
	}

	if (p_opts->use_original_extended_sa_rates_only &&
	    p_opts->max_rate_enum > IB_PATH_RECORD_RATE_300_GBS) {
		if (p_opts->max_rate_enum != SA_RATE_MAX_ENUM)
			log_report(" Warning:"
				   " use_original_extended_sa_rates_only"
				   " defines a rate limit lower"
				   " than max_rate_enum.\n\t "
				   " Setting max_rate_enum to %d\n",
				   IB_PATH_RECORD_RATE_300_GBS);

		p_opts->max_rate_enum = IB_PATH_RECORD_RATE_300_GBS;
	}

	if (p_opts->max_rate_enum < IB_MIN_RATE ||
	   (p_opts->max_rate_enum > IB_MAX_RATE &&
	    p_opts->max_rate_enum != SA_RATE_MAX_ENUM)) {
		log_report("Illegal max_rate_enum %u, setting to default "
			   "value (%u), support all rates\n",
			   p_opts->max_rate_enum, SA_RATE_MAX_ENUM);
		p_opts->max_rate_enum = SA_RATE_MAX_ENUM;
	}

	if (p_opts->pfrn_sl > IB_PFRN_MAX_SL) {
		log_report("Illegal pfrn_sl %u, setting to default "
			   "value (%u)\n",
			   p_opts->pfrn_sl, OSM_DEFAULT_PFRN_SL);
		p_opts->pfrn_sl = OSM_DEFAULT_PFRN_SL;
	}

	if (p_opts->pfrn_mask_clear_timeout > IB_PFRN_MAX_MASK_CLEAR) {
		log_report("Illegal pfrn_mask_clear_timeout %u, "
			   "setting to default value (%u)\n",
			   p_opts->pfrn_mask_clear_timeout,
			   OSM_DEFAULT_PFRN_MASK_CLEAR);
		p_opts->pfrn_mask_clear_timeout = OSM_DEFAULT_PFRN_MASK_CLEAR;
	}

	if (p_opts->pfrn_mask_force_clear_timeout > IB_PFRN_MAX_MASK_FORCE_CLEAR) {
		log_report("Illegal pfrn_mask_force_clear_timeout %u, "
			   "setting to default value (%u)\n",
			   p_opts->pfrn_mask_force_clear_timeout,
			   OSM_DEFAULT_PFRN_MASK_FORCE_CLEAR);
		p_opts->pfrn_mask_force_clear_timeout =
		    OSM_DEFAULT_PFRN_MASK_FORCE_CLEAR;
	}

	if (p_opts->ar_tree_asymmetric_flow >= OSM_ASYMMETRIC_TREE_MAX) {
		log_report("Illegal ar_tree_asymmetric_flow %u, setting to default value %u\n",
			p_opts->ar_tree_asymmetric_flow,
			OSM_ASYMMETRIC_TREE_1_SUB_GROUP_ENABLED);
		p_opts->ar_tree_asymmetric_flow = OSM_ASYMMETRIC_TREE_1_SUB_GROUP_ENABLED;
	}

	if (p_opts->ar_tree_asymmetric_flow_threshold > 100) {
		log_report("Illegal ar_tree_asymmetric_flow_threshold %u, setting to 0\n",
			p_opts->ar_tree_asymmetric_flow_threshold);
		p_opts->ar_tree_asymmetric_flow_threshold = 0;
	}

	if (p_opts->fabric_mode_policy_file != NULL) {
		FILE *policy_file = fopen(p_opts->fabric_mode_policy_file, "r");
		if (!policy_file) {
			if (p_opts->fabric_mode_profile &&
			    strcmp(p_opts->fabric_mode_profile, OSM_FABRIC_MODE_PROFILE_IGNORE)) {
				log_report("Fabric Mode policy file provided: "
					   "%s doesn't exist. Set fabric_mode_profile to %s\n",
					   p_opts->fabric_mode_policy_file,
					   OSM_FABRIC_MODE_PROFILE_IGNORE);
				free(p_opts->fabric_mode_profile);
				p_opts->fabric_mode_profile = strdup(OSM_FABRIC_MODE_PROFILE_IGNORE);
			}
		} else
			fclose(policy_file);
	}

#if OSM_ENABLE_NVLINK
	if (!p_opts->nvl_partition_reroute_wait ||
	    p_opts->nvl_partition_reroute_wait > OSM_DEFAULT_NVL_PRTN_REROUTE_WAIT_MAX) {
		log_report("Illegal nvl_partition_reroute_wait %u, "
			   "setting to default value %u\n",
			   p_opts->nvl_partition_reroute_wait, OSM_DEFAULT_NVL_PRTN_REROUTE_WAIT);
		p_opts->nvl_partition_reroute_wait= OSM_DEFAULT_NVL_PRTN_REROUTE_WAIT;
	}
#endif

	return 0;
}

static const char * subn_selection_function_type_str(osm_rtr_selection_function_t selection_func)
{
	switch (selection_func) {
		case OSM_CRC32:
			return "crc32";
		default:
			break;
	}

	return "unknown function";
}

static char * subn_rtr_selection_algo_parameters_str(uint32_t algo_param)
{
	switch (algo_param) {
		case (OSM_RTR_SELECTION_PARAM_SGID | OSM_RTR_SELECTION_PARAM_DGID):
			return "sgid,dgid";

		case OSM_RTR_SELECTION_PARAM_SGID:
			return "sgid";

		case OSM_RTR_SELECTION_PARAM_DGID:
			return "dgid";

		default:
			OSM_ASSERT(0);
	}

	return "sgid,dgid";
}

int osm_subn_parse_conf_file(char *file_name, osm_subn_opt_t * p_opts,
			     boolean_t log)
{
	char line[1024];
	FILE *opts_file;
	char *p_key, *p_val, *pound_sign;
	const opt_rec_t *r;
	void *p_field1, *p_field2;
	int token_matched;

	opts_file = fopen(file_name, "r");
	if (!opts_file) {
		if (errno == ENOENT)
			return 1;
		if (log)
			printf("cannot open file \'%s\': %s\n",
			       file_name, strerror(errno));
		return -1;
	}
	if (log)
		printf(" Reading Cached Option File: %s\n", file_name);

	p_opts->config_file = file_name;
	if (!p_opts->file_opts && !(p_opts->file_opts = malloc(sizeof(*p_opts)))) {
		fclose(opts_file);
		return -1;
	}
	memcpy(p_opts->file_opts, p_opts, sizeof(*p_opts));
	p_opts->file_opts->file_opts = NULL;

	while (fgets(line, 1023, opts_file) != NULL) {
		pound_sign = strchr(line,'#');
		token_matched = 0;
		/* Truncate any comments. */
		if (pound_sign)
			*pound_sign = '\0';

		/* get the first token */
		p_key = strtok_r(line, " \t\n", &p_val);
		if (!p_key)
			continue;

		p_val = clean_val(p_val);

		for (r = opt_tbl; r->name; r++) {
			if (strcmp(r->name, p_key))
				continue;

			token_matched = 1;
			p_field1 = (void *)p_opts->file_opts + r->opt_offset;
			p_field2 = (void *)p_opts + r->opt_offset;
			/* don't call setup function first time */
			r->parse_fn(NULL, p_key, p_val, p_field1, p_field2,
				    log, NULL);
			break;
		}

		if (!token_matched && log)
			log_report(" Unrecognized token: \"%s\"\n", p_key);
	}
	fclose(opts_file);

	osm_subn_verify_config(p_opts);

	return 0;
}

int osm_subn_rescan_conf_files(IN osm_subn_t * p_subn)
{
	char line[1024];
	osm_subn_opt_t *p_opts = &p_subn->opt;
	const opt_rec_t *r;
	FILE *opts_file;
	char *p_key, *p_val, *pound_sign;
	void *p_field1, *p_field2;
	int token_matched;
	unsigned long prev_perflog_max_size = 0;

	if (!p_opts->config_file)
		return 0;

	opts_file = fopen(p_opts->config_file, "r");
	if (!opts_file) {
		if (errno == ENOENT)
			return 1;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"cannot open file \'%s\': %s\n",
			p_opts->config_file, strerror(errno));
		return -1;
	}

	update_prev_qos_options(p_opts);

	while (fgets(line, 1023, opts_file) != NULL) {
		pound_sign = strchr(line,'#');
		token_matched = 0;

		/* Truncate any comments. */
		if (pound_sign)
			*pound_sign = '\0';

		/* get the first token */
		p_key = strtok_r(line, " \t\n", &p_val);
		if (!p_key)
			continue;

		p_val = clean_val(p_val);

		for (r = opt_tbl; r->name; r++) {
			if (strcmp(r->name, p_key))
				continue;

			token_matched = 1;

			p_field1 = (void *)p_opts->file_opts + r->opt_offset;
			/*
			 * the SM will update the field if the field is defined as updatable or
			 * the field is a string and its current value is NULL
			 */
			if ((!r->can_update) &&
			   (!(r->parse_fn == opts_parse_charp && *(char **)p_field1 == NULL)))
				continue;

			p_field2 = (void *)p_opts + r->opt_offset;
			r->parse_fn(p_subn, p_key, p_val, p_field1, p_field2,
				    TRUE, r->setup_fn);
			break;
		}
		if (!token_matched)
                       log_report(" Unrecognized token: \"%s\"\n", p_key);
	}
	fclose(opts_file);

	osm_subn_verify_config(p_opts);

	p_subn->min_path_bit = 0;

	p_subn->max_path_bit = (1 << osm_subn_get_max_lmc(p_subn)) - 1;

	p_subn->p_osm->statistic_max_size = p_subn->opt.osm_stats_dump_limit << 20;

	prev_perflog_max_size = p_subn->p_osm->perflog_mgr.perflog_max_size;
	p_subn->p_osm->perflog_mgr.perflog_max_size = p_subn->opt.osm_perflog_dump_limit << 20;
	if (prev_perflog_max_size != p_subn->p_osm->perflog_mgr.perflog_max_size) {
		p_subn->p_osm->perflog_mgr.current_str =
			(char *) realloc(p_subn->p_osm->perflog_mgr.current_str,
					 p_subn->p_osm->perflog_mgr.perflog_max_size);
		if (!p_subn->p_osm->perflog_mgr.current_str) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7550: Memory allocation failed\n");
		}
	}

	/*
	 * Acaquire lock in order to protect data structures that might be accessed
	 * by other threads.
	 */
	CL_PLOCK_EXCL_ACQUIRE(p_subn->p_osm->sm.p_lock);

	if (osm_tenant_mgr_rescan(p_subn) > 0) {
		osm_virt_mgr_revalidate_all_vports(&p_subn->p_osm->sm);
		osm_tenant_mgr_count_mc_groups_in_fabric(p_subn);
	}

	parse_prefix_routes_file(p_subn);

	parse_topo_config_file(p_subn);

	parse_per_mod_logging_file(p_subn);

	parse_hm_policy_file(p_subn);

	parse_held_back_file(p_subn);

	parse_verbose_bypass_policy(p_subn->p_osm);

	if (p_subn->p_osm->node_name_map)
		close_node_name_map(p_subn->p_osm->node_name_map);

	p_subn->p_osm->node_name_map = open_node_name_map(p_subn->opt.node_name_map_name);

	enable_activity_subjects(&p_subn->p_osm->sm.activity_mgr, &p_subn->p_osm->sm);

	osm_subn_parse_service_keys_file(p_subn);

	CL_PLOCK_RELEASE(p_subn->p_osm->sm.p_lock);

	return 0;
}

void osm_subn_output_conf(FILE *out, IN osm_subn_opt_t * p_opts)
{
	int i;

	fprintf(out,
		"#\n# DEVICE ATTRIBUTES OPTIONS\n#\n"
		"# The port GUID on which the OpenSM is running.\n"
		"# Multiple port GUIDs can be specified, separated by commas\n"
		"guid 0x%016" PRIx64 , cl_ntoh64(p_opts->guid[0]));

	for (i = 1; i < p_opts->num_guids; i++)
		fprintf(out, ", 0x%016" PRIx64 , cl_ntoh64(p_opts->guid[i]));

	fprintf(out, "\n\n");

	fprintf(out,
		"# M_Key value sent to all ports qualifying all Set(PortInfo)\n"
		"m_key 0x%016" PRIx64 "\n\n"
		"# The lease period used for the M_Key on this subnet in [sec]\n"
		"m_key_lease_period %u\n\n"
		"# The protection level used for the M_Key on this subnet\n"
		"m_key_protection_level %u\n\n"
		"# If TRUE, SM tries to determine the m_key of unknown ports from guid2mkey file.\n"
		"# If FALSE, SM won't try to determine the m_key of unknown ports.\n"
		"# Preconfigured m_key will be used instead\n"
		"m_key_lookup %s\n\n"
		"# If TRUE, SM will set a different m_key for each port in the fabric.\n"
		"# When this option is enabled, SM will force m_key_protection_level 2\n"
		"# on all devices\n"
		"m_key_per_port %s\n\n"
		"# SM_Key value of the SM used for SM authentication\n"
		"sm_key 0x%016" PRIx64 "\n\n"
		"# SM_Key value to qualify rcv SA queries as 'trusted'\n"
		"sa_key 0x%016" PRIx64 "\n\n"
		"# Note that for both values above (sm_key and sa_key)\n"
		"# OpenSM version 3.2.1 and below used the default value '1'\n"
		"# in a host byte order, it is fixed now but you may need to\n"
		"# change the values to interoperate with old OpenSM running\n"
		"# on a little-endian machine.\n\n"
		"# Subnet prefix used on this subnet\n"
		"subnet_prefix 0x%016" PRIx64 "\n\n"
		"# The LMC value used on this subnet\n"
		"lmc %u\n\n"
		"# Sets the maximum number of LIDs per router to support when calculating\n"
		"# inter-subnet path records and sets LMC for routers accordingly.\n"
		"# If global LMC is not zero, lids_per_rtr is limited by global LMC.\n"
		"# If set to 0, OpenSM will use number of LIDs according to global LMC.\n"
		"lids_per_rtr %u\n\n"
		"# lmc_esp0 determines whether the LMC value used on the subnet is used for\n"
		"# enhanced switch port 0. If TRUE, the LMC value for the subnet is used for\n"
		"# ESP0. Otherwise, LMC value for ESP0s is 0.\n"
		"lmc_esp0 %s\n\n"
		"# sm_sl determines SMSL used for SM/SA communication\n"
		"sm_sl %u\n\n"
		"# The code of maximal time a packet can live in a switch\n"
		"# The actual time is 4.096usec * 2^<packet_life_time>\n"
		"# The value 0x14 disables this mechanism\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"packet_life_time 0x%02x\n\n"
		"# The number of sequential packets dropped that cause the port\n"
		"# to enter the VLStalled state. The result of setting this value to\n"
		"# zero is undefined.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"vl_stall_count 0x%02x\n\n"
		"# The number of sequential packets dropped that cause the port\n"
		"# to enter the VLStalled state. This value is for switch ports\n"
		"# driving a CA or router port. The result of setting this value\n"
		"# to zero is undefined.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"leaf_vl_stall_count 0x%02x\n\n"
		"# The code of maximal time a packet can wait at the head of\n"
		"# transmission queue.\n"
		"# The actual time is 4.096usec * 2^<head_of_queue_lifetime>\n"
		"# The value 0x14 disables this mechanism\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"head_of_queue_lifetime 0x%02x\n\n"
		"# The maximal time a packet can wait at the head of queue on\n"
		"# switch port connected to a CA or router port\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"leaf_head_of_queue_lifetime 0x%02x\n\n"
		"# Enable QOS config VL\n"
		"# When enabled use device_configuration_file to set VL weights per port group\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"qos_config_vl_enabled %s\n\n"
		"# Limit the maximal operational VLs\n"
		"max_op_vls %u\n\n"
		"# Limit the maximal operational VLs for CAs. If 0, max_op_vls is used\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_op_vls_ca %u\n\n"
		"# Limit the maximal operational VLs for Switches. If 0, max_op_vls is used\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_op_vls_sw %u\n\n"
		"# Limit the maximal operational VLs for Routers. If 0, max_op_vls is used\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_op_vls_rtr %u\n\n"
		"# Limit the maximal neighbor MTU\n"
		"# If 0, MTU Cap is used\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_neighbor_mtu %u\n\n"
		"# Suppress Multicast Pkey Traps\n"
		"# If bit 0 is set, suppression is enabled\n"
		"# Currently only one bit is supported\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"suppress_mc_pkey_traps 0x%x\n\n"
		"# Force PortInfo:LinkSpeedEnabled on switch ports\n"
		"# If 0, don't modify PortInfo:LinkSpeedEnabled on switch port\n"
		"# Otherwise, use value for PortInfo:LinkSpeedEnabled on switch port\n"
		"# Values are (IB Spec 1.2.1, 14.2.5.6 Table 146 \"PortInfo\")\n"
		"#    1: 2.5 Gbps\n"
		"#    3: 2.5 or 5.0 Gbps\n"
		"#    5: 2.5 or 10.0 Gbps\n"
		"#    7: 2.5 or 5.0 or 10.0 Gbps\n"
		"#    2,4,6,8-14 Reserved\n"
		"#    Default 15: set to PortInfo:LinkSpeedSupported\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"force_link_speed %u\n\n"
		"# Force PortInfo:LinkSpeedExtEnabled on ports\n"
		"# If 0, don't modify PortInfo:LinkSpeedExtEnabled on port\n"
		"# Otherwise, use value for PortInfo:LinkSpeedExtEnabled on port\n"
		"# Values are (MgtWG RefIDs #4722 and #9366)\n"
		"#    1: 14.0625 Gbps\n"
		"#    2: 25.78125 Gbps\n"
		"#    3: 14.0625 Gbps or 25.78125 Gbps\n"
		"#    4: 53.125 Gbps\n"
		"#    5: 14.0625 Gbps or 53.125 Gbps\n"
		"#    6: 25.78125 Gbps or 53.125 Gbps\n"
		"#    7: 14.0625 Gbps, 25.78125 Gbps or 53.125 Gbps\n"
		"#    8: 106.25 Gbps\n"
		"#    9: 14.0625 Gbps or 106.25 Gbps\n"
		"#    10: 106.25 Gbps or 25.78125 Gbps or 106.25 Gbps\n"
		"#    11: 14.0625 Gbps or 25.78125 Gbps or 106.25 Gbps\n"
		"#    12: 53.125 Gbps or 106.25 Gbps\n"
		"#    13: 14.0625 Gbps or 53.125 Gbps or 106.25 Gbps\n"
		"#    14: 25.78125 Gbps or 53.125 Gbps or 106.25 Gbps\n"
		"#    15: 14.0625 Gbps, 25.78125 Gbps or 53.125 Gbps or 106.25 Gbps\n"
		"#    30: Disable extended link speeds\n"
		"#    Default 31: set to PortInfo:LinkSpeedExtSupported\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"force_link_speed_ext %u\n\n"
		"# Force PortInfo:LinkSpeedExtEnabled2 on ports\n"
		"# If 0, don't modify PortInfo:LinkSpeedExtEnabled2 on port\n"
		"# Otherwise, use value for PortInfo:LinkSpeedExtEnabled2 on port\n"
		"# Values are (IB Spec 1.7, 14.2.6.6 Table 170)\n"
		"#    1: 212.50 Gbps\n"
		"#    Default 7: set to PortInfo:LinkSpeedExtSupported2\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"force_link_speed_ext2 %u\n\n"
		"# Force PortInfo:LinkWidthEnabled on switch ports\n"
		"# If 0, don't modify PortInfo:LinkWidthEnabled on switch port\n"
		"# Otherwise, use value for PortInfo:LinkWidthEnabled on switch port\n"
		"# Values are (IB Spec 1.2.1, 14.2.5.6 Table 146 \"PortInfo\"\n"
		"# augmented by MgtWG RefIDs #9306-9309)\n"
		"#    1: 1x\n"
		"#    2: 4x\n"
		"#    3: 1x or 4x\n"
		"#    4: 8x\n"
		"#    5: 1x or 8x\n"
		"#    6: 4x or 8x\n"
		"#    7: 1x or 4x or 8x\n"
		"#    8: 12x\n"
		"#    9: 1x or 12x\n"
		"#    10: 4x or 12x\n"
		"#    11: 1x or 4x or 12x\n"
		"#    12: 8x or 12x\n"
		"#    13: 1x or 8x or 12x\n"
		"#    14: 4x or 8x or 12x\n"
		"#    15: 1x or 4x or 8x or 12x\n"
		"#    16: 2x\n"
		"#    17: 1x or 2x\n"
		"#    18: 2x or 4x\n"
		"#    19: 1x or 2x or 4x\n"
		"#    20: 2x or 8x\n"
		"#    21: 1x or 2x or 8x\n"
		"#    22: 2x or 4x or 8x\n"
		"#    23: 1x or 2x or 4x or 8x\n"
		"#    24: 2x or 12x\n"
		"#    25: 1x or 2x or 12x\n"
		"#    26: 2x or 4x or 12x\n"
		"#    27: 1x or 2x or 4x or 12x\n"
		"#    28: 2x or 8x or 12x\n"
		"#    29: 1x or 2x or 8x or 12x\n"
		"#    30: 2x or 4x or 8x or 12x\n"
		"#    31: 1x or 2x or 4x or 8x or 12x\n"
		"#    32-254 Reserved\n"
		"#    Default 255: set to PortInfo:LinkWidthSupported\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"force_link_width %u\n\n"
		"# FDR10 on ports on devices that support FDR10 (DEPRECATED)\n"
		"# This option is DEPRECATED. Please use support_mepi_speeds/mepi_enabled_speeds\n"
		"# Values are:\n"
		"#    0: don't use fdr10 (no MLNX ExtendedPortInfo MADs)\n"
		"#    1: enable fdr10 when supported\n"
		"#    2: disable fdr10 when supported\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"fdr10 %u\n\n"
		"# Should mlnx extended port info MADs be used\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"support_mepi_speeds %s\n\n"
		"# Bitmap of enabled speeds in mlnx extended port info MADs\n"
		"# Bits mapping:\n"
		"#    0  : FDR10\n"
		"#    1-7: Not Used\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"mepi_enabled_speeds %u\n\n"
		"# The subnet_timeout code that will be set for all the ports\n"
		"# The actual timeout is 4.096usec * 2^<subnet_timeout>\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"subnet_timeout %u\n\n"
		"# Threshold of local phy errors for sending Trap 129\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"local_phy_errors_threshold 0x%02x\n\n"
		"# Threshold of credit overrun errors for sending Trap 130\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"overrun_errors_threshold 0x%02x\n\n"
		"# Use SwitchInfo:MulticastFDBTop if advertised in PortInfo:CapabilityMask\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"use_mfttop %s\n\n",
		cl_ntoh64(p_opts->m_key),
		cl_ntoh16(p_opts->m_key_lease_period),
		p_opts->m_key_protect_bits,
		p_opts->m_key_lookup ? "TRUE" : "FALSE",
		p_opts->m_key_per_port ? "TRUE" : "FALSE",
		cl_ntoh64(p_opts->sm_key),
		cl_ntoh64(p_opts->sa_key),
		cl_ntoh64(p_opts->subnet_prefix),
		p_opts->lmc,
		p_opts->lids_per_rtr,
		p_opts->lmc_esp0 ? "TRUE" : "FALSE",
		p_opts->sm_sl,
		p_opts->packet_life_time,
		p_opts->vl_stall_count,
		p_opts->leaf_vl_stall_count,
		p_opts->head_of_queue_lifetime,
		p_opts->leaf_head_of_queue_lifetime,
		p_opts->qos_config_vl_enabled ? "TRUE" : "FALSE",
		p_opts->max_op_vls,
		p_opts->max_op_vls_ca,
		p_opts->max_op_vls_sw,
		p_opts->max_op_vls_rtr,
		p_opts->max_neighbor_mtu,
		p_opts->suppress_mc_pkey_traps,
		p_opts->force_link_speed,
		p_opts->force_link_speed_ext,
		p_opts->force_link_speed_ext2,
		p_opts->force_link_width,
		p_opts->fdr10,
		p_opts->support_mepi_speeds ? "TRUE" : "FALSE",
		p_opts->mepi_enabled_speeds,
		p_opts->subnet_timeout,
		p_opts->local_phy_errors_threshold,
		p_opts->overrun_errors_threshold,
		p_opts->use_mfttop ? "TRUE" : "FALSE");

	fprintf(out,
		"#\n# PARTITIONING OPTIONS\n#\n"
		"# Partition configuration file to be used\n"
		"partition_config_file %s\n\n"
		"# Disable partition enforcement by switches (DEPRECATED)\n"
		"# This option is DEPRECATED. Please use part_enforce instead\n"
		"no_partition_enforcement %s\n\n"
		"# Partition enforcement type (for switches)\n"
		"# Values are both, out, in and off\n"
		"# Default is both (outbound and inbound enforcement)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"part_enforce %s\n\n"
		"# Allow both full and limited membership on the same partition\n"
		"allow_both_pkeys %s\n\n"
		"# Keep current and take into account old pkey indexes\n"
		"# during the calculation of physical ports pkey tables\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"keep_pkey_indexes %s\n\n"
		"# SM assigned GUID byte where GUID is formed from OpenFabrics OUI\n"
		"# followed by 40 bits xy 00 ab cd ef where xy is the SM assigned GUID byte\n"
		"# and ab cd ef is an SM autogenerated 24 bits\n"
		"# SM assigned GUID byte should be configured as subnet unique\n"
		"sm_assigned_guid 0x%02x\n\n",
		p_opts->partition_config_file,
		p_opts->no_partition_enforcement ? "TRUE" : "FALSE",
		p_opts->part_enforce,
		p_opts->allow_both_pkeys ? "TRUE" : "FALSE",
		p_opts->keep_pkey_indexes ?  "TRUE" : "FALSE",
		p_opts->sm_assigned_guid);

	fprintf(out,
		"#\n# SWEEP OPTIONS\n#\n"
		"# The number of seconds between subnet sweeps (0 disables it)\n"
		"sweep_interval %u\n\n"
		"# If TRUE, cause all LIDs to be reassigned\n"
		"reassign_lids %s\n\n"
		"# If TRUE, forces every sweep to be a heavy sweep\n"
		"# This option is DEPRECATED. Please use force_heavy_sweep_window\n"
		"force_heavy_sweep %s\n\n"
		"# Forces heavy sweep after a number of light sweeps\n"
		"# (-1 disables this option and 0 will cause every sweep to be heavy)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"force_heavy_sweep_window %d\n\n"
		"# If TRUE, every trap 128 and 144 will cause a heavy sweep.\n"
		"# NOTE: successive identical traps (>10) are suppressed\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"sweep_on_trap %s\n\n",
		p_opts->sweep_interval,
		p_opts->reassign_lids ? "TRUE" : "FALSE",
		p_opts->force_heavy_sweep ? "TRUE" : "FALSE",
		p_opts->force_heavy_sweep_window,
		p_opts->sweep_on_trap ? "TRUE" : "FALSE");

	fprintf(out,
		"#\n# ROUTING OPTIONS\n#\n"
		"# If TRUE, count switches as link subscriptions\n"
		"port_profile_switch_nodes %s\n\n",
		p_opts->port_profile_switch_nodes ? "TRUE" : "FALSE");

	fprintf(out,
		"# Name of file with port GUIDs to be ignored by port profiling\n"
		"port_prof_ignore_file %s\n\n", p_opts->port_prof_ignore_file ?
		p_opts->port_prof_ignore_file : null_str);

	fprintf(out,
		"# The file holding routing weighting factors per output port\n"
		"hop_weights_file %s\n\n",
		p_opts->hop_weights_file ? p_opts->hop_weights_file : null_str);

	fprintf(out,
		"# The file holding non-default port order per switch for routing\n"
		"port_search_ordering_file %s\n\n",
		p_opts->port_search_ordering_file ?
		p_opts->port_search_ordering_file : null_str);

	fprintf(out,
		"# multicast group primary root GUID (deprecated)\n"
		"mc_primary_root_guid 0x%016" PRIx64 "\n\n",
		cl_ntoh64(p_opts->mc_primary_root_guid));

	fprintf(out,
		"# multicast group secondary root GUID (deprecated)\n"
		"mc_secondary_root_guid 0x%016" PRIx64 "\n\n",
		cl_ntoh64(p_opts->mc_secondary_root_guid));

	fprintf(out,
		"# The file holding map from MGID to switch that OpenSM will\n"
		"# prefer to use as root for the multicast tree.\n"
		"# Line format: <MGID(IPv6)> (<GUID>|\"<NodeDescription\")\n"
		"# MGID 0 implies the OpenSM will try to use the give switch\n"
		"# as root for MGIDs that were not defined in this file.\n"
		"mc_roots_file %s\n\n",
		p_opts->mc_roots_file ? p_opts->mc_roots_file : null_str);

	fprintf(out,
		"# Comma-separated list of port GUIDs on which SM is allowed to run on.\n"
		"# When specified, OpenSM ignores any SM running on ports that\n"
		"# are not in this list.\n"
		"# 0: Do not allow any other SM.\n"
		"# (null): Feature disabled.\n"
		"allowed_sm_guids %s\n\n", p_opts->allowed_sm_guids ?
		p_opts->allowed_sm_guids : null_str);

	fprintf(out,
		"# Routing engine\n"
		"# Multiple routing engines can be specified separated by\n"
		"# commas so that specific ordering of routing algorithms will\n"
		"# be tried if earlier routing engines fail.\n"
		"# Supported engines: minhop, updn, dnup, file, ftree, lash,\n"
		"#    dor, torus-2QoS, kdor-hc, kdor-ghc ,\n"
		"#    dfsssp (EXPERIMENTAL),\n"
		"#    sssp (EXPERIMENTAL), chain,\n"
		"#    pqft (EXPERIMENTAL),\n"
		"#    dfp, dfp2, ar_updn, ar_ftree, ar_torus, ar_dor (AR)\n"
		"routing_engine %s\n\n", p_opts->routing_engine_names ?
		p_opts->routing_engine_names : null_str);

	fprintf(out,
		"# Allows SM response to SA queries while routing. Particularly,\n"
		"# SA Service Record.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"enable_queries_during_routing %s\n\n",
		p_opts->enable_queries_during_routing ? "TRUE" : "FALSE");

	fprintf(out,
		"# Routing engines will avoid throttled switch-to-switch links.\n"
		"# (Currently supported by: dfsssp, sssp; use FALSE if unsure)\n"
		"avoid_throttled_links %s\n\n",
		p_opts->avoid_throttled_links ? "TRUE" : "FALSE");

	fprintf(out,
		"# Connect roots (use FALSE if unsure)\n"
		"# This option is obsolete for fat-tree and updn routing\n"
		"# engines, use calc_missing_routes instead.\n"
		"connect_roots %s\n\n",
		p_opts->connect_roots ? "TRUE" : "FALSE");

	fprintf(out,
		"# Calculate missing routes\n"
		"# This option will cause OpenSM to find and recalculate\n"
		"# missing routes without creating credit-loops.\n"
		"# This option is valid only for updn and ftree routing\n"
		"# engines.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"calculate_missing_routes %s\n\n",
		p_opts->calc_missing_routes ? "TRUE" : "FALSE");

	fprintf(out,
		"# Maximum number of CAs on switch to allow considering it\n"
		"# as spine instead of leaf by a routing algorithm.\n"
		"max_cas_on_spine %d\n\n",
		p_opts->max_cas_on_spine);

	fprintf(out,
		"# Find root using coloring algorithm for tree-based topologies.\n"
		"# Supported by updn, ar_updn routing engines.\n"
		"# Coloring algorithm uses BFS from leaf switches toward root switches.\n"
		"# The algorithm determines leaf switches according to the number\n"
		"# of hosts connected to the switch. Number of hosts to determine\n"
		"# that a switch is a leaf is specified by max_cas_on_spine\n"
		"find_roots_color_algorithm %s\n\n",
		p_opts->find_roots_color_algorithm ? "TRUE" : "FALSE");

	fprintf(out,
		"# Find root using coloring algorithm for DF+.\n"
		"# Supported by dfp and dfp2 routing engines.\n"
		"# Coloring algorithm uses BFS from leaf switches toward root switches.\n"
		"# The algorithm determines leaf switches according to the number\n"
		"# of hosts connected to the switch. Number of hosts to determine\n"
		"# that a switch is a leaf, is specified by dfp_max_cas_on_spine. (experimental)\n"
		"dfp_find_roots_color_algorithm %s\n\n",
		p_opts->dfp_find_roots_color_algorithm ? "TRUE" : "FALSE");

	fprintf(out,
		"# Use unicast routing cache (use FALSE if unsure)\n"
		"use_ucast_cache %s\n\n",
		p_opts->use_ucast_cache ? "TRUE" : "FALSE");

	fprintf(out,
		"# Lid matrix dump file name\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"lid_matrix_dump_file %s\n\n", p_opts->lid_matrix_dump_file ?
		p_opts->lid_matrix_dump_file : null_str);

	fprintf(out,
		"# LFTs file name\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"lfts_file %s\n\n",
		p_opts->lfts_file ? p_opts->lfts_file : null_str);

	fprintf(out,
		"# The file holds the root node GUIDs of the topology.\n"
		"# Single GUID or port group in each line.\n"
		"# Feature supported by updn, ar_updn, ftree, ar_ftree and dfp2 routing engines.\n"
		"root_guid_file %s\n\n",
		p_opts->root_guid_file ? p_opts->root_guid_file : null_str);

	fprintf(out,
		"# The file holding the port groups policy\npgrp_policy_file %s\n\n",
		p_opts->pgrp_policy_file ? p_opts->pgrp_policy_file : null_str);

	fprintf(out,
		"# The file holding the topology policy\ntopo_policy_file %s\n\n",
		p_opts->topo_policy_file ? p_opts->topo_policy_file : null_str);

	fprintf(out,
		"# The file holding the routing chains policy\nrch_policy_file %s\n\n",
		p_opts->rch_policy_file ? p_opts->rch_policy_file : null_str);

	fprintf(out,
		"# The file holding the device configuration parameters\n"
		"device_configuration_file %s\n\n",
		p_opts->device_configuration_file ?
		p_opts->device_configuration_file : null_str);

	fprintf(out,
		"# Enable Fast Recovery configuration. If enabled, use\n"
		"# fast_recovery_conf_file for setting feature relevant parameters\n"
		"# Supported values:\n"
		"#    0: Ignore fast recovery related MADs and configuration\n"
		"#    1: Disable fast recovery\n"
		"#    2: Enable fast recovery\n"
		"fast_recovery_enabled %u\n\n"
		"# The file holding the fast recovery configuration parameters\n"
		"fast_recovery_conf_file %s\n\n",
		p_opts->fast_recovery_enabled,
		p_opts->fast_recovery_conf_file ?
		p_opts->fast_recovery_conf_file : null_str);

	fprintf(out,
		"# The ftree ca order dump file\n"
		"ftree_ca_order_dump_file %s\n\n",
		p_opts->ftree_ca_order_dump_file ? p_opts->ftree_ca_order_dump_file : null_str);

	fprintf(out,
		"# The file holding the node GUIDs list to held back\n"
		"held_back_sw_file %s\n\n",
		p_opts->held_back_sw_file ? p_opts->held_back_sw_file : null_str);

	fprintf(out,
		"# The file holding the fat-tree compute node GUIDs\n"
		"# One GUID in each line\n"
		"cn_guid_file %s\n\n",
		p_opts->cn_guid_file ? p_opts->cn_guid_file : null_str);

	fprintf(out,
		"# The file holding the fat-tree I/O node GUIDs\n"
		"# One GUID in each line.\n"
		"# If only io_guid file is provided, the rest of nodes\n"
		"# are considered as compute nodes.\n"
		"io_guid_file %s\n\n",
		p_opts->io_guid_file ? p_opts->io_guid_file : null_str);


	fprintf(out,
		"# If TRUE, enables alternative indexing policy for ftree routing\n"
		"# in quasi-ftree topologies that can improve shift-pattern support.\n"
		"# The switch indexing starts from root switch and leaf switches\n"
		"# are termination points of BFS algorithm\n"
		"# If FALSE, the indexing starts from leaf switch (default)\n"
		"quasi_ftree_indexing %s\n\n",
		p_opts-> quasi_ftree_indexing ? "TRUE" : "FALSE");

	fprintf(out,
		"# Number of reverse hops allowed for I/O nodes\n"
		"# Used for connectivity between I/O nodes connected to Top Switches\n"
		"# This option is obsolete and recommended not using this option when\n"
		"# calculate_missing_routes option is on\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_reverse_hops %d\n\n",
		p_opts->max_reverse_hops);

	fprintf(out,
		"# Number of threads to be used for parallel minhop/updn/dor calculations\n"
		"# If 0, number of threads will be equal to number of processors\n"
		"routing_threads_num %d\n\n",
		p_opts->routing_threads_num);

	fprintf(out,
		"# Max number of threads that are allowed to run\n"
		"# on the same processor during parallel computing\n"
		"# If 0, threads assignment per processor is up to\n"
		"# operating system initial assignment\n"
		"max_threads_per_core %d\n\n",
		p_opts->max_threads_per_core);

	fprintf(out,
		"# The file holding the node ids which will be used by Up/Down\n"
		"# algorithm instead of GUIDs (one GUID and id in each line)\n"
		"ids_guid_file %s\n\n",
		p_opts->ids_guid_file ? p_opts->ids_guid_file : null_str);

	fprintf(out,
		"# The file holding GUID routing order GUIDs (for MinHop and Up/Down)\n"
		"guid_routing_order_file %s\n\n",
		p_opts->guid_routing_order_file ? p_opts->guid_routing_order_file : null_str);

	fprintf(out,
		"# During rerouting stage of the heavy sweep\n"
		"# SM will keep routing balanced (default: FALSE)\n"
		"# If set to TRUE, rebalancing will happen during IDLE time.\n"
		"# Applicable to minhop/dor/updn routing engines only.\n"
		"offsweep_balancing_enabled %s\n\n",
		p_opts->offsweep_balancing_enabled ? "TRUE" : "FALSE");

	fprintf(out,
		"# During IDLE time processing\n"
		"# SM will start rebalancing of the routing\n"
		"# after the predefined number of seconds (default: 180s)\n"
		"# The option is applicable when offsweep_balancing_enabled is TRUE\n"
		"offsweep_balancing_window %d\n\n",
		p_opts->offsweep_balancing_window);

	fprintf(out,
		"# The definition of the expected Fat Tree topology:\n"
		"# PGFT(h; M1, M2,...Mh; W1,W2,...Wh; P1,P1,...Ph) or\n"
		"# QFT(h; M1, M2,...Mh; W1,W2,...Wh; P1,P1,...Ph)\n"
		"pqft_structure %s\n\n",
		p_opts->pqft_structure ? p_opts->pqft_structure : null_str);

	fprintf(out,
		"# Do mesh topology analysis (for LASH algorithm)\n"
		"do_mesh_analysis %s\n\n",
		p_opts->do_mesh_analysis ? "TRUE" : "FALSE");

	fprintf(out,
		"# Starting VL for LASH algorithm\n"
		"lash_start_vl %u\n\n",
		p_opts->lash_start_vl);

	fprintf(out,
		"# Port Shifting (use FALSE if unsure)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"port_shifting %s\n\n",
		p_opts->port_shifting ? "TRUE" : "FALSE");

	fprintf(out,
		"# Assign ports in a random order instead of round-robin\n"
		"# If zero, disable (default). Otherwise, use the value as a random seed\n"
		"scatter_ports %d\n\n",
		p_opts->scatter_ports);

	fprintf(out,
		"# Enable LID tracking mode for Up/Down\n"
		"updn_lid_tracking_mode %s\n\n",
		p_opts->updn_lid_tracking_mode ? "TRUE" : "FALSE");

	fprintf(out,
		"# When enabled in updn_lid_tracking_mode,\n"
		"# all routes to DLID are consolidated\n"
		"# as early as possible in the fabric\n"
		"updn_lid_tracking_converge_routes %s\n\n",
		p_opts->updn_lid_tracking_converge_routes ? "TRUE" : "FALSE");

	fprintf(out,
		"# When enabled in updn_lid_tracking_mode,\n"
		"# comparing total routes on the adjacent switch\n"
		"# takes precedence over number of routes\n"
		"# via adjacent switch to the neighbor switch of the destination\n"
		"updn_lid_tracking_prefer_total_routes %s\n\n",
		p_opts->updn_lid_tracking_prefer_total_routes ? "TRUE" : "FALSE");


	fprintf(out,
		"# Maximum number of CAs on switch to allow considering it\n"
		"# as spine instead of leaf by DF+ routing algorithm.\n"
		"dfp_max_cas_on_spine %d\n\n",
		p_opts->dfp_max_cas_on_spine);

	fprintf(out,
		"# Multicast optimization (use FALSE if unsure)\n"
		"multicast_optimization %s\n\n",
		p_opts->const_multicast ? "TRUE" : "FALSE");

	fprintf(out,
		"# Don't use scatter for ports defined in\n"
		"# guid_routing_order file\n"
		"guid_routing_order_no_scatter %s\n\n",
		p_opts->guid_routing_order_no_scatter ? "TRUE" : "FALSE");

	fprintf(out,
		"# Use scatter when routing to the switch LIDs\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"use_scatter_for_switch_lid %s\n\n",
		p_opts->use_scatter_for_switch_lid ? "TRUE" : "FALSE");

	fprintf(out,
		"# SA database file name\n"
		"sa_db_file %s\n\n",
		p_opts->sa_db_file ? p_opts->sa_db_file : null_str);

	fprintf(out,
		"# If TRUE, causes OpenSM to dump SA database at the end of\n"
		"# every light sweep, regardless of the verbosity level\n"
		"sa_db_dump %s\n\n",
		p_opts->sa_db_dump ? "TRUE" : "FALSE");

	fprintf(out,
		"# If TRUE, causes OpenSM to dump SM database at the end of\n"
		"# every heavy sweep\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"sm_db_dump %s\n\n",
		p_opts->sm_db_dump ? "TRUE" : "FALSE");

	fprintf(out,
		"# Load Inform Info From SA\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"load_inform_info_from_sadb %s\n",
		p_opts->load_inform_info_from_sadb ? "TRUE": "FALSE");

	fprintf(out,
		"# If TRUE, allows OpenSM to respond to SA full World Path Record\n"
		"# queries (path records between each pair of existing ports).\n"
		"# Otherwise, such requests are ignored.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"sa_pr_full_world_queries_allowed %s\n\n",
		p_opts->sa_pr_full_world_queries_allowed ? "TRUE" : "FALSE");

	fprintf(out,
		"# Torus-2QoS configuration file name\n"
		"torus_config %s\n\n",
		p_opts->torus_conf_file ? p_opts->torus_conf_file : null_str);

	fprintf(out,
		"#\n# HANDOVER - MULTIPLE SMs OPTIONS\n#\n"
		"# SM priority used for deciding who is the master\n"
		"# Range goes from 0 (lowest priority) to 15 (highest).\n"
		"sm_priority %u\n\n"
		"# SM MASTER priority that assigned to SM when it becomes MASTER\n"
		"# Should be higher than sm_priority.\n"
		"master_sm_priority %u\n\n"
		"# Force LID for local SM when in MASTER state\n"
		"# Selected LID must match selected LMC\n"
		"# 0 disables the feature\n"
		"master_sm_lid %u\n\n"
		"# If TRUE, and the SM runs on more than one port,\n"
		"# failover to one of the axillary SM ports when the main one goes down,\n"
		"# if there is one in active state.\n"
		"enable_sm_port_failover %s\n\n"
		"# If TRUE, other SMs on the subnet should be ignored\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"ignore_other_sm %s\n\n"
		"# Timeout in [msec] between two polls of active master SM\n"
		"sminfo_polling_timeout %u\n\n"
		"# Number of failing polls of remote SM that declares it dead\n"
		"polling_retry_number %u\n\n"
		"# If TRUE, honor the guid2lid file when coming out of standby\n"
		"# state, if such file exists and is valid\n"
		"honor_guid2lid_file %s\n\n",
		p_opts->sm_priority,
		p_opts->master_sm_priority,
		p_opts->master_sm_lid,
		p_opts->enable_sm_port_failover ? "TRUE" : "FALSE",
		p_opts->ignore_other_sm ? "TRUE" : "FALSE",
		p_opts->sminfo_polling_timeout,
		p_opts->polling_retry_number,
		p_opts->honor_guid2lid_file ? "TRUE" : "FALSE");

	fprintf(out,
		"#\n# TIMING AND THREADING OPTIONS\n#\n"
		"# Maximum number of SMPs sent in parallel\n"
		"max_wire_smps %u\n\n"
		"# Maximum number of timeout-based SMPs allowed to be outstanding\n"
		"# A value less than or equal to max_wire_smps disables this mechanism\n"
		"max_wire_smps2 %u\n\n"
		"# The timeout in [usec] used for sending SMPs above max_wire_smps limit and below max_wire_smps2 limit\n"
		"max_smps_timeout %u\n\n"
		"# The maximum time in [msec] allowed for a transaction to complete\n"
		"transaction_timeout %u\n\n"
		"# The maximum number of retries allowed for a transaction to complete\n"
		"transaction_retries %u\n\n"
		"# The maximum time in [msec] allowed for a \"long\" transaction to complete.\n"
		"# Currently, long transactions are only used for \n"
		"# optimized SL2VLMappingTable and PortInfo for port 0 MADs\n"
		"long_transaction_timeout %u\n\n"
		"# Maximal time in [msec] a message can stay in the incoming message queue.\n"
		"# If there is more than one message in the queue and the last message\n"
		"# stayed in the queue more than this value, any SA request will be\n"
		"# immediately dropped, but BUSY status is not currently returned.\n"
		"max_msg_fifo_timeout %u\n\n"
		"# Use a single thread for handling SA queries\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"single_thread %s\n\n"
		"# Maximum number of messages that can reside in the incoming message queue,\n"
		"# before dropping SubnAdmGet/SubnAdmGetTbl requests.\n"
		"max_msg_fifo_len %d\n\n",
		p_opts->max_wire_smps,
		p_opts->max_wire_smps2,
		p_opts->max_smps_timeout,
		p_opts->transaction_timeout,
		p_opts->transaction_retries,
		p_opts->long_transaction_timeout,
		p_opts->max_msg_fifo_timeout,
		p_opts->single_thread ? "TRUE" : "FALSE",
		p_opts->max_msg_fifo_len);

	fprintf(out,
		"#\n# MISC OPTIONS\n#\n"
		"# Daemon mode\n"
		"daemon %s\n\n"
		"# SM Inactive\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"sm_inactive %s\n\n"
		"# Babbling Port Policy\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"babbling_port_policy %s\n\n"
		"# Drop subscription on report failure (o13-17.2.1)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"drop_subscr_on_report_fail %s\n\n"
		"# Drop event subscriptions (InformInfo and ServiceRecord) on port removal and SM coming out of STANDBY\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"drop_event_subscriptions %s\n\n"
		"# Drop event subscriptions (InformInfo and ServiceRecord) on ports that have no routing to SM\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"drop_unreachable_event_subscriptions %s\n\n"
		"# Validate IPoIB non-broadcast group creation parameters against\n"
		"# broadcast group parameters per IETF RFC 4391 (default TRUE)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"ipoib_mcgroup_creation_validation %s\n\n"
		"# Validate multicast join parameters against multicast group\n"
		"# parameters when MC group already exists\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"mcgroup_join_validation %s\n\n"
		"# Use original extended SA rates only\n"
		"# The original extended SA rates are up through 300 Gbps (12x EDR)\n"
		"# Set to TRUE for subnets with old kernels/drivers that don't understand\n"
		"# the new SA rates for 2x link width and/or HDR link speed (19-22)\n"
		"# default is FALSE\n"
		"# Notice: use_original_extended_sa_rates_only is deprecated by max_rate_enum!\n"
		"# Use max_rate_enum 18 instead.\n"
		"use_original_extended_sa_rates_only %s\n\n"
		"# Enumeration of the maximal rate subnet supports. Option is needed for\n"
		"# subnets with old kernels/drivers that don't understand new SA rates.\n"
		"# For example:\n"
		"# 18: Rate of 300Gbps - support speeds up to 12xEDR\n"
		"# 22: Rate of 600Gbps - support speeds up to HDR\n"
		"# For further reference refer to IB specification chapter 15.2.5.16.1\n"
		"max_rate_enum %u\n\n"
		"# SA Reports\n"
		"# 0 - No reports\n"
		"# 1 - GID/MGID IN/OUT and Local changes reports only\n"
		"# 2 - All reports\n"
		"reports %u\n\n"
		"# Maximum number of SA Reports() queued to be sent\n"
		"# 0 disables this feature\n"
		"max_sa_reports_queued %u\n\n"
		"# Maximum number of SA Reports() sent in parallel\n"
		"# 0 disables this report limiting feature\n"
		"max_sa_reports_on_wire %u\n\n"
		"# Is the enhanced trust model active.\n"
		"# Defines a minimal set of SA operations,\n"
		"# that are allowed for untrusted clients\n"
		"# default is FALSE\n"
		"sa_enhanced_trust_model %s\n\n"
		"# If the sa_enhanced_trust_model is active and sa_etm_allow_untrusted_proxy_requests\n"
		"# is set to FALSE, we don't allow untrusted proxy requests\n"
		"# default is TRUE\n"
		"sa_etm_allow_untrusted_proxy_requests %s\n\n"
		"# If the sa_enhanced_trust_model is active and sa_etm_allow_untrusted_guidinfo_rec\n"
		"# is set to FALSE, we drop guidInfoRecord requests if sm_key == 0 or incorrect\n"
		"# default is TRUE\n"
		"sa_etm_allow_untrusted_guidinfo_rec %s\n\n"
		"# If the sa_enhanced_trust_model is active and sa_etm_allow_guidinfo_rec_by_vf\n"
		"# is set to FALSE, we allow for clients to send guidInfoRecord\n"
		"# SET/DELETE requests, only from physical ports\n"
		"# (doesn't matter if the client is trusted or not)\n"
		"# default is TRUE\n"
		"sa_etm_allow_guidinfo_rec_by_vf %s\n\n"
		"# Max number of multicast groups per port/vport that can be registered,\n"
		"# active only if the sa_enhanced_trust_model is set to TRUE.\n"
		"# Default: 0 (unlimited)\n"
		"sa_etm_max_num_mcgs %u\n\n"
		"# Max number of services (ServiceRecords) per port/vport that can be registered,\n"
		"# active only if the sa_enhanced_trust_model is set to TRUE.\n"
		"# Default: 0 (unlimited)\n"
		"sa_etm_max_num_srvcs %u\n\n"
		"# Max number of event subscriptions (InformInfo) per port/vport that can be registered,\n"
		"# active only if the sa_enhanced_trust_model is set to TRUE.\n"
		"# Default: 0 (unlimited)\n"
		"sa_etm_max_num_event_subs %u\n\n"
		"# Max number of SA requests per 100 msec\n"
		"# default is 0 (unlimited)\n"
		"sa_rate_threshold %u\n\n"
		"# If enabled, every SA request is checked for SGID spoofing\n"
		"# default is TRUE\n"
		"sa_check_sgid_spoofing %s\n\n"
		"# Use Optimized SLtoVLMapping programming (bit 0) if supported by device\n"
		"use_optimized_slvl %s\n\n"
		"# Use Optimized SLtoVLMapping programming (bit 1) if supported by device\n"
		"use_optimized_port_mask_slvl %s\n\n"
		"# Sync in memory files used for high availability with storage\n"
		"fsync_high_avail_files %s\n\n"
		"# Default MC group MTU for dynamic group creation\n"
		"# 0 disables this feature\n"
		"# Otherwise, value is valid IB encoded MTU\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"default_mcg_mtu %u\n\n"
		"# Default MC group rate for dynamic group creation\n"
		"# 0 disables this feature\n"
		"# Otherwise, value is valid IB encoded rate\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"default_mcg_rate %u\n"
		"# Validate SMP read only fields\n"
		"validate_smps %s\n\n",
		p_opts->daemon ? "TRUE" : "FALSE",
		p_opts->sm_inactive ? "TRUE" : "FALSE",
		p_opts->babbling_port_policy ? "TRUE" : "FALSE",
		p_opts->drop_subscr_on_report_fail ? "TRUE" : "FALSE",
		p_opts->drop_event_subscriptions ? "TRUE" : "FALSE",
		p_opts->drop_unreachable_event_subscriptions ? "TRUE" : "FALSE",
		p_opts->ipoib_mcgroup_creation_validation ? "TRUE" : "FALSE",
		p_opts->mcgroup_join_validation ? "TRUE" : "FALSE",
		p_opts->use_original_extended_sa_rates_only ? "TRUE" : "FALSE",
		p_opts->max_rate_enum,
		p_opts->reports,
		p_opts->max_sa_reports_queued,
		p_opts->max_sa_reports_on_wire,
		p_opts->sa_enhanced_trust_model ? "TRUE" : "FALSE",
		p_opts->sa_etm_allow_untrusted_proxy_requests ? "TRUE" : "FALSE",
		p_opts->sa_etm_allow_untrusted_guidinfo_rec ? "TRUE" : "FALSE",
		p_opts->sa_etm_allow_guidinfo_rec_by_vf ? "TRUE" : "FALSE",
		p_opts->sa_etm_max_num_mcgs,
		p_opts->sa_etm_max_num_srvcs,
		p_opts->sa_etm_max_num_event_subs,
		p_opts->sa_rate_threshold,
		p_opts->sa_check_sgid_spoofing ? "TRUE" : "FALSE",
		p_opts->use_optimized_slvl ? "TRUE" : "FALSE",
		p_opts->use_optimized_port_mask_slvl ? "TRUE" : "FALSE",
		p_opts->fsync_high_avail_files ? "TRUE" : "FALSE",
		p_opts->default_mcg_mtu,
		p_opts->default_mcg_rate,
		p_opts->validate_smps ? "TRUE" : "FALSE");

#ifdef ENABLE_OSM_PERF_MGR
	fprintf(out,
		"#\n# Performance Manager Options\n#\n"
		"# perfmgr enable\n"
		"perfmgr %s\n\n"
		"# perfmgr redirection enable\n"
		"perfmgr_redir %s\n\n"
		"# sweep time in seconds\n"
		"perfmgr_sweep_time_s %u\n\n"
		"# Max outstanding queries\n"
		"perfmgr_max_outstanding_queries %u\n"
		"perfmgr_ignore_cas %s\n\n"
		"# Remove missing nodes from DB\n"
		"perfmgr_rm_nodes %s\n\n"
		"# Log error counters to opensm.log\n"
		"perfmgr_log_errors %s\n\n"
		"# Query PerfMgrGet(ClassPortInfo) for extended capabilities\n"
		"perfmgr_query_cpi %s\n\n"
		"# Log xmit_wait errors\n"
		"perfmgr_xmit_wait_log %s\n\n"
		"# If logging xmit_wait's; set threshold (default %u)\n"
		"perfmgr_xmit_wait_threshold %u\n\n"
		,
		p_opts->perfmgr ? "TRUE" : "FALSE",
		p_opts->perfmgr_redir ? "TRUE" : "FALSE",
		p_opts->perfmgr_sweep_time_s,
		p_opts->perfmgr_max_outstanding_queries,
		p_opts->perfmgr_ignore_cas ? "TRUE" : "FALSE",
		p_opts->perfmgr_rm_nodes ? "TRUE" : "FALSE",
		p_opts->perfmgr_log_errors ? "TRUE" : "FALSE",
		p_opts->perfmgr_query_cpi ? "TRUE" : "FALSE",
		p_opts->perfmgr_xmit_wait_log ? "TRUE" : "FALSE",
		OSM_PERFMGR_DEFAULT_XMIT_WAIT_THRESHOLD,
		p_opts->perfmgr_xmit_wait_threshold);

	fprintf(out,
		"#\n# Event DB Options\n#\n"
		"# Dump file to dump the events to\n"
		"event_db_dump_file %s\n\n", p_opts->event_db_dump_file ?
		p_opts->event_db_dump_file : null_str);
#endif				/* ENABLE_OSM_PERF_MGR */

	fprintf(out,
		"#\n#Plugin Options\n#\n"
		"# Plugin name(s)\n"
		"# SM should be restarted upon parameter change\n"
		"plugin_name %s\n\n"
		"# Options string that would be passed to the plugin(s)\n"
		"# SM should be restarted upon parameter change\n"
		"plugin_options %s\n\n",
		p_opts->plugin_name ? p_opts->plugin_name : null_str,
		p_opts->plugin_options ? p_opts->plugin_options : null_str);

	fprintf(out,
		"#\n# Event Plugin Options\n#\n"
		"# Event plugin name(s)\n"
		"# SM should be restarted upon parameter change\n"
		"event_plugin_name %s\n\n"
		"# Options string that would be passed to the plugin(s)\n"
		"# the option string is space sensitive (only 1 space between words)\n"
		"# SM should be restarted upon parameter change\n"
		"event_plugin_options %s\n\n",
		p_opts->event_plugin_name ?
		p_opts->event_plugin_name : null_str,
		p_opts->event_plugin_options ?
		p_opts->event_plugin_options : null_str);

	fprintf(out,
		"#\n# Node name map for mapping nodes to more descriptive node descriptions\n"
		"# (man ibnetdiscover for more information)\n#\n"
		"node_name_map_name %s\n\n", p_opts->node_name_map_name ?
		p_opts->node_name_map_name : null_str);

	fprintf(out,
		"#\n# DEBUG FEATURES\n#\n"
		"# Enable OpenSM crash daemon\n"
		"enable_crashd %s\n\n"
		"# The log flags used\n"
		"log_flags 0x%02x\n\n"
		"# Force flush of the log file after each log message\n"
		"force_log_flush %s\n\n"
		"# Log file to be used\n"
		"log_file %s\n\n"
		"# Limit the size of the log file in MB. If overrun, log is restarted\n"
		"log_max_size %u\n\n"
		"# If TRUE, will accumulate the log over multiple OpenSM sessions\n"
		"accum_log_file %s\n\n"
		"# Per module logging configuration file\n"
		"# Each line in config file contains <module_name><separator><log_flags>\n"
		"# where module_name is file name including .c\n"
		"# separator is either = , space, or tab\n"
		"# log_flags are the same flags as used in the coarse/overall logging\n"
		"per_module_logging_file %s\n\n"
		"# The directory to hold the file OpenSM dumps\n"
		"dump_files_dir %s\n\n"
		"# If TRUE, enables new high-risk options and hardware specific quirks\n"
		"enable_quirks %s\n\n"
		"# If TRUE, disables client reregistration\n"
		"# Client reregistration is used to request clients to resend subscription requests.\n"
		"# Disabling client reregistration could affect application recovery from subnet\n"
		"# events (such as SM failover).\n"
		"# This option is kept for backward compatibility. Not recommended for use.\n"
		"# This parameter is deprecated by client_rereg_mode parameter.\n"
		"no_clients_rereg %s\n\n"
		"# Control parameter for sending client reregistration options.\n"
		"# 0 - Sending client reregistration disabled.\n"
		"#     This option is kept for backward compatibility. Not recommended for use.\n"
		"# 1 - Send client reregistration during LID assignment.\n"
		"#     This option is kept for backward compatibility. Not recommended for use.\n"
		"# 2 - Send Client reregistration during link activation.\n"
		"client_rereg_mode %d\n\n"
		"# If TRUE, OpenSM should disable multicast support and\n"
		"# no multicast routing is performed\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"disable_multicast %s\n\n"
		"# If non-zero, OpenSM will use the provided switch node GUID as the\n"
		"# If TRUE, OpenSM will exit on fatal initialization issues\n"
		"exit_on_fatal %s\n\n" "# console [off|local"
#ifdef ENABLE_OSM_CONSOLE_LOOPBACK
		"|loopback"
#endif
#ifdef ENABLE_OSM_CONSOLE_SOCKET
		"|socket]\n"
#else
		"]\n"
#endif
		"console %s\n\n"
		"# Telnet port for console (default %d)\n"
		"console_port %d\n\n",
		p_opts->enable_crashd ? "TRUE" : "FALSE",
		p_opts->log_flags,
		p_opts->force_log_flush ? "TRUE" : "FALSE",
		p_opts->log_file,
		p_opts->log_max_size,
		p_opts->accum_log_file ? "TRUE" : "FALSE",
		p_opts->per_module_logging_file ?
			p_opts->per_module_logging_file : null_str,
		p_opts->dump_files_dir,
		p_opts->enable_quirks ? "TRUE" : "FALSE",
		p_opts->no_clients_rereg ? "TRUE" : "FALSE",
		p_opts->client_rereg_mode,
		p_opts->disable_multicast ? "TRUE" : "FALSE",
		p_opts->exit_on_fatal ? "TRUE" : "FALSE",
		p_opts->console,
		OSM_DEFAULT_CONSOLE_PORT, p_opts->console_port);

	fprintf(out,
		"#\n# QoS OPTIONS\n#\n"
		"# Enable QoS setup\n"
		"qos %s\n\n"
		"# QoS policy file to be used\n"
		"qos_policy_file %s\n\n"
		"# Suppress QoS MAD status errors\n"
		"suppress_sl2vl_mad_status_errors %s\n\n"
		"# Override multicast SL provided in join/create request\n"
		"# OpenSM uses the given SL to override the SL in the request\n"
		"# 0xff disables the feature\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"override_create_mcg_sl 0x%x\n\n",
		p_opts->qos ? "TRUE" : "FALSE", p_opts->qos_policy_file,
		p_opts->suppress_sl2vl_mad_status_errors ? "TRUE" : "FALSE",
		p_opts->override_create_mcg_sl);

	subn_dump_qos_options(out,
			      "QoS default options", "qos",
			      &p_opts->qos_options);
	subn_dump_qos_options(out,
			      "QoS CA options", "qos_ca",
			      &p_opts->qos_ca_options);
	subn_dump_qos_options(out,
			      "QoS Switch Port 0 options", "qos_sw0",
			      &p_opts->qos_sw0_options);
	subn_dump_qos_options(out,
			      "QoS Switch external ports options", "qos_swe",
			      &p_opts->qos_swe_options);
	subn_dump_qos_options(out,
			      "QoS Switch-to-switch external port options", "qos_sw2sw",
			      &p_opts->qos_sw2sw_options);
	subn_dump_qos_options(out,
			      "QoS Router ports options", "qos_rtr",
			      &p_opts->qos_rtr_options);

	fprintf(out,
		"#\n# Enhanced QoS OPTIONS\n#\n"
		"# The file holding the enhanced QoS policy\n"
		"enhanced_qos_policy_file %s\n\n"
		"# If TRUE, will set unlimited rate (0) for unspecified SLs\n"
		"# in vport0 bandwidth rule. If no rule is defined,\n"
		"# unlimited rate will be set in all SLs of vport0\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"enhanced_qos_vport0_unlimit_default_rl %s\n\n",
		p_opts->enhanced_qos_policy_file ? p_opts->enhanced_qos_policy_file : null_str,
		p_opts->enhanced_qos_vport0_unlimit_default_rl ? "TRUE" : "FALSE");

	fprintf(out,
		"#\n# Congestion Control OPTIONS (EXPERIMENTAL)\n#\n\n"
		"# Enable Congestion Control Configuration\n"
		"# 0: Ignore congestion control\n"
		"# 1: Disable congestion control\n"
		"# 2: Enable congestion control\n"
		"mlnx_congestion_control %d\n\n"
		"# The file holding the congestion control policy\n"
		"congestion_control_policy_file %s\n\n"
		"# The directory holding the PPCC algorithm profiles\n"
		"ppcc_algo_dir %s\n\n"
		"# CCKey to use when configuring congestion control\n"
		"# note that this does not configure a new CCkey, only the CCkey to use\n"
		"# This parameter is deprecated.\n"
		"# Use the parameters below in order to configure CC key per device.\n"
		"cc_key 0x%016" PRIx64 "\n\n"
		"# Congestion Control Max outstanding MAD\n"
		"cc_max_outstanding_mads %u\n\n",
		p_opts->mlnx_congestion_control,
		p_opts->congestion_control_policy_file ?
		p_opts->congestion_control_policy_file : null_str,
		p_opts->ppcc_algo_dir ?
		p_opts->ppcc_algo_dir : null_str,
		cl_ntoh64(p_opts->cc_key),
		p_opts->cc_max_outstanding_mads);

	fprintf(out,
		"# Enable Congestion Control Key Configuration. If enabled, CC keys are\n"
		"# configured using a seed indicated by key_mgr_seed.\n"
		"# Supported values:\n"
		"#    0: Ignore CCKey\n"
		"#    1: Disable CCKey\n"
		"#    2: Enable CCKey\n"
		"cc_key_enable %u\n\n"
		"# The lease period used for CC Keys in [sec]\n"
		"cc_key_lease_period %u\n\n"
		"# The protection level used for CC Keys. Supported values:\n"
		"#    0: Protection is provided. However, CC managers are allowed\n"
		"#	 to read the key by KeyInfo GET.\n"
		"#    1: Protect subnet ports with CC key.\n"
		"cc_key_protect_bit %u\n\n",
		p_opts->cc_key_enable,
		p_opts->cc_key_lease_period,
		p_opts->cc_key_protect_bit);

	fprintf(out,
		"# Enable Vendor Specific Key Configuration. If enabled, VS keys are\n"
		"# configured using a seed indicated by key_mgr_seed.\n"
		"# Supported values:\n"
		"#    0: Ignore VSKey\n"
		"#    1: Disable VSKey\n"
		"#    2: Enable VSKey\n"
		"vs_key_enable %u\n\n"
		"# The lease period used for VS keys in [sec]\n"
		"vs_key_lease_period %u\n\n"
		"# The protection level used for VS Keys. A mode defined by protect bit\n"
		"# and protection scope. Supported values:\n"
		"#    0: Protection is provided. However, VS managers are allowed\n"
		"#	 to read the key by KeyInfo GET.\n"
		"#    1: Protect subnet ports with VS key for both Informational\n"
		"#       and Configurational MADs\n"
		"vs_key_ci_protect_bits %u\n\n",
		p_opts->vs_key_enable,
		p_opts->vs_key_lease_period,
		p_opts->vs_key_ci_protect_bits);

	fprintf(out,
		"# Maximum number of vendor-specific mads in the network at once\n"
		"vs_max_outstanding_mads %u\n\n",
		p_opts->vs_max_outstanding_mads);

	fprintf(out,
		"# Enable Node to Node Key Configuration. If enabled, N2N keys are\n"
		"# configured using a seed indicated by key_mgr_seed.\n"
		"# Supported values:\n"
		"#    0: Ignore N2NKey\n"
		"#    1: Disable N2NKey\n"
		"#    2: Enable N2NKey\n"
		"n2n_key_enable %u\n\n"
		"# The lease period used for N2N keys in [sec]\n"
		"n2n_key_lease_period %u\n\n"
		"# The protection level used for N2N Keys. A mode defined by protect bit\n"
		"# and protection scope. Supported values:\n"
		"#    0: Protection is provided. However, N2N managers are allowed\n"
		"#       to read the key by KeyInfo GET.\n"
		"#    1: Protect subnet ports with N2N key for both Informational\n"
		"#       and Configurational MADs\n"
		"n2n_key_protect_bit %u\n\n"
		"# Maximum number of N2N MADs in the network at once\n"
		"n2n_max_outstanding_mads %u\n\n",
		p_opts->n2n_key_enable,
		p_opts->n2n_key_lease_period,
		p_opts->n2n_key_protect_bit,
		p_opts->n2n_max_outstanding_mads);

	fprintf(out,
		"# Parameter used by key manager for CC/VS key configuration.\n"
		"# if 0, uses mkey as a key for CC and VS Classes. Otherwise, use\n"
		"# this parameter as a seed for CC/VS key generation.\n"
		"key_mgr_seed 0x%016" PRIx64 "\n\n",
		p_opts->key_mgr_seed);

	fprintf(out, "\n");

	const char *osm_hm_action_str[OSM_HM_ACTION_MAX] = {"ignore", "report",
							    "isolate", "no_discover",
							    "disable"};

	fprintf(out,
		"# Enable Unhealthy Ports Configuration\n"
		"hm_unhealthy_ports_checks %s\n\n"
		"# Ports Health Policy Files\n"
		"hm_ports_health_policy_file %s\n\n",
		p_opts->hm_unhealthy_ports_checks ? "TRUE" : "FALSE",
		p_opts->hm_ports_health_policy_file);

	fprintf(out,
		"#\n# Unhealthy Ports Reboot condition options\n#\n"
		"# CA/RTR Reboot Action (ignore, report or isolate)\n"
		"hm_ca_reboot_action %s\n\n"
		"# Switch Reboot Action (ignore, report, isolate or no_discover)\n"
		"hm_sw_reboot_action %s\n\n"
		"# Number of reboots in period to declare a node as unhealthy\n"
		"hm_num_reboots %d\n\n"
		"# The period for counting number of reboots in seconds\n"
		"hm_reboots_period_secs %d\n\n",
		osm_hm_action_str[p_opts->hm_ca_reboot_action],
		osm_hm_action_str[p_opts->hm_sw_reboot_action],
		p_opts->hm_num_reboots,
		p_opts->hm_reboots_period_secs);

	fprintf(out,
		"#\n# Unhealthy Ports Unresponsive condition options\n#\n"
		"# CA/RTR Unresponsive Action (ignore, report or isolate)\n"
		"hm_ca_unresponsive_action %s\n\n"
		"# Switch Unresponsive Action (ignore, report, isolate or no_discover)\n"
		"hm_sw_unresponsive_action %s\n\n"
		"# The number of sweeps that had that port unresponsive\n"
		"hm_num_no_resp_sweeps %d\n\n"
		"# The number of sweeps of which any port exceeding\n"
		"# hm_num_no_resp_sweeps is declared unhealthy\n"
		"hm_num_no_resp_sweeps_window %d\n\n",
		osm_hm_action_str[p_opts->hm_ca_unresponsive_action],
		osm_hm_action_str[p_opts->hm_sw_unresponsive_action],
		p_opts->hm_num_no_resp_sweeps,
		p_opts->hm_num_no_resp_sweeps_window);

	fprintf(out,
		"#\n# Unhealthy Ports Noisy condition options\n#\n"
		"# CA/RTR Noisy Action (ignore, report or isolate)\n"
		"hm_ca_noisy_action %s\n\n"
		"# Switch Noisy Action (ignore, report, isolate or no_discover)\n"
		"hm_sw_noisy_action %s\n\n"
		"# Number of traps received in period to declare the port as\n"
		"# unhealthy.\n"
		"hm_num_traps %d\n\n"
		"# The period for counting number of received traps in seconds\n"
		"hm_num_traps_period_secs %d\n\n",
		osm_hm_action_str[p_opts->hm_ca_noisy_action],
		osm_hm_action_str[p_opts->hm_sw_noisy_action],
		p_opts->hm_num_traps,
		p_opts->hm_num_traps_period_secs);

	fprintf(out,
		"#\n# Unhealthy Ports SetErr condition options\n#\n"
		"# CA/RTR SetErr Action (ignore, report or isolate)\n"
		"hm_ca_seterr_action %s\n\n"
		"# Switch SetErr Action (ignore, report, isolate or no_discover)\n"
		"hm_sw_seterr_action %s\n\n"
		"# The number of sweeps that had that port report back an\n"
		"# error for a Set\n"
		"hm_num_set_err_sweeps %d\n\n"
		"# The number of sweeps of which any port exceeding\n"
		"# hm_num_set_err_sweeps is declared unhealthy\n"
		"hm_num_set_err_sweeps_window %d\n",
		osm_hm_action_str[p_opts->hm_ca_seterr_action],
		osm_hm_action_str[p_opts->hm_sw_seterr_action],
		p_opts->hm_num_set_err_sweeps,
		p_opts->hm_num_set_err_sweeps_window);

	fprintf(out,
		"#\n# Unhealthy Ports Flapping Link condition options\n#\n"
		"# CA/RTR Flapping Link Action (ignore, report or isolate)\n"
		"hm_ca_flapping_action %s\n\n"
		"# Switch Flapping Link Action (ignore, report, isolate or no_discover)\n"
		"hm_sw_flapping_action %s\n\n"
		"# The number of sweeps in which the link was flapping\n"
		"hm_num_flapping_sweeps %d\n\n"
		"# The number of sweeps of which any port exceeding\n"
		"# hm_num_flapping_sweeps is declared unhealthy\n"
		"hm_num_flapping_sweeps_window %d\n",
		osm_hm_action_str[p_opts->hm_ca_flapping_action],
		osm_hm_action_str[p_opts->hm_sw_flapping_action],
		p_opts->hm_num_flapping_sweeps,
		p_opts->hm_num_flapping_sweeps_window);

	fprintf(out,
		"#\n# Unhealthy Ports Illegal condition options\n#\n"
		"# CA/RTR Illegal Action (ignore, report or isolate)\n"
		"hm_ca_illegal_action %s\n\n"
		"# Switch Illegal Action (ignore, report, isolate or no_discover)\n"
		"hm_sw_illegal_action %s\n\n"
		"# Number of illegal SMPs a port may return to be declared\n"
		"# unhealthy\n"
		"hm_num_illegal %d\n",
		osm_hm_action_str[p_opts->hm_ca_illegal_action],
		osm_hm_action_str[p_opts->hm_sw_illegal_action],
		p_opts->hm_num_illegal);

	fprintf(out,
		"#\n# Unhealthy Ports Manual condition options\n#\n"
		"# CA/RTR Manual Action (ignore, report or isolate)\n"
		"hm_ca_manual_action %s\n\n"
		"# Switch Manual Action (ignore, report, isolate or no_discover)\n"
		"hm_sw_manual_action %s\n",
		osm_hm_action_str[p_opts->hm_ca_manual_action],
		osm_hm_action_str[p_opts->hm_sw_manual_action]);

	fprintf(out, "\n");

	fprintf(out,
		"# Prefix routes file name\n"
		"prefix_routes_file %s\n\n",
		p_opts->prefix_routes_file);

	fprintf(out,
		"#\n# IPv6 Solicited Node Multicast (SNM) Options\n#\n"
		"consolidate_ipv6_snm_req %s\n\n",
		p_opts->consolidate_ipv6_snm_req ? "TRUE" : "FALSE");

	fprintf(out,
		"#\n# Multicast IPv4 Mux mask\n#\n"
		"# e.g. multiplexing mask for IP to MAC mapping use 0x007fffff\n"
		"consolidate_ipv4_mask 0x%08X\n\n",
		p_opts->consolidate_ipv4_mask);

#ifdef ENABLE_MLNX_OS
	fprintf(out,
		"#\n# Soft limit of number of CA ports that OpenSM is allowed\n"
		"# to manage (0 for unlimited). Must be <= ca_port_num_limit\n#\n"
		"ca_port_soft_limit %d\n\n",
		p_opts->ca_port_soft_limit);

	fprintf(out,
		"#\n# Max number of CA ports that OpenSM is allowed\n"
		"# to manage (0 for unlimited)\n#\n"
		"ca_port_num_limit %d\n\n",
		p_opts->ca_port_num_limit);

	fprintf(out,
		"# Switch name\n"
		"switch_name %s\n\n",
		p_opts->switch_name);
#endif
	fprintf(out, "# Log prefix\nlog_prefix %s\n\n", p_opts->log_prefix);

	fprintf(out,
		"# Max Failed Sequential Discovery Loops\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_seq_redisc %d\n\n",
		p_opts->max_seq_redisc);

	fprintf(out,
		"# Send client reregister on Alias GUID migration\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"rereg_on_guid_migr %s\n\n",
		p_opts->rereg_on_guid_migr ? "TRUE" : "FALSE");

	fprintf(out,
		"# Sending GID IN/OUT notices on Alias GUIDs reregister/delete\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"aguid_inout_notice %s\n\n",
		p_opts->aguid_inout_notice ? "TRUE" : "FALSE");

	fprintf(out,
		"# SM assigned Alias GUIDs algorithm\n"
		"sm_assign_guid_func %s\n\n",
		p_opts->sm_assign_guid_func);

	fprintf(out,
		"# Maximum number of attempts to find alternative direct route\n"
		"# towards unresponsive ports\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_alt_dr_path_retries %d\n\n",
		p_opts->max_alt_dr_path_retries);

	fprintf(out,
		"# Maximum topologies sw may be part of\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"max_topologies_per_sw %u\n\n",
		p_opts->max_topologies_per_sw);

	fprintf(out,
		"# Auto join HCA without topology to its sw topology\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"auto_join_hca_wo_topo %s\n\n",
		p_opts->auto_join_hca_wo_topo ? "TRUE" : "FALSE");

	fprintf(out,
		"# Enable incremental multicast routing\n"
		"enable_inc_mc_routing %s\n\n",
		p_opts->enable_inc_mc_routing ? "TRUE" : "FALSE");

	fprintf(out,
		"# Allow resetting SM port after changing port's properties\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"allow_sm_port_reset %s\n\n",
		p_opts->allow_sm_port_reset ? "TRUE" : "FALSE");

	fprintf(out,
		"# Reset port after changing port's enabled\n"
		"# speed (no_reset, logical, physical)\n"
		"port_speed_change_action %s\n\n",
		osm_port_reset_type_str(p_opts->port_speed_change_action));

	fprintf(out,
		"# Reset port after changing port's enabled\n"
		"# extended speed (no_reset, logical, physical)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"port_ext_speed_change_action %s\n\n",
		osm_port_reset_type_str(p_opts->port_ext_speed_change_action));

	fprintf(out,
		"# Reset port after changing port's enabled\n"
		"# extended speed 2 (no_reset, logical, physical)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"port_ext_speed2_change_action %s\n\n",
		osm_port_reset_type_str(p_opts->port_ext_speed2_change_action));

	fprintf(out,
		"# Reset port after changing port's enabled\n"
		"# Mellanox extended speed (no_reset, logical, physical)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"port_mepi_speed_change_action %s\n\n",
		osm_port_reset_type_str(p_opts->port_mepi_speed_change_action));

	fprintf(out,
		"# Reset port after changing port's MTU\n"
		"# (no_reset, logical, physical)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"port_mtu_change_action %s\n\n",
		osm_port_reset_type_str(p_opts->port_mtu_change_action));

	fprintf(out,
		"# Reset port after changing port's Operational VLs\n"
		"# (no_reset, logical, physical)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"port_vl_change_action %s\n\n",
		osm_port_reset_type_str(p_opts->port_vl_change_action));

	fprintf(out,
		"# Support Mellanox enhanced link mode\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"support_mlnx_enhanced_link %s\n\n",
		p_opts->support_mlnx_enhanced_link ? "TRUE" : "FALSE");

	fprintf(out,
		"# Enable/Disable Mellanox enhanced link option on IB ports\n"
		"mlnx_enhanced_link_enable %s\n\n",
		p_opts->mlnx_enhanced_link_enable ? "TRUE" : "FALSE");

	fprintf(out,
		"# Adaptive Timeout SL mask\n"
		"# Allow QPs on this SLs to implement adaptive timeout\n"
		"adaptive_timeout_sl_mask 0x%x\n\n",
		p_opts->adaptive_timeout_sl_mask);

	fprintf(out,
		"# Reset port after changing port's AME bit\n"
		"# (no_reset, logical, physical)\n"
		"port_ame_bit_change_action %s\n\n",
		osm_port_reset_type_str(p_opts->port_ame_bit_change_action));

	fprintf(out,
		"# Enable caching of MLNX ExtendedPortInfo information\n"
		"# When caching is enabled, OpenSM will not send MEPI(Get)\n"
		"# during heavy sweep to the existing ports in ACTIVE state.\n"
		"mepi_cache_enabled %s\n\n",
		p_opts->mepi_cache_enabled ? "TRUE" : "FALSE");

	fprintf(out,
		"# Virtualization support\n"
		"# 0: Ignore Virtualization - No virtualization support\n"
		"# 1: Disable Virtualization - Disable virtualization on all\n"
		"#			       Virtualization supporting ports\n"
		"# 2: Enable Virtualization - Enable virtualization on all\n"
		"#			      Virtualization supporting ports\n"
		"virt_enabled %d\n\n",
		p_opts->virt_enabled);

	fprintf(out,
		"# Maximum number of ports to be processed simultaneously\n"
		"# by Virtualization Manager (0 - process all pending ports)\n"
		"virt_max_ports_in_process %d\n\n",
		p_opts->virt_max_ports_in_process);

	fprintf(out,
		"# Default value for hop limit to be returned in path records\n"
		"# where either the source or destination are virtual ports\n"
		"virt_default_hop_limit %d\n\n",
		p_opts->virt_default_hop_limit);

	fprintf(out,
		"# Parameter indicates whether PIR and NR responses shall contain the virtual port\n"
		"# records and virtual node records in addition to the physical records.\n"
		"# Notice: only virtual ports and virtual nodes with a unique LID\n"
		"# appear in these records.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"enable_virt_rec_ext %s\n\n",
		p_opts->enable_virt_rec_ext ? "TRUE" : "FALSE");

	fprintf(out,
		"# Provides better outgoing port selection when routing path-bit LIDs:\n"
		"# Base LID + N (N in 1..2^LMC-1)\n"
		"# Preserves balanced routing of HCA Base LIDs\n"
		"# Effective with minhop/updn routing engine\n"
		"# It is recommended to use scatter_ports 0 when using this option\n"
		"improved_lmc_path_distribution %s\n\n",
		p_opts->improved_lmc_path_distribution ? "TRUE" : "FALSE");

	fprintf(out,
		"# Log flags for redirection to the syslog\n"
		"# Default - 0x0 - only messages specifically marked by\n"
		"# OpenSM as directed to syslog will be redirected.\n"
		"syslog_log_flags 0x%02x\n\n",
		p_opts->syslog_log_flags);

	fprintf(out,
		"# Perform heavy sweep in case of HUP signal received and\n"
		"# no configuration change recognized\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"sweep_every_hup_signal %s\n\n",
		p_opts->sweep_every_hup_signal ? "TRUE" : "FALSE");

	fprintf(out,
		"# Time interval [in min] between statistics dumps.\n"
		"# The value 0 implies no statistic dump\n"
		"# Max value is 71,582\n"
		"osm_stats_interval %d\n\n",
		p_opts->osm_stats_interval);

	fprintf(out,
		"# Max size [in MB] of statistic dump file.\n"
		"# The value 0 implies no size limitation.\n"
		"# Max value is 4095 (4GB).\n"
		"osm_stats_dump_limit %lu\n\n",
		p_opts->osm_stats_dump_limit = OSM_DEFAULT_STATISTIC_DUMP_FILE_SIZE);
	fprintf(out,
		"# Max size [in MB] of perflog dump file.\n"
		"# The value 0 implies no size limitation.\n"
		"# Max value is 4095 (4GB).\n"
		"osm_perflog_dump_limit %lu\n\n",
		p_opts->osm_perflog_dump_limit = OSM_DEFAULT_PERFLOG_DUMP_FILE_SIZE);
	fprintf(out,
		"# Enable performance logging.\n"
		"# when enabled, SM dumps the time it took for specific stages.\n"
		"enable_performance_logging %s\n\n",
		p_opts->enable_performance_logging ? "TRUE" : "FALSE");
	fprintf(out,
		"# Indication whether to dump MADs statistics per SM port\n"
		"osm_stats_dump_per_sm_port %s\n\n",
		p_opts->osm_stats_dump_per_sm_port ? "TRUE" : "FALSE");
	fprintf(out,
		"# Enable router alias GUID configuration\n"
		"# Values are\n"
		"#    0: Ignore Router Mode Alias GUIDs.\n"
		"#    1: Enable -  Configure Router Mode Alias GUIDs.\n"
		"#    2: Disable - Clear Router Mode Alias GUIDs.\n"
		"rtr_aguid_enable %u\n\n",
		p_opts->rtr_aguid_enable);
	fprintf(out,
		"# Router inter-subnet PathRecord FlowLabel\n"
		"rtr_pr_flow_label %u\n\n",
		p_opts->rtr_pr_flow_label);
	fprintf(out,
		"# Router inter-subnet PathRecord TClass\n"
		"rtr_pr_tclass %u\n\n",
		p_opts->rtr_pr_tclass);
	fprintf(out,
		"# Router inter-subnet PathRecord SL\n"
		"rtr_pr_sl %u\n\n",
		p_opts->rtr_pr_sl);
	fprintf(out,
		"# Router inter-subnet PathRecord MTU\n"
		"# 255 stands for calculating the maximal common MTU according "
		"to the path between end port and router in the local subnet.\n"
		"rtr_pr_mtu %u\n\n",
		p_opts->rtr_pr_mtu);
	fprintf(out,
		"# Router inter-subnet PathRecord Rate\n"
		"# 255 stands for calculating the maximal common rate according to"
		" the path between the end port and router in the local subnet.\n"
		"rtr_pr_rate %u\n\n",
		p_opts->rtr_pr_rate);

	fprintf(out,
		"# SHArP support\n"
		"# 0: Ignore SHArP - No SHArP support\n"
		"# 1: Disable SHArP - Disable SHArP on all supporting switches\n"
		"# 2: Enable SHArP - Enable SHArP on all supporting switches\n"
		"sharp_enabled %d\n\n",
		p_opts->sharp_enabled);

	fprintf(out,
		"# Default value for hop limit to be returned in path records\n"
		"# where either the source or destination are alias GUID\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"aguid_default_hop_limit %d\n\n",
		p_opts->aguid_default_hop_limit);

	fprintf(out,
		"# ROUTER POLICY OPTIONS\n"
		"router_policy_enabled %s\n"
		"router_policy_file  %s\n\n",
		p_opts->router_policy_enabled ? "TRUE": "FALSE",
		p_opts->router_policy_file);

	fprintf(out,
		"# The file holding the log verbosity bypass policy\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"verbose_bypass_policy_file %s\n\n",
		p_opts->verbose_bypass_policy_file ? p_opts->verbose_bypass_policy_file : null_str);

	fprintf(out,
		"# Full path file name for SA DB dump file\n"
		"sa_db_dump_file %s\n\n",
		p_opts->sa_db_dump_file ? p_opts->sa_db_dump_file : null_str);

	fprintf(out,
		"# Enable dumping subnet LST file\n"
		"enable_subnet_lst %s\n\n",
		p_opts->enable_subnet_lst ? "TRUE" : "FALSE");

	fprintf(out,
		"# Full path file name for subnet LST dump file\n"
		"subnet_lst_file %s\n\n",
		p_opts->subnet_lst_file ? p_opts->subnet_lst_file : null_str);

	fprintf(out,
		"# Full path file name for virtualization dump file\n"
		"virt_dump_file %s\n\n",
		p_opts->virt_dump_file ? p_opts->virt_dump_file : null_str);

	fprintf(out,
		"# DOR routing hyper-cube mode\n"
		"dor_hyper_cube_mode %s\n\n",
		p_opts->dor_hyper_cube_mode ? "TRUE" : "FALSE");

	fprintf(out,
		"# Additional devices which support general info\n"
		"# List of device IDs per vendor ID using the following format:\n"
		"# <dev_id> = <dev ID value> or device ID range <min_dev_id-max_dev_id>\n"
		"# <specific vendor devices list> = <vend_id>:<dev_id>,<dev-id>\n"
		"# <any vendor devices list> = <dev_id>,<dev-id>\n"
		"# <vendor devices list> = either <specific vendor devices list> or <any vendor devices list>\n"
		"# Multi vendors list = [<vendor devices list>][<vendor devices list>]\n"
		"# Example: To define dev_id 0x20 and 0x22 for vendor 0x0001 and\n"
		"#	    dev IDs 0x30-0x40 for any vendor use the below format:\n"
		"#		[0x0001:0x20,0x22][0x30-0x40]\n"
		"additional_gi_supporting_devices %s\n\n",
		p_opts->additional_gi_supporting_devices ?
		p_opts->additional_gi_supporting_devices : null_str);

	fprintf(out,
		"# Additional devices which support MlnxExtPortInfo\n"
		"# List of device IDs per vendor ID using the following format:\n"
		"# <dev_id> = <dev ID value> or device ID range <min_dev_id-max_dev_id>\n"
		"# <specific vendor devices list> = <vend_id>:<dev_id>,<dev-id>\n"
		"# <any vendor devices list> = <dev_id>,<dev-id>\n"
		"# <vendor devices list> = either <specific vendor devices list> or <any vendor devices list>\n"
		"# Multi vendors list = [<vendor devices list>][<vendor devices list>]\n"
		"# Example: To define dev_id 0x20 and 0x22 for vendor 0x0001 and\n"
		"#	    dev IDs 0x30-0x40 for any vendor use the below format:\n"
		"#		[0x0001:0x20,0x22][0x30-0x40]\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"additional_mepi_force_devices %s\n\n",
		p_opts->additional_mepi_force_devices ?
		p_opts->additional_mepi_force_devices : null_str);

	fprintf(out,
		"# Disable default general info support\n"
		"# When this flag is set, OpenSM will send general info MADs only to\n"
		"# devices specified by additional_gi_supporting_devices parameter\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"disable_default_gi_support %s\n\n",
		p_opts->disable_default_gi_support ? "TRUE" : "FALSE");

	fprintf(out,
		"# The SM mode\n"
		"# 0: Standalone - SM can't manage InfiniBand fabric.\n"
		"#		  SM will manage SM node only when SM port is down\n"
		"# 1: Regular    - SM can manage InfiniBand fabric.\n"
		"#		  SM won't be active when SM port goes down\n"
		"# 2: Both       - SM can manage InfiniBand fabric.\n"
		"#		  SM keeps managing SM node when SM port goes down\n"
		"standalone %u\n\n", p_opts->stand_alone);

	fprintf(out,
		"# Comma-separated list of subjects to include in activity report\n"
		"# Possible subjects to include in the list:\n"
		"# mc, prtn, virt\n"
		"# 'all' - Include all subjects\n"
		"# 'none' - No subjects (Default)\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"activity_report_subjects %s\n\n", p_opts->activity_report_subjects);

	fprintf(out,
		"# AR SL mask - 16-bit bitmask indicating which SLs should be configured for AR\n"
		"ar_sl_mask 0x%04X\n\n",
		p_opts->ar_sl_mask);

	fprintf(out,
		"# Enable adaptive routing only to devices that support packet reordering.\n"
		"# When enabled, state in ARLFT entries for devices which do not support packet\n"
		"# reordering is set to static.\n"
		"# When disabled, ARLFT entries remain as determined by the routing engine.\n"
		"enable_ar_by_device_cap %s\n\n",
		p_opts->enable_ar_by_device_cap ? "TRUE" : "FALSE");

	fprintf(out,
		"# Enable adaptive routing group copy optimization.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"enable_ar_group_copy %s\n\n",
		p_opts->enable_ar_group_copy ? "TRUE" : "FALSE");

	fprintf(out,
		"# Advanced routing - Adaptive routing mode\n"
		"# Adaptive Routing enables the switch to select the output port based on the port's load.\n"
		"# Supported values:\n"
		"# 0 - Adaptive routing disabled.\n"
		"# 1 - Enable local adaptive routing (switches select exit port based on local buffer utilization).\n"
		"# 2 - Enable adaptive routing with notifications (deprecated).\n"
		"# 3 - Auto mode in which adaptive routing is determined by the routing engine. (default)\n"
		"# NOTE: Debug option. Changing the default value is not recommended.\n"
		"ar_mode %d\n\n",
		p_opts->ar_mode);

	fprintf(out,
		"# Advanced routing - Fast link fault recovery feature\n"
		"# The feature is required for traffic resiliency upon switch-to-switch link failures.\n"
		"# Supported values:\n"
		"# 0 - Fast link fault recovery disabled.\n"
		"# 1 - Enable local fast link fault recovery only.\n"
		"# 2 - Enable legacy fast link fault recovery with notifications (deprecated)\n"
		"# 3 - Auto mode in which enhanced fast link fault recovery support is determined by the routing engine. (default)\n"
		"#     ar_updn, ar_ftree and dfp2 will enable enhanced fast link fault recovery in the switches.\n"
		"#     Otherwise, local fast link fault recovery only will be enabled in the switches.\n"
		"# NOTE: Debug option. Changing the default value is not recommended.\n"
		"shield_mode %d\n\n",
		p_opts->shield_mode);

	fprintf(out,
		"# pFRN is supported by the following engines: ar_updn, ar_ftree, dfp2\n"
		"# SL for pFRN communication\n"
		"pfrn_sl %u\n\n"
		"# Time (in seconds) since last pFRN for a specific subgroup\n"
		"# was received, after which the entire mask must be cleared\n"
		"# Must be a multiple of 30 seconds\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"pfrn_mask_clear_timeout %u\n\n"
		"# Maximal time (in seconds) since last mask clear, after which mask\n"
		"# must be cleared.\n"
		"# Must be a multiple of 240 seconds\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"pfrn_mask_force_clear_timeout %u\n\n"
		"# Enable pFRN over routers\n"
		"# 0: Ignore - Do not change pFRN configuration on routers\n"
		"# 1: Disable - Disable pFRN over routers\n"
		"# 2: Enable - Enable pFRN over routers\n"
		"pfrn_over_router_enabled %d\n\n",
		p_opts->pfrn_sl,
		p_opts->pfrn_mask_clear_timeout,
		p_opts->pfrn_mask_force_clear_timeout,
		p_opts->pfrn_over_router_enabled);

	fprintf(out,
		"# Number of threads to be used for processing SMPs\n"
		"smp_threads %d\n\n",
		p_opts->smp_threads);

	fprintf(out,
		"# Number of threads to be used for processing GMPs\n"
		"gmp_threads %d\n\n",
		p_opts->gmp_threads);

	fprintf(out,
		"# Number of threads to be used for processing key violation trap mads\n"
		"# by VS and CC managers\n"
		"gmp_traps_threads_num %d\n\n",
		p_opts->gmp_traps_threads_num);

	fprintf(out,
		"# Maximum number of SMPs sent in parallel to the same port.\n"
		"# Currently, the supported MADs are:\n"
		"# portInfo/Extended portInfo, LFTs, AR LFTs,\n"
		"# AR group table, AR copy group table, RN subgroup direction,\n"
		"# SLVL table, VL Arbitration\n"
		"max_wire_smps_per_device %d\n\n",
		p_opts->max_wire_smps_per_device);

	fprintf(out,
		"# AR Asymmetric trees max flow algorithm\n"
		"# Supported values:\n"
		"# 0 - Disable the algorithm.\n"
		"# 1 - Enable with 1 subgroup support.\n"
		"# 2 - Enable with 2 subgroups on leaf switches.\n"
		"# 3 - Enable asymmetric tree algorithm (experimental).\n"
		"ar_tree_asymmetric_flow %d\n\n",
		p_opts->ar_tree_asymmetric_flow);
	fprintf(out,
		"# In use if ar_tree_asymmetric_flow is set to 3.\n"
		"# Threshold (percent) of BW drop between spine and core\n"
		"# links before excluding spines (experimental).\n"
		"# If a spine's spine to core BW percent drops below the threshold\n"
		"# due to link failures that spine will be eligible for\n"
		"# removal from the AR group.\n"
		"# Example: ar_tree_asymmetric_flow_threshold 34\n"
		"#          If spine has 32 links towards the core, then it would\n"
		"#          need to have 11 links removed before excluding the switch\n"
		"#          (34%% of 32 = 10.88)\n"
		"# 0 - Disable feature.\n"
		"ar_tree_asymmetric_flow_threshold %d\n\n",
		p_opts->ar_tree_asymmetric_flow_threshold);
	fprintf(out,
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"# Threshold limit of nodes to exclude (experimental).\n"
		"# If the number of eligible spines to be excluded, (as determined by\n"
		"# the ar_tree_asymmetric_flow_threshold parameter above),\n"
		"# is higher than this limit, no spines will be removed\n"
		"# The default value is 2\n"
		"ar_tree_asymmetric_flow_threshold_limit %d\n\n",
		p_opts->ar_tree_asymmetric_flow_threshold_limit);

	fprintf(out,
		"# Bit mask of flags to control various options and behavior\n"
		"# of routing engines.\n"
		"# Supported values:\n"
		"# 0x1 - Enable switch rank adjustments for tree based routing engines.\n"
		"routing_flags 0x%" PRIx64 "\n\n",
		p_opts->routing_flags);

	/* optional string attributes ... */
	fprintf(out,
		"# Advanced routing - Advanced routing engine\n"
		"# Supported values:\n"
		"# none - Advanced routing is not enabled.\n"
		"# ar_lag - Ports groups are created out of \"parallel\" links. "
		"Links that connect the same pair of switches.\n"
		"# ar_tree - All the ports with minimal hops to the destination are in the same group. "
		"Must run together with UPDN routing engine.\n"
		"# auto - The advanced routing engine is selected based on the routing engine. "
		"Works for ar_updn, ar_ftree, ar_torus, ar_dor engines.\n"
		"adv_routing_engine %s\n\n",
		p_opts->adv_routing_engine_name);

	fprintf(out,
		"# AR Transport mask - indicates which transport types are enabled for AR\n"
		"# Bit 0 = UD, Bit 1 = RC, Bit 2 = UC, Bit 3 = DCT, Bits 4-7 are reserved.\n"
		"ar_transport_mask 0x%04X\n\n",
		p_opts->ar_transport_mask);

	fprintf(out,
		"# Enable adaptive routing data dump to file.\n"
		"dump_ar %s\n\n",
		p_opts->dump_ar ? "TRUE" : "FALSE");

	fprintf(out,
		"# Load GUID to AR group ID from cache file.\n"
		"# When enabled, it can reduce AR group configuration changes after restart.\n"
		"cache_ar_group_id %s\n\n",
		p_opts->cache_ar_group_id ? "TRUE" : "FALSE");

	fprintf(out,
		"# Mode of taking down up turns in intermediate groups in dfp routing engine\n"
		"# Supported values:\n"
		"# 0 - Allow down up turns in intermediate groups.\n"
		"#     Avoid using this value with shield based on pFRN.\n"
		"# 1 - Disable all down up turns.\n"
		"# 2 - Allow down up turns through intermediate spine,"
		" when there is no direct route to the destination group.\n"
		"dfp_down_up_turns_mode %d\n\n",
		p_opts->dfp_down_up_turns_mode);

	fprintf(out,
		"# Enable VL packing from CA to switch.\n"
		"# When enabled, use single VL per QoS on traffic between CA and switch.\n"
		"enable_vl_packing %s\n\n",
		p_opts->enable_vl_packing ? "TRUE" : "FALSE");

	fprintf(out,
		"# The file holding map from service name to service key.\n"
		"service_name2key_map_file %s\n\n",
		p_opts->service_name2key_map_file ?
		   p_opts->service_name2key_map_file : null_str);

	fprintf(out,
		"# The file holding the topo configuration.\n"
		"topo_config_file %s\n\n",
		p_opts->topo_config_file ? p_opts->topo_config_file : null_str);

	fprintf(out,
		"# If set to true, the SM will adjust its operational\n"
		"# mode to take into account the topo_config file.\n"
		"topo_config_enabled %s\n\n",
		p_opts->topo_config_enabled ? "TRUE" : "FALSE");

	fprintf(out,
		"# Hash function to be used by router selection algorithm.\n"
		"# Supported values: crc32\n"
		"rtr_selection_function %s\n\n",
		subn_selection_function_type_str(p_opts->rtr_selection_function));

	fprintf(out,
		"# Parameter used by selection function for router selection\n"
		"rtr_selection_seed 0x%016" PRIx64 "\n\n",
		p_opts->rtr_selection_seed);

	fprintf(out,
		"# Parameters to be used by selection function for selecting routing path.\n"
		"# Should be in comma-separated values format.\n"
		"# Supported values: sgid, dgid\n"
		"rtr_selection_algo_parameters %s\n\n",
		subn_rtr_selection_algo_parameters_str(p_opts->rtr_selection_params_mask));

	fprintf(out,
		"# Get MFT tables from switches on first master sweep\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"get_mft_tables %s\n\n",
		p_opts->get_mft_tables ? "TRUE" : "FALSE");

	fprintf(out,
		"# HBF (Hash-Based Forwarding) SL mask - 16-bit bitmask indicating "
		"which SLs should be configured for HBF\n"
		"hbf_sl_mask 0x%04X\n\n",
		p_opts->hbf_sl_mask);

	fprintf(out,
		"# HBF (Hash-Based Forwarding) hash type: 0 - CRC, 1 - XOR\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"hbf_hash_type %u\n\n",
		p_opts->hbf_hash_type);

	fprintf(out,
		"# HBF (Hash-Based Forwarding) hash seed type: 0 - Configurable, 1 - Random\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"hbf_seed_type %u\n\n",
		p_opts->hbf_seed_type);

	fprintf(out,
		"# HBF (Hash-Based Forwarding) hash seed. Values 0 - 0xFFFFFFFF\n"
		"# 0xFFFFFFFF stands for taking the 32 LSB of the node GUID as HBF hash seed.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"hbf_seed 0x%08" PRIX32 "\n\n",
		p_opts->hbf_seed);

	fprintf(out,
		"# HBF (Hash-Based Forwarding) hash fields - 64-bit bitmask indicating\n"
		"# the fields that affect hash calculation.\n"
		"# Bit 1: DLID\n"
		"# Bit 2: SL\n"
		"# Bit 3: VL\n"
		"# Bit 4: LRH LNH\n"
		"# GRH fields:\n"
		"# Bit 10: SGID\n"
		"# Bit 11: DGID\n"
		"# Layer4 fields:\n"
		"# Bit 20: BTH destination QP\n"
		"# Bit 21: DETH source QP\n"
		"# Bit 22: DCETH V1 source QP\n"
		"# Bit 23: DCETH V1 ISID\n"
		"# Bit 30: Input port\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"hbf_hash_fields 0x%016" PRIX64 "\n\n",
		cl_ntoh64(p_opts->hbf_hash_fields));

	fprintf(out,
		"# Weighted Hash-Based Forwarding (WHBF) configures the ratio\n"
		"# between three subgroups of the Adaptive Routing (AR) Group.\n"
		"# format is a tuple of three integers in a range 0 to 0xffff:\n"
		"#  <subgroup 0 weight>,<subgroup 1 weight>,<subgroup 2 weight>\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"hbf_weights %s\n\n",
		p_opts->hbf_weights);

#if OSM_ENABLE_NVLINK
	fprintf(out,
		"# Enable NVLink subnet configuration. When enabled, SM discovers\n"
		"# GPUs and set planes and rails according to the predefined policy.\n"
		"nvlink_enable %s\n\n",
		p_opts->nvlink_enable ? "TRUE" : "FALSE");

	fprintf(out,
		"# If TRUE, cause all NVLink rails to be reassigned\n"
		"nvl_reassign_rails %s\n\n",
		p_opts->nvl_reassign_rails ? "TRUE" : "FALSE");

	fprintf(out,
		"# NVL HBF (NVLink Hash-Based Forwarding)\n"
		"# Supported values for hash type: 0 - CRC, 1 - Packet Hash\n"
		"nvl_hbf_hash_type %u\n\n",
		p_opts->nvl_hbf_hash_type);

	fprintf(out,
		"# NVL HBF (NVLink Hash-Based Forwarding) bitmask of the packet hash\n"
		"# field to be used when calculating the hash for the NVLink HBF.\n"
		"# The bits that are set shall be consecutive.\n"
		"# Reserved when hash_type = 0 (CRC)\n"
		"nvl_hbf_packet_hash_bitmask %u\n\n",
		p_opts->nvl_hbf_packet_hash_bitmask);

	fprintf(out,
		"# NVL HBF (NVLink Hash-Based Forwarding) seed.\n"
		"# Reserved when hash_type = 1 (Packet Hash)\n"
		"nvl_hbf_seed 0x%08" PRIX32 "\n\n",
		p_opts->nvl_hbf_seed);

	fprintf(out,
		"# NVL HBF (NVLink Hash-Based Forwarding) hash fields - 64-bit bitmask indicating\n"
		"# the fields that affect hash calculation.\n"
		"# Reserved when hash_type = 1 (Packet Hash)\n"
		"# NVLRH fields for UC:\n"
		"# Bit 0: SLID\n"
		"# Bit 1: DLID\n"
		"# Bit 2: SL\n"
		"# Bit 3: VL\n"
		"# Bit 4: pipe_id\n"
		"# Bit 5: nvlrh_hash[0]\n"
		"# Bit 6: nvlrh_hash[1]\n"
		"# Bit 7: nvlrh_hash[2]\n"
		"# Bit 8: nvlrh_hash[3]\n"
		"# Bit 9: nvlrh_hash[4]\n"
		"# Bit 10: nvlrh_hash[5]\n"
		"# Bit 11: nvlrh_hash[6]\n"
		"# Bit 12: nvlrh_hash[7]\n"
		"# Bit 16: Input port\n"
		"# NVLRH fields for MC:\n"
		"# Bit 32: SLID\n"
		"# Bit 33: DLID\n"
		"# Bit 34: SL\n"
		"# Bit 35: VL\n"
		"# Bit 36: pipe_id\n"
		"# Bit 37: nvlrh_hash[0]\n"
		"# Bit 38: nvlrh_hash[1]\n"
		"# Bit 39: nvlrh_hash[2]\n"
		"# Bit 40: nvlrh_hash[3]\n"
		"# Bit 41: nvlrh_hash[4]\n"
		"# Bit 42: nvlrh_hash[5]\n"
		"# Bit 43: nvlrh_hash[6]\n"
		"# Bit 44: nvlrh_hash[7]\n"
		"# Bit 48: Input port\n"
		"nvl_hbf_hash_fields 0x%016" PRIX64 "\n\n",
		p_opts->nvl_hbf_fields_enable);

	fprintf(out,
		"# Time [in STO units] that SM should wait between configuring partition routing\n"
		"# tables to activating the partition. Supported values range [1,100]\n"
		"nvl_partition_reroute_wait %u\n\n",
		p_opts->nvl_partition_reroute_wait);

#endif

	fprintf(out,
		"# Sets the number of LIDs per aggregation node connected to a prism switch.\n"
		"# Should be set to log2(number of planes per prism switch).\n"
		"# If global LMC or lids_per_rtr are not 0, prism_an_lmc is limited by global LMC.\n"
		"# If set to 0, OpenSM will use number of LIDs according to global LMC.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"prism_an_lmc %u\n\n",
		p_opts->prism_an_lmc);

	fprintf(out,
		"# If set to true, the SM will enable the tenants manager\n"
		"tenants_policy_enabled %s\n\n",
		p_opts->tenants_policy_enabled ? "TRUE" : "FALSE");
	fprintf(out,
		"# The file holding the tenant configuration.\n"
		"tenants_policy_file %s\n\n",
		p_opts->tenants_policy_file ? p_opts->tenants_policy_file : null_str);

	fprintf(out,
		"# When enabled, SM replies to LID-routed SMPs with direct-route.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"reply_lid_smps_in_dr %s\n\n",
		p_opts->reply_lid_smps_in_dr ? "TRUE" : "FALSE");

	fprintf(out,
		"# If set to true, SM can respond to LID-routed traps sent by unknown senders.\n"
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"respond_unknown_lid_traps %s\n\n",
		p_opts->respond_unknown_lid_traps ? "TRUE" : "FALSE");

	fprintf(out,
		"# Fabric Mode feature\n"
		"# Indicates which profile from fabric mode policy file to use.\n"
		"# If set to none, feature is disabled, and SM will not change current\n"
		"# device configuration\n"
		"fabric_mode_profile %s\n\n"
		"# The file holding the Fabric Mode policy\n"
		"fabric_mode_policy_file %s\n\n",
		p_opts->fabric_mode_profile ? p_opts->fabric_mode_profile : null_str,
		p_opts->fabric_mode_policy_file ? p_opts->fabric_mode_policy_file : null_str);

	fprintf(out,
		"# ISSU feature mode\n"
		"# Supported values:\n"
		"# 0 - Disable the ISSU feature.\n"
		"# 1 - Enable ISSU in active mode (full ISSU capabilities are enabled).\n"
		"# 2 - Enable ISSU in passive mode (monitor, but don't initiate upgrade).\n"
		"issu_mode %d\n\n",
		p_opts->issu_mode);
	fprintf(out,
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"# Minimum wait time in seconds between SM receiving\n"
		"# an upgrade request and SM initiating the upgrade.\n"
		"# If set to 0, the SM will use the builtin default.\n"
		"issu_pre_upgrade_time %d\n\n",
		p_opts->issu_pre_upgrade_time);
	fprintf(out,
		"# NOTE: Debug option. Changing the value is not recommended.\n"
		"# Upgrade timeout wait time in seconds.\n"
		"# If set to 0, the SM will use the builtin default\n"
		"issu_timeout %d\n",
		p_opts->issu_timeout);

	/* optional string attributes ... */
}

int osm_subn_write_conf_file(char *file_name, IN osm_subn_opt_t * p_opts)
{
	FILE *opts_file;

	opts_file = fopen(file_name, "w");
	if (!opts_file) {
		printf("cannot open file \'%s\' for writing: %s\n",
			file_name, strerror(errno));
		return -1;
	}

	osm_subn_output_conf(opts_file, p_opts);

	fclose(opts_file);

	return 0;
}

void osm_update_allowed_sm_guids(osm_subn_t *p_subn, const char *str_sm_guids)
{
	char *val_str, *str, *p;
	uint64_t val_ho = 0;
	uint32_t elements_num = 0;
	int ret = 0;
	cl_u64_vector_t *p_vector = &p_subn->allowed_sm_guid_tbl;

	if (str_sm_guids && *str_sm_guids) {
		/*
		 * Insert GUID 0 (invalid value to not overlap with regular GUIDs)
		 * to indicate that the feature is enabled.
		 * This way if the list is not empty the feature is enabled.
		 */
		cl_u64_vector_insert(p_vector, 0, NULL);
		str = strdup(str_sm_guids);
		val_str = strtok_r(str, ", \t\n", &p);
		while (val_str && *val_str) {
			ret = opts_strtoull(&val_ho, val_str, "allowed_sm_guids", UINT64_MAX);
			if (ret || !(val_ho)) {
				if (!ret && !val_ho && !elements_num) {
					val_str = strtok_r(NULL, ", \t\n", &p);
					if (!(val_str && *val_str))
						break;
				}

				/* If the input is not valid, return to default value */
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7549: allowed_sm_guids param is not valid. "
					"Should be GUIDs in comma separated values format. "
					"Disabling the feature\n");
				cl_u64_vector_remove_all(p_vector);
				break;
			}

			cl_u64_vector_insert(p_vector, cl_hton64(val_ho), NULL);
			elements_num++;
			val_str = strtok_r(NULL, ", \t\n", &p);
		}
		free(str);
	}

	cl_u64_vector_sort(p_vector);
}

boolean_t lookup_held_back_switch(IN osm_subn_t *subn, IN osm_node_t * p_node)
{
	osm_port_t *smport;
	osm_node_t *p_remote_node;

	cl_map_item_t *p_item = cl_qmap_get(&subn->held_back_sw_tbl,
					    osm_node_get_node_guid(p_node));
	if (p_item == cl_qmap_end(&subn->held_back_sw_tbl))
		p_item = cl_qmap_get(&subn->held_back_sw_tbl,
				     p_node->node_info.sys_guid);
	if (p_item == cl_qmap_end(&subn->held_back_sw_tbl))
		return FALSE;

	smport = osm_get_port_by_guid(subn, subn->sm_port_guid);
	OSM_ASSERT(smport);
	if (osm_node_get_type(smport->p_node) == IB_NODE_TYPE_SWITCH) {
		if (smport->p_node == p_node)
			return FALSE;
	} else {
		if (smport->p_physp->p_remote_physp) {
			p_remote_node = smport->p_physp->p_remote_physp->p_node;
			if (p_remote_node == p_node)
				return FALSE;
		}
	}
	return TRUE;
}

int osm_subn_add_leaf(IN OUT osm_subn_t * p_subn, IN osm_switch_t * p_sw)
{
	OSM_ASSERT(p_sw);

	if (cl_list_insert_tail(&p_subn->leaf_switches_list, p_sw)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7523: Not enough memory for leaf switches array\n");
		goto Exit;
	}

	return 0;
Exit:
	return -1;

}

void osm_subn_clear_leafs(IN OUT osm_subn_t * p_subn)
{
	cl_list_remove_all(&p_subn->leaf_switches_list);
}

uint16_t osm_subn_get_max_mlid(IN osm_subn_t * p_subn)
{
	return p_subn->max_mlid;
}

void osm_subn_set_mbox(IN osm_subn_t * p_subn, struct osm_mgrp_box * mbox)
{
	p_subn->mboxes[mbox->mlid - IB_LID_MCAST_START_HO] = mbox;
	if (p_subn->max_mlid < mbox->mlid)
		p_subn->max_mlid = mbox->mlid;
}

void osm_subn_remove_mbox(IN osm_subn_t * p_subn, uint16_t mlid)
{
	int i;
	p_subn->mboxes[mlid - IB_LID_MCAST_START_HO] = NULL;
	if (p_subn->max_mlid == mlid) {
		for (i = mlid - IB_LID_MCAST_START_HO; i >= 0; i--) {
			if (p_subn->mboxes[i])
				break;
		}
		p_subn->max_mlid = i + IB_LID_MCAST_START_HO;
	}
}

void osm_subn_mark_mbox_by_sw(IN osm_subn_t * p_subn, IN osm_switch_t * p_sw,
			      IN uint8_t port_num)
{
	const osm_mcast_tbl_t *p_tbl = osm_switch_get_mcast_tbl_ptr(p_sw);
	osm_mgrp_box_t *mbox;
	uint16_t max_mlid;
	int i;

	if (p_subn->force_multicast_rerouting)
		return;

	max_mlid = MIN(p_tbl->max_mlid_ho, p_subn->max_mlid);
	for (i = max_mlid - IB_LID_MCAST_START_HO; i >= 0; i--) {
		mbox = (osm_mgrp_box_t *) p_subn->mboxes[i];
		if (!mbox || mbox->imr_state == OSM_MGRP_BOX_IMR_NOT_SUPPORTED)
			continue;

		/* If port_num is 0, check all the ports of the switch */
		if (port_num) {
			if (osm_mcast_tbl_is_port(p_tbl,
						  i + IB_LID_MCAST_START_HO,
						  port_num))
				mbox->imr_state =
				    OSM_MGRP_BOX_IMR_NOT_SUPPORTED;
		} else {
			if (osm_mcast_tbl_is_any_port(p_tbl,
						      i + IB_LID_MCAST_START_HO))
				mbox->imr_state =
				    OSM_MGRP_BOX_IMR_NOT_SUPPORTED;
		}
	}
}

void osm_subn_clear_mtrees(IN osm_subn_t * p_subn)
{
	int i;
	for (i = 0; i <= p_subn->max_mcast_lid_ho - IB_LID_MCAST_START_HO; i++)
		if (p_subn->mboxes[i])
			osm_mgrp_box_clear_mtree(p_subn, p_subn->mboxes[i]);
}

void osm_subn_clear_mc_roots_tbl(IN osm_subn_t * p_subn)
{
	cl_fmap_item_t *p_item;
	cl_fmap_item_t *p_next_item;
	osm_mc_roots_list_t *p_mc_roots_list;

	p_next_item = cl_fmap_head(&p_subn->mc_roots_tbl);
	while (p_next_item != cl_fmap_end(&p_subn->mc_roots_tbl)) {
		p_item = p_next_item;
		p_next_item = cl_fmap_next(p_item);
		cl_fmap_remove_item(&p_subn->mc_roots_tbl, p_item);
		p_mc_roots_list = (osm_mc_roots_list_t *)p_item;
		cl_list_remove_all(&p_mc_roots_list->switch_list);
		cl_list_destroy(&p_mc_roots_list->switch_list);
		free(p_item);
	}
}

static int subn_add_mc_root(IN osm_subn_t * p_subn, IN ib_gid_t * p_mgid,
			    IN osm_switch_t * p_sw)
{
	cl_fmap_item_t *p_item;

	cl_list_iterator_t list_iter;

	osm_mc_roots_list_t *p_mc_roots_list;
	char gid_str[INET6_ADDRSTRLEN];

	p_item = cl_fmap_get(&p_subn->mc_roots_tbl, p_mgid);
	if (p_item == cl_fmap_end(&p_subn->mc_roots_tbl)) {
		p_mc_roots_list =
		    (osm_mc_roots_list_t *)calloc(1, sizeof(osm_mc_roots_list_t));
		if (!p_mc_roots_list) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7529: Not enough memory to add mulitcast root\n");
			return -1;
		}
		if (cl_list_init(&p_mc_roots_list->switch_list, 1)) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 752A: Failed to initialize list\n");
			free(p_mc_roots_list);
			return -1;
		}

		memcpy(&p_mc_roots_list->mgid, p_mgid, sizeof(ib_gid_t));
		cl_fmap_insert(&p_subn->mc_roots_tbl, &p_mc_roots_list->mgid,
			       &p_mc_roots_list->map_item);
	} else
		p_mc_roots_list = (osm_mc_roots_list_t *)p_item;

	/* check that the switch is not in the switch */
	for (list_iter = cl_list_head(&p_mc_roots_list->switch_list);
	     list_iter != cl_list_end(&p_mc_roots_list->switch_list);
	     list_iter = cl_list_next(list_iter)) {
		if (p_sw == (osm_switch_t *)cl_list_obj(list_iter))
			return 0;
	}

	if (cl_list_insert_tail(&p_mc_roots_list->switch_list, p_sw)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 752B: Failed to add switch 0x%" PRIx64 " to "
			"multicast roots switches list for MGID %s\n",
			osm_node_get_node_guid_ho(p_sw->p_node),
			inet_ntop(AF_INET6, p_mgid->raw, gid_str,
				  sizeof(gid_str)));
			return -1;
	}

	return 0;
}

static inline
void regex_get_substr(char * dest, const char * source, regmatch_t * p_match)
{
	int len = p_match->rm_eo - p_match->rm_so;

	memcpy(dest, source + p_match->rm_so, len);
	dest[len] = '\0';
}

int osm_subn_parse_mc_roots_file(IN osm_subn_t * p_subn)
{
	char line[1024];
	char str[256];
	char *file_name;
	FILE *file;
	unsigned line_num;
	regex_t regex_comment, regex_line;
	regmatch_t regex_match[5];
	ib_gid_t mgid;
	static ib_gid_t mgid0 = {{ 0 }};
	ib_net64_t guid;
	cl_map_item_t *p_item;
	osm_switch_t *p_sw;
	int rc = 0;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* Clear MC roots table */
	osm_subn_clear_mc_roots_tbl(p_subn);

	file_name = p_subn->opt.mc_roots_file;
	if (!file_name) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"mc roots file name is not specified.\n");
		goto skip_parsing;
	}

	file = fopen(file_name, "r");
	if (!file) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 752C: Can't open mc roots file \'%s\'\n",
			file_name);
		rc = -1;
		goto skip_parsing;
	}

	/* Create regex for pattern for comment line */
	rc = regcomp(&regex_comment, "^\\s*\\(#.*\\)\\?$", 0);
	if (rc) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7530: Failed to initialize parser for mc roots "
			"file\n");
		fclose(file);
		goto skip_parsing;
	}
	/* Create regex for pattern <MGID> <GUID|NodeDescription> */
	rc = regcomp(&regex_line,
		     "^\\([0-9a-fA-F:]\\+\\)"
		     "\\s\\+"
		     "\\(\\(0x[0-9a-fA-F]\\+\\)\\|\\\"\\([^\\\"]\\+\\)\\\"\\)"
		     "\\s*$",
		     0);
	if (rc) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7531: Failed to initialize parser for mc roots "
			"file\n");
		regfree(&regex_comment);
		fclose(file);
		goto skip_parsing;
	}

	line_num = 0;
	while (fgets(line, sizeof(line) - 1, file) != NULL) {
		line_num++;
		p_sw = NULL;

		/* Comment or empty line */
		if (regexec(&regex_comment, line, 0, NULL, 0) == 0)
			continue;

		/* Read valid line */
		if (regexec(&regex_line, line, 5, &regex_match[0], 0) != 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 752D: Failed to parse file \'%s\'"
				"line %d, invalid line format\n",
				file_name, line_num);
			rc = -1;
			break;
		}

		/* Check line length */
		if (regex_match[0].rm_eo >= 256) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 752E: Failed to parse file \'%s\'"
				"line %d, line too long\n",
				file_name, line_num);
			rc = -1;
			break;
		}

		/* Parse MGID */
		regex_get_substr(str, line, &regex_match[1]);
		if (inet_pton(AF_INET6, str, &mgid) <= 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 752F: Failed to parse file \'%s\'"
				"line %d, invalid MGID\n",
				file_name, line_num);
			rc = -1;
			break;
		}

		/* Check if switch is given by GUID */
		if (regex_match[3].rm_so != -1) {
			regex_get_substr(str, line, &regex_match[3]);
			guid = cl_hton64(strtoull(str, NULL, 16));
			p_sw = osm_get_switch_by_guid(p_subn, guid);
			if (p_sw)
				subn_add_mc_root(p_subn, &mgid, p_sw);
			else {
				OSM_LOG(&p_subn->p_osm->log,
					OSM_LOG_VERBOSE,
					"Could not find switch with "
					"GUID 0x%" PRIx64 "\n",
					cl_ntoh64(guid));
			}
			continue;
		}

		/* Check if switch is given by NodeDescription */
		if (regex_match[4].rm_so != -1) {
			regex_get_substr(str, line, &regex_match[4]);
			for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
			     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
			     p_item = cl_qmap_next(p_item)) {
				p_sw = (osm_switch_t *)p_item;
				if (!strcmp(p_sw->p_node->print_desc, str))
					break;
				else
					p_sw = NULL;
			}

			if (p_sw)
				subn_add_mc_root(p_subn, &mgid, p_sw);
			else
				OSM_LOG(&p_subn->p_osm->log,
					OSM_LOG_VERBOSE,
					"Could not find switch with "
					"NodeDescription \'%s\'\n",
					str);
			continue;
		}
	}

	/* Release resources */
	fclose(file);
	regfree(&regex_comment);
	regfree(&regex_line);

	if (rc)
		osm_subn_clear_mc_roots_tbl(p_subn);

skip_parsing:

	/* Add primary and secondary roots as default roots */
	if (p_subn->opt.mc_primary_root_guid) {
		p_sw = osm_get_switch_by_guid(p_subn,
					      p_subn->opt.mc_primary_root_guid);
		if (p_sw)
			subn_add_mc_root(p_subn, &mgid0, p_sw);
		else
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Could not find primary mc root switch "
				"0x%" PRIx64 "\n",
				cl_ntoh64(p_subn->opt.mc_primary_root_guid));
	}

	if (p_subn->opt.mc_secondary_root_guid) {
		p_sw = osm_get_switch_by_guid(p_subn,
					      p_subn->opt.mc_secondary_root_guid);
		if (p_sw)
			subn_add_mc_root(p_subn, &mgid0, p_sw);
		else
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Could not find secondary mc root switch "
				"0x%" PRIx64 "\n",
				cl_ntoh64(p_subn->opt.mc_secondary_root_guid));
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return rc;
}

void osm_subn_clear_service_name_key_tbl(IN osm_subn_t * p_subn)
{
	cl_fmap_item_t *p_item;
	cl_fmap_item_t *p_next_item;

	p_next_item = cl_fmap_head(&p_subn->service_keys_tbl);
	while (p_next_item != cl_fmap_end(&p_subn->service_keys_tbl)) {
		p_item = p_next_item;
		p_next_item = cl_fmap_next(p_item);
		cl_fmap_remove_item(&p_subn->service_keys_tbl, p_item);
		free(p_item);
	}
}

static int subn_add_service_name(IN osm_subn_t * p_subn,
				 IN const char * service_name,
				 IN uint8_t * service_key,
				 IN const char * service_key_str)
{
	cl_fmap_item_t *p_item;
	osm_service_key_t *p_service_key;
	char service_key_str_exist[INET6_ADDRSTRLEN] = {'\0'};
	const char *result;

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"Add service name %s with key %s.\n",
		service_name, service_key_str);

	p_item = cl_fmap_get(&p_subn->service_keys_tbl, service_name);
	if (p_item == cl_fmap_end(&p_subn->service_keys_tbl)) {
		p_service_key =
		    (osm_service_key_t *)calloc(1, sizeof(osm_service_key_t));
		if (!p_service_key) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7545: Not enough memory to add service name key\n");
			return -1;
		}
		strncpy(p_service_key->service_name, service_name,
			IB_SERVICE_NAME_SIZE - 1);
		memcpy(p_service_key->service_key, service_key,
		    sizeof(p_service_key->service_key));
		cl_fmap_insert(&p_subn->service_keys_tbl,
			       p_service_key->service_name,
			       &p_service_key->map_item);
	} else {
		 /* Service name already exists. */
		p_service_key = (osm_service_key_t *)p_item;
		result = inet_ntop(AF_INET6, p_service_key->service_key,
				   service_key_str_exist, sizeof(service_key_str_exist));
		if (result) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
				"Service with name '%s' already exists "
				"with key %s. Multiple keys are not supported. "
				"The key will be replaced with the new value.\n",
				service_name, service_key_str_exist);
		}
		memcpy(p_service_key->service_key, service_key,
		    sizeof(p_service_key->service_key));
	}
	return 0;
}

int osm_subn_parse_service_keys_file(IN osm_subn_t * p_subn)
{
	char line[1024];
	char service_name[IB_SERVICE_NAME_SIZE] = {'\0'};
	char service_key_str[INET6_ADDRSTRLEN];
	uint8_t service_key[16];
	char *file_name;
	FILE *file;
	unsigned line_num;
	regex_t regex_comment, regex_line;
	regmatch_t regex_match[4];
	int rc = 0;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* Clear service name key table */
	osm_subn_clear_service_name_key_tbl(p_subn);
	file_name = p_subn->opt.service_name2key_map_file;
	if (!file_name) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"service name key file name is not specified.\n");
		goto Exit;
	}
	file = fopen(file_name, "r");
	if (!file) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7538: Can't open service name key file \'%s\'\n",
			file_name);
		rc = -1;
		goto Exit;
	}
	/* Create regex for pattern for comment line */
	rc = regcomp(&regex_comment, "^\\s*\\(#.*\\)\\?$", 0);
	if (rc) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7539: Failed to initialize parser for service name key "
			"file\n");
		fclose(file);
		goto Exit;
	}
	/* Create regex for pattern <service name> <service key> */
	rc = regcomp(&regex_line,
		     "^([a-zA-Z0-9\\-_.]{1,63})\\s+([0-9a-fA-F:]+)\\s*$",
		     REG_EXTENDED);
	if (rc) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7541: Failed to initialize parser for "
			"service name key file\n");
		regfree(&regex_comment);
		fclose(file);
		goto Exit;
	}
	line_num = 0;
	while (fgets(line, sizeof(line) - 1, file) != NULL) {
		line_num++;

		/* Comment or empty line */
		if (regexec(&regex_comment, line, 0, NULL, 0) == 0)
			continue;

		/* Read valid line */
		if (regexec(&regex_line, line, 4, &regex_match[0], 0) != 0) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7542: Failed to parse file \'%s\'"
				" line %d, invalid line format: %s\n",
				file_name, line_num, line);
			rc = -1;
			break;
		}

		/* Check line length */
		if (regex_match[0].rm_eo >= 256) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7543: Failed to parse file \'%s\'"
				"line %d, line too long\n",
				file_name, line_num);
			rc = -1;
			break;
		}

		/* Parse service name */
		if (regex_match[1].rm_so != -1) {
			regex_get_substr(service_name, line, &regex_match[1]);
		}

		if (regex_match[2].rm_so != -1) {
			regex_get_substr(service_key_str, line, &regex_match[2]);
			if (inet_pton(AF_INET6, service_key_str, service_key) <= 0) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 7544: Failed to parse file \'%s\'"
					"line %d, invalid service key (not in ipv6 format).\n",
					file_name, line_num);
				rc = -1;
				break;
			}
			subn_add_service_name(p_subn, service_name, service_key,
					      service_key_str);
		}
		memset(service_name, 0, sizeof(service_name));
	}
	/* Release resources */
	fclose(file);
	regfree(&regex_comment);
	regfree(&regex_line);

	if (rc)
		osm_subn_clear_service_name_key_tbl(p_subn);
Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log)
	return rc;
}

static void apply_vlids_to_sw_lft(cl_map_item_t *p_item, void *ctx)
{
	osm_switch_apply_vlids_to_lft((osm_switch_t*)p_item, (osm_subn_t*)ctx);
}

boolean_t osm_subn_apply_vlids_to_lfts(IN osm_subn_t *p_subn)
{
	osm_ucast_mgr_t *p_ucast_mgr = &p_subn->p_osm->sm.ucast_mgr;
	osm_lid_mgr_t *p_lid_mgr = &p_subn->p_osm->sm.lid_mgr;
	uint16_t max_vlid = osm_vlid_mgr_get_max_vlid(&p_lid_mgr->vlid_mgr);

	/* If no virtual lids nothing to apply */
	if (!max_vlid) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"No Vlids to apply to lft tables\n");
		goto EXIT;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE,
			"Applying Vlids to lft tables\n");

	cl_qmap_apply_func(&p_subn->sw_guid_tbl, apply_vlids_to_sw_lft, p_subn);

	if (p_subn->max_lid < max_vlid)
		p_subn->max_lid = p_ucast_mgr->max_lid = max_vlid;

	osm_vlid_mgr_reset_update_since_apply(&p_lid_mgr->vlid_mgr);
EXIT:
	return max_vlid ? TRUE : FALSE;
}

ib_net64_t osm_subn_calc_mkey(osm_subn_t *p_subn, ib_net64_t guid)
{
	uint8_t sha1_result[CL_SHA1_RESULTLEN];
	cl_sha1_ctxt_t sha1_ctxt;
	ib_net64_t m_key = p_subn->opt.m_key;
	size_t i;

	if (!p_subn->opt.m_key_per_port)
		goto exit;

	m_key = 0;
	if (!guid)
		/* For unknown GUIDs, use m_key 0 */
		goto exit;

	cl_sha1_init(&sha1_ctxt);
	cl_sha1_update(&sha1_ctxt, (uint8_t *)&p_subn->opt.m_key, sizeof(p_subn->opt.m_key));
	cl_sha1_update(&sha1_ctxt, (uint8_t *)&guid, sizeof(guid));
	cl_sha1_digest(&sha1_ctxt, sha1_result);

	/* Convert SHA1 result to 64 bit using xor */
	for (i = 0; i < CL_SHA1_RESULTLEN; i++)
		m_key ^= (uint64_t)sha1_result[i] << ((i % sizeof(m_key)) * 8);

exit:
	return m_key;
}

static osm_rtr_selection_function_t subn_selection_function_type(const char * selection_func)
{
	if (!strcmp(selection_func, "crc32"))
		return OSM_CRC32;

	return OSM_FUNC_UNKNOWN;
}

void osm_update_rtr_selection_algo_parameters(osm_subn_t * p_subn, char * selection_param)
{
	char *algo_parameters = strdup((const char *)selection_param), *p;
	const char *str;
	uint32_t tmp_algo_params = 0;

	str = strtok_r(algo_parameters, ",\t\n", &p);
	while (str && *str) {
		if (!strcmp(str,"sgid")) {
			tmp_algo_params |= OSM_RTR_SELECTION_PARAM_SGID;
		} else if (!strcmp(str, "dgid")) {
			tmp_algo_params |= OSM_RTR_SELECTION_PARAM_DGID;
		} else {
			/*if input not valid, doesn't update selection
			 * algorithm parameters*/
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 7546: Router selection algorithm parameters "
				 "are not valid. Should be sgid and/or dgid in "
				 "comma separated values format\n");
			goto Exit;
		}
		str = strtok_r(NULL, ",\t\n", &p);
	}

	if (tmp_algo_params)
		p_subn->opt.rtr_selection_params_mask = tmp_algo_params;
Exit:
	free(algo_parameters);
}

void osm_update_rtr_selection_function(osm_subn_t * p_subn, char * selection_func)
{
	osm_rtr_selection_function_t new_f =
	    subn_selection_function_type((const char *) selection_func);

	if (new_f == OSM_FUNC_UNKNOWN) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 7547: Router selection function is not valid. "
			"Supported values: crc32.\n");
	} else {
		p_subn->opt.rtr_selection_function = new_f;
	}
}

uint8_t osm_get_max_op_vls_by_node(IN const osm_subn_t * p_subn,
				   IN const struct osm_node * p_node)
{
	if (osm_node_get_type(p_node) == IB_NODE_TYPE_CA && p_subn->opt.max_op_vls_ca)
		return p_subn->opt.max_op_vls_ca;
	else if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH && p_subn->opt.max_op_vls_sw)
		return p_subn->opt.max_op_vls_sw;
	else if (osm_node_get_type(p_node) == IB_NODE_TYPE_ROUTER && p_subn->opt.max_op_vls_rtr)
		return p_subn->opt.max_op_vls_rtr;
	return p_subn->opt.max_op_vls;
}

boolean_t osm_is_sm_guid_allowed(IN const osm_subn_t *p_subn, ib_net64_t guid)
{
	const cl_u64_vector_t *p_vector = &p_subn->allowed_sm_guid_tbl;
	uint32_t vector_size = cl_u64_vector_get_size(p_vector);

	/* In case the vector is empty the allowed SM GUIDs feature is disabled */
	if (!vector_size)
		return TRUE;

	if (cl_u64_vector_lookup(p_vector, guid) != vector_size)
		return TRUE;

	return FALSE;
}

void osm_setup_hbf_weights(osm_subn_t *p_subn, const char *hbf_weights_str)
{
	regex_t weights_exp;
	int res = 0, len;
	regmatch_t regex_match[OSM_AR_HBF_WEIGHT_SUBGROUP_NUM + 1];
	char str[100] = {'\0'};
	unsigned int i, wrong_param = 0;

	if (!strcmp(hbf_weights_str, OSM_DEFAULT_WHBF_CONFIG)) {
		/* leave to router engine to decide */
		return;
	}

	res = regcomp(&weights_exp,
		      "^([0-9]+),([0-9]+),([0-9]+)$",
		      REG_EXTENDED);
	if (res) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 754A: Failed to initialize parser for "
			"hbf_weights parameter.\n");
		return;
	}

	if (regexec(&weights_exp, hbf_weights_str, 4, &regex_match[0], 0)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 754B: Failed to parse hbf_weights parameter, "
			"invalid format.\n");
		regfree(&weights_exp);
		return;
	} else {
		len = regex_match[1].rm_eo - regex_match[1].rm_so;
		for (i = 0; i < OSM_AR_HBF_WEIGHT_SUBGROUP_NUM; i++) {
			len = regex_match[i+1].rm_eo - regex_match[i+1].rm_so;
			if (len > 5 ) {
				wrong_param = 1;
			} else {
				regex_get_substr(str, hbf_weights_str, &regex_match[i+1]);
				if (atoi(str) > 0xffff)
					wrong_param = 1;
			}
			if (wrong_param) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 754F: Failed to parse hbf_weights parameter, "
					"invalid format, weight #%u is greater than 0xffff.\n",
					i + 1);
				regfree(&weights_exp);
				return;
			}

		}
		p_subn->ar_conf_params.hbf_weights_configured = 0;
		for (i = 0; i < OSM_AR_HBF_WEIGHT_SUBGROUP_NUM; i++) {
			regex_get_substr(str, hbf_weights_str, &regex_match[i+1]);
			p_subn->ar_conf_params.hbf_weights[i] = atoi(str);
			/* all weights zero - no whbf */
			p_subn->ar_conf_params.hbf_weights_configured |=
				p_subn->ar_conf_params.hbf_weights[i];
		}
	}
	regfree(&weights_exp);
	p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
}

void osm_subn_clear_ports_mads_queue(osm_subn_t * p_subn)
{
	cl_qmap_t *p_port_guid_tbl = &p_subn->port_guid_tbl;
	osm_port_t *p_port;

	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item))
		osm_port_remove_all_mads(p_port, p_subn->p_osm->sm.p_mad_pool);
}

void osm_subn_send_profiles_config_block(struct osm_sm * p_sm,
					 struct osm_switch * p_sw,
					 uint8_t feature,
					 uint8_t block_num)
{
	ib_profiles_config_t calc_port_profiles;
	ib_profiles_config_t *p_ports_profiles;
	osm_profiles_context_t context;
	uint16_t attr_mod;
	uint8_t profile;
	int port_num, port_num_start, port_num_end;
	osm_physp_t *p_physp, *p_physp0;
	ib_api_status_t status;

	/* Iterate over port numbers relevant for this block and build payload */
	memset(&calc_port_profiles, 0, sizeof(ib_profiles_config_t));

	/* For disabling Fast Recovery, set all profiles to 0 */
	if (feature == IB_PROFILES_FEATURE_FAST_RECOVERY &&
	    p_sm->p_subn->opt.fast_recovery_enabled == OSM_CAPABILITY_DISABLED)
		goto send_mad;

	/* Set port range to iterate on for this block */
	port_num_start = block_num * IB_PROFILE_CONFIG_NUM_PROFILES_PER_BLOCK;
	port_num_end = port_num_start + IB_PROFILE_CONFIG_NUM_PROFILES_PER_BLOCK;
	if (p_sw->p_node->physp_tbl_size < port_num_end)
		port_num_end = p_sw->p_node->physp_tbl_size;

	for (port_num = port_num_start; port_num < port_num_end; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp)
			continue;

		if (ib_port_info_get_port_state(&p_physp->port_info) == IB_LINK_DOWN)
			continue;

		/*
		 * Skip special ports, so their port profile remains 0, which means
		 * avoiding configuration of Fast Recovery feature for those ports.
		 */
		if (feature == IB_PROFILES_FEATURE_FAST_RECOVERY &&
		    ib_mlnx_epi_get_is_special_port_bit(&p_physp->ext_port_info))
			continue;

		profile = osm_physp_get_profile(p_physp, feature);
		ib_profiles_config_set_profile(&calc_port_profiles, port_num, profile);
	}

send_mad:
	/*
	 * Avoid sending the MAD if calculated profiles are identical to those
	 * previously received from subnet
	 */
	p_ports_profiles = osm_switch_get_ports_profiles(p_sw, feature);
	if (!p_sm->p_subn->need_update && !p_sw->need_update &&
	    !memcmp(&calc_port_profiles, &p_ports_profiles[block_num],
		    sizeof(ib_profiles_config_t))) {
		return;
	}

	/* Send MAD */
	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 7501: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;
	context.feature = feature;
	context.block = block_num;

	/*
	 * Attribute modifier of ProfilesConfig:
	 *	8 MSB feature
	 *	8 LSB block number
	 */
	attr_mod = (feature << IB_PROFILES_FEATURE_MASK_SHIFT) | block_num;

	/*
	 * Set off block bit of bitmask as indication profiles of this block
	 * were not configured yet for the feature.
	 * Thus, if not all profileConfig blocks were set, avoids sending feature
	 * MADs that are using those profile.
	 */
	osm_switch_set_profiles_configured(p_sw, context.block, FALSE, feature);

	/* NVLHBFConfig and correspond ProfilesConfig are sent as a apart of AR flow */
	if (feature == IB_PROFILES_FEATURE_NVL_HBF)
		p_sw->ar_data.pending_mads++;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0, (const uint8_t *)&calc_port_profiles,
			     sizeof(ib_profiles_config_t),
			     IB_MAD_ATTR_PROFILES_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS)
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR 7502: "
			"Setting ProfilesConfig to 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp0->port_guid),	ib_get_err_str(status));
}

struct osm_switch * osm_subn_get_sm_switch(osm_subn_t *p_subn)
{
	osm_port_t *p_sm_port;
	osm_node_t *p_sw_node, *p_sm_node;
	uint8_t port_num;

	p_sm_port = osm_get_port_by_guid(p_subn, p_subn->sm_port_guid);
	if (!p_sm_port) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 750E: "
			"Could not find SM port\n");
		return NULL;
	}

	p_sm_node = p_sm_port->p_node;
	if (osm_node_get_type(p_sm_node) == IB_NODE_TYPE_SWITCH)
		return p_sm_node->sw;

	/* Get SM's remote switch */
	port_num = osm_physp_get_port_num(osm_port_get_physp(p_sm_port));
	p_sw_node = osm_node_get_remote_node(p_sm_node, port_num, NULL);
	if (!p_sw_node) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR 750F: "
			"Could not find SM remote node\n");
		return NULL;
	}

	return p_sw_node->sw;
}
